
drop table if exists pts_collection;
create table pts_collection ( id serial primary key, gid integer, pt pcpoint(1) );
begin;
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,0,0,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,1,0.01,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,2,0.02,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,3,0.03,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,4,0.04,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,5,0.05,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,6,0.06,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,7,0.07,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,8,0.08,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,9,0.09,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,10,0.1,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,11,0.11,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,12,0.12,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,13,0.13,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,14,0.14,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,15,0.15,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,16,0.16,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,17,0.17,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,18,0.18,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[0,19,0.19,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,20,0.2,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,21,0.21,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,22,0.22,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,23,0.23,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,24,0.24,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,25,0.25,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,26,0.26,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,27,0.27,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,28,0.28,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,29,0.29,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,30,0.3,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,31,0.31,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,32,0.32,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,33,0.33,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,34,0.34,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,35,0.35,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,36,0.36,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,37,0.37,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,38,0.38,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[0,39,0.39,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,40,0.4,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,41,0.41,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,42,0.42,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,43,0.43,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,44,0.44,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,45,0.45,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,46,0.46,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,47,0.47,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,48,0.48,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,49,0.49,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,50,0.5,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,51,0.51,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,52,0.52,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,53,0.53,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,54,0.54,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,55,0.55,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,56,0.56,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,57,0.57,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,58,0.58,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[0,59,0.59,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,60,0.6,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,61,0.61,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,62,0.62,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,63,0.63,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,64,0.64,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,65,0.65,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,66,0.66,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,67,0.67,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,68,0.68,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,69,0.69,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,70,0.7,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,71,0.71,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,72,0.72,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,73,0.73,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,74,0.74,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,75,0.75,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,76,0.76,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,77,0.77,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,78,0.78,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[0,79,0.79,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,80,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,81,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,82,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,83,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,84,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,85,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,86,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,87,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,88,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,89,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,90,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,91,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,92,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,93,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,94,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,95,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,96,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,97,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,98,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[0,99,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,100,1,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,101,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,102,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,103,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,104,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,105,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,106,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,107,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,108,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,109,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,110,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,111,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,112,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,113,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,114,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,115,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,116,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,117,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,118,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[0,119,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,120,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,121,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,122,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,123,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,124,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,125,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,126,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,127,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,128,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,129,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,130,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,131,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,132,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,133,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,134,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,135,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,136,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,137,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,138,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[0,139,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,140,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,141,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,142,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,143,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,144,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,145,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,146,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,147,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,148,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,149,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,150,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,151,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,152,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,153,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,154,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,155,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,156,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,157,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,158,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[0,159,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,160,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,161,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,162,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,163,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,164,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,165,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,166,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,167,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,168,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,169,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,170,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,171,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,172,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,173,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,174,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,175,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,176,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,177,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,178,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[0,179,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,180,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,181,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,182,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,183,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,184,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,185,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,186,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,187,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,188,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,189,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,190,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,191,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,192,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,193,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,194,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,195,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,196,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,197,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,198,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[0,199,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,0,0.01,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,1,0.02,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,2,0.03,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,3,0.04,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,4,0.05,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,5,0.06,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,6,0.07,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,7,0.08,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,8,0.09,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,9,0.1,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,10,0.11,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,11,0.12,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,12,0.13,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,13,0.14,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,14,0.15,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,15,0.16,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,16,0.17,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,17,0.18,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,18,0.19,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[1,19,0.2,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,20,0.21,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,21,0.22,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,22,0.23,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,23,0.24,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,24,0.25,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,25,0.26,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,26,0.27,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,27,0.28,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,28,0.29,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,29,0.3,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,30,0.31,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,31,0.32,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,32,0.33,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,33,0.34,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,34,0.35,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,35,0.36,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,36,0.37,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,37,0.38,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,38,0.39,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[1,39,0.4,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,40,0.41,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,41,0.42,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,42,0.43,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,43,0.44,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,44,0.45,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,45,0.46,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,46,0.47,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,47,0.48,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,48,0.49,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,49,0.5,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,50,0.51,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,51,0.52,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,52,0.53,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,53,0.54,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,54,0.55,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,55,0.56,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,56,0.57,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,57,0.58,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,58,0.59,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[1,59,0.6,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,60,0.61,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,61,0.62,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,62,0.63,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,63,0.64,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,64,0.65,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,65,0.66,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,66,0.67,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,67,0.68,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,68,0.69,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,69,0.7,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,70,0.71,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,71,0.72,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,72,0.73,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,73,0.74,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,74,0.75,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,75,0.76,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,76,0.77,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,77,0.78,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,78,0.79,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[1,79,0.8,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,80,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,81,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,82,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,83,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,84,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,85,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,86,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,87,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,88,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,89,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,90,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,91,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,92,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,93,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,94,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,95,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,96,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,97,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,98,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[1,99,1,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,100,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,101,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,102,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,103,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,104,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,105,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,106,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,107,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,108,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,109,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,110,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,111,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,112,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,113,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,114,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,115,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,116,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,117,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,118,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[1,119,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,120,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,121,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,122,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,123,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,124,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,125,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,126,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,127,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,128,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,129,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,130,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,131,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,132,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,133,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,134,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,135,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,136,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,137,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,138,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[1,139,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,140,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,141,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,142,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,143,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,144,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,145,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,146,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,147,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,148,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,149,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,150,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,151,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,152,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,153,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,154,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,155,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,156,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,157,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,158,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[1,159,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,160,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,161,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,162,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,163,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,164,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,165,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,166,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,167,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,168,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,169,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,170,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,171,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,172,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,173,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,174,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,175,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,176,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,177,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,178,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[1,179,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,180,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,181,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,182,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,183,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,184,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,185,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,186,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,187,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,188,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,189,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,190,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,191,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,192,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,193,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,194,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,195,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,196,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,197,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,198,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[1,199,2,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,0,0.02,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,1,0.03,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,2,0.04,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,3,0.05,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,4,0.06,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,5,0.07,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,6,0.08,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,7,0.09,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,8,0.1,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,9,0.11,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,10,0.12,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,11,0.13,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,12,0.14,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,13,0.15,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,14,0.16,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,15,0.17,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,16,0.18,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,17,0.19,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,18,0.2,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[2,19,0.21,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,20,0.22,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,21,0.23,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,22,0.24,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,23,0.25,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,24,0.26,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,25,0.27,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,26,0.28,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,27,0.29,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,28,0.3,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,29,0.31,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,30,0.32,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,31,0.33,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,32,0.34,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,33,0.35,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,34,0.36,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,35,0.37,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,36,0.38,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,37,0.39,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,38,0.4,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[2,39,0.41,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,40,0.42,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,41,0.43,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,42,0.44,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,43,0.45,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,44,0.46,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,45,0.47,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,46,0.48,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,47,0.49,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,48,0.5,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,49,0.51,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,50,0.52,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,51,0.53,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,52,0.54,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,53,0.55,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,54,0.56,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,55,0.57,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,56,0.58,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,57,0.59,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,58,0.6,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[2,59,0.61,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,60,0.62,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,61,0.63,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,62,0.64,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,63,0.65,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,64,0.66,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,65,0.67,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,66,0.68,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,67,0.69,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,68,0.7,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,69,0.71,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,70,0.72,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,71,0.73,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,72,0.74,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,73,0.75,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,74,0.76,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,75,0.77,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,76,0.78,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,77,0.79,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,78,0.8,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[2,79,0.81,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,80,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,81,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,82,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,83,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,84,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,85,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,86,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,87,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,88,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,89,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,90,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,91,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,92,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,93,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,94,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,95,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,96,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,97,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,98,1,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[2,99,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,100,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,101,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,102,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,103,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,104,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,105,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,106,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,107,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,108,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,109,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,110,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,111,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,112,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,113,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,114,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,115,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,116,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,117,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,118,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[2,119,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,120,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,121,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,122,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,123,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,124,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,125,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,126,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,127,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,128,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,129,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,130,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,131,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,132,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,133,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,134,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,135,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,136,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,137,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,138,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[2,139,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,140,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,141,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,142,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,143,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,144,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,145,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,146,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,147,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,148,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,149,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,150,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,151,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,152,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,153,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,154,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,155,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,156,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,157,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,158,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[2,159,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,160,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,161,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,162,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,163,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,164,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,165,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,166,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,167,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,168,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,169,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,170,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,171,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,172,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,173,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,174,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,175,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,176,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,177,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,178,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[2,179,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,180,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,181,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,182,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,183,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,184,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,185,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,186,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,187,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,188,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,189,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,190,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,191,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,192,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,193,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,194,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,195,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,196,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,197,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,198,2,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[2,199,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,0,0.03,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,1,0.04,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,2,0.05,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,3,0.06,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,4,0.07,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,5,0.08,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,6,0.09,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,7,0.1,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,8,0.11,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,9,0.12,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,10,0.13,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,11,0.14,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,12,0.15,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,13,0.16,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,14,0.17,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,15,0.18,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,16,0.19,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,17,0.2,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,18,0.21,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[3,19,0.22,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,20,0.23,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,21,0.24,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,22,0.25,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,23,0.26,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,24,0.27,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,25,0.28,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,26,0.29,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,27,0.3,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,28,0.31,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,29,0.32,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,30,0.33,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,31,0.34,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,32,0.35,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,33,0.36,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,34,0.37,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,35,0.38,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,36,0.39,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,37,0.4,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,38,0.41,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[3,39,0.42,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,40,0.43,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,41,0.44,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,42,0.45,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,43,0.46,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,44,0.47,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,45,0.48,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,46,0.49,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,47,0.5,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,48,0.51,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,49,0.52,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,50,0.53,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,51,0.54,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,52,0.55,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,53,0.56,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,54,0.57,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,55,0.58,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,56,0.59,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,57,0.6,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,58,0.61,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[3,59,0.62,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,60,0.63,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,61,0.64,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,62,0.65,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,63,0.66,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,64,0.67,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,65,0.68,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,66,0.69,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,67,0.7,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,68,0.71,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,69,0.72,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,70,0.73,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,71,0.74,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,72,0.75,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,73,0.76,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,74,0.77,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,75,0.78,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,76,0.79,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,77,0.8,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,78,0.81,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[3,79,0.82,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,80,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,81,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,82,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,83,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,84,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,85,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,86,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,87,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,88,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,89,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,90,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,91,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,92,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,93,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,94,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,95,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,96,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,97,1,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,98,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[3,99,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,100,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,101,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,102,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,103,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,104,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,105,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,106,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,107,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,108,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,109,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,110,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,111,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,112,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,113,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,114,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,115,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,116,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,117,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,118,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[3,119,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,120,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,121,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,122,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,123,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,124,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,125,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,126,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,127,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,128,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,129,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,130,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,131,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,132,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,133,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,134,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,135,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,136,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,137,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,138,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[3,139,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,140,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,141,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,142,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,143,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,144,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,145,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,146,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,147,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,148,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,149,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,150,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,151,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,152,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,153,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,154,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,155,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,156,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,157,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,158,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[3,159,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,160,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,161,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,162,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,163,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,164,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,165,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,166,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,167,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,168,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,169,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,170,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,171,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,172,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,173,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,174,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,175,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,176,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,177,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,178,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[3,179,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,180,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,181,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,182,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,183,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,184,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,185,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,186,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,187,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,188,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,189,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,190,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,191,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,192,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,193,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,194,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,195,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,196,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,197,2,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,198,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[3,199,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,0,0.04,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,1,0.05,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,2,0.06,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,3,0.07,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,4,0.08,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,5,0.09,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,6,0.1,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,7,0.11,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,8,0.12,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,9,0.13,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,10,0.14,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,11,0.15,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,12,0.16,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,13,0.17,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,14,0.18,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,15,0.19,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,16,0.2,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,17,0.21,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,18,0.22,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[4,19,0.23,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,20,0.24,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,21,0.25,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,22,0.26,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,23,0.27,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,24,0.28,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,25,0.29,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,26,0.3,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,27,0.31,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,28,0.32,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,29,0.33,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,30,0.34,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,31,0.35,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,32,0.36,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,33,0.37,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,34,0.38,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,35,0.39,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,36,0.4,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,37,0.41,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,38,0.42,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[4,39,0.43,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,40,0.44,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,41,0.45,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,42,0.46,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,43,0.47,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,44,0.48,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,45,0.49,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,46,0.5,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,47,0.51,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,48,0.52,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,49,0.53,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,50,0.54,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,51,0.55,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,52,0.56,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,53,0.57,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,54,0.58,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,55,0.59,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,56,0.6,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,57,0.61,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,58,0.62,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[4,59,0.63,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,60,0.64,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,61,0.65,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,62,0.66,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,63,0.67,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,64,0.68,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,65,0.69,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,66,0.7,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,67,0.71,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,68,0.72,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,69,0.73,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,70,0.74,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,71,0.75,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,72,0.76,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,73,0.77,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,74,0.78,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,75,0.79,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,76,0.8,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,77,0.81,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,78,0.82,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[4,79,0.83,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,80,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,81,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,82,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,83,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,84,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,85,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,86,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,87,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,88,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,89,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,90,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,91,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,92,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,93,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,94,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,95,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,96,1,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,97,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,98,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[4,99,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,100,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,101,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,102,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,103,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,104,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,105,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,106,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,107,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,108,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,109,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,110,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,111,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,112,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,113,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,114,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,115,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,116,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,117,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,118,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[4,119,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,120,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,121,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,122,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,123,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,124,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,125,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,126,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,127,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,128,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,129,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,130,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,131,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,132,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,133,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,134,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,135,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,136,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,137,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,138,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[4,139,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,140,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,141,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,142,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,143,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,144,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,145,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,146,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,147,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,148,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,149,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,150,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,151,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,152,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,153,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,154,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,155,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,156,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,157,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,158,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[4,159,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,160,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,161,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,162,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,163,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,164,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,165,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,166,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,167,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,168,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,169,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,170,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,171,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,172,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,173,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,174,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,175,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,176,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,177,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,178,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[4,179,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,180,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,181,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,182,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,183,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,184,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,185,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,186,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,187,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,188,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,189,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,190,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,191,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,192,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,193,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,194,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,195,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,196,2,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,197,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,198,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[4,199,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,0,0.05,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,1,0.06,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,2,0.07,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,3,0.08,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,4,0.09,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,5,0.1,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,6,0.11,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,7,0.12,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,8,0.13,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,9,0.14,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,10,0.15,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,11,0.16,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,12,0.17,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,13,0.18,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,14,0.19,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,15,0.2,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,16,0.21,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,17,0.22,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,18,0.23,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[5,19,0.24,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,20,0.25,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,21,0.26,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,22,0.27,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,23,0.28,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,24,0.29,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,25,0.3,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,26,0.31,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,27,0.32,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,28,0.33,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,29,0.34,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,30,0.35,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,31,0.36,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,32,0.37,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,33,0.38,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,34,0.39,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,35,0.4,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,36,0.41,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,37,0.42,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,38,0.43,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[5,39,0.44,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,40,0.45,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,41,0.46,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,42,0.47,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,43,0.48,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,44,0.49,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,45,0.5,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,46,0.51,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,47,0.52,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,48,0.53,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,49,0.54,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,50,0.55,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,51,0.56,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,52,0.57,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,53,0.58,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,54,0.59,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,55,0.6,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,56,0.61,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,57,0.62,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,58,0.63,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[5,59,0.64,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,60,0.65,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,61,0.66,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,62,0.67,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,63,0.68,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,64,0.69,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,65,0.7,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,66,0.71,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,67,0.72,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,68,0.73,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,69,0.74,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,70,0.75,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,71,0.76,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,72,0.77,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,73,0.78,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,74,0.79,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,75,0.8,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,76,0.81,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,77,0.82,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,78,0.83,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[5,79,0.84,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,80,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,81,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,82,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,83,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,84,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,85,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,86,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,87,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,88,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,89,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,90,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,91,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,92,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,93,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,94,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,95,1,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,96,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,97,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,98,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[5,99,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,100,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,101,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,102,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,103,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,104,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,105,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,106,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,107,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,108,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,109,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,110,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,111,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,112,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,113,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,114,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,115,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,116,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,117,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,118,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[5,119,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,120,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,121,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,122,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,123,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,124,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,125,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,126,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,127,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,128,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,129,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,130,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,131,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,132,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,133,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,134,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,135,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,136,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,137,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,138,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[5,139,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,140,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,141,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,142,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,143,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,144,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,145,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,146,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,147,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,148,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,149,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,150,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,151,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,152,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,153,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,154,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,155,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,156,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,157,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,158,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[5,159,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,160,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,161,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,162,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,163,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,164,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,165,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,166,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,167,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,168,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,169,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,170,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,171,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,172,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,173,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,174,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,175,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,176,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,177,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,178,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[5,179,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,180,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,181,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,182,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,183,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,184,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,185,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,186,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,187,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,188,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,189,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,190,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,191,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,192,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,193,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,194,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,195,2,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,196,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,197,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,198,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[5,199,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,0,0.06,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,1,0.07,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,2,0.08,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,3,0.09,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,4,0.1,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,5,0.11,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,6,0.12,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,7,0.13,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,8,0.14,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,9,0.15,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,10,0.16,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,11,0.17,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,12,0.18,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,13,0.19,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,14,0.2,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,15,0.21,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,16,0.22,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,17,0.23,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,18,0.24,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[6,19,0.25,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,20,0.26,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,21,0.27,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,22,0.28,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,23,0.29,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,24,0.3,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,25,0.31,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,26,0.32,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,27,0.33,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,28,0.34,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,29,0.35,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,30,0.36,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,31,0.37,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,32,0.38,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,33,0.39,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,34,0.4,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,35,0.41,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,36,0.42,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,37,0.43,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,38,0.44,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[6,39,0.45,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,40,0.46,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,41,0.47,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,42,0.48,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,43,0.49,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,44,0.5,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,45,0.51,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,46,0.52,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,47,0.53,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,48,0.54,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,49,0.55,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,50,0.56,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,51,0.57,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,52,0.58,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,53,0.59,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,54,0.6,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,55,0.61,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,56,0.62,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,57,0.63,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,58,0.64,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[6,59,0.65,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,60,0.66,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,61,0.67,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,62,0.68,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,63,0.69,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,64,0.7,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,65,0.71,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,66,0.72,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,67,0.73,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,68,0.74,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,69,0.75,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,70,0.76,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,71,0.77,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,72,0.78,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,73,0.79,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,74,0.8,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,75,0.81,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,76,0.82,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,77,0.83,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,78,0.84,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[6,79,0.85,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,80,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,81,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,82,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,83,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,84,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,85,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,86,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,87,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,88,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,89,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,90,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,91,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,92,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,93,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,94,1,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,95,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,96,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,97,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,98,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[6,99,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,100,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,101,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,102,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,103,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,104,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,105,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,106,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,107,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,108,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,109,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,110,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,111,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,112,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,113,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,114,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,115,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,116,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,117,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,118,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[6,119,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,120,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,121,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,122,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,123,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,124,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,125,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,126,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,127,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,128,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,129,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,130,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,131,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,132,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,133,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,134,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,135,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,136,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,137,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,138,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[6,139,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,140,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,141,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,142,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,143,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,144,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,145,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,146,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,147,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,148,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,149,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,150,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,151,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,152,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,153,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,154,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,155,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,156,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,157,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,158,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[6,159,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,160,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,161,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,162,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,163,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,164,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,165,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,166,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,167,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,168,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,169,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,170,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,171,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,172,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,173,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,174,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,175,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,176,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,177,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,178,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[6,179,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,180,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,181,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,182,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,183,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,184,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,185,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,186,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,187,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,188,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,189,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,190,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,191,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,192,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,193,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,194,2,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,195,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,196,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,197,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,198,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[6,199,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,0,0.07,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,1,0.08,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,2,0.09,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,3,0.1,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,4,0.11,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,5,0.12,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,6,0.13,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,7,0.14,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,8,0.15,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,9,0.16,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,10,0.17,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,11,0.18,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,12,0.19,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,13,0.2,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,14,0.21,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,15,0.22,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,16,0.23,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,17,0.24,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,18,0.25,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[7,19,0.26,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,20,0.27,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,21,0.28,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,22,0.29,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,23,0.3,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,24,0.31,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,25,0.32,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,26,0.33,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,27,0.34,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,28,0.35,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,29,0.36,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,30,0.37,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,31,0.38,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,32,0.39,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,33,0.4,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,34,0.41,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,35,0.42,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,36,0.43,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,37,0.44,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,38,0.45,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[7,39,0.46,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,40,0.47,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,41,0.48,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,42,0.49,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,43,0.5,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,44,0.51,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,45,0.52,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,46,0.53,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,47,0.54,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,48,0.55,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,49,0.56,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,50,0.57,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,51,0.58,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,52,0.59,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,53,0.6,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,54,0.61,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,55,0.62,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,56,0.63,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,57,0.64,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,58,0.65,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[7,59,0.66,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,60,0.67,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,61,0.68,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,62,0.69,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,63,0.7,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,64,0.71,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,65,0.72,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,66,0.73,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,67,0.74,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,68,0.75,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,69,0.76,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,70,0.77,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,71,0.78,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,72,0.79,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,73,0.8,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,74,0.81,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,75,0.82,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,76,0.83,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,77,0.84,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,78,0.85,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[7,79,0.86,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,80,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,81,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,82,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,83,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,84,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,85,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,86,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,87,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,88,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,89,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,90,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,91,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,92,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,93,1,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,94,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,95,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,96,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,97,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,98,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[7,99,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,100,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,101,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,102,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,103,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,104,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,105,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,106,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,107,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,108,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,109,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,110,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,111,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,112,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,113,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,114,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,115,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,116,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,117,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,118,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[7,119,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,120,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,121,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,122,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,123,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,124,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,125,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,126,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,127,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,128,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,129,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,130,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,131,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,132,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,133,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,134,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,135,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,136,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,137,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,138,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[7,139,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,140,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,141,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,142,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,143,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,144,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,145,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,146,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,147,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,148,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,149,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,150,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,151,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,152,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,153,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,154,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,155,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,156,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,157,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,158,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[7,159,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,160,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,161,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,162,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,163,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,164,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,165,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,166,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,167,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,168,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,169,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,170,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,171,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,172,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,173,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,174,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,175,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,176,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,177,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,178,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[7,179,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,180,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,181,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,182,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,183,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,184,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,185,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,186,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,187,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,188,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,189,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,190,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,191,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,192,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,193,2,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,194,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,195,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,196,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,197,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,198,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[7,199,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,0,0.08,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,1,0.09,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,2,0.1,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,3,0.11,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,4,0.12,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,5,0.13,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,6,0.14,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,7,0.15,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,8,0.16,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,9,0.17,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,10,0.18,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,11,0.19,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,12,0.2,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,13,0.21,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,14,0.22,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,15,0.23,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,16,0.24,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,17,0.25,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,18,0.26,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[8,19,0.27,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,20,0.28,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,21,0.29,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,22,0.3,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,23,0.31,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,24,0.32,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,25,0.33,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,26,0.34,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,27,0.35,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,28,0.36,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,29,0.37,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,30,0.38,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,31,0.39,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,32,0.4,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,33,0.41,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,34,0.42,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,35,0.43,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,36,0.44,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,37,0.45,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,38,0.46,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[8,39,0.47,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,40,0.48,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,41,0.49,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,42,0.5,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,43,0.51,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,44,0.52,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,45,0.53,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,46,0.54,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,47,0.55,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,48,0.56,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,49,0.57,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,50,0.58,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,51,0.59,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,52,0.6,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,53,0.61,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,54,0.62,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,55,0.63,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,56,0.64,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,57,0.65,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,58,0.66,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[8,59,0.67,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,60,0.68,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,61,0.69,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,62,0.7,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,63,0.71,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,64,0.72,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,65,0.73,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,66,0.74,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,67,0.75,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,68,0.76,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,69,0.77,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,70,0.78,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,71,0.79,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,72,0.8,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,73,0.81,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,74,0.82,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,75,0.83,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,76,0.84,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,77,0.85,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,78,0.86,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[8,79,0.87,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,80,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,81,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,82,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,83,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,84,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,85,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,86,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,87,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,88,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,89,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,90,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,91,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,92,1,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,93,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,94,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,95,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,96,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,97,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,98,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[8,99,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,100,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,101,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,102,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,103,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,104,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,105,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,106,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,107,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,108,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,109,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,110,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,111,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,112,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,113,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,114,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,115,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,116,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,117,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,118,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[8,119,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,120,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,121,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,122,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,123,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,124,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,125,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,126,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,127,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,128,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,129,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,130,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,131,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,132,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,133,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,134,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,135,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,136,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,137,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,138,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[8,139,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,140,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,141,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,142,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,143,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,144,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,145,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,146,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,147,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,148,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,149,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,150,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,151,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,152,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,153,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,154,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,155,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,156,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,157,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,158,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[8,159,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,160,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,161,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,162,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,163,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,164,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,165,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,166,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,167,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,168,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,169,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,170,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,171,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,172,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,173,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,174,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,175,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,176,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,177,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,178,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[8,179,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,180,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,181,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,182,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,183,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,184,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,185,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,186,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,187,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,188,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,189,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,190,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,191,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,192,2,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,193,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,194,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,195,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,196,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,197,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,198,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[8,199,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,0,0.09,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,1,0.1,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,2,0.11,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,3,0.12,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,4,0.13,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,5,0.14,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,6,0.15,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,7,0.16,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,8,0.17,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,9,0.18,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,10,0.19,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,11,0.2,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,12,0.21,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,13,0.22,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,14,0.23,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,15,0.24,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,16,0.25,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,17,0.26,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,18,0.27,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[9,19,0.28,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,20,0.29,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,21,0.3,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,22,0.31,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,23,0.32,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,24,0.33,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,25,0.34,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,26,0.35,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,27,0.36,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,28,0.37,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,29,0.38,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,30,0.39,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,31,0.4,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,32,0.41,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,33,0.42,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,34,0.43,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,35,0.44,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,36,0.45,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,37,0.46,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,38,0.47,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[9,39,0.48,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,40,0.49,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,41,0.5,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,42,0.51,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,43,0.52,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,44,0.53,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,45,0.54,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,46,0.55,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,47,0.56,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,48,0.57,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,49,0.58,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,50,0.59,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,51,0.6,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,52,0.61,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,53,0.62,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,54,0.63,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,55,0.64,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,56,0.65,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,57,0.66,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,58,0.67,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[9,59,0.68,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,60,0.69,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,61,0.7,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,62,0.71,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,63,0.72,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,64,0.73,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,65,0.74,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,66,0.75,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,67,0.76,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,68,0.77,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,69,0.78,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,70,0.79,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,71,0.8,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,72,0.81,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,73,0.82,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,74,0.83,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,75,0.84,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,76,0.85,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,77,0.86,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,78,0.87,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[9,79,0.88,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,80,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,81,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,82,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,83,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,84,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,85,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,86,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,87,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,88,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,89,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,90,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,91,1,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,92,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,93,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,94,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,95,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,96,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,97,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,98,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[9,99,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,100,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,101,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,102,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,103,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,104,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,105,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,106,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,107,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,108,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,109,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,110,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,111,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,112,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,113,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,114,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,115,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,116,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,117,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,118,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[9,119,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,120,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,121,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,122,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,123,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,124,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,125,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,126,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,127,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,128,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,129,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,130,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,131,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,132,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,133,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,134,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,135,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,136,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,137,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,138,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[9,139,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,140,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,141,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,142,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,143,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,144,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,145,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,146,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,147,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,148,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,149,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,150,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,151,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,152,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,153,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,154,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,155,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,156,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,157,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,158,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[9,159,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,160,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,161,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,162,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,163,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,164,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,165,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,166,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,167,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,168,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,169,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,170,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,171,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,172,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,173,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,174,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,175,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,176,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,177,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,178,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[9,179,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,180,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,181,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,182,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,183,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,184,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,185,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,186,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,187,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,188,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,189,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,190,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,191,2,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,192,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,193,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,194,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,195,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,196,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,197,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,198,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[9,199,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,0,0.1,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,1,0.11,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,2,0.12,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,3,0.13,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,4,0.14,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,5,0.15,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,6,0.16,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,7,0.17,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,8,0.18,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,9,0.19,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,10,0.2,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,11,0.21,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,12,0.22,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,13,0.23,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,14,0.24,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,15,0.25,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,16,0.26,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,17,0.27,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,18,0.28,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[10,19,0.29,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,20,0.3,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,21,0.31,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,22,0.32,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,23,0.33,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,24,0.34,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,25,0.35,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,26,0.36,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,27,0.37,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,28,0.38,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,29,0.39,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,30,0.4,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,31,0.41,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,32,0.42,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,33,0.43,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,34,0.44,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,35,0.45,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,36,0.46,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,37,0.47,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,38,0.48,0]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[10,39,0.49,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,40,0.5,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,41,0.51,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,42,0.52,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,43,0.53,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,44,0.54,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,45,0.55,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,46,0.56,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,47,0.57,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,48,0.58,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,49,0.59,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,50,0.6,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,51,0.61,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,52,0.62,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,53,0.63,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,54,0.64,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,55,0.65,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,56,0.66,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,57,0.67,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,58,0.68,0]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[10,59,0.69,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,60,0.7,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,61,0.71,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,62,0.72,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,63,0.73,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,64,0.74,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,65,0.75,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,66,0.76,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,67,0.77,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,68,0.78,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,69,0.79,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,70,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,71,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,72,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,73,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,74,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,75,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,76,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,77,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,78,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[10,79,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,80,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,81,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,82,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,83,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,84,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,85,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,86,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,87,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,88,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,89,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,90,1,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,91,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,92,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,93,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,94,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,95,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,96,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,97,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,98,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[10,99,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,100,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,101,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,102,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,103,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,104,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,105,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,106,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,107,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,108,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,109,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,110,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,111,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,112,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,113,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,114,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,115,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,116,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,117,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,118,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[10,119,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,120,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,121,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,122,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,123,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,124,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,125,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,126,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,127,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,128,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,129,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,130,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,131,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,132,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,133,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,134,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,135,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,136,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,137,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,138,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[10,139,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,140,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,141,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,142,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,143,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,144,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,145,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,146,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,147,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,148,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,149,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,150,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,151,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,152,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,153,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,154,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,155,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,156,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,157,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,158,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[10,159,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,160,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,161,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,162,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,163,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,164,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,165,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,166,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,167,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,168,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,169,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,170,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,171,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,172,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,173,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,174,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,175,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,176,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,177,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,178,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[10,179,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,180,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,181,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,182,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,183,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,184,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,185,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,186,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,187,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,188,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,189,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,190,2,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,191,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,192,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,193,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,194,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,195,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,196,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,197,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,198,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[10,199,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,0,0.11,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,1,0.12,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,2,0.13,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,3,0.14,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,4,0.15,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,5,0.16,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,6,0.17,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,7,0.18,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,8,0.19,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,9,0.2,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,10,0.21,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,11,0.22,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,12,0.23,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,13,0.24,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,14,0.25,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,15,0.26,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,16,0.27,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,17,0.28,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,18,0.29,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[11,19,0.3,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,20,0.31,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,21,0.32,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,22,0.33,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,23,0.34,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,24,0.35,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,25,0.36,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,26,0.37,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,27,0.38,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,28,0.39,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,29,0.4,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,30,0.41,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,31,0.42,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,32,0.43,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,33,0.44,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,34,0.45,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,35,0.46,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,36,0.47,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,37,0.48,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,38,0.49,1]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[11,39,0.5,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,40,0.51,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,41,0.52,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,42,0.53,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,43,0.54,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,44,0.55,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,45,0.56,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,46,0.57,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,47,0.58,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,48,0.59,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,49,0.6,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,50,0.61,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,51,0.62,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,52,0.63,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,53,0.64,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,54,0.65,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,55,0.66,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,56,0.67,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,57,0.68,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,58,0.69,1]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[11,59,0.7,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,60,0.71,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,61,0.72,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,62,0.73,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,63,0.74,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,64,0.75,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,65,0.76,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,66,0.77,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,67,0.78,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,68,0.79,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,69,0.8,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,70,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,71,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,72,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,73,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,74,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,75,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,76,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,77,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,78,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[11,79,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,80,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,81,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,82,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,83,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,84,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,85,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,86,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,87,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,88,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,89,1,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,90,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,91,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,92,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,93,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,94,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,95,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,96,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,97,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,98,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[11,99,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,100,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,101,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,102,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,103,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,104,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,105,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,106,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,107,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,108,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,109,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,110,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,111,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,112,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,113,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,114,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,115,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,116,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,117,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,118,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[11,119,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,120,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,121,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,122,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,123,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,124,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,125,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,126,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,127,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,128,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,129,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,130,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,131,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,132,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,133,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,134,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,135,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,136,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,137,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,138,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[11,139,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,140,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,141,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,142,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,143,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,144,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,145,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,146,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,147,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,148,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,149,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,150,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,151,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,152,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,153,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,154,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,155,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,156,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,157,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,158,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[11,159,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,160,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,161,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,162,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,163,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,164,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,165,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,166,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,167,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,168,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,169,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,170,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,171,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,172,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,173,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,174,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,175,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,176,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,177,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,178,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[11,179,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,180,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,181,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,182,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,183,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,184,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,185,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,186,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,187,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,188,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,189,2,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,190,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,191,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,192,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,193,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,194,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,195,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,196,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,197,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,198,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[11,199,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,0,0.12,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,1,0.13,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,2,0.14,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,3,0.15,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,4,0.16,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,5,0.17,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,6,0.18,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,7,0.19,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,8,0.2,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,9,0.21,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,10,0.22,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,11,0.23,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,12,0.24,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,13,0.25,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,14,0.26,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,15,0.27,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,16,0.28,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,17,0.29,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,18,0.3,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[12,19,0.31,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,20,0.32,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,21,0.33,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,22,0.34,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,23,0.35,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,24,0.36,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,25,0.37,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,26,0.38,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,27,0.39,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,28,0.4,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,29,0.41,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,30,0.42,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,31,0.43,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,32,0.44,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,33,0.45,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,34,0.46,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,35,0.47,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,36,0.48,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,37,0.49,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,38,0.5,2]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[12,39,0.51,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,40,0.52,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,41,0.53,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,42,0.54,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,43,0.55,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,44,0.56,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,45,0.57,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,46,0.58,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,47,0.59,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,48,0.6,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,49,0.61,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,50,0.62,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,51,0.63,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,52,0.64,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,53,0.65,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,54,0.66,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,55,0.67,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,56,0.68,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,57,0.69,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,58,0.7,2]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[12,59,0.71,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,60,0.72,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,61,0.73,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,62,0.74,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,63,0.75,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,64,0.76,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,65,0.77,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,66,0.78,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,67,0.79,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,68,0.8,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,69,0.81,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,70,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,71,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,72,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,73,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,74,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,75,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,76,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,77,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,78,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[12,79,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,80,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,81,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,82,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,83,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,84,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,85,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,86,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,87,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,88,1,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,89,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,90,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,91,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,92,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,93,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,94,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,95,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,96,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,97,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,98,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[12,99,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,100,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,101,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,102,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,103,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,104,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,105,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,106,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,107,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,108,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,109,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,110,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,111,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,112,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,113,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,114,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,115,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,116,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,117,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,118,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[12,119,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,120,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,121,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,122,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,123,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,124,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,125,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,126,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,127,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,128,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,129,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,130,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,131,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,132,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,133,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,134,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,135,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,136,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,137,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,138,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[12,139,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,140,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,141,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,142,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,143,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,144,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,145,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,146,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,147,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,148,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,149,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,150,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,151,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,152,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,153,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,154,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,155,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,156,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,157,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,158,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[12,159,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,160,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,161,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,162,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,163,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,164,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,165,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,166,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,167,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,168,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,169,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,170,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,171,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,172,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,173,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,174,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,175,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,176,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,177,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,178,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[12,179,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,180,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,181,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,182,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,183,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,184,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,185,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,186,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,187,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,188,2,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,189,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,190,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,191,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,192,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,193,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,194,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,195,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,196,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,197,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,198,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[12,199,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,0,0.13,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,1,0.14,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,2,0.15,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,3,0.16,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,4,0.17,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,5,0.18,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,6,0.19,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,7,0.2,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,8,0.21,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,9,0.22,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,10,0.23,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,11,0.24,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,12,0.25,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,13,0.26,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,14,0.27,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,15,0.28,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,16,0.29,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,17,0.3,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,18,0.31,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[13,19,0.32,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,20,0.33,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,21,0.34,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,22,0.35,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,23,0.36,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,24,0.37,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,25,0.38,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,26,0.39,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,27,0.4,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,28,0.41,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,29,0.42,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,30,0.43,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,31,0.44,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,32,0.45,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,33,0.46,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,34,0.47,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,35,0.48,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,36,0.49,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,37,0.5,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,38,0.51,3]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[13,39,0.52,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,40,0.53,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,41,0.54,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,42,0.55,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,43,0.56,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,44,0.57,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,45,0.58,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,46,0.59,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,47,0.6,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,48,0.61,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,49,0.62,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,50,0.63,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,51,0.64,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,52,0.65,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,53,0.66,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,54,0.67,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,55,0.68,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,56,0.69,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,57,0.7,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,58,0.71,3]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[13,59,0.72,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,60,0.73,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,61,0.74,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,62,0.75,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,63,0.76,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,64,0.77,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,65,0.78,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,66,0.79,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,67,0.8,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,68,0.81,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,69,0.82,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,70,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,71,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,72,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,73,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,74,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,75,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,76,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,77,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,78,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[13,79,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,80,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,81,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,82,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,83,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,84,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,85,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,86,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,87,1,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,88,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,89,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,90,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,91,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,92,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,93,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,94,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,95,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,96,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,97,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,98,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[13,99,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,100,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,101,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,102,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,103,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,104,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,105,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,106,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,107,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,108,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,109,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,110,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,111,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,112,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,113,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,114,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,115,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,116,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,117,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,118,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[13,119,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,120,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,121,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,122,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,123,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,124,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,125,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,126,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,127,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,128,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,129,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,130,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,131,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,132,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,133,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,134,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,135,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,136,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,137,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,138,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[13,139,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,140,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,141,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,142,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,143,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,144,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,145,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,146,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,147,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,148,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,149,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,150,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,151,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,152,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,153,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,154,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,155,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,156,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,157,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,158,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[13,159,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,160,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,161,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,162,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,163,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,164,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,165,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,166,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,167,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,168,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,169,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,170,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,171,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,172,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,173,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,174,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,175,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,176,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,177,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,178,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[13,179,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,180,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,181,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,182,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,183,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,184,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,185,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,186,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,187,2,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,188,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,189,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,190,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,191,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,192,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,193,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,194,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,195,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,196,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,197,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,198,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[13,199,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,0,0.14,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,1,0.15,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,2,0.16,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,3,0.17,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,4,0.18,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,5,0.19,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,6,0.2,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,7,0.21,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,8,0.22,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,9,0.23,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,10,0.24,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,11,0.25,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,12,0.26,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,13,0.27,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,14,0.28,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,15,0.29,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,16,0.3,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,17,0.31,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,18,0.32,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[14,19,0.33,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,20,0.34,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,21,0.35,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,22,0.36,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,23,0.37,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,24,0.38,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,25,0.39,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,26,0.4,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,27,0.41,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,28,0.42,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,29,0.43,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,30,0.44,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,31,0.45,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,32,0.46,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,33,0.47,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,34,0.48,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,35,0.49,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,36,0.5,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,37,0.51,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,38,0.52,4]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[14,39,0.53,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,40,0.54,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,41,0.55,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,42,0.56,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,43,0.57,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,44,0.58,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,45,0.59,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,46,0.6,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,47,0.61,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,48,0.62,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,49,0.63,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,50,0.64,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,51,0.65,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,52,0.66,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,53,0.67,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,54,0.68,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,55,0.69,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,56,0.7,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,57,0.71,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,58,0.72,4]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[14,59,0.73,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,60,0.74,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,61,0.75,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,62,0.76,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,63,0.77,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,64,0.78,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,65,0.79,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,66,0.8,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,67,0.81,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,68,0.82,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,69,0.83,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,70,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,71,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,72,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,73,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,74,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,75,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,76,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,77,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,78,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[14,79,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,80,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,81,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,82,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,83,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,84,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,85,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,86,1,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,87,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,88,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,89,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,90,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,91,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,92,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,93,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,94,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,95,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,96,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,97,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,98,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[14,99,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,100,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,101,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,102,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,103,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,104,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,105,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,106,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,107,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,108,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,109,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,110,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,111,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,112,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,113,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,114,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,115,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,116,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,117,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,118,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[14,119,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,120,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,121,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,122,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,123,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,124,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,125,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,126,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,127,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,128,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,129,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,130,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,131,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,132,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,133,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,134,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,135,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,136,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,137,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,138,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[14,139,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,140,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,141,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,142,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,143,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,144,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,145,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,146,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,147,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,148,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,149,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,150,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,151,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,152,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,153,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,154,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,155,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,156,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,157,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,158,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[14,159,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,160,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,161,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,162,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,163,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,164,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,165,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,166,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,167,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,168,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,169,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,170,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,171,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,172,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,173,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,174,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,175,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,176,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,177,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,178,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[14,179,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,180,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,181,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,182,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,183,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,184,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,185,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,186,2,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,187,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,188,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,189,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,190,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,191,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,192,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,193,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,194,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,195,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,196,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,197,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,198,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[14,199,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,0,0.15,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,1,0.16,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,2,0.17,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,3,0.18,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,4,0.19,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,5,0.2,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,6,0.21,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,7,0.22,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,8,0.23,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,9,0.24,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,10,0.25,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,11,0.26,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,12,0.27,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,13,0.28,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,14,0.29,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,15,0.3,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,16,0.31,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,17,0.32,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,18,0.33,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[15,19,0.34,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,20,0.35,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,21,0.36,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,22,0.37,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,23,0.38,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,24,0.39,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,25,0.4,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,26,0.41,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,27,0.42,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,28,0.43,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,29,0.44,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,30,0.45,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,31,0.46,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,32,0.47,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,33,0.48,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,34,0.49,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,35,0.5,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,36,0.51,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,37,0.52,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,38,0.53,5]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[15,39,0.54,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,40,0.55,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,41,0.56,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,42,0.57,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,43,0.58,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,44,0.59,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,45,0.6,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,46,0.61,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,47,0.62,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,48,0.63,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,49,0.64,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,50,0.65,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,51,0.66,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,52,0.67,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,53,0.68,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,54,0.69,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,55,0.7,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,56,0.71,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,57,0.72,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,58,0.73,5]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[15,59,0.74,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,60,0.75,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,61,0.76,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,62,0.77,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,63,0.78,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,64,0.79,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,65,0.8,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,66,0.81,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,67,0.82,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,68,0.83,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,69,0.84,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,70,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,71,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,72,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,73,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,74,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,75,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,76,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,77,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,78,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[15,79,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,80,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,81,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,82,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,83,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,84,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,85,1,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,86,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,87,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,88,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,89,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,90,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,91,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,92,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,93,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,94,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,95,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,96,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,97,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,98,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[15,99,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,100,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,101,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,102,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,103,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,104,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,105,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,106,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,107,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,108,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,109,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,110,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,111,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,112,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,113,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,114,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,115,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,116,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,117,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,118,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[15,119,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,120,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,121,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,122,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,123,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,124,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,125,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,126,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,127,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,128,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,129,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,130,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,131,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,132,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,133,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,134,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,135,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,136,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,137,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,138,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[15,139,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,140,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,141,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,142,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,143,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,144,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,145,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,146,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,147,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,148,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,149,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,150,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,151,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,152,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,153,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,154,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,155,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,156,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,157,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,158,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[15,159,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,160,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,161,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,162,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,163,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,164,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,165,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,166,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,167,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,168,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,169,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,170,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,171,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,172,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,173,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,174,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,175,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,176,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,177,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,178,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[15,179,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,180,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,181,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,182,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,183,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,184,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,185,2,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,186,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,187,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,188,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,189,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,190,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,191,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,192,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,193,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,194,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,195,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,196,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,197,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,198,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[15,199,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,0,0.16,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,1,0.17,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,2,0.18,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,3,0.19,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,4,0.2,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,5,0.21,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,6,0.22,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,7,0.23,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,8,0.24,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,9,0.25,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,10,0.26,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,11,0.27,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,12,0.28,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,13,0.29,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,14,0.3,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,15,0.31,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,16,0.32,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,17,0.33,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,18,0.34,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[16,19,0.35,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,20,0.36,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,21,0.37,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,22,0.38,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,23,0.39,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,24,0.4,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,25,0.41,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,26,0.42,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,27,0.43,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,28,0.44,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,29,0.45,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,30,0.46,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,31,0.47,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,32,0.48,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,33,0.49,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,34,0.5,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,35,0.51,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,36,0.52,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,37,0.53,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,38,0.54,6]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[16,39,0.55,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,40,0.56,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,41,0.57,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,42,0.58,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,43,0.59,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,44,0.6,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,45,0.61,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,46,0.62,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,47,0.63,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,48,0.64,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,49,0.65,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,50,0.66,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,51,0.67,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,52,0.68,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,53,0.69,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,54,0.7,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,55,0.71,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,56,0.72,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,57,0.73,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,58,0.74,6]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[16,59,0.75,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,60,0.76,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,61,0.77,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,62,0.78,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,63,0.79,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,64,0.8,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,65,0.81,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,66,0.82,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,67,0.83,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,68,0.84,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,69,0.85,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,70,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,71,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,72,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,73,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,74,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,75,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,76,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,77,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,78,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[16,79,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,80,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,81,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,82,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,83,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,84,1,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,85,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,86,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,87,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,88,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,89,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,90,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,91,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,92,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,93,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,94,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,95,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,96,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,97,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,98,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[16,99,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,100,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,101,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,102,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,103,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,104,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,105,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,106,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,107,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,108,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,109,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,110,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,111,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,112,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,113,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,114,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,115,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,116,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,117,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,118,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[16,119,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,120,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,121,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,122,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,123,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,124,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,125,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,126,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,127,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,128,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,129,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,130,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,131,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,132,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,133,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,134,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,135,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,136,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,137,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,138,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[16,139,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,140,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,141,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,142,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,143,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,144,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,145,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,146,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,147,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,148,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,149,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,150,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,151,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,152,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,153,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,154,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,155,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,156,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,157,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,158,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[16,159,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,160,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,161,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,162,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,163,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,164,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,165,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,166,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,167,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,168,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,169,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,170,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,171,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,172,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,173,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,174,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,175,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,176,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,177,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,178,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[16,179,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,180,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,181,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,182,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,183,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,184,2,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,185,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,186,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,187,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,188,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,189,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,190,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,191,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,192,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,193,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,194,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,195,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,196,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,197,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,198,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[16,199,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,0,0.17,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,1,0.18,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,2,0.19,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,3,0.2,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,4,0.21,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,5,0.22,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,6,0.23,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,7,0.24,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,8,0.25,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,9,0.26,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,10,0.27,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,11,0.28,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,12,0.29,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,13,0.3,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,14,0.31,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,15,0.32,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,16,0.33,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,17,0.34,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,18,0.35,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[17,19,0.36,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,20,0.37,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,21,0.38,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,22,0.39,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,23,0.4,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,24,0.41,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,25,0.42,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,26,0.43,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,27,0.44,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,28,0.45,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,29,0.46,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,30,0.47,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,31,0.48,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,32,0.49,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,33,0.5,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,34,0.51,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,35,0.52,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,36,0.53,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,37,0.54,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,38,0.55,7]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[17,39,0.56,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,40,0.57,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,41,0.58,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,42,0.59,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,43,0.6,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,44,0.61,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,45,0.62,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,46,0.63,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,47,0.64,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,48,0.65,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,49,0.66,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,50,0.67,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,51,0.68,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,52,0.69,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,53,0.7,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,54,0.71,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,55,0.72,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,56,0.73,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,57,0.74,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,58,0.75,7]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[17,59,0.76,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,60,0.77,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,61,0.78,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,62,0.79,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,63,0.8,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,64,0.81,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,65,0.82,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,66,0.83,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,67,0.84,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,68,0.85,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,69,0.86,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,70,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,71,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,72,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,73,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,74,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,75,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,76,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,77,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,78,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[17,79,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,80,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,81,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,82,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,83,1,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,84,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,85,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,86,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,87,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,88,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,89,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,90,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,91,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,92,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,93,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,94,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,95,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,96,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,97,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,98,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[17,99,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,100,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,101,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,102,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,103,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,104,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,105,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,106,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,107,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,108,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,109,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,110,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,111,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,112,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,113,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,114,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,115,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,116,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,117,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,118,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[17,119,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,120,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,121,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,122,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,123,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,124,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,125,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,126,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,127,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,128,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,129,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,130,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,131,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,132,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,133,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,134,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,135,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,136,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,137,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,138,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[17,139,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,140,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,141,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,142,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,143,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,144,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,145,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,146,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,147,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,148,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,149,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,150,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,151,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,152,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,153,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,154,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,155,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,156,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,157,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,158,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[17,159,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,160,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,161,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,162,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,163,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,164,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,165,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,166,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,167,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,168,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,169,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,170,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,171,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,172,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,173,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,174,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,175,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,176,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,177,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,178,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[17,179,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,180,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,181,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,182,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,183,2,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,184,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,185,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,186,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,187,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,188,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,189,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,190,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,191,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,192,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,193,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,194,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,195,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,196,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,197,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,198,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[17,199,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,0,0.18,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,1,0.19,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,2,0.2,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,3,0.21,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,4,0.22,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,5,0.23,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,6,0.24,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,7,0.25,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,8,0.26,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,9,0.27,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,10,0.28,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,11,0.29,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,12,0.3,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,13,0.31,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,14,0.32,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,15,0.33,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,16,0.34,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,17,0.35,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,18,0.36,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[18,19,0.37,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,20,0.38,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,21,0.39,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,22,0.4,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,23,0.41,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,24,0.42,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,25,0.43,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,26,0.44,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,27,0.45,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,28,0.46,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,29,0.47,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,30,0.48,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,31,0.49,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,32,0.5,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,33,0.51,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,34,0.52,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,35,0.53,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,36,0.54,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,37,0.55,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,38,0.56,8]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[18,39,0.57,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,40,0.58,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,41,0.59,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,42,0.6,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,43,0.61,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,44,0.62,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,45,0.63,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,46,0.64,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,47,0.65,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,48,0.66,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,49,0.67,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,50,0.68,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,51,0.69,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,52,0.7,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,53,0.71,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,54,0.72,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,55,0.73,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,56,0.74,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,57,0.75,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,58,0.76,8]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[18,59,0.77,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,60,0.78,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,61,0.79,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,62,0.8,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,63,0.81,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,64,0.82,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,65,0.83,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,66,0.84,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,67,0.85,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,68,0.86,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,69,0.87,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,70,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,71,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,72,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,73,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,74,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,75,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,76,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,77,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,78,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[18,79,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,80,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,81,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,82,1,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,83,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,84,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,85,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,86,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,87,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,88,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,89,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,90,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,91,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,92,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,93,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,94,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,95,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,96,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,97,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,98,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[18,99,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,100,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,101,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,102,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,103,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,104,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,105,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,106,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,107,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,108,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,109,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,110,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,111,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,112,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,113,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,114,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,115,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,116,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,117,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,118,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[18,119,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,120,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,121,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,122,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,123,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,124,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,125,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,126,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,127,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,128,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,129,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,130,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,131,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,132,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,133,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,134,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,135,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,136,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,137,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,138,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[18,139,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,140,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,141,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,142,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,143,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,144,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,145,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,146,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,147,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,148,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,149,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,150,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,151,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,152,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,153,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,154,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,155,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,156,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,157,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,158,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[18,159,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,160,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,161,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,162,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,163,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,164,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,165,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,166,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,167,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,168,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,169,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,170,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,171,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,172,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,173,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,174,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,175,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,176,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,177,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,178,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[18,179,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,180,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,181,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,182,2,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,183,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,184,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,185,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,186,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,187,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,188,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,189,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,190,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,191,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,192,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,193,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,194,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,195,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,196,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,197,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,198,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[18,199,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,0,0.19,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,1,0.2,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,2,0.21,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,3,0.22,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,4,0.23,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,5,0.24,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,6,0.25,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,7,0.26,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,8,0.27,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,9,0.28,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,10,0.29,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,11,0.3,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,12,0.31,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,13,0.32,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,14,0.33,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,15,0.34,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,16,0.35,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,17,0.36,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,18,0.37,9]) );
insert into pts_collection ( gid, pt ) values ( 0, pc_makepoint(1, ARRAY[19,19,0.38,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,20,0.39,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,21,0.4,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,22,0.41,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,23,0.42,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,24,0.43,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,25,0.44,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,26,0.45,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,27,0.46,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,28,0.47,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,29,0.48,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,30,0.49,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,31,0.5,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,32,0.51,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,33,0.52,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,34,0.53,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,35,0.54,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,36,0.55,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,37,0.56,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,38,0.57,9]) );
insert into pts_collection ( gid, pt ) values ( 1, pc_makepoint(1, ARRAY[19,39,0.58,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,40,0.59,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,41,0.6,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,42,0.61,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,43,0.62,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,44,0.63,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,45,0.64,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,46,0.65,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,47,0.66,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,48,0.67,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,49,0.68,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,50,0.69,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,51,0.7,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,52,0.71,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,53,0.72,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,54,0.73,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,55,0.74,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,56,0.75,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,57,0.76,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,58,0.77,9]) );
insert into pts_collection ( gid, pt ) values ( 2, pc_makepoint(1, ARRAY[19,59,0.78,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,60,0.79,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,61,0.8,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,62,0.81,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,63,0.82,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,64,0.83,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,65,0.84,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,66,0.85,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,67,0.86,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,68,0.87,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,69,0.88,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,70,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,71,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,72,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,73,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,74,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,75,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,76,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,77,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,78,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 3, pc_makepoint(1, ARRAY[19,79,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,80,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,81,1,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,82,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,83,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,84,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,85,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,86,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,87,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,88,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,89,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,90,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,91,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,92,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,93,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,94,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,95,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,96,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,97,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,98,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 4, pc_makepoint(1, ARRAY[19,99,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,100,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,101,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,102,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,103,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,104,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,105,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,106,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,107,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,108,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,109,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,110,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,111,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,112,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,113,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,114,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,115,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,116,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,117,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,118,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 5, pc_makepoint(1, ARRAY[19,119,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,120,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,121,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,122,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,123,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,124,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,125,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,126,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,127,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,128,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,129,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,130,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,131,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,132,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,133,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,134,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,135,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,136,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,137,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,138,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 6, pc_makepoint(1, ARRAY[19,139,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,140,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,141,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,142,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,143,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,144,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,145,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,146,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,147,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,148,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,149,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,150,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,151,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,152,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,153,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,154,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,155,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,156,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,157,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,158,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 7, pc_makepoint(1, ARRAY[19,159,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,160,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,161,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,162,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,163,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,164,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,165,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,166,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,167,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,168,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,169,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,170,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,171,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,172,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,173,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,174,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,175,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,176,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,177,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,178,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 8, pc_makepoint(1, ARRAY[19,179,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,180,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,181,2,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,182,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,183,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,184,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,185,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,186,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,187,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,188,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,189,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,190,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,191,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,192,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,193,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,194,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,195,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,196,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,197,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,198,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 9, pc_makepoint(1, ARRAY[19,199,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,0,0.2,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,1,0.21,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,2,0.22,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,3,0.23,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,4,0.24,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,5,0.25,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,6,0.26,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,7,0.27,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,8,0.28,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,9,0.29,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,10,0.3,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,11,0.31,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,12,0.32,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,13,0.33,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,14,0.34,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,15,0.35,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,16,0.36,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,17,0.37,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,18,0.38,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[20,19,0.39,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,20,0.4,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,21,0.41,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,22,0.42,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,23,0.43,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,24,0.44,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,25,0.45,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,26,0.46,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,27,0.47,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,28,0.48,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,29,0.49,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,30,0.5,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,31,0.51,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,32,0.52,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,33,0.53,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,34,0.54,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,35,0.55,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,36,0.56,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,37,0.57,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,38,0.58,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[20,39,0.59,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,40,0.6,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,41,0.61,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,42,0.62,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,43,0.63,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,44,0.64,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,45,0.65,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,46,0.66,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,47,0.67,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,48,0.68,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,49,0.69,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,50,0.7,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,51,0.71,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,52,0.72,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,53,0.73,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,54,0.74,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,55,0.75,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,56,0.76,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,57,0.77,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,58,0.78,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[20,59,0.79,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,60,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,61,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,62,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,63,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,64,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,65,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,66,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,67,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,68,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,69,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,70,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,71,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,72,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,73,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,74,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,75,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,76,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,77,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,78,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[20,79,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,80,1,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,81,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,82,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,83,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,84,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,85,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,86,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,87,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,88,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,89,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,90,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,91,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,92,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,93,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,94,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,95,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,96,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,97,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,98,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[20,99,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,100,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,101,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,102,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,103,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,104,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,105,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,106,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,107,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,108,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,109,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,110,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,111,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,112,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,113,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,114,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,115,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,116,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,117,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,118,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[20,119,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,120,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,121,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,122,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,123,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,124,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,125,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,126,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,127,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,128,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,129,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,130,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,131,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,132,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,133,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,134,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,135,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,136,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,137,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,138,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[20,139,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,140,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,141,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,142,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,143,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,144,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,145,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,146,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,147,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,148,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,149,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,150,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,151,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,152,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,153,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,154,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,155,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,156,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,157,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,158,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[20,159,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,160,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,161,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,162,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,163,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,164,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,165,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,166,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,167,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,168,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,169,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,170,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,171,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,172,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,173,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,174,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,175,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,176,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,177,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,178,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[20,179,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,180,2,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,181,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,182,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,183,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,184,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,185,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,186,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,187,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,188,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,189,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,190,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,191,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,192,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,193,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,194,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,195,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,196,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,197,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,198,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[20,199,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,0,0.21,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,1,0.22,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,2,0.23,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,3,0.24,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,4,0.25,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,5,0.26,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,6,0.27,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,7,0.28,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,8,0.29,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,9,0.3,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,10,0.31,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,11,0.32,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,12,0.33,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,13,0.34,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,14,0.35,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,15,0.36,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,16,0.37,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,17,0.38,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,18,0.39,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[21,19,0.4,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,20,0.41,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,21,0.42,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,22,0.43,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,23,0.44,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,24,0.45,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,25,0.46,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,26,0.47,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,27,0.48,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,28,0.49,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,29,0.5,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,30,0.51,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,31,0.52,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,32,0.53,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,33,0.54,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,34,0.55,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,35,0.56,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,36,0.57,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,37,0.58,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,38,0.59,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[21,39,0.6,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,40,0.61,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,41,0.62,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,42,0.63,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,43,0.64,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,44,0.65,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,45,0.66,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,46,0.67,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,47,0.68,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,48,0.69,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,49,0.7,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,50,0.71,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,51,0.72,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,52,0.73,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,53,0.74,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,54,0.75,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,55,0.76,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,56,0.77,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,57,0.78,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,58,0.79,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[21,59,0.8,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,60,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,61,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,62,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,63,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,64,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,65,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,66,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,67,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,68,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,69,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,70,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,71,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,72,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,73,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,74,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,75,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,76,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,77,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,78,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[21,79,1,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,80,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,81,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,82,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,83,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,84,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,85,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,86,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,87,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,88,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,89,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,90,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,91,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,92,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,93,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,94,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,95,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,96,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,97,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,98,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[21,99,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,100,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,101,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,102,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,103,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,104,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,105,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,106,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,107,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,108,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,109,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,110,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,111,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,112,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,113,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,114,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,115,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,116,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,117,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,118,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[21,119,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,120,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,121,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,122,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,123,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,124,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,125,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,126,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,127,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,128,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,129,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,130,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,131,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,132,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,133,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,134,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,135,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,136,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,137,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,138,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[21,139,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,140,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,141,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,142,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,143,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,144,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,145,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,146,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,147,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,148,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,149,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,150,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,151,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,152,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,153,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,154,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,155,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,156,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,157,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,158,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[21,159,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,160,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,161,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,162,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,163,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,164,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,165,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,166,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,167,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,168,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,169,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,170,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,171,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,172,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,173,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,174,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,175,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,176,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,177,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,178,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[21,179,2,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,180,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,181,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,182,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,183,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,184,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,185,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,186,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,187,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,188,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,189,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,190,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,191,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,192,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,193,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,194,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,195,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,196,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,197,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,198,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[21,199,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,0,0.22,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,1,0.23,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,2,0.24,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,3,0.25,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,4,0.26,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,5,0.27,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,6,0.28,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,7,0.29,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,8,0.3,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,9,0.31,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,10,0.32,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,11,0.33,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,12,0.34,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,13,0.35,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,14,0.36,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,15,0.37,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,16,0.38,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,17,0.39,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,18,0.4,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[22,19,0.41,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,20,0.42,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,21,0.43,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,22,0.44,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,23,0.45,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,24,0.46,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,25,0.47,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,26,0.48,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,27,0.49,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,28,0.5,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,29,0.51,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,30,0.52,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,31,0.53,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,32,0.54,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,33,0.55,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,34,0.56,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,35,0.57,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,36,0.58,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,37,0.59,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,38,0.6,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[22,39,0.61,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,40,0.62,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,41,0.63,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,42,0.64,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,43,0.65,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,44,0.66,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,45,0.67,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,46,0.68,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,47,0.69,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,48,0.7,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,49,0.71,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,50,0.72,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,51,0.73,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,52,0.74,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,53,0.75,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,54,0.76,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,55,0.77,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,56,0.78,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,57,0.79,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,58,0.8,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[22,59,0.81,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,60,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,61,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,62,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,63,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,64,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,65,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,66,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,67,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,68,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,69,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,70,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,71,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,72,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,73,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,74,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,75,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,76,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,77,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,78,1,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[22,79,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,80,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,81,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,82,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,83,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,84,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,85,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,86,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,87,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,88,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,89,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,90,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,91,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,92,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,93,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,94,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,95,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,96,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,97,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,98,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[22,99,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,100,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,101,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,102,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,103,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,104,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,105,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,106,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,107,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,108,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,109,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,110,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,111,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,112,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,113,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,114,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,115,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,116,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,117,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,118,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[22,119,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,120,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,121,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,122,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,123,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,124,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,125,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,126,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,127,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,128,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,129,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,130,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,131,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,132,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,133,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,134,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,135,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,136,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,137,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,138,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[22,139,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,140,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,141,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,142,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,143,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,144,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,145,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,146,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,147,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,148,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,149,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,150,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,151,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,152,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,153,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,154,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,155,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,156,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,157,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,158,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[22,159,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,160,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,161,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,162,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,163,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,164,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,165,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,166,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,167,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,168,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,169,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,170,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,171,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,172,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,173,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,174,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,175,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,176,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,177,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,178,2,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[22,179,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,180,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,181,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,182,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,183,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,184,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,185,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,186,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,187,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,188,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,189,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,190,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,191,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,192,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,193,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,194,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,195,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,196,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,197,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,198,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[22,199,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,0,0.23,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,1,0.24,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,2,0.25,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,3,0.26,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,4,0.27,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,5,0.28,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,6,0.29,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,7,0.3,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,8,0.31,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,9,0.32,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,10,0.33,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,11,0.34,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,12,0.35,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,13,0.36,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,14,0.37,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,15,0.38,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,16,0.39,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,17,0.4,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,18,0.41,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[23,19,0.42,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,20,0.43,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,21,0.44,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,22,0.45,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,23,0.46,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,24,0.47,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,25,0.48,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,26,0.49,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,27,0.5,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,28,0.51,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,29,0.52,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,30,0.53,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,31,0.54,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,32,0.55,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,33,0.56,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,34,0.57,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,35,0.58,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,36,0.59,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,37,0.6,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,38,0.61,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[23,39,0.62,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,40,0.63,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,41,0.64,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,42,0.65,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,43,0.66,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,44,0.67,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,45,0.68,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,46,0.69,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,47,0.7,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,48,0.71,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,49,0.72,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,50,0.73,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,51,0.74,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,52,0.75,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,53,0.76,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,54,0.77,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,55,0.78,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,56,0.79,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,57,0.8,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,58,0.81,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[23,59,0.82,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,60,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,61,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,62,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,63,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,64,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,65,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,66,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,67,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,68,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,69,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,70,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,71,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,72,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,73,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,74,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,75,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,76,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,77,1,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,78,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[23,79,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,80,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,81,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,82,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,83,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,84,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,85,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,86,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,87,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,88,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,89,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,90,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,91,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,92,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,93,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,94,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,95,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,96,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,97,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,98,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[23,99,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,100,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,101,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,102,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,103,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,104,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,105,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,106,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,107,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,108,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,109,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,110,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,111,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,112,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,113,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,114,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,115,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,116,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,117,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,118,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[23,119,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,120,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,121,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,122,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,123,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,124,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,125,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,126,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,127,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,128,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,129,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,130,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,131,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,132,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,133,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,134,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,135,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,136,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,137,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,138,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[23,139,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,140,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,141,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,142,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,143,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,144,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,145,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,146,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,147,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,148,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,149,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,150,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,151,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,152,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,153,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,154,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,155,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,156,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,157,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,158,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[23,159,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,160,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,161,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,162,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,163,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,164,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,165,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,166,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,167,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,168,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,169,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,170,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,171,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,172,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,173,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,174,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,175,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,176,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,177,2,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,178,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[23,179,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,180,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,181,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,182,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,183,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,184,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,185,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,186,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,187,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,188,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,189,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,190,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,191,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,192,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,193,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,194,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,195,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,196,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,197,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,198,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[23,199,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,0,0.24,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,1,0.25,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,2,0.26,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,3,0.27,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,4,0.28,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,5,0.29,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,6,0.3,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,7,0.31,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,8,0.32,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,9,0.33,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,10,0.34,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,11,0.35,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,12,0.36,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,13,0.37,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,14,0.38,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,15,0.39,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,16,0.4,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,17,0.41,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,18,0.42,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[24,19,0.43,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,20,0.44,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,21,0.45,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,22,0.46,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,23,0.47,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,24,0.48,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,25,0.49,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,26,0.5,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,27,0.51,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,28,0.52,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,29,0.53,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,30,0.54,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,31,0.55,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,32,0.56,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,33,0.57,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,34,0.58,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,35,0.59,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,36,0.6,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,37,0.61,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,38,0.62,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[24,39,0.63,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,40,0.64,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,41,0.65,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,42,0.66,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,43,0.67,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,44,0.68,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,45,0.69,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,46,0.7,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,47,0.71,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,48,0.72,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,49,0.73,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,50,0.74,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,51,0.75,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,52,0.76,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,53,0.77,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,54,0.78,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,55,0.79,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,56,0.8,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,57,0.81,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,58,0.82,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[24,59,0.83,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,60,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,61,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,62,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,63,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,64,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,65,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,66,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,67,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,68,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,69,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,70,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,71,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,72,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,73,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,74,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,75,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,76,1,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,77,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,78,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[24,79,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,80,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,81,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,82,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,83,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,84,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,85,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,86,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,87,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,88,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,89,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,90,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,91,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,92,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,93,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,94,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,95,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,96,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,97,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,98,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[24,99,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,100,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,101,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,102,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,103,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,104,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,105,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,106,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,107,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,108,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,109,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,110,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,111,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,112,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,113,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,114,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,115,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,116,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,117,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,118,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[24,119,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,120,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,121,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,122,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,123,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,124,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,125,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,126,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,127,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,128,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,129,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,130,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,131,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,132,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,133,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,134,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,135,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,136,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,137,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,138,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[24,139,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,140,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,141,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,142,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,143,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,144,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,145,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,146,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,147,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,148,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,149,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,150,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,151,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,152,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,153,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,154,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,155,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,156,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,157,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,158,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[24,159,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,160,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,161,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,162,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,163,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,164,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,165,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,166,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,167,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,168,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,169,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,170,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,171,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,172,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,173,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,174,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,175,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,176,2,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,177,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,178,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[24,179,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,180,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,181,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,182,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,183,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,184,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,185,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,186,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,187,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,188,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,189,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,190,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,191,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,192,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,193,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,194,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,195,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,196,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,197,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,198,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[24,199,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,0,0.25,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,1,0.26,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,2,0.27,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,3,0.28,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,4,0.29,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,5,0.3,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,6,0.31,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,7,0.32,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,8,0.33,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,9,0.34,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,10,0.35,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,11,0.36,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,12,0.37,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,13,0.38,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,14,0.39,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,15,0.4,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,16,0.41,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,17,0.42,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,18,0.43,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[25,19,0.44,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,20,0.45,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,21,0.46,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,22,0.47,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,23,0.48,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,24,0.49,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,25,0.5,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,26,0.51,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,27,0.52,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,28,0.53,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,29,0.54,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,30,0.55,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,31,0.56,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,32,0.57,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,33,0.58,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,34,0.59,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,35,0.6,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,36,0.61,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,37,0.62,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,38,0.63,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[25,39,0.64,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,40,0.65,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,41,0.66,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,42,0.67,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,43,0.68,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,44,0.69,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,45,0.7,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,46,0.71,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,47,0.72,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,48,0.73,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,49,0.74,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,50,0.75,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,51,0.76,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,52,0.77,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,53,0.78,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,54,0.79,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,55,0.8,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,56,0.81,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,57,0.82,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,58,0.83,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[25,59,0.84,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,60,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,61,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,62,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,63,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,64,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,65,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,66,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,67,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,68,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,69,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,70,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,71,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,72,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,73,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,74,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,75,1,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,76,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,77,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,78,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[25,79,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,80,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,81,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,82,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,83,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,84,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,85,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,86,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,87,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,88,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,89,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,90,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,91,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,92,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,93,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,94,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,95,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,96,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,97,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,98,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[25,99,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,100,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,101,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,102,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,103,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,104,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,105,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,106,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,107,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,108,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,109,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,110,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,111,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,112,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,113,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,114,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,115,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,116,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,117,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,118,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[25,119,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,120,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,121,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,122,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,123,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,124,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,125,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,126,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,127,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,128,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,129,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,130,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,131,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,132,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,133,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,134,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,135,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,136,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,137,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,138,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[25,139,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,140,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,141,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,142,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,143,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,144,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,145,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,146,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,147,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,148,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,149,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,150,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,151,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,152,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,153,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,154,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,155,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,156,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,157,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,158,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[25,159,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,160,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,161,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,162,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,163,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,164,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,165,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,166,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,167,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,168,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,169,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,170,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,171,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,172,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,173,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,174,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,175,2,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,176,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,177,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,178,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[25,179,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,180,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,181,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,182,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,183,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,184,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,185,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,186,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,187,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,188,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,189,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,190,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,191,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,192,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,193,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,194,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,195,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,196,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,197,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,198,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[25,199,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,0,0.26,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,1,0.27,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,2,0.28,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,3,0.29,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,4,0.3,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,5,0.31,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,6,0.32,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,7,0.33,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,8,0.34,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,9,0.35,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,10,0.36,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,11,0.37,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,12,0.38,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,13,0.39,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,14,0.4,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,15,0.41,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,16,0.42,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,17,0.43,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,18,0.44,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[26,19,0.45,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,20,0.46,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,21,0.47,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,22,0.48,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,23,0.49,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,24,0.5,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,25,0.51,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,26,0.52,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,27,0.53,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,28,0.54,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,29,0.55,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,30,0.56,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,31,0.57,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,32,0.58,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,33,0.59,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,34,0.6,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,35,0.61,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,36,0.62,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,37,0.63,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,38,0.64,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[26,39,0.65,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,40,0.66,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,41,0.67,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,42,0.68,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,43,0.69,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,44,0.7,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,45,0.71,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,46,0.72,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,47,0.73,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,48,0.74,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,49,0.75,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,50,0.76,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,51,0.77,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,52,0.78,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,53,0.79,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,54,0.8,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,55,0.81,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,56,0.82,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,57,0.83,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,58,0.84,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[26,59,0.85,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,60,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,61,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,62,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,63,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,64,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,65,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,66,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,67,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,68,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,69,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,70,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,71,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,72,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,73,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,74,1,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,75,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,76,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,77,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,78,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[26,79,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,80,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,81,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,82,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,83,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,84,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,85,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,86,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,87,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,88,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,89,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,90,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,91,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,92,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,93,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,94,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,95,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,96,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,97,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,98,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[26,99,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,100,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,101,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,102,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,103,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,104,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,105,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,106,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,107,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,108,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,109,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,110,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,111,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,112,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,113,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,114,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,115,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,116,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,117,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,118,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[26,119,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,120,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,121,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,122,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,123,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,124,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,125,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,126,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,127,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,128,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,129,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,130,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,131,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,132,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,133,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,134,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,135,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,136,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,137,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,138,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[26,139,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,140,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,141,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,142,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,143,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,144,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,145,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,146,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,147,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,148,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,149,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,150,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,151,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,152,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,153,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,154,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,155,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,156,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,157,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,158,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[26,159,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,160,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,161,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,162,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,163,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,164,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,165,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,166,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,167,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,168,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,169,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,170,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,171,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,172,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,173,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,174,2,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,175,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,176,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,177,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,178,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[26,179,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,180,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,181,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,182,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,183,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,184,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,185,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,186,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,187,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,188,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,189,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,190,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,191,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,192,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,193,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,194,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,195,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,196,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,197,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,198,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[26,199,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,0,0.27,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,1,0.28,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,2,0.29,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,3,0.3,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,4,0.31,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,5,0.32,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,6,0.33,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,7,0.34,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,8,0.35,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,9,0.36,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,10,0.37,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,11,0.38,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,12,0.39,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,13,0.4,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,14,0.41,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,15,0.42,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,16,0.43,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,17,0.44,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,18,0.45,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[27,19,0.46,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,20,0.47,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,21,0.48,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,22,0.49,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,23,0.5,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,24,0.51,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,25,0.52,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,26,0.53,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,27,0.54,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,28,0.55,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,29,0.56,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,30,0.57,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,31,0.58,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,32,0.59,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,33,0.6,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,34,0.61,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,35,0.62,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,36,0.63,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,37,0.64,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,38,0.65,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[27,39,0.66,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,40,0.67,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,41,0.68,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,42,0.69,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,43,0.7,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,44,0.71,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,45,0.72,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,46,0.73,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,47,0.74,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,48,0.75,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,49,0.76,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,50,0.77,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,51,0.78,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,52,0.79,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,53,0.8,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,54,0.81,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,55,0.82,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,56,0.83,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,57,0.84,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,58,0.85,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[27,59,0.86,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,60,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,61,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,62,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,63,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,64,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,65,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,66,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,67,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,68,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,69,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,70,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,71,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,72,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,73,1,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,74,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,75,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,76,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,77,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,78,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[27,79,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,80,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,81,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,82,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,83,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,84,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,85,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,86,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,87,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,88,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,89,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,90,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,91,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,92,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,93,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,94,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,95,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,96,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,97,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,98,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[27,99,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,100,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,101,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,102,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,103,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,104,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,105,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,106,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,107,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,108,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,109,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,110,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,111,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,112,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,113,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,114,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,115,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,116,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,117,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,118,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[27,119,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,120,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,121,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,122,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,123,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,124,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,125,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,126,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,127,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,128,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,129,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,130,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,131,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,132,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,133,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,134,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,135,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,136,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,137,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,138,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[27,139,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,140,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,141,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,142,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,143,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,144,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,145,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,146,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,147,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,148,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,149,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,150,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,151,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,152,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,153,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,154,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,155,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,156,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,157,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,158,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[27,159,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,160,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,161,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,162,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,163,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,164,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,165,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,166,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,167,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,168,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,169,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,170,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,171,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,172,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,173,2,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,174,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,175,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,176,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,177,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,178,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[27,179,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,180,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,181,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,182,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,183,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,184,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,185,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,186,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,187,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,188,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,189,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,190,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,191,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,192,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,193,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,194,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,195,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,196,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,197,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,198,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[27,199,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,0,0.28,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,1,0.29,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,2,0.3,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,3,0.31,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,4,0.32,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,5,0.33,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,6,0.34,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,7,0.35,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,8,0.36,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,9,0.37,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,10,0.38,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,11,0.39,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,12,0.4,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,13,0.41,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,14,0.42,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,15,0.43,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,16,0.44,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,17,0.45,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,18,0.46,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[28,19,0.47,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,20,0.48,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,21,0.49,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,22,0.5,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,23,0.51,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,24,0.52,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,25,0.53,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,26,0.54,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,27,0.55,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,28,0.56,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,29,0.57,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,30,0.58,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,31,0.59,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,32,0.6,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,33,0.61,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,34,0.62,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,35,0.63,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,36,0.64,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,37,0.65,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,38,0.66,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[28,39,0.67,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,40,0.68,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,41,0.69,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,42,0.7,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,43,0.71,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,44,0.72,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,45,0.73,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,46,0.74,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,47,0.75,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,48,0.76,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,49,0.77,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,50,0.78,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,51,0.79,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,52,0.8,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,53,0.81,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,54,0.82,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,55,0.83,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,56,0.84,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,57,0.85,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,58,0.86,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[28,59,0.87,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,60,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,61,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,62,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,63,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,64,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,65,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,66,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,67,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,68,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,69,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,70,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,71,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,72,1,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,73,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,74,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,75,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,76,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,77,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,78,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[28,79,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,80,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,81,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,82,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,83,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,84,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,85,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,86,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,87,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,88,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,89,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,90,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,91,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,92,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,93,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,94,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,95,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,96,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,97,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,98,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[28,99,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,100,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,101,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,102,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,103,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,104,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,105,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,106,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,107,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,108,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,109,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,110,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,111,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,112,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,113,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,114,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,115,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,116,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,117,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,118,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[28,119,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,120,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,121,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,122,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,123,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,124,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,125,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,126,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,127,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,128,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,129,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,130,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,131,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,132,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,133,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,134,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,135,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,136,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,137,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,138,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[28,139,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,140,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,141,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,142,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,143,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,144,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,145,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,146,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,147,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,148,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,149,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,150,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,151,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,152,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,153,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,154,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,155,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,156,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,157,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,158,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[28,159,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,160,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,161,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,162,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,163,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,164,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,165,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,166,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,167,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,168,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,169,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,170,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,171,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,172,2,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,173,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,174,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,175,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,176,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,177,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,178,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[28,179,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,180,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,181,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,182,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,183,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,184,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,185,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,186,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,187,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,188,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,189,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,190,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,191,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,192,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,193,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,194,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,195,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,196,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,197,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,198,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[28,199,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,0,0.29,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,1,0.3,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,2,0.31,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,3,0.32,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,4,0.33,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,5,0.34,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,6,0.35,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,7,0.36,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,8,0.37,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,9,0.38,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,10,0.39,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,11,0.4,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,12,0.41,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,13,0.42,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,14,0.43,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,15,0.44,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,16,0.45,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,17,0.46,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,18,0.47,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[29,19,0.48,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,20,0.49,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,21,0.5,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,22,0.51,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,23,0.52,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,24,0.53,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,25,0.54,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,26,0.55,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,27,0.56,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,28,0.57,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,29,0.58,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,30,0.59,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,31,0.6,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,32,0.61,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,33,0.62,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,34,0.63,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,35,0.64,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,36,0.65,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,37,0.66,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,38,0.67,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[29,39,0.68,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,40,0.69,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,41,0.7,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,42,0.71,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,43,0.72,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,44,0.73,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,45,0.74,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,46,0.75,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,47,0.76,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,48,0.77,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,49,0.78,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,50,0.79,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,51,0.8,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,52,0.81,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,53,0.82,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,54,0.83,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,55,0.84,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,56,0.85,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,57,0.86,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,58,0.87,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[29,59,0.88,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,60,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,61,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,62,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,63,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,64,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,65,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,66,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,67,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,68,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,69,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,70,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,71,1,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,72,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,73,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,74,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,75,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,76,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,77,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,78,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[29,79,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,80,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,81,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,82,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,83,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,84,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,85,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,86,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,87,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,88,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,89,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,90,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,91,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,92,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,93,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,94,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,95,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,96,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,97,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,98,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[29,99,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,100,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,101,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,102,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,103,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,104,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,105,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,106,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,107,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,108,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,109,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,110,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,111,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,112,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,113,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,114,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,115,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,116,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,117,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,118,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[29,119,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,120,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,121,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,122,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,123,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,124,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,125,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,126,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,127,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,128,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,129,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,130,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,131,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,132,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,133,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,134,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,135,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,136,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,137,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,138,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[29,139,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,140,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,141,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,142,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,143,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,144,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,145,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,146,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,147,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,148,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,149,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,150,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,151,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,152,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,153,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,154,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,155,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,156,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,157,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,158,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[29,159,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,160,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,161,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,162,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,163,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,164,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,165,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,166,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,167,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,168,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,169,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,170,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,171,2,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,172,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,173,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,174,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,175,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,176,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,177,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,178,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[29,179,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,180,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,181,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,182,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,183,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,184,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,185,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,186,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,187,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,188,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,189,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,190,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,191,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,192,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,193,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,194,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,195,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,196,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,197,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,198,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[29,199,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,0,0.3,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,1,0.31,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,2,0.32,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,3,0.33,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,4,0.34,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,5,0.35,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,6,0.36,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,7,0.37,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,8,0.38,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,9,0.39,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,10,0.4,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,11,0.41,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,12,0.42,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,13,0.43,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,14,0.44,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,15,0.45,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,16,0.46,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,17,0.47,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,18,0.48,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[30,19,0.49,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,20,0.5,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,21,0.51,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,22,0.52,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,23,0.53,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,24,0.54,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,25,0.55,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,26,0.56,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,27,0.57,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,28,0.58,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,29,0.59,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,30,0.6,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,31,0.61,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,32,0.62,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,33,0.63,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,34,0.64,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,35,0.65,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,36,0.66,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,37,0.67,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,38,0.68,0]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[30,39,0.69,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,40,0.7,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,41,0.71,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,42,0.72,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,43,0.73,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,44,0.74,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,45,0.75,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,46,0.76,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,47,0.77,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,48,0.78,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,49,0.79,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,50,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,51,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,52,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,53,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,54,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,55,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,56,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,57,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,58,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[30,59,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,60,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,61,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,62,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,63,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,64,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,65,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,66,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,67,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,68,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,69,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,70,1,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,71,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,72,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,73,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,74,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,75,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,76,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,77,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,78,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[30,79,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,80,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,81,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,82,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,83,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,84,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,85,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,86,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,87,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,88,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,89,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,90,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,91,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,92,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,93,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,94,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,95,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,96,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,97,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,98,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[30,99,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,100,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,101,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,102,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,103,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,104,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,105,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,106,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,107,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,108,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,109,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,110,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,111,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,112,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,113,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,114,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,115,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,116,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,117,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,118,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[30,119,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,120,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,121,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,122,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,123,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,124,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,125,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,126,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,127,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,128,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,129,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,130,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,131,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,132,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,133,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,134,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,135,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,136,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,137,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,138,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[30,139,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,140,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,141,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,142,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,143,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,144,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,145,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,146,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,147,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,148,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,149,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,150,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,151,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,152,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,153,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,154,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,155,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,156,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,157,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,158,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[30,159,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,160,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,161,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,162,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,163,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,164,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,165,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,166,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,167,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,168,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,169,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,170,2,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,171,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,172,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,173,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,174,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,175,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,176,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,177,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,178,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[30,179,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,180,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,181,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,182,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,183,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,184,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,185,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,186,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,187,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,188,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,189,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,190,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,191,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,192,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,193,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,194,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,195,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,196,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,197,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,198,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[30,199,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,0,0.31,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,1,0.32,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,2,0.33,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,3,0.34,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,4,0.35,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,5,0.36,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,6,0.37,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,7,0.38,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,8,0.39,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,9,0.4,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,10,0.41,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,11,0.42,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,12,0.43,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,13,0.44,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,14,0.45,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,15,0.46,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,16,0.47,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,17,0.48,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,18,0.49,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[31,19,0.5,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,20,0.51,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,21,0.52,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,22,0.53,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,23,0.54,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,24,0.55,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,25,0.56,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,26,0.57,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,27,0.58,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,28,0.59,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,29,0.6,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,30,0.61,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,31,0.62,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,32,0.63,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,33,0.64,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,34,0.65,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,35,0.66,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,36,0.67,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,37,0.68,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,38,0.69,1]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[31,39,0.7,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,40,0.71,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,41,0.72,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,42,0.73,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,43,0.74,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,44,0.75,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,45,0.76,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,46,0.77,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,47,0.78,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,48,0.79,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,49,0.8,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,50,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,51,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,52,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,53,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,54,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,55,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,56,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,57,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,58,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[31,59,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,60,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,61,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,62,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,63,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,64,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,65,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,66,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,67,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,68,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,69,1,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,70,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,71,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,72,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,73,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,74,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,75,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,76,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,77,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,78,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[31,79,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,80,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,81,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,82,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,83,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,84,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,85,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,86,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,87,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,88,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,89,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,90,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,91,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,92,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,93,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,94,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,95,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,96,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,97,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,98,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[31,99,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,100,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,101,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,102,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,103,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,104,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,105,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,106,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,107,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,108,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,109,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,110,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,111,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,112,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,113,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,114,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,115,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,116,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,117,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,118,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[31,119,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,120,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,121,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,122,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,123,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,124,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,125,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,126,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,127,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,128,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,129,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,130,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,131,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,132,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,133,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,134,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,135,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,136,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,137,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,138,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[31,139,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,140,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,141,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,142,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,143,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,144,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,145,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,146,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,147,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,148,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,149,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,150,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,151,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,152,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,153,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,154,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,155,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,156,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,157,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,158,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[31,159,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,160,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,161,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,162,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,163,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,164,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,165,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,166,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,167,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,168,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,169,2,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,170,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,171,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,172,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,173,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,174,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,175,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,176,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,177,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,178,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[31,179,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,180,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,181,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,182,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,183,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,184,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,185,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,186,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,187,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,188,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,189,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,190,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,191,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,192,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,193,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,194,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,195,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,196,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,197,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,198,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[31,199,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,0,0.32,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,1,0.33,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,2,0.34,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,3,0.35,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,4,0.36,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,5,0.37,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,6,0.38,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,7,0.39,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,8,0.4,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,9,0.41,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,10,0.42,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,11,0.43,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,12,0.44,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,13,0.45,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,14,0.46,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,15,0.47,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,16,0.48,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,17,0.49,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,18,0.5,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[32,19,0.51,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,20,0.52,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,21,0.53,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,22,0.54,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,23,0.55,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,24,0.56,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,25,0.57,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,26,0.58,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,27,0.59,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,28,0.6,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,29,0.61,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,30,0.62,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,31,0.63,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,32,0.64,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,33,0.65,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,34,0.66,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,35,0.67,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,36,0.68,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,37,0.69,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,38,0.7,2]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[32,39,0.71,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,40,0.72,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,41,0.73,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,42,0.74,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,43,0.75,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,44,0.76,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,45,0.77,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,46,0.78,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,47,0.79,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,48,0.8,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,49,0.81,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,50,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,51,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,52,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,53,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,54,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,55,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,56,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,57,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,58,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[32,59,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,60,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,61,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,62,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,63,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,64,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,65,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,66,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,67,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,68,1,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,69,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,70,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,71,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,72,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,73,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,74,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,75,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,76,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,77,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,78,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[32,79,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,80,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,81,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,82,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,83,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,84,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,85,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,86,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,87,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,88,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,89,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,90,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,91,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,92,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,93,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,94,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,95,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,96,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,97,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,98,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[32,99,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,100,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,101,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,102,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,103,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,104,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,105,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,106,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,107,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,108,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,109,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,110,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,111,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,112,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,113,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,114,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,115,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,116,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,117,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,118,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[32,119,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,120,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,121,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,122,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,123,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,124,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,125,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,126,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,127,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,128,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,129,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,130,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,131,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,132,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,133,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,134,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,135,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,136,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,137,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,138,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[32,139,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,140,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,141,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,142,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,143,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,144,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,145,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,146,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,147,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,148,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,149,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,150,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,151,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,152,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,153,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,154,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,155,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,156,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,157,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,158,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[32,159,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,160,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,161,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,162,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,163,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,164,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,165,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,166,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,167,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,168,2,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,169,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,170,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,171,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,172,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,173,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,174,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,175,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,176,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,177,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,178,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[32,179,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,180,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,181,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,182,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,183,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,184,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,185,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,186,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,187,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,188,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,189,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,190,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,191,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,192,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,193,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,194,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,195,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,196,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,197,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,198,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[32,199,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,0,0.33,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,1,0.34,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,2,0.35,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,3,0.36,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,4,0.37,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,5,0.38,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,6,0.39,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,7,0.4,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,8,0.41,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,9,0.42,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,10,0.43,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,11,0.44,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,12,0.45,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,13,0.46,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,14,0.47,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,15,0.48,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,16,0.49,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,17,0.5,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,18,0.51,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[33,19,0.52,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,20,0.53,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,21,0.54,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,22,0.55,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,23,0.56,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,24,0.57,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,25,0.58,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,26,0.59,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,27,0.6,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,28,0.61,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,29,0.62,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,30,0.63,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,31,0.64,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,32,0.65,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,33,0.66,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,34,0.67,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,35,0.68,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,36,0.69,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,37,0.7,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,38,0.71,3]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[33,39,0.72,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,40,0.73,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,41,0.74,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,42,0.75,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,43,0.76,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,44,0.77,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,45,0.78,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,46,0.79,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,47,0.8,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,48,0.81,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,49,0.82,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,50,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,51,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,52,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,53,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,54,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,55,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,56,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,57,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,58,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[33,59,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,60,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,61,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,62,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,63,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,64,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,65,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,66,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,67,1,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,68,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,69,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,70,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,71,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,72,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,73,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,74,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,75,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,76,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,77,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,78,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[33,79,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,80,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,81,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,82,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,83,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,84,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,85,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,86,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,87,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,88,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,89,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,90,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,91,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,92,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,93,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,94,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,95,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,96,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,97,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,98,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[33,99,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,100,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,101,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,102,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,103,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,104,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,105,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,106,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,107,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,108,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,109,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,110,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,111,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,112,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,113,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,114,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,115,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,116,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,117,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,118,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[33,119,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,120,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,121,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,122,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,123,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,124,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,125,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,126,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,127,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,128,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,129,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,130,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,131,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,132,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,133,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,134,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,135,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,136,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,137,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,138,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[33,139,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,140,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,141,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,142,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,143,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,144,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,145,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,146,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,147,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,148,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,149,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,150,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,151,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,152,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,153,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,154,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,155,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,156,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,157,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,158,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[33,159,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,160,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,161,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,162,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,163,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,164,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,165,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,166,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,167,2,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,168,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,169,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,170,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,171,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,172,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,173,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,174,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,175,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,176,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,177,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,178,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[33,179,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,180,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,181,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,182,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,183,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,184,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,185,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,186,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,187,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,188,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,189,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,190,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,191,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,192,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,193,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,194,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,195,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,196,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,197,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,198,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[33,199,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,0,0.34,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,1,0.35,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,2,0.36,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,3,0.37,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,4,0.38,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,5,0.39,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,6,0.4,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,7,0.41,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,8,0.42,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,9,0.43,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,10,0.44,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,11,0.45,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,12,0.46,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,13,0.47,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,14,0.48,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,15,0.49,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,16,0.5,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,17,0.51,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,18,0.52,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[34,19,0.53,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,20,0.54,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,21,0.55,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,22,0.56,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,23,0.57,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,24,0.58,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,25,0.59,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,26,0.6,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,27,0.61,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,28,0.62,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,29,0.63,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,30,0.64,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,31,0.65,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,32,0.66,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,33,0.67,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,34,0.68,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,35,0.69,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,36,0.7,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,37,0.71,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,38,0.72,4]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[34,39,0.73,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,40,0.74,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,41,0.75,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,42,0.76,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,43,0.77,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,44,0.78,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,45,0.79,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,46,0.8,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,47,0.81,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,48,0.82,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,49,0.83,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,50,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,51,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,52,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,53,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,54,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,55,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,56,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,57,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,58,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[34,59,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,60,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,61,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,62,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,63,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,64,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,65,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,66,1,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,67,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,68,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,69,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,70,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,71,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,72,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,73,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,74,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,75,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,76,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,77,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,78,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[34,79,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,80,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,81,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,82,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,83,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,84,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,85,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,86,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,87,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,88,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,89,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,90,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,91,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,92,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,93,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,94,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,95,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,96,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,97,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,98,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[34,99,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,100,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,101,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,102,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,103,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,104,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,105,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,106,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,107,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,108,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,109,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,110,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,111,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,112,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,113,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,114,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,115,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,116,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,117,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,118,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[34,119,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,120,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,121,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,122,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,123,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,124,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,125,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,126,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,127,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,128,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,129,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,130,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,131,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,132,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,133,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,134,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,135,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,136,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,137,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,138,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[34,139,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,140,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,141,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,142,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,143,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,144,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,145,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,146,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,147,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,148,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,149,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,150,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,151,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,152,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,153,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,154,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,155,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,156,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,157,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,158,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[34,159,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,160,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,161,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,162,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,163,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,164,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,165,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,166,2,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,167,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,168,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,169,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,170,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,171,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,172,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,173,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,174,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,175,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,176,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,177,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,178,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[34,179,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,180,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,181,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,182,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,183,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,184,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,185,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,186,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,187,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,188,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,189,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,190,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,191,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,192,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,193,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,194,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,195,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,196,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,197,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,198,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[34,199,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,0,0.35,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,1,0.36,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,2,0.37,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,3,0.38,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,4,0.39,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,5,0.4,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,6,0.41,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,7,0.42,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,8,0.43,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,9,0.44,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,10,0.45,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,11,0.46,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,12,0.47,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,13,0.48,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,14,0.49,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,15,0.5,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,16,0.51,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,17,0.52,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,18,0.53,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[35,19,0.54,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,20,0.55,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,21,0.56,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,22,0.57,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,23,0.58,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,24,0.59,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,25,0.6,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,26,0.61,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,27,0.62,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,28,0.63,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,29,0.64,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,30,0.65,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,31,0.66,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,32,0.67,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,33,0.68,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,34,0.69,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,35,0.7,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,36,0.71,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,37,0.72,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,38,0.73,5]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[35,39,0.74,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,40,0.75,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,41,0.76,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,42,0.77,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,43,0.78,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,44,0.79,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,45,0.8,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,46,0.81,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,47,0.82,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,48,0.83,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,49,0.84,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,50,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,51,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,52,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,53,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,54,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,55,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,56,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,57,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,58,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[35,59,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,60,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,61,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,62,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,63,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,64,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,65,1,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,66,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,67,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,68,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,69,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,70,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,71,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,72,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,73,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,74,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,75,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,76,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,77,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,78,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[35,79,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,80,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,81,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,82,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,83,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,84,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,85,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,86,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,87,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,88,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,89,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,90,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,91,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,92,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,93,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,94,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,95,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,96,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,97,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,98,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[35,99,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,100,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,101,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,102,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,103,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,104,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,105,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,106,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,107,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,108,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,109,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,110,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,111,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,112,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,113,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,114,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,115,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,116,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,117,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,118,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[35,119,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,120,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,121,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,122,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,123,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,124,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,125,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,126,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,127,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,128,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,129,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,130,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,131,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,132,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,133,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,134,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,135,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,136,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,137,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,138,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[35,139,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,140,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,141,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,142,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,143,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,144,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,145,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,146,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,147,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,148,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,149,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,150,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,151,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,152,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,153,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,154,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,155,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,156,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,157,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,158,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[35,159,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,160,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,161,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,162,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,163,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,164,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,165,2,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,166,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,167,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,168,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,169,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,170,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,171,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,172,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,173,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,174,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,175,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,176,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,177,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,178,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[35,179,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,180,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,181,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,182,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,183,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,184,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,185,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,186,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,187,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,188,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,189,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,190,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,191,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,192,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,193,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,194,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,195,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,196,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,197,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,198,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[35,199,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,0,0.36,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,1,0.37,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,2,0.38,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,3,0.39,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,4,0.4,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,5,0.41,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,6,0.42,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,7,0.43,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,8,0.44,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,9,0.45,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,10,0.46,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,11,0.47,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,12,0.48,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,13,0.49,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,14,0.5,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,15,0.51,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,16,0.52,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,17,0.53,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,18,0.54,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[36,19,0.55,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,20,0.56,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,21,0.57,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,22,0.58,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,23,0.59,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,24,0.6,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,25,0.61,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,26,0.62,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,27,0.63,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,28,0.64,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,29,0.65,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,30,0.66,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,31,0.67,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,32,0.68,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,33,0.69,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,34,0.7,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,35,0.71,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,36,0.72,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,37,0.73,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,38,0.74,6]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[36,39,0.75,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,40,0.76,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,41,0.77,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,42,0.78,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,43,0.79,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,44,0.8,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,45,0.81,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,46,0.82,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,47,0.83,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,48,0.84,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,49,0.85,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,50,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,51,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,52,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,53,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,54,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,55,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,56,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,57,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,58,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[36,59,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,60,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,61,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,62,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,63,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,64,1,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,65,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,66,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,67,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,68,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,69,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,70,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,71,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,72,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,73,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,74,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,75,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,76,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,77,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,78,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[36,79,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,80,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,81,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,82,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,83,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,84,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,85,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,86,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,87,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,88,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,89,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,90,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,91,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,92,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,93,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,94,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,95,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,96,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,97,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,98,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[36,99,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,100,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,101,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,102,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,103,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,104,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,105,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,106,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,107,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,108,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,109,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,110,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,111,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,112,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,113,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,114,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,115,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,116,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,117,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,118,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[36,119,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,120,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,121,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,122,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,123,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,124,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,125,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,126,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,127,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,128,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,129,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,130,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,131,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,132,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,133,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,134,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,135,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,136,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,137,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,138,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[36,139,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,140,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,141,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,142,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,143,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,144,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,145,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,146,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,147,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,148,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,149,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,150,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,151,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,152,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,153,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,154,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,155,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,156,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,157,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,158,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[36,159,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,160,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,161,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,162,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,163,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,164,2,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,165,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,166,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,167,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,168,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,169,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,170,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,171,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,172,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,173,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,174,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,175,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,176,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,177,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,178,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[36,179,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,180,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,181,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,182,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,183,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,184,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,185,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,186,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,187,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,188,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,189,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,190,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,191,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,192,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,193,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,194,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,195,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,196,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,197,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,198,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[36,199,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,0,0.37,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,1,0.38,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,2,0.39,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,3,0.4,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,4,0.41,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,5,0.42,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,6,0.43,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,7,0.44,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,8,0.45,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,9,0.46,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,10,0.47,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,11,0.48,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,12,0.49,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,13,0.5,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,14,0.51,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,15,0.52,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,16,0.53,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,17,0.54,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,18,0.55,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[37,19,0.56,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,20,0.57,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,21,0.58,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,22,0.59,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,23,0.6,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,24,0.61,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,25,0.62,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,26,0.63,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,27,0.64,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,28,0.65,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,29,0.66,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,30,0.67,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,31,0.68,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,32,0.69,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,33,0.7,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,34,0.71,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,35,0.72,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,36,0.73,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,37,0.74,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,38,0.75,7]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[37,39,0.76,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,40,0.77,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,41,0.78,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,42,0.79,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,43,0.8,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,44,0.81,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,45,0.82,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,46,0.83,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,47,0.84,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,48,0.85,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,49,0.86,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,50,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,51,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,52,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,53,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,54,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,55,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,56,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,57,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,58,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[37,59,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,60,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,61,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,62,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,63,1,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,64,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,65,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,66,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,67,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,68,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,69,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,70,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,71,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,72,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,73,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,74,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,75,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,76,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,77,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,78,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[37,79,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,80,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,81,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,82,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,83,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,84,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,85,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,86,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,87,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,88,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,89,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,90,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,91,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,92,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,93,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,94,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,95,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,96,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,97,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,98,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[37,99,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,100,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,101,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,102,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,103,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,104,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,105,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,106,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,107,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,108,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,109,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,110,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,111,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,112,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,113,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,114,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,115,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,116,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,117,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,118,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[37,119,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,120,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,121,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,122,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,123,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,124,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,125,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,126,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,127,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,128,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,129,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,130,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,131,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,132,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,133,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,134,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,135,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,136,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,137,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,138,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[37,139,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,140,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,141,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,142,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,143,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,144,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,145,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,146,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,147,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,148,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,149,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,150,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,151,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,152,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,153,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,154,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,155,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,156,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,157,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,158,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[37,159,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,160,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,161,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,162,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,163,2,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,164,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,165,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,166,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,167,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,168,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,169,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,170,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,171,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,172,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,173,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,174,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,175,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,176,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,177,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,178,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[37,179,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,180,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,181,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,182,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,183,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,184,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,185,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,186,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,187,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,188,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,189,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,190,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,191,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,192,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,193,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,194,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,195,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,196,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,197,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,198,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[37,199,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,0,0.38,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,1,0.39,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,2,0.4,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,3,0.41,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,4,0.42,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,5,0.43,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,6,0.44,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,7,0.45,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,8,0.46,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,9,0.47,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,10,0.48,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,11,0.49,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,12,0.5,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,13,0.51,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,14,0.52,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,15,0.53,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,16,0.54,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,17,0.55,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,18,0.56,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[38,19,0.57,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,20,0.58,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,21,0.59,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,22,0.6,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,23,0.61,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,24,0.62,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,25,0.63,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,26,0.64,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,27,0.65,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,28,0.66,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,29,0.67,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,30,0.68,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,31,0.69,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,32,0.7,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,33,0.71,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,34,0.72,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,35,0.73,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,36,0.74,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,37,0.75,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,38,0.76,8]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[38,39,0.77,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,40,0.78,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,41,0.79,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,42,0.8,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,43,0.81,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,44,0.82,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,45,0.83,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,46,0.84,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,47,0.85,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,48,0.86,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,49,0.87,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,50,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,51,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,52,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,53,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,54,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,55,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,56,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,57,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,58,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[38,59,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,60,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,61,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,62,1,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,63,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,64,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,65,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,66,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,67,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,68,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,69,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,70,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,71,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,72,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,73,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,74,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,75,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,76,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,77,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,78,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[38,79,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,80,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,81,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,82,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,83,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,84,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,85,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,86,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,87,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,88,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,89,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,90,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,91,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,92,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,93,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,94,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,95,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,96,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,97,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,98,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[38,99,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,100,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,101,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,102,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,103,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,104,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,105,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,106,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,107,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,108,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,109,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,110,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,111,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,112,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,113,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,114,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,115,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,116,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,117,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,118,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[38,119,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,120,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,121,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,122,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,123,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,124,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,125,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,126,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,127,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,128,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,129,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,130,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,131,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,132,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,133,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,134,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,135,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,136,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,137,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,138,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[38,139,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,140,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,141,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,142,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,143,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,144,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,145,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,146,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,147,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,148,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,149,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,150,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,151,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,152,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,153,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,154,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,155,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,156,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,157,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,158,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[38,159,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,160,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,161,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,162,2,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,163,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,164,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,165,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,166,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,167,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,168,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,169,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,170,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,171,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,172,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,173,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,174,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,175,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,176,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,177,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,178,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[38,179,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,180,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,181,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,182,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,183,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,184,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,185,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,186,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,187,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,188,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,189,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,190,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,191,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,192,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,193,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,194,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,195,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,196,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,197,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,198,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[38,199,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,0,0.39,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,1,0.4,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,2,0.41,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,3,0.42,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,4,0.43,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,5,0.44,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,6,0.45,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,7,0.46,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,8,0.47,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,9,0.48,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,10,0.49,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,11,0.5,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,12,0.51,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,13,0.52,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,14,0.53,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,15,0.54,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,16,0.55,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,17,0.56,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,18,0.57,9]) );
insert into pts_collection ( gid, pt ) values ( 20, pc_makepoint(1, ARRAY[39,19,0.58,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,20,0.59,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,21,0.6,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,22,0.61,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,23,0.62,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,24,0.63,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,25,0.64,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,26,0.65,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,27,0.66,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,28,0.67,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,29,0.68,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,30,0.69,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,31,0.7,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,32,0.71,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,33,0.72,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,34,0.73,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,35,0.74,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,36,0.75,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,37,0.76,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,38,0.77,9]) );
insert into pts_collection ( gid, pt ) values ( 21, pc_makepoint(1, ARRAY[39,39,0.78,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,40,0.79,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,41,0.8,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,42,0.81,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,43,0.82,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,44,0.83,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,45,0.84,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,46,0.85,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,47,0.86,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,48,0.87,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,49,0.88,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,50,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,51,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,52,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,53,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,54,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,55,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,56,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,57,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,58,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 22, pc_makepoint(1, ARRAY[39,59,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,60,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,61,1,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,62,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,63,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,64,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,65,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,66,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,67,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,68,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,69,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,70,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,71,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,72,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,73,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,74,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,75,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,76,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,77,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,78,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 23, pc_makepoint(1, ARRAY[39,79,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,80,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,81,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,82,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,83,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,84,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,85,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,86,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,87,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,88,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,89,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,90,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,91,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,92,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,93,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,94,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,95,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,96,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,97,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,98,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 24, pc_makepoint(1, ARRAY[39,99,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,100,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,101,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,102,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,103,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,104,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,105,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,106,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,107,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,108,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,109,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,110,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,111,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,112,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,113,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,114,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,115,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,116,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,117,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,118,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 25, pc_makepoint(1, ARRAY[39,119,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,120,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,121,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,122,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,123,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,124,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,125,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,126,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,127,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,128,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,129,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,130,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,131,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,132,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,133,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,134,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,135,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,136,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,137,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,138,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 26, pc_makepoint(1, ARRAY[39,139,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,140,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,141,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,142,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,143,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,144,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,145,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,146,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,147,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,148,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,149,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,150,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,151,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,152,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,153,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,154,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,155,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,156,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,157,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,158,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 27, pc_makepoint(1, ARRAY[39,159,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,160,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,161,2,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,162,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,163,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,164,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,165,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,166,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,167,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,168,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,169,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,170,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,171,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,172,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,173,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,174,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,175,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,176,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,177,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,178,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 28, pc_makepoint(1, ARRAY[39,179,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,180,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,181,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,182,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,183,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,184,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,185,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,186,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,187,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,188,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,189,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,190,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,191,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,192,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,193,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,194,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,195,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,196,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,197,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,198,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 29, pc_makepoint(1, ARRAY[39,199,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,0,0.4,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,1,0.41,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,2,0.42,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,3,0.43,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,4,0.44,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,5,0.45,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,6,0.46,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,7,0.47,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,8,0.48,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,9,0.49,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,10,0.5,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,11,0.51,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,12,0.52,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,13,0.53,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,14,0.54,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,15,0.55,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,16,0.56,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,17,0.57,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,18,0.58,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[40,19,0.59,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,20,0.6,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,21,0.61,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,22,0.62,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,23,0.63,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,24,0.64,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,25,0.65,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,26,0.66,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,27,0.67,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,28,0.68,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,29,0.69,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,30,0.7,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,31,0.71,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,32,0.72,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,33,0.73,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,34,0.74,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,35,0.75,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,36,0.76,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,37,0.77,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,38,0.78,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[40,39,0.79,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,40,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,41,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,42,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,43,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,44,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,45,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,46,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,47,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,48,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,49,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,50,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,51,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,52,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,53,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,54,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,55,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,56,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,57,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,58,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[40,59,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,60,1,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,61,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,62,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,63,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,64,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,65,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,66,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,67,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,68,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,69,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,70,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,71,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,72,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,73,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,74,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,75,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,76,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,77,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,78,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[40,79,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,80,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,81,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,82,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,83,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,84,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,85,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,86,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,87,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,88,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,89,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,90,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,91,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,92,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,93,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,94,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,95,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,96,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,97,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,98,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[40,99,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,100,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,101,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,102,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,103,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,104,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,105,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,106,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,107,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,108,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,109,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,110,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,111,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,112,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,113,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,114,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,115,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,116,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,117,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,118,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[40,119,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,120,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,121,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,122,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,123,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,124,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,125,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,126,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,127,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,128,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,129,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,130,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,131,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,132,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,133,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,134,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,135,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,136,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,137,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,138,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[40,139,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,140,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,141,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,142,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,143,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,144,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,145,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,146,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,147,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,148,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,149,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,150,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,151,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,152,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,153,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,154,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,155,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,156,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,157,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,158,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[40,159,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,160,2,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,161,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,162,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,163,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,164,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,165,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,166,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,167,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,168,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,169,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,170,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,171,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,172,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,173,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,174,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,175,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,176,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,177,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,178,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[40,179,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,180,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,181,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,182,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,183,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,184,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,185,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,186,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,187,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,188,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,189,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,190,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,191,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,192,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,193,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,194,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,195,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,196,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,197,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,198,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[40,199,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,0,0.41,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,1,0.42,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,2,0.43,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,3,0.44,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,4,0.45,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,5,0.46,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,6,0.47,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,7,0.48,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,8,0.49,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,9,0.5,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,10,0.51,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,11,0.52,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,12,0.53,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,13,0.54,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,14,0.55,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,15,0.56,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,16,0.57,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,17,0.58,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,18,0.59,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[41,19,0.6,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,20,0.61,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,21,0.62,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,22,0.63,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,23,0.64,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,24,0.65,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,25,0.66,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,26,0.67,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,27,0.68,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,28,0.69,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,29,0.7,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,30,0.71,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,31,0.72,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,32,0.73,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,33,0.74,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,34,0.75,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,35,0.76,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,36,0.77,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,37,0.78,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,38,0.79,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[41,39,0.8,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,40,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,41,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,42,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,43,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,44,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,45,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,46,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,47,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,48,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,49,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,50,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,51,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,52,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,53,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,54,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,55,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,56,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,57,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,58,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[41,59,1,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,60,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,61,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,62,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,63,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,64,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,65,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,66,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,67,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,68,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,69,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,70,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,71,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,72,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,73,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,74,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,75,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,76,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,77,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,78,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[41,79,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,80,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,81,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,82,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,83,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,84,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,85,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,86,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,87,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,88,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,89,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,90,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,91,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,92,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,93,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,94,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,95,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,96,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,97,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,98,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[41,99,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,100,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,101,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,102,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,103,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,104,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,105,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,106,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,107,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,108,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,109,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,110,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,111,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,112,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,113,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,114,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,115,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,116,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,117,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,118,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[41,119,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,120,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,121,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,122,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,123,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,124,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,125,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,126,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,127,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,128,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,129,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,130,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,131,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,132,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,133,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,134,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,135,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,136,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,137,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,138,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[41,139,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,140,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,141,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,142,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,143,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,144,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,145,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,146,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,147,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,148,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,149,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,150,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,151,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,152,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,153,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,154,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,155,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,156,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,157,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,158,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[41,159,2,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,160,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,161,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,162,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,163,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,164,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,165,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,166,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,167,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,168,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,169,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,170,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,171,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,172,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,173,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,174,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,175,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,176,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,177,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,178,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[41,179,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,180,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,181,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,182,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,183,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,184,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,185,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,186,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,187,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,188,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,189,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,190,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,191,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,192,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,193,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,194,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,195,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,196,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,197,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,198,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[41,199,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,0,0.42,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,1,0.43,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,2,0.44,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,3,0.45,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,4,0.46,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,5,0.47,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,6,0.48,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,7,0.49,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,8,0.5,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,9,0.51,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,10,0.52,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,11,0.53,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,12,0.54,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,13,0.55,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,14,0.56,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,15,0.57,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,16,0.58,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,17,0.59,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,18,0.6,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[42,19,0.61,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,20,0.62,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,21,0.63,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,22,0.64,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,23,0.65,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,24,0.66,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,25,0.67,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,26,0.68,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,27,0.69,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,28,0.7,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,29,0.71,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,30,0.72,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,31,0.73,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,32,0.74,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,33,0.75,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,34,0.76,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,35,0.77,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,36,0.78,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,37,0.79,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,38,0.8,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[42,39,0.81,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,40,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,41,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,42,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,43,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,44,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,45,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,46,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,47,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,48,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,49,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,50,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,51,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,52,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,53,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,54,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,55,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,56,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,57,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,58,1,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[42,59,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,60,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,61,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,62,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,63,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,64,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,65,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,66,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,67,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,68,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,69,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,70,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,71,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,72,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,73,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,74,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,75,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,76,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,77,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,78,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[42,79,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,80,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,81,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,82,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,83,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,84,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,85,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,86,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,87,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,88,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,89,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,90,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,91,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,92,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,93,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,94,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,95,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,96,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,97,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,98,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[42,99,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,100,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,101,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,102,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,103,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,104,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,105,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,106,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,107,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,108,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,109,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,110,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,111,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,112,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,113,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,114,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,115,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,116,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,117,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,118,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[42,119,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,120,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,121,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,122,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,123,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,124,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,125,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,126,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,127,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,128,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,129,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,130,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,131,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,132,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,133,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,134,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,135,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,136,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,137,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,138,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[42,139,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,140,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,141,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,142,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,143,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,144,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,145,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,146,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,147,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,148,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,149,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,150,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,151,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,152,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,153,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,154,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,155,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,156,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,157,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,158,2,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[42,159,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,160,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,161,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,162,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,163,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,164,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,165,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,166,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,167,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,168,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,169,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,170,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,171,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,172,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,173,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,174,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,175,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,176,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,177,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,178,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[42,179,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,180,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,181,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,182,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,183,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,184,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,185,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,186,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,187,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,188,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,189,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,190,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,191,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,192,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,193,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,194,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,195,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,196,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,197,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,198,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[42,199,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,0,0.43,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,1,0.44,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,2,0.45,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,3,0.46,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,4,0.47,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,5,0.48,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,6,0.49,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,7,0.5,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,8,0.51,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,9,0.52,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,10,0.53,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,11,0.54,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,12,0.55,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,13,0.56,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,14,0.57,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,15,0.58,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,16,0.59,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,17,0.6,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,18,0.61,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[43,19,0.62,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,20,0.63,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,21,0.64,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,22,0.65,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,23,0.66,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,24,0.67,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,25,0.68,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,26,0.69,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,27,0.7,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,28,0.71,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,29,0.72,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,30,0.73,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,31,0.74,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,32,0.75,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,33,0.76,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,34,0.77,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,35,0.78,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,36,0.79,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,37,0.8,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,38,0.81,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[43,39,0.82,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,40,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,41,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,42,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,43,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,44,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,45,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,46,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,47,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,48,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,49,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,50,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,51,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,52,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,53,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,54,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,55,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,56,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,57,1,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,58,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[43,59,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,60,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,61,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,62,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,63,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,64,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,65,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,66,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,67,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,68,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,69,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,70,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,71,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,72,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,73,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,74,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,75,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,76,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,77,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,78,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[43,79,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,80,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,81,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,82,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,83,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,84,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,85,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,86,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,87,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,88,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,89,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,90,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,91,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,92,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,93,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,94,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,95,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,96,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,97,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,98,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[43,99,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,100,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,101,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,102,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,103,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,104,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,105,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,106,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,107,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,108,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,109,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,110,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,111,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,112,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,113,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,114,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,115,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,116,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,117,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,118,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[43,119,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,120,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,121,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,122,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,123,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,124,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,125,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,126,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,127,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,128,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,129,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,130,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,131,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,132,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,133,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,134,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,135,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,136,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,137,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,138,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[43,139,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,140,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,141,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,142,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,143,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,144,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,145,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,146,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,147,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,148,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,149,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,150,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,151,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,152,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,153,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,154,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,155,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,156,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,157,2,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,158,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[43,159,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,160,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,161,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,162,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,163,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,164,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,165,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,166,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,167,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,168,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,169,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,170,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,171,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,172,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,173,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,174,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,175,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,176,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,177,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,178,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[43,179,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,180,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,181,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,182,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,183,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,184,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,185,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,186,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,187,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,188,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,189,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,190,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,191,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,192,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,193,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,194,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,195,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,196,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,197,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,198,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[43,199,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,0,0.44,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,1,0.45,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,2,0.46,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,3,0.47,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,4,0.48,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,5,0.49,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,6,0.5,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,7,0.51,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,8,0.52,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,9,0.53,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,10,0.54,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,11,0.55,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,12,0.56,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,13,0.57,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,14,0.58,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,15,0.59,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,16,0.6,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,17,0.61,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,18,0.62,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[44,19,0.63,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,20,0.64,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,21,0.65,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,22,0.66,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,23,0.67,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,24,0.68,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,25,0.69,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,26,0.7,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,27,0.71,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,28,0.72,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,29,0.73,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,30,0.74,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,31,0.75,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,32,0.76,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,33,0.77,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,34,0.78,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,35,0.79,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,36,0.8,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,37,0.81,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,38,0.82,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[44,39,0.83,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,40,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,41,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,42,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,43,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,44,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,45,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,46,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,47,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,48,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,49,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,50,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,51,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,52,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,53,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,54,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,55,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,56,1,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,57,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,58,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[44,59,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,60,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,61,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,62,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,63,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,64,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,65,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,66,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,67,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,68,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,69,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,70,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,71,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,72,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,73,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,74,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,75,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,76,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,77,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,78,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[44,79,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,80,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,81,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,82,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,83,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,84,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,85,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,86,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,87,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,88,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,89,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,90,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,91,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,92,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,93,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,94,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,95,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,96,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,97,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,98,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[44,99,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,100,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,101,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,102,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,103,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,104,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,105,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,106,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,107,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,108,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,109,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,110,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,111,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,112,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,113,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,114,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,115,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,116,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,117,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,118,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[44,119,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,120,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,121,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,122,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,123,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,124,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,125,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,126,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,127,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,128,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,129,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,130,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,131,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,132,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,133,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,134,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,135,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,136,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,137,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,138,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[44,139,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,140,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,141,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,142,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,143,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,144,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,145,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,146,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,147,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,148,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,149,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,150,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,151,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,152,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,153,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,154,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,155,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,156,2,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,157,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,158,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[44,159,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,160,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,161,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,162,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,163,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,164,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,165,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,166,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,167,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,168,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,169,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,170,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,171,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,172,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,173,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,174,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,175,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,176,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,177,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,178,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[44,179,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,180,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,181,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,182,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,183,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,184,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,185,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,186,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,187,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,188,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,189,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,190,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,191,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,192,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,193,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,194,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,195,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,196,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,197,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,198,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[44,199,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,0,0.45,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,1,0.46,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,2,0.47,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,3,0.48,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,4,0.49,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,5,0.5,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,6,0.51,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,7,0.52,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,8,0.53,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,9,0.54,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,10,0.55,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,11,0.56,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,12,0.57,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,13,0.58,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,14,0.59,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,15,0.6,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,16,0.61,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,17,0.62,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,18,0.63,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[45,19,0.64,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,20,0.65,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,21,0.66,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,22,0.67,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,23,0.68,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,24,0.69,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,25,0.7,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,26,0.71,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,27,0.72,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,28,0.73,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,29,0.74,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,30,0.75,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,31,0.76,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,32,0.77,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,33,0.78,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,34,0.79,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,35,0.8,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,36,0.81,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,37,0.82,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,38,0.83,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[45,39,0.84,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,40,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,41,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,42,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,43,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,44,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,45,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,46,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,47,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,48,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,49,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,50,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,51,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,52,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,53,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,54,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,55,1,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,56,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,57,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,58,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[45,59,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,60,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,61,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,62,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,63,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,64,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,65,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,66,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,67,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,68,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,69,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,70,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,71,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,72,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,73,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,74,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,75,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,76,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,77,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,78,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[45,79,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,80,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,81,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,82,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,83,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,84,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,85,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,86,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,87,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,88,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,89,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,90,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,91,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,92,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,93,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,94,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,95,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,96,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,97,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,98,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[45,99,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,100,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,101,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,102,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,103,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,104,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,105,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,106,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,107,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,108,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,109,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,110,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,111,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,112,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,113,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,114,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,115,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,116,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,117,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,118,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[45,119,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,120,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,121,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,122,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,123,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,124,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,125,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,126,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,127,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,128,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,129,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,130,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,131,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,132,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,133,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,134,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,135,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,136,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,137,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,138,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[45,139,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,140,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,141,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,142,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,143,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,144,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,145,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,146,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,147,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,148,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,149,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,150,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,151,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,152,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,153,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,154,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,155,2,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,156,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,157,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,158,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[45,159,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,160,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,161,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,162,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,163,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,164,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,165,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,166,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,167,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,168,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,169,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,170,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,171,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,172,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,173,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,174,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,175,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,176,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,177,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,178,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[45,179,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,180,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,181,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,182,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,183,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,184,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,185,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,186,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,187,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,188,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,189,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,190,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,191,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,192,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,193,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,194,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,195,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,196,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,197,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,198,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[45,199,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,0,0.46,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,1,0.47,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,2,0.48,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,3,0.49,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,4,0.5,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,5,0.51,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,6,0.52,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,7,0.53,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,8,0.54,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,9,0.55,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,10,0.56,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,11,0.57,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,12,0.58,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,13,0.59,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,14,0.6,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,15,0.61,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,16,0.62,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,17,0.63,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,18,0.64,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[46,19,0.65,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,20,0.66,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,21,0.67,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,22,0.68,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,23,0.69,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,24,0.7,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,25,0.71,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,26,0.72,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,27,0.73,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,28,0.74,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,29,0.75,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,30,0.76,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,31,0.77,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,32,0.78,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,33,0.79,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,34,0.8,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,35,0.81,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,36,0.82,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,37,0.83,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,38,0.84,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[46,39,0.85,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,40,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,41,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,42,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,43,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,44,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,45,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,46,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,47,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,48,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,49,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,50,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,51,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,52,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,53,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,54,1,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,55,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,56,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,57,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,58,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[46,59,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,60,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,61,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,62,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,63,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,64,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,65,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,66,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,67,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,68,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,69,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,70,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,71,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,72,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,73,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,74,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,75,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,76,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,77,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,78,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[46,79,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,80,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,81,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,82,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,83,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,84,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,85,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,86,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,87,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,88,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,89,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,90,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,91,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,92,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,93,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,94,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,95,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,96,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,97,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,98,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[46,99,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,100,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,101,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,102,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,103,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,104,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,105,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,106,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,107,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,108,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,109,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,110,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,111,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,112,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,113,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,114,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,115,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,116,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,117,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,118,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[46,119,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,120,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,121,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,122,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,123,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,124,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,125,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,126,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,127,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,128,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,129,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,130,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,131,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,132,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,133,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,134,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,135,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,136,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,137,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,138,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[46,139,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,140,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,141,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,142,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,143,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,144,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,145,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,146,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,147,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,148,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,149,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,150,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,151,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,152,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,153,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,154,2,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,155,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,156,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,157,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,158,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[46,159,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,160,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,161,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,162,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,163,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,164,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,165,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,166,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,167,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,168,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,169,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,170,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,171,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,172,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,173,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,174,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,175,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,176,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,177,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,178,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[46,179,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,180,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,181,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,182,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,183,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,184,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,185,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,186,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,187,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,188,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,189,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,190,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,191,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,192,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,193,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,194,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,195,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,196,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,197,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,198,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[46,199,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,0,0.47,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,1,0.48,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,2,0.49,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,3,0.5,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,4,0.51,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,5,0.52,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,6,0.53,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,7,0.54,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,8,0.55,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,9,0.56,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,10,0.57,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,11,0.58,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,12,0.59,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,13,0.6,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,14,0.61,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,15,0.62,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,16,0.63,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,17,0.64,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,18,0.65,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[47,19,0.66,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,20,0.67,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,21,0.68,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,22,0.69,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,23,0.7,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,24,0.71,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,25,0.72,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,26,0.73,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,27,0.74,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,28,0.75,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,29,0.76,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,30,0.77,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,31,0.78,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,32,0.79,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,33,0.8,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,34,0.81,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,35,0.82,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,36,0.83,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,37,0.84,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,38,0.85,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[47,39,0.86,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,40,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,41,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,42,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,43,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,44,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,45,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,46,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,47,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,48,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,49,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,50,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,51,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,52,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,53,1,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,54,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,55,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,56,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,57,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,58,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[47,59,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,60,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,61,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,62,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,63,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,64,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,65,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,66,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,67,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,68,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,69,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,70,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,71,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,72,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,73,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,74,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,75,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,76,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,77,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,78,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[47,79,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,80,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,81,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,82,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,83,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,84,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,85,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,86,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,87,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,88,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,89,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,90,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,91,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,92,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,93,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,94,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,95,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,96,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,97,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,98,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[47,99,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,100,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,101,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,102,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,103,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,104,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,105,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,106,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,107,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,108,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,109,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,110,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,111,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,112,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,113,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,114,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,115,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,116,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,117,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,118,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[47,119,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,120,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,121,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,122,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,123,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,124,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,125,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,126,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,127,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,128,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,129,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,130,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,131,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,132,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,133,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,134,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,135,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,136,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,137,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,138,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[47,139,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,140,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,141,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,142,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,143,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,144,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,145,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,146,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,147,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,148,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,149,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,150,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,151,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,152,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,153,2,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,154,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,155,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,156,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,157,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,158,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[47,159,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,160,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,161,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,162,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,163,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,164,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,165,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,166,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,167,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,168,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,169,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,170,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,171,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,172,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,173,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,174,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,175,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,176,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,177,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,178,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[47,179,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,180,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,181,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,182,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,183,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,184,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,185,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,186,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,187,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,188,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,189,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,190,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,191,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,192,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,193,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,194,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,195,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,196,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,197,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,198,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[47,199,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,0,0.48,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,1,0.49,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,2,0.5,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,3,0.51,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,4,0.52,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,5,0.53,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,6,0.54,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,7,0.55,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,8,0.56,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,9,0.57,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,10,0.58,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,11,0.59,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,12,0.6,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,13,0.61,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,14,0.62,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,15,0.63,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,16,0.64,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,17,0.65,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,18,0.66,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[48,19,0.67,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,20,0.68,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,21,0.69,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,22,0.7,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,23,0.71,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,24,0.72,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,25,0.73,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,26,0.74,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,27,0.75,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,28,0.76,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,29,0.77,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,30,0.78,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,31,0.79,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,32,0.8,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,33,0.81,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,34,0.82,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,35,0.83,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,36,0.84,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,37,0.85,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,38,0.86,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[48,39,0.87,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,40,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,41,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,42,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,43,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,44,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,45,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,46,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,47,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,48,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,49,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,50,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,51,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,52,1,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,53,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,54,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,55,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,56,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,57,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,58,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[48,59,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,60,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,61,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,62,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,63,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,64,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,65,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,66,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,67,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,68,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,69,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,70,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,71,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,72,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,73,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,74,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,75,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,76,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,77,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,78,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[48,79,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,80,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,81,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,82,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,83,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,84,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,85,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,86,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,87,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,88,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,89,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,90,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,91,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,92,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,93,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,94,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,95,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,96,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,97,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,98,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[48,99,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,100,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,101,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,102,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,103,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,104,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,105,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,106,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,107,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,108,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,109,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,110,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,111,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,112,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,113,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,114,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,115,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,116,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,117,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,118,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[48,119,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,120,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,121,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,122,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,123,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,124,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,125,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,126,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,127,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,128,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,129,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,130,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,131,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,132,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,133,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,134,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,135,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,136,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,137,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,138,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[48,139,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,140,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,141,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,142,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,143,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,144,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,145,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,146,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,147,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,148,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,149,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,150,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,151,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,152,2,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,153,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,154,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,155,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,156,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,157,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,158,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[48,159,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,160,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,161,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,162,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,163,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,164,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,165,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,166,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,167,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,168,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,169,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,170,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,171,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,172,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,173,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,174,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,175,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,176,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,177,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,178,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[48,179,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,180,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,181,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,182,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,183,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,184,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,185,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,186,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,187,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,188,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,189,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,190,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,191,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,192,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,193,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,194,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,195,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,196,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,197,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,198,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[48,199,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,0,0.49,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,1,0.5,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,2,0.51,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,3,0.52,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,4,0.53,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,5,0.54,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,6,0.55,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,7,0.56,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,8,0.57,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,9,0.58,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,10,0.59,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,11,0.6,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,12,0.61,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,13,0.62,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,14,0.63,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,15,0.64,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,16,0.65,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,17,0.66,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,18,0.67,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[49,19,0.68,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,20,0.69,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,21,0.7,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,22,0.71,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,23,0.72,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,24,0.73,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,25,0.74,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,26,0.75,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,27,0.76,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,28,0.77,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,29,0.78,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,30,0.79,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,31,0.8,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,32,0.81,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,33,0.82,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,34,0.83,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,35,0.84,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,36,0.85,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,37,0.86,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,38,0.87,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[49,39,0.88,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,40,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,41,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,42,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,43,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,44,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,45,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,46,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,47,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,48,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,49,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,50,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,51,1,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,52,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,53,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,54,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,55,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,56,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,57,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,58,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[49,59,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,60,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,61,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,62,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,63,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,64,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,65,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,66,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,67,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,68,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,69,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,70,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,71,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,72,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,73,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,74,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,75,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,76,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,77,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,78,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[49,79,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,80,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,81,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,82,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,83,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,84,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,85,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,86,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,87,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,88,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,89,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,90,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,91,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,92,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,93,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,94,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,95,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,96,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,97,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,98,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[49,99,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,100,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,101,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,102,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,103,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,104,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,105,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,106,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,107,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,108,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,109,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,110,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,111,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,112,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,113,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,114,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,115,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,116,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,117,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,118,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[49,119,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,120,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,121,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,122,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,123,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,124,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,125,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,126,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,127,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,128,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,129,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,130,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,131,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,132,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,133,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,134,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,135,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,136,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,137,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,138,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[49,139,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,140,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,141,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,142,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,143,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,144,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,145,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,146,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,147,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,148,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,149,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,150,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,151,2,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,152,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,153,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,154,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,155,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,156,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,157,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,158,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[49,159,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,160,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,161,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,162,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,163,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,164,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,165,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,166,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,167,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,168,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,169,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,170,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,171,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,172,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,173,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,174,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,175,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,176,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,177,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,178,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[49,179,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,180,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,181,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,182,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,183,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,184,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,185,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,186,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,187,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,188,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,189,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,190,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,191,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,192,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,193,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,194,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,195,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,196,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,197,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,198,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[49,199,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,0,0.5,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,1,0.51,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,2,0.52,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,3,0.53,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,4,0.54,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,5,0.55,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,6,0.56,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,7,0.57,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,8,0.58,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,9,0.59,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,10,0.6,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,11,0.61,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,12,0.62,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,13,0.63,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,14,0.64,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,15,0.65,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,16,0.66,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,17,0.67,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,18,0.68,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[50,19,0.69,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,20,0.7,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,21,0.71,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,22,0.72,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,23,0.73,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,24,0.74,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,25,0.75,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,26,0.76,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,27,0.77,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,28,0.78,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,29,0.79,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,30,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,31,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,32,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,33,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,34,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,35,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,36,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,37,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,38,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[50,39,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,40,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,41,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,42,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,43,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,44,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,45,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,46,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,47,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,48,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,49,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,50,1,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,51,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,52,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,53,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,54,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,55,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,56,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,57,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,58,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[50,59,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,60,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,61,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,62,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,63,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,64,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,65,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,66,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,67,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,68,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,69,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,70,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,71,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,72,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,73,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,74,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,75,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,76,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,77,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,78,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[50,79,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,80,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,81,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,82,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,83,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,84,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,85,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,86,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,87,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,88,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,89,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,90,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,91,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,92,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,93,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,94,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,95,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,96,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,97,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,98,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[50,99,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,100,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,101,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,102,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,103,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,104,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,105,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,106,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,107,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,108,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,109,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,110,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,111,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,112,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,113,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,114,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,115,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,116,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,117,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,118,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[50,119,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,120,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,121,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,122,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,123,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,124,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,125,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,126,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,127,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,128,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,129,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,130,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,131,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,132,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,133,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,134,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,135,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,136,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,137,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,138,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[50,139,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,140,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,141,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,142,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,143,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,144,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,145,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,146,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,147,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,148,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,149,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,150,2,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,151,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,152,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,153,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,154,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,155,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,156,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,157,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,158,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[50,159,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,160,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,161,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,162,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,163,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,164,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,165,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,166,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,167,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,168,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,169,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,170,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,171,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,172,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,173,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,174,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,175,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,176,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,177,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,178,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[50,179,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,180,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,181,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,182,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,183,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,184,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,185,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,186,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,187,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,188,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,189,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,190,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,191,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,192,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,193,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,194,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,195,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,196,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,197,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,198,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[50,199,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,0,0.51,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,1,0.52,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,2,0.53,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,3,0.54,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,4,0.55,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,5,0.56,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,6,0.57,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,7,0.58,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,8,0.59,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,9,0.6,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,10,0.61,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,11,0.62,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,12,0.63,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,13,0.64,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,14,0.65,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,15,0.66,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,16,0.67,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,17,0.68,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,18,0.69,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[51,19,0.7,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,20,0.71,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,21,0.72,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,22,0.73,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,23,0.74,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,24,0.75,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,25,0.76,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,26,0.77,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,27,0.78,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,28,0.79,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,29,0.8,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,30,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,31,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,32,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,33,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,34,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,35,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,36,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,37,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,38,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[51,39,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,40,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,41,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,42,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,43,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,44,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,45,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,46,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,47,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,48,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,49,1,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,50,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,51,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,52,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,53,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,54,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,55,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,56,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,57,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,58,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[51,59,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,60,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,61,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,62,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,63,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,64,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,65,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,66,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,67,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,68,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,69,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,70,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,71,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,72,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,73,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,74,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,75,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,76,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,77,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,78,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[51,79,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,80,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,81,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,82,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,83,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,84,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,85,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,86,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,87,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,88,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,89,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,90,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,91,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,92,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,93,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,94,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,95,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,96,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,97,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,98,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[51,99,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,100,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,101,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,102,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,103,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,104,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,105,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,106,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,107,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,108,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,109,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,110,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,111,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,112,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,113,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,114,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,115,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,116,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,117,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,118,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[51,119,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,120,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,121,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,122,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,123,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,124,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,125,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,126,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,127,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,128,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,129,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,130,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,131,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,132,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,133,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,134,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,135,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,136,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,137,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,138,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[51,139,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,140,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,141,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,142,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,143,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,144,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,145,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,146,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,147,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,148,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,149,2,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,150,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,151,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,152,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,153,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,154,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,155,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,156,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,157,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,158,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[51,159,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,160,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,161,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,162,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,163,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,164,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,165,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,166,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,167,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,168,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,169,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,170,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,171,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,172,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,173,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,174,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,175,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,176,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,177,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,178,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[51,179,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,180,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,181,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,182,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,183,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,184,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,185,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,186,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,187,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,188,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,189,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,190,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,191,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,192,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,193,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,194,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,195,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,196,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,197,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,198,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[51,199,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,0,0.52,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,1,0.53,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,2,0.54,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,3,0.55,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,4,0.56,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,5,0.57,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,6,0.58,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,7,0.59,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,8,0.6,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,9,0.61,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,10,0.62,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,11,0.63,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,12,0.64,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,13,0.65,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,14,0.66,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,15,0.67,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,16,0.68,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,17,0.69,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,18,0.7,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[52,19,0.71,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,20,0.72,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,21,0.73,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,22,0.74,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,23,0.75,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,24,0.76,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,25,0.77,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,26,0.78,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,27,0.79,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,28,0.8,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,29,0.81,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,30,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,31,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,32,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,33,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,34,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,35,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,36,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,37,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,38,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[52,39,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,40,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,41,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,42,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,43,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,44,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,45,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,46,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,47,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,48,1,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,49,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,50,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,51,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,52,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,53,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,54,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,55,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,56,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,57,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,58,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[52,59,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,60,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,61,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,62,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,63,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,64,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,65,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,66,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,67,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,68,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,69,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,70,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,71,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,72,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,73,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,74,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,75,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,76,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,77,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,78,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[52,79,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,80,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,81,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,82,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,83,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,84,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,85,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,86,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,87,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,88,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,89,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,90,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,91,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,92,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,93,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,94,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,95,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,96,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,97,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,98,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[52,99,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,100,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,101,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,102,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,103,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,104,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,105,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,106,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,107,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,108,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,109,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,110,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,111,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,112,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,113,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,114,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,115,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,116,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,117,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,118,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[52,119,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,120,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,121,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,122,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,123,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,124,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,125,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,126,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,127,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,128,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,129,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,130,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,131,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,132,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,133,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,134,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,135,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,136,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,137,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,138,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[52,139,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,140,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,141,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,142,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,143,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,144,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,145,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,146,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,147,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,148,2,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,149,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,150,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,151,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,152,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,153,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,154,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,155,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,156,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,157,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,158,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[52,159,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,160,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,161,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,162,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,163,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,164,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,165,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,166,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,167,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,168,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,169,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,170,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,171,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,172,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,173,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,174,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,175,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,176,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,177,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,178,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[52,179,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,180,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,181,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,182,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,183,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,184,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,185,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,186,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,187,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,188,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,189,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,190,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,191,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,192,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,193,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,194,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,195,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,196,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,197,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,198,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[52,199,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,0,0.53,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,1,0.54,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,2,0.55,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,3,0.56,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,4,0.57,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,5,0.58,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,6,0.59,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,7,0.6,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,8,0.61,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,9,0.62,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,10,0.63,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,11,0.64,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,12,0.65,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,13,0.66,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,14,0.67,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,15,0.68,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,16,0.69,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,17,0.7,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,18,0.71,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[53,19,0.72,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,20,0.73,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,21,0.74,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,22,0.75,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,23,0.76,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,24,0.77,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,25,0.78,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,26,0.79,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,27,0.8,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,28,0.81,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,29,0.82,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,30,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,31,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,32,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,33,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,34,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,35,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,36,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,37,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,38,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[53,39,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,40,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,41,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,42,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,43,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,44,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,45,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,46,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,47,1,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,48,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,49,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,50,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,51,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,52,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,53,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,54,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,55,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,56,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,57,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,58,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[53,59,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,60,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,61,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,62,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,63,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,64,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,65,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,66,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,67,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,68,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,69,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,70,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,71,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,72,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,73,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,74,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,75,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,76,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,77,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,78,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[53,79,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,80,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,81,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,82,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,83,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,84,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,85,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,86,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,87,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,88,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,89,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,90,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,91,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,92,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,93,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,94,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,95,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,96,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,97,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,98,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[53,99,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,100,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,101,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,102,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,103,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,104,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,105,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,106,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,107,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,108,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,109,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,110,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,111,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,112,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,113,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,114,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,115,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,116,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,117,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,118,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[53,119,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,120,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,121,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,122,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,123,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,124,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,125,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,126,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,127,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,128,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,129,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,130,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,131,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,132,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,133,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,134,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,135,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,136,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,137,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,138,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[53,139,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,140,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,141,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,142,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,143,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,144,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,145,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,146,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,147,2,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,148,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,149,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,150,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,151,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,152,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,153,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,154,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,155,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,156,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,157,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,158,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[53,159,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,160,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,161,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,162,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,163,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,164,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,165,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,166,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,167,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,168,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,169,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,170,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,171,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,172,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,173,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,174,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,175,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,176,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,177,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,178,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[53,179,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,180,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,181,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,182,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,183,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,184,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,185,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,186,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,187,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,188,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,189,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,190,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,191,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,192,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,193,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,194,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,195,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,196,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,197,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,198,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[53,199,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,0,0.54,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,1,0.55,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,2,0.56,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,3,0.57,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,4,0.58,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,5,0.59,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,6,0.6,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,7,0.61,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,8,0.62,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,9,0.63,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,10,0.64,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,11,0.65,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,12,0.66,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,13,0.67,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,14,0.68,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,15,0.69,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,16,0.7,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,17,0.71,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,18,0.72,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[54,19,0.73,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,20,0.74,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,21,0.75,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,22,0.76,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,23,0.77,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,24,0.78,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,25,0.79,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,26,0.8,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,27,0.81,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,28,0.82,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,29,0.83,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,30,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,31,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,32,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,33,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,34,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,35,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,36,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,37,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,38,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[54,39,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,40,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,41,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,42,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,43,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,44,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,45,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,46,1,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,47,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,48,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,49,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,50,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,51,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,52,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,53,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,54,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,55,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,56,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,57,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,58,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[54,59,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,60,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,61,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,62,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,63,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,64,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,65,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,66,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,67,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,68,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,69,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,70,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,71,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,72,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,73,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,74,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,75,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,76,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,77,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,78,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[54,79,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,80,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,81,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,82,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,83,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,84,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,85,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,86,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,87,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,88,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,89,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,90,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,91,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,92,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,93,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,94,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,95,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,96,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,97,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,98,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[54,99,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,100,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,101,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,102,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,103,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,104,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,105,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,106,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,107,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,108,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,109,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,110,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,111,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,112,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,113,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,114,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,115,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,116,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,117,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,118,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[54,119,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,120,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,121,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,122,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,123,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,124,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,125,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,126,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,127,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,128,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,129,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,130,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,131,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,132,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,133,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,134,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,135,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,136,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,137,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,138,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[54,139,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,140,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,141,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,142,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,143,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,144,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,145,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,146,2,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,147,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,148,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,149,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,150,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,151,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,152,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,153,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,154,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,155,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,156,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,157,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,158,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[54,159,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,160,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,161,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,162,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,163,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,164,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,165,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,166,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,167,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,168,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,169,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,170,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,171,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,172,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,173,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,174,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,175,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,176,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,177,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,178,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[54,179,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,180,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,181,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,182,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,183,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,184,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,185,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,186,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,187,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,188,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,189,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,190,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,191,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,192,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,193,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,194,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,195,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,196,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,197,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,198,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[54,199,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,0,0.55,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,1,0.56,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,2,0.57,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,3,0.58,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,4,0.59,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,5,0.6,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,6,0.61,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,7,0.62,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,8,0.63,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,9,0.64,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,10,0.65,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,11,0.66,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,12,0.67,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,13,0.68,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,14,0.69,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,15,0.7,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,16,0.71,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,17,0.72,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,18,0.73,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[55,19,0.74,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,20,0.75,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,21,0.76,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,22,0.77,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,23,0.78,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,24,0.79,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,25,0.8,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,26,0.81,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,27,0.82,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,28,0.83,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,29,0.84,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,30,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,31,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,32,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,33,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,34,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,35,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,36,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,37,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,38,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[55,39,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,40,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,41,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,42,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,43,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,44,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,45,1,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,46,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,47,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,48,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,49,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,50,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,51,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,52,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,53,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,54,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,55,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,56,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,57,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,58,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[55,59,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,60,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,61,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,62,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,63,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,64,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,65,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,66,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,67,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,68,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,69,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,70,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,71,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,72,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,73,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,74,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,75,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,76,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,77,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,78,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[55,79,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,80,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,81,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,82,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,83,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,84,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,85,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,86,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,87,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,88,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,89,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,90,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,91,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,92,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,93,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,94,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,95,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,96,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,97,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,98,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[55,99,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,100,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,101,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,102,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,103,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,104,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,105,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,106,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,107,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,108,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,109,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,110,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,111,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,112,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,113,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,114,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,115,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,116,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,117,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,118,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[55,119,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,120,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,121,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,122,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,123,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,124,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,125,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,126,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,127,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,128,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,129,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,130,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,131,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,132,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,133,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,134,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,135,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,136,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,137,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,138,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[55,139,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,140,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,141,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,142,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,143,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,144,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,145,2,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,146,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,147,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,148,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,149,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,150,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,151,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,152,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,153,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,154,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,155,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,156,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,157,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,158,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[55,159,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,160,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,161,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,162,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,163,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,164,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,165,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,166,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,167,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,168,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,169,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,170,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,171,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,172,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,173,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,174,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,175,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,176,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,177,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,178,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[55,179,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,180,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,181,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,182,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,183,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,184,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,185,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,186,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,187,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,188,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,189,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,190,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,191,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,192,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,193,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,194,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,195,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,196,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,197,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,198,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[55,199,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,0,0.56,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,1,0.57,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,2,0.58,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,3,0.59,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,4,0.6,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,5,0.61,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,6,0.62,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,7,0.63,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,8,0.64,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,9,0.65,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,10,0.66,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,11,0.67,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,12,0.68,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,13,0.69,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,14,0.7,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,15,0.71,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,16,0.72,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,17,0.73,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,18,0.74,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[56,19,0.75,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,20,0.76,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,21,0.77,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,22,0.78,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,23,0.79,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,24,0.8,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,25,0.81,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,26,0.82,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,27,0.83,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,28,0.84,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,29,0.85,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,30,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,31,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,32,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,33,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,34,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,35,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,36,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,37,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,38,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[56,39,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,40,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,41,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,42,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,43,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,44,1,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,45,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,46,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,47,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,48,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,49,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,50,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,51,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,52,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,53,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,54,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,55,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,56,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,57,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,58,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[56,59,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,60,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,61,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,62,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,63,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,64,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,65,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,66,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,67,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,68,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,69,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,70,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,71,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,72,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,73,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,74,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,75,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,76,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,77,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,78,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[56,79,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,80,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,81,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,82,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,83,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,84,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,85,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,86,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,87,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,88,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,89,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,90,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,91,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,92,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,93,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,94,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,95,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,96,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,97,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,98,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[56,99,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,100,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,101,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,102,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,103,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,104,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,105,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,106,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,107,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,108,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,109,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,110,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,111,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,112,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,113,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,114,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,115,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,116,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,117,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,118,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[56,119,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,120,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,121,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,122,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,123,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,124,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,125,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,126,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,127,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,128,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,129,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,130,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,131,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,132,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,133,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,134,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,135,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,136,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,137,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,138,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[56,139,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,140,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,141,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,142,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,143,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,144,2,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,145,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,146,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,147,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,148,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,149,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,150,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,151,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,152,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,153,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,154,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,155,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,156,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,157,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,158,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[56,159,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,160,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,161,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,162,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,163,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,164,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,165,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,166,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,167,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,168,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,169,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,170,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,171,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,172,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,173,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,174,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,175,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,176,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,177,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,178,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[56,179,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,180,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,181,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,182,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,183,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,184,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,185,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,186,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,187,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,188,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,189,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,190,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,191,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,192,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,193,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,194,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,195,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,196,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,197,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,198,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[56,199,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,0,0.57,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,1,0.58,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,2,0.59,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,3,0.6,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,4,0.61,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,5,0.62,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,6,0.63,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,7,0.64,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,8,0.65,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,9,0.66,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,10,0.67,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,11,0.68,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,12,0.69,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,13,0.7,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,14,0.71,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,15,0.72,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,16,0.73,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,17,0.74,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,18,0.75,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[57,19,0.76,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,20,0.77,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,21,0.78,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,22,0.79,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,23,0.8,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,24,0.81,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,25,0.82,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,26,0.83,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,27,0.84,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,28,0.85,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,29,0.86,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,30,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,31,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,32,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,33,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,34,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,35,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,36,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,37,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,38,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[57,39,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,40,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,41,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,42,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,43,1,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,44,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,45,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,46,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,47,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,48,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,49,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,50,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,51,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,52,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,53,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,54,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,55,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,56,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,57,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,58,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[57,59,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,60,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,61,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,62,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,63,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,64,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,65,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,66,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,67,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,68,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,69,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,70,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,71,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,72,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,73,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,74,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,75,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,76,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,77,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,78,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[57,79,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,80,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,81,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,82,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,83,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,84,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,85,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,86,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,87,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,88,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,89,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,90,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,91,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,92,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,93,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,94,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,95,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,96,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,97,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,98,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[57,99,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,100,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,101,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,102,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,103,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,104,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,105,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,106,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,107,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,108,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,109,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,110,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,111,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,112,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,113,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,114,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,115,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,116,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,117,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,118,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[57,119,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,120,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,121,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,122,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,123,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,124,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,125,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,126,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,127,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,128,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,129,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,130,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,131,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,132,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,133,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,134,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,135,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,136,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,137,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,138,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[57,139,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,140,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,141,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,142,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,143,2,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,144,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,145,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,146,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,147,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,148,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,149,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,150,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,151,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,152,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,153,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,154,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,155,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,156,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,157,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,158,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[57,159,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,160,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,161,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,162,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,163,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,164,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,165,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,166,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,167,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,168,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,169,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,170,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,171,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,172,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,173,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,174,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,175,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,176,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,177,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,178,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[57,179,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,180,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,181,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,182,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,183,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,184,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,185,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,186,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,187,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,188,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,189,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,190,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,191,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,192,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,193,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,194,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,195,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,196,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,197,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,198,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[57,199,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,0,0.58,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,1,0.59,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,2,0.6,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,3,0.61,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,4,0.62,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,5,0.63,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,6,0.64,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,7,0.65,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,8,0.66,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,9,0.67,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,10,0.68,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,11,0.69,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,12,0.7,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,13,0.71,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,14,0.72,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,15,0.73,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,16,0.74,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,17,0.75,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,18,0.76,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[58,19,0.77,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,20,0.78,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,21,0.79,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,22,0.8,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,23,0.81,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,24,0.82,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,25,0.83,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,26,0.84,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,27,0.85,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,28,0.86,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,29,0.87,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,30,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,31,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,32,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,33,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,34,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,35,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,36,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,37,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,38,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[58,39,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,40,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,41,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,42,1,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,43,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,44,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,45,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,46,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,47,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,48,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,49,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,50,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,51,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,52,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,53,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,54,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,55,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,56,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,57,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,58,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[58,59,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,60,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,61,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,62,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,63,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,64,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,65,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,66,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,67,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,68,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,69,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,70,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,71,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,72,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,73,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,74,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,75,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,76,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,77,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,78,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[58,79,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,80,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,81,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,82,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,83,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,84,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,85,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,86,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,87,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,88,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,89,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,90,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,91,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,92,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,93,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,94,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,95,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,96,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,97,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,98,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[58,99,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,100,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,101,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,102,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,103,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,104,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,105,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,106,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,107,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,108,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,109,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,110,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,111,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,112,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,113,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,114,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,115,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,116,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,117,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,118,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[58,119,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,120,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,121,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,122,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,123,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,124,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,125,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,126,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,127,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,128,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,129,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,130,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,131,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,132,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,133,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,134,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,135,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,136,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,137,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,138,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[58,139,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,140,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,141,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,142,2,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,143,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,144,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,145,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,146,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,147,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,148,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,149,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,150,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,151,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,152,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,153,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,154,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,155,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,156,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,157,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,158,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[58,159,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,160,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,161,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,162,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,163,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,164,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,165,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,166,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,167,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,168,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,169,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,170,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,171,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,172,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,173,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,174,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,175,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,176,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,177,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,178,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[58,179,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,180,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,181,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,182,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,183,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,184,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,185,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,186,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,187,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,188,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,189,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,190,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,191,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,192,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,193,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,194,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,195,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,196,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,197,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,198,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[58,199,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,0,0.59,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,1,0.6,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,2,0.61,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,3,0.62,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,4,0.63,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,5,0.64,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,6,0.65,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,7,0.66,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,8,0.67,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,9,0.68,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,10,0.69,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,11,0.7,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,12,0.71,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,13,0.72,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,14,0.73,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,15,0.74,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,16,0.75,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,17,0.76,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,18,0.77,9]) );
insert into pts_collection ( gid, pt ) values ( 40, pc_makepoint(1, ARRAY[59,19,0.78,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,20,0.79,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,21,0.8,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,22,0.81,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,23,0.82,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,24,0.83,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,25,0.84,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,26,0.85,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,27,0.86,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,28,0.87,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,29,0.88,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,30,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,31,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,32,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,33,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,34,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,35,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,36,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,37,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,38,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 41, pc_makepoint(1, ARRAY[59,39,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,40,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,41,1,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,42,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,43,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,44,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,45,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,46,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,47,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,48,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,49,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,50,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,51,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,52,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,53,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,54,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,55,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,56,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,57,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,58,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 42, pc_makepoint(1, ARRAY[59,59,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,60,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,61,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,62,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,63,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,64,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,65,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,66,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,67,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,68,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,69,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,70,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,71,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,72,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,73,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,74,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,75,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,76,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,77,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,78,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 43, pc_makepoint(1, ARRAY[59,79,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,80,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,81,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,82,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,83,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,84,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,85,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,86,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,87,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,88,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,89,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,90,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,91,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,92,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,93,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,94,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,95,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,96,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,97,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,98,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 44, pc_makepoint(1, ARRAY[59,99,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,100,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,101,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,102,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,103,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,104,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,105,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,106,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,107,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,108,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,109,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,110,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,111,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,112,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,113,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,114,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,115,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,116,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,117,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,118,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 45, pc_makepoint(1, ARRAY[59,119,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,120,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,121,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,122,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,123,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,124,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,125,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,126,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,127,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,128,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,129,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,130,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,131,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,132,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,133,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,134,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,135,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,136,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,137,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,138,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 46, pc_makepoint(1, ARRAY[59,139,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,140,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,141,2,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,142,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,143,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,144,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,145,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,146,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,147,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,148,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,149,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,150,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,151,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,152,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,153,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,154,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,155,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,156,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,157,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,158,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 47, pc_makepoint(1, ARRAY[59,159,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,160,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,161,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,162,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,163,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,164,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,165,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,166,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,167,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,168,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,169,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,170,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,171,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,172,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,173,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,174,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,175,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,176,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,177,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,178,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 48, pc_makepoint(1, ARRAY[59,179,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,180,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,181,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,182,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,183,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,184,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,185,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,186,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,187,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,188,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,189,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,190,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,191,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,192,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,193,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,194,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,195,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,196,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,197,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,198,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 49, pc_makepoint(1, ARRAY[59,199,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,0,0.6,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,1,0.61,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,2,0.62,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,3,0.63,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,4,0.64,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,5,0.65,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,6,0.66,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,7,0.67,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,8,0.68,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,9,0.69,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,10,0.7,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,11,0.71,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,12,0.72,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,13,0.73,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,14,0.74,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,15,0.75,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,16,0.76,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,17,0.77,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,18,0.78,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[60,19,0.79,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,20,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,21,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,22,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,23,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,24,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,25,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,26,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,27,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,28,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,29,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,30,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,31,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,32,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,33,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,34,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,35,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,36,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,37,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,38,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[60,39,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,40,1,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,41,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,42,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,43,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,44,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,45,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,46,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,47,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,48,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,49,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,50,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,51,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,52,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,53,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,54,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,55,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,56,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,57,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,58,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[60,59,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,60,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,61,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,62,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,63,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,64,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,65,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,66,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,67,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,68,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,69,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,70,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,71,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,72,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,73,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,74,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,75,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,76,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,77,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,78,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[60,79,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,80,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,81,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,82,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,83,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,84,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,85,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,86,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,87,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,88,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,89,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,90,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,91,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,92,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,93,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,94,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,95,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,96,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,97,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,98,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[60,99,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,100,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,101,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,102,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,103,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,104,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,105,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,106,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,107,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,108,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,109,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,110,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,111,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,112,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,113,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,114,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,115,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,116,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,117,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,118,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[60,119,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,120,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,121,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,122,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,123,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,124,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,125,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,126,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,127,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,128,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,129,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,130,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,131,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,132,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,133,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,134,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,135,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,136,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,137,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,138,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[60,139,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,140,2,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,141,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,142,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,143,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,144,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,145,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,146,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,147,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,148,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,149,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,150,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,151,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,152,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,153,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,154,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,155,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,156,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,157,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,158,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[60,159,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,160,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,161,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,162,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,163,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,164,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,165,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,166,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,167,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,168,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,169,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,170,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,171,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,172,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,173,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,174,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,175,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,176,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,177,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,178,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[60,179,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,180,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,181,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,182,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,183,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,184,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,185,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,186,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,187,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,188,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,189,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,190,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,191,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,192,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,193,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,194,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,195,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,196,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,197,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,198,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[60,199,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,0,0.61,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,1,0.62,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,2,0.63,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,3,0.64,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,4,0.65,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,5,0.66,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,6,0.67,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,7,0.68,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,8,0.69,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,9,0.7,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,10,0.71,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,11,0.72,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,12,0.73,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,13,0.74,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,14,0.75,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,15,0.76,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,16,0.77,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,17,0.78,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,18,0.79,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[61,19,0.8,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,20,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,21,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,22,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,23,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,24,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,25,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,26,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,27,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,28,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,29,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,30,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,31,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,32,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,33,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,34,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,35,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,36,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,37,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,38,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[61,39,1,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,40,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,41,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,42,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,43,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,44,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,45,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,46,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,47,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,48,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,49,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,50,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,51,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,52,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,53,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,54,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,55,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,56,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,57,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,58,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[61,59,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,60,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,61,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,62,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,63,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,64,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,65,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,66,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,67,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,68,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,69,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,70,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,71,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,72,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,73,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,74,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,75,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,76,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,77,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,78,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[61,79,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,80,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,81,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,82,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,83,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,84,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,85,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,86,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,87,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,88,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,89,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,90,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,91,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,92,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,93,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,94,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,95,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,96,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,97,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,98,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[61,99,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,100,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,101,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,102,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,103,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,104,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,105,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,106,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,107,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,108,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,109,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,110,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,111,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,112,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,113,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,114,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,115,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,116,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,117,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,118,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[61,119,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,120,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,121,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,122,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,123,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,124,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,125,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,126,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,127,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,128,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,129,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,130,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,131,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,132,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,133,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,134,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,135,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,136,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,137,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,138,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[61,139,2,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,140,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,141,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,142,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,143,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,144,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,145,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,146,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,147,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,148,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,149,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,150,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,151,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,152,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,153,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,154,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,155,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,156,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,157,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,158,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[61,159,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,160,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,161,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,162,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,163,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,164,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,165,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,166,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,167,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,168,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,169,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,170,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,171,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,172,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,173,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,174,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,175,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,176,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,177,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,178,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[61,179,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,180,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,181,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,182,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,183,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,184,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,185,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,186,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,187,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,188,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,189,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,190,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,191,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,192,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,193,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,194,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,195,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,196,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,197,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,198,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[61,199,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,0,0.62,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,1,0.63,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,2,0.64,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,3,0.65,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,4,0.66,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,5,0.67,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,6,0.68,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,7,0.69,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,8,0.7,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,9,0.71,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,10,0.72,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,11,0.73,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,12,0.74,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,13,0.75,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,14,0.76,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,15,0.77,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,16,0.78,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,17,0.79,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,18,0.8,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[62,19,0.81,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,20,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,21,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,22,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,23,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,24,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,25,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,26,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,27,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,28,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,29,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,30,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,31,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,32,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,33,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,34,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,35,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,36,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,37,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,38,1,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[62,39,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,40,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,41,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,42,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,43,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,44,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,45,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,46,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,47,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,48,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,49,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,50,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,51,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,52,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,53,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,54,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,55,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,56,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,57,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,58,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[62,59,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,60,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,61,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,62,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,63,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,64,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,65,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,66,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,67,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,68,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,69,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,70,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,71,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,72,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,73,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,74,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,75,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,76,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,77,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,78,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[62,79,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,80,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,81,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,82,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,83,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,84,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,85,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,86,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,87,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,88,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,89,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,90,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,91,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,92,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,93,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,94,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,95,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,96,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,97,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,98,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[62,99,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,100,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,101,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,102,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,103,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,104,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,105,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,106,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,107,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,108,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,109,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,110,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,111,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,112,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,113,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,114,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,115,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,116,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,117,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,118,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[62,119,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,120,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,121,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,122,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,123,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,124,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,125,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,126,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,127,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,128,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,129,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,130,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,131,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,132,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,133,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,134,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,135,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,136,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,137,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,138,2,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[62,139,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,140,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,141,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,142,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,143,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,144,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,145,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,146,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,147,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,148,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,149,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,150,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,151,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,152,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,153,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,154,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,155,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,156,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,157,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,158,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[62,159,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,160,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,161,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,162,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,163,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,164,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,165,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,166,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,167,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,168,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,169,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,170,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,171,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,172,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,173,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,174,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,175,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,176,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,177,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,178,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[62,179,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,180,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,181,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,182,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,183,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,184,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,185,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,186,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,187,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,188,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,189,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,190,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,191,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,192,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,193,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,194,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,195,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,196,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,197,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,198,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[62,199,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,0,0.63,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,1,0.64,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,2,0.65,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,3,0.66,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,4,0.67,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,5,0.68,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,6,0.69,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,7,0.7,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,8,0.71,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,9,0.72,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,10,0.73,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,11,0.74,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,12,0.75,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,13,0.76,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,14,0.77,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,15,0.78,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,16,0.79,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,17,0.8,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,18,0.81,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[63,19,0.82,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,20,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,21,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,22,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,23,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,24,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,25,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,26,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,27,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,28,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,29,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,30,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,31,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,32,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,33,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,34,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,35,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,36,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,37,1,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,38,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[63,39,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,40,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,41,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,42,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,43,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,44,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,45,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,46,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,47,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,48,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,49,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,50,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,51,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,52,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,53,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,54,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,55,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,56,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,57,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,58,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[63,59,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,60,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,61,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,62,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,63,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,64,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,65,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,66,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,67,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,68,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,69,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,70,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,71,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,72,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,73,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,74,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,75,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,76,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,77,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,78,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[63,79,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,80,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,81,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,82,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,83,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,84,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,85,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,86,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,87,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,88,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,89,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,90,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,91,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,92,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,93,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,94,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,95,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,96,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,97,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,98,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[63,99,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,100,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,101,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,102,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,103,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,104,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,105,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,106,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,107,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,108,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,109,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,110,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,111,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,112,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,113,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,114,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,115,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,116,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,117,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,118,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[63,119,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,120,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,121,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,122,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,123,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,124,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,125,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,126,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,127,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,128,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,129,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,130,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,131,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,132,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,133,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,134,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,135,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,136,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,137,2,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,138,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[63,139,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,140,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,141,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,142,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,143,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,144,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,145,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,146,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,147,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,148,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,149,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,150,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,151,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,152,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,153,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,154,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,155,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,156,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,157,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,158,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[63,159,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,160,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,161,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,162,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,163,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,164,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,165,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,166,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,167,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,168,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,169,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,170,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,171,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,172,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,173,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,174,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,175,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,176,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,177,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,178,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[63,179,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,180,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,181,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,182,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,183,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,184,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,185,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,186,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,187,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,188,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,189,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,190,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,191,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,192,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,193,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,194,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,195,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,196,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,197,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,198,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[63,199,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,0,0.64,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,1,0.65,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,2,0.66,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,3,0.67,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,4,0.68,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,5,0.69,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,6,0.7,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,7,0.71,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,8,0.72,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,9,0.73,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,10,0.74,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,11,0.75,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,12,0.76,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,13,0.77,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,14,0.78,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,15,0.79,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,16,0.8,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,17,0.81,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,18,0.82,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[64,19,0.83,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,20,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,21,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,22,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,23,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,24,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,25,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,26,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,27,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,28,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,29,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,30,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,31,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,32,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,33,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,34,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,35,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,36,1,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,37,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,38,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[64,39,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,40,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,41,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,42,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,43,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,44,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,45,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,46,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,47,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,48,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,49,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,50,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,51,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,52,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,53,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,54,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,55,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,56,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,57,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,58,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[64,59,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,60,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,61,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,62,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,63,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,64,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,65,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,66,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,67,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,68,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,69,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,70,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,71,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,72,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,73,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,74,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,75,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,76,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,77,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,78,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[64,79,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,80,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,81,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,82,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,83,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,84,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,85,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,86,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,87,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,88,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,89,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,90,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,91,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,92,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,93,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,94,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,95,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,96,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,97,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,98,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[64,99,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,100,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,101,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,102,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,103,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,104,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,105,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,106,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,107,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,108,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,109,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,110,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,111,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,112,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,113,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,114,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,115,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,116,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,117,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,118,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[64,119,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,120,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,121,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,122,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,123,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,124,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,125,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,126,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,127,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,128,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,129,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,130,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,131,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,132,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,133,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,134,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,135,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,136,2,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,137,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,138,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[64,139,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,140,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,141,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,142,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,143,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,144,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,145,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,146,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,147,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,148,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,149,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,150,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,151,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,152,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,153,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,154,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,155,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,156,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,157,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,158,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[64,159,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,160,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,161,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,162,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,163,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,164,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,165,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,166,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,167,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,168,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,169,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,170,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,171,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,172,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,173,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,174,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,175,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,176,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,177,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,178,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[64,179,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,180,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,181,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,182,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,183,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,184,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,185,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,186,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,187,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,188,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,189,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,190,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,191,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,192,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,193,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,194,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,195,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,196,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,197,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,198,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[64,199,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,0,0.65,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,1,0.66,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,2,0.67,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,3,0.68,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,4,0.69,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,5,0.7,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,6,0.71,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,7,0.72,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,8,0.73,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,9,0.74,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,10,0.75,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,11,0.76,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,12,0.77,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,13,0.78,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,14,0.79,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,15,0.8,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,16,0.81,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,17,0.82,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,18,0.83,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[65,19,0.84,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,20,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,21,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,22,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,23,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,24,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,25,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,26,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,27,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,28,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,29,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,30,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,31,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,32,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,33,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,34,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,35,1,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,36,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,37,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,38,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[65,39,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,40,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,41,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,42,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,43,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,44,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,45,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,46,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,47,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,48,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,49,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,50,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,51,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,52,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,53,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,54,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,55,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,56,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,57,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,58,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[65,59,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,60,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,61,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,62,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,63,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,64,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,65,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,66,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,67,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,68,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,69,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,70,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,71,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,72,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,73,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,74,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,75,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,76,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,77,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,78,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[65,79,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,80,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,81,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,82,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,83,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,84,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,85,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,86,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,87,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,88,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,89,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,90,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,91,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,92,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,93,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,94,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,95,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,96,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,97,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,98,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[65,99,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,100,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,101,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,102,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,103,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,104,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,105,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,106,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,107,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,108,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,109,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,110,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,111,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,112,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,113,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,114,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,115,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,116,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,117,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,118,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[65,119,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,120,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,121,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,122,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,123,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,124,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,125,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,126,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,127,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,128,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,129,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,130,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,131,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,132,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,133,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,134,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,135,2,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,136,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,137,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,138,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[65,139,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,140,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,141,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,142,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,143,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,144,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,145,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,146,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,147,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,148,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,149,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,150,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,151,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,152,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,153,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,154,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,155,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,156,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,157,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,158,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[65,159,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,160,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,161,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,162,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,163,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,164,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,165,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,166,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,167,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,168,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,169,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,170,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,171,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,172,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,173,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,174,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,175,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,176,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,177,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,178,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[65,179,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,180,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,181,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,182,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,183,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,184,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,185,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,186,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,187,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,188,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,189,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,190,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,191,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,192,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,193,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,194,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,195,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,196,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,197,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,198,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[65,199,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,0,0.66,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,1,0.67,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,2,0.68,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,3,0.69,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,4,0.7,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,5,0.71,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,6,0.72,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,7,0.73,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,8,0.74,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,9,0.75,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,10,0.76,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,11,0.77,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,12,0.78,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,13,0.79,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,14,0.8,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,15,0.81,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,16,0.82,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,17,0.83,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,18,0.84,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[66,19,0.85,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,20,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,21,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,22,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,23,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,24,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,25,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,26,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,27,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,28,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,29,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,30,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,31,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,32,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,33,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,34,1,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,35,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,36,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,37,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,38,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[66,39,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,40,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,41,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,42,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,43,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,44,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,45,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,46,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,47,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,48,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,49,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,50,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,51,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,52,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,53,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,54,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,55,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,56,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,57,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,58,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[66,59,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,60,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,61,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,62,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,63,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,64,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,65,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,66,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,67,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,68,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,69,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,70,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,71,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,72,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,73,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,74,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,75,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,76,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,77,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,78,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[66,79,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,80,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,81,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,82,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,83,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,84,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,85,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,86,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,87,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,88,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,89,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,90,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,91,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,92,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,93,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,94,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,95,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,96,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,97,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,98,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[66,99,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,100,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,101,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,102,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,103,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,104,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,105,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,106,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,107,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,108,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,109,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,110,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,111,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,112,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,113,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,114,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,115,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,116,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,117,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,118,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[66,119,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,120,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,121,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,122,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,123,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,124,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,125,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,126,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,127,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,128,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,129,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,130,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,131,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,132,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,133,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,134,2,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,135,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,136,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,137,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,138,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[66,139,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,140,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,141,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,142,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,143,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,144,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,145,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,146,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,147,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,148,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,149,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,150,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,151,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,152,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,153,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,154,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,155,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,156,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,157,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,158,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[66,159,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,160,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,161,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,162,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,163,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,164,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,165,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,166,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,167,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,168,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,169,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,170,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,171,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,172,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,173,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,174,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,175,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,176,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,177,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,178,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[66,179,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,180,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,181,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,182,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,183,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,184,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,185,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,186,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,187,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,188,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,189,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,190,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,191,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,192,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,193,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,194,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,195,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,196,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,197,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,198,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[66,199,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,0,0.67,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,1,0.68,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,2,0.69,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,3,0.7,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,4,0.71,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,5,0.72,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,6,0.73,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,7,0.74,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,8,0.75,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,9,0.76,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,10,0.77,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,11,0.78,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,12,0.79,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,13,0.8,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,14,0.81,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,15,0.82,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,16,0.83,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,17,0.84,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,18,0.85,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[67,19,0.86,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,20,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,21,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,22,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,23,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,24,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,25,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,26,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,27,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,28,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,29,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,30,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,31,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,32,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,33,1,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,34,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,35,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,36,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,37,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,38,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[67,39,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,40,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,41,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,42,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,43,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,44,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,45,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,46,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,47,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,48,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,49,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,50,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,51,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,52,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,53,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,54,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,55,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,56,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,57,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,58,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[67,59,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,60,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,61,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,62,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,63,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,64,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,65,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,66,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,67,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,68,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,69,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,70,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,71,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,72,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,73,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,74,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,75,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,76,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,77,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,78,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[67,79,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,80,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,81,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,82,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,83,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,84,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,85,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,86,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,87,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,88,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,89,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,90,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,91,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,92,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,93,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,94,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,95,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,96,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,97,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,98,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[67,99,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,100,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,101,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,102,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,103,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,104,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,105,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,106,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,107,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,108,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,109,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,110,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,111,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,112,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,113,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,114,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,115,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,116,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,117,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,118,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[67,119,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,120,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,121,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,122,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,123,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,124,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,125,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,126,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,127,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,128,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,129,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,130,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,131,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,132,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,133,2,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,134,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,135,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,136,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,137,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,138,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[67,139,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,140,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,141,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,142,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,143,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,144,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,145,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,146,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,147,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,148,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,149,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,150,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,151,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,152,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,153,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,154,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,155,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,156,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,157,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,158,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[67,159,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,160,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,161,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,162,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,163,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,164,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,165,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,166,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,167,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,168,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,169,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,170,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,171,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,172,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,173,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,174,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,175,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,176,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,177,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,178,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[67,179,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,180,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,181,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,182,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,183,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,184,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,185,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,186,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,187,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,188,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,189,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,190,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,191,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,192,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,193,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,194,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,195,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,196,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,197,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,198,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[67,199,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,0,0.68,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,1,0.69,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,2,0.7,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,3,0.71,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,4,0.72,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,5,0.73,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,6,0.74,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,7,0.75,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,8,0.76,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,9,0.77,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,10,0.78,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,11,0.79,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,12,0.8,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,13,0.81,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,14,0.82,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,15,0.83,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,16,0.84,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,17,0.85,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,18,0.86,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[68,19,0.87,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,20,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,21,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,22,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,23,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,24,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,25,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,26,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,27,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,28,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,29,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,30,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,31,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,32,1,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,33,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,34,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,35,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,36,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,37,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,38,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[68,39,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,40,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,41,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,42,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,43,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,44,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,45,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,46,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,47,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,48,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,49,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,50,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,51,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,52,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,53,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,54,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,55,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,56,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,57,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,58,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[68,59,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,60,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,61,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,62,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,63,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,64,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,65,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,66,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,67,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,68,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,69,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,70,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,71,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,72,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,73,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,74,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,75,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,76,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,77,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,78,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[68,79,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,80,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,81,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,82,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,83,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,84,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,85,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,86,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,87,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,88,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,89,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,90,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,91,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,92,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,93,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,94,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,95,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,96,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,97,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,98,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[68,99,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,100,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,101,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,102,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,103,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,104,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,105,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,106,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,107,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,108,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,109,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,110,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,111,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,112,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,113,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,114,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,115,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,116,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,117,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,118,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[68,119,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,120,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,121,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,122,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,123,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,124,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,125,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,126,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,127,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,128,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,129,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,130,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,131,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,132,2,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,133,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,134,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,135,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,136,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,137,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,138,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[68,139,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,140,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,141,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,142,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,143,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,144,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,145,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,146,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,147,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,148,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,149,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,150,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,151,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,152,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,153,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,154,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,155,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,156,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,157,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,158,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[68,159,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,160,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,161,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,162,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,163,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,164,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,165,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,166,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,167,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,168,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,169,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,170,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,171,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,172,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,173,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,174,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,175,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,176,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,177,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,178,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[68,179,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,180,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,181,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,182,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,183,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,184,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,185,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,186,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,187,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,188,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,189,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,190,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,191,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,192,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,193,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,194,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,195,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,196,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,197,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,198,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[68,199,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,0,0.69,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,1,0.7,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,2,0.71,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,3,0.72,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,4,0.73,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,5,0.74,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,6,0.75,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,7,0.76,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,8,0.77,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,9,0.78,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,10,0.79,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,11,0.8,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,12,0.81,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,13,0.82,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,14,0.83,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,15,0.84,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,16,0.85,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,17,0.86,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,18,0.87,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[69,19,0.88,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,20,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,21,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,22,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,23,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,24,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,25,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,26,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,27,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,28,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,29,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,30,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,31,1,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,32,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,33,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,34,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,35,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,36,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,37,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,38,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[69,39,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,40,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,41,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,42,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,43,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,44,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,45,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,46,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,47,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,48,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,49,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,50,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,51,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,52,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,53,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,54,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,55,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,56,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,57,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,58,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[69,59,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,60,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,61,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,62,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,63,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,64,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,65,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,66,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,67,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,68,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,69,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,70,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,71,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,72,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,73,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,74,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,75,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,76,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,77,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,78,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[69,79,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,80,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,81,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,82,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,83,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,84,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,85,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,86,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,87,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,88,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,89,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,90,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,91,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,92,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,93,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,94,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,95,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,96,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,97,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,98,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[69,99,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,100,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,101,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,102,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,103,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,104,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,105,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,106,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,107,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,108,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,109,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,110,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,111,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,112,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,113,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,114,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,115,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,116,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,117,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,118,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[69,119,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,120,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,121,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,122,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,123,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,124,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,125,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,126,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,127,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,128,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,129,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,130,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,131,2,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,132,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,133,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,134,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,135,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,136,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,137,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,138,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[69,139,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,140,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,141,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,142,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,143,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,144,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,145,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,146,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,147,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,148,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,149,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,150,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,151,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,152,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,153,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,154,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,155,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,156,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,157,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,158,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[69,159,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,160,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,161,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,162,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,163,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,164,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,165,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,166,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,167,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,168,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,169,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,170,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,171,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,172,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,173,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,174,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,175,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,176,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,177,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,178,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[69,179,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,180,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,181,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,182,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,183,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,184,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,185,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,186,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,187,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,188,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,189,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,190,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,191,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,192,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,193,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,194,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,195,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,196,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,197,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,198,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[69,199,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,0,0.7,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,1,0.71,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,2,0.72,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,3,0.73,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,4,0.74,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,5,0.75,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,6,0.76,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,7,0.77,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,8,0.78,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,9,0.79,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,10,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,11,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,12,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,13,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,14,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,15,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,16,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,17,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,18,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[70,19,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,20,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,21,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,22,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,23,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,24,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,25,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,26,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,27,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,28,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,29,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,30,1,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,31,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,32,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,33,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,34,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,35,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,36,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,37,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,38,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[70,39,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,40,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,41,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,42,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,43,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,44,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,45,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,46,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,47,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,48,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,49,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,50,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,51,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,52,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,53,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,54,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,55,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,56,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,57,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,58,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[70,59,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,60,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,61,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,62,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,63,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,64,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,65,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,66,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,67,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,68,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,69,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,70,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,71,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,72,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,73,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,74,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,75,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,76,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,77,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,78,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[70,79,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,80,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,81,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,82,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,83,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,84,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,85,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,86,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,87,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,88,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,89,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,90,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,91,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,92,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,93,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,94,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,95,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,96,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,97,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,98,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[70,99,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,100,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,101,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,102,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,103,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,104,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,105,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,106,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,107,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,108,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,109,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,110,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,111,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,112,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,113,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,114,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,115,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,116,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,117,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,118,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[70,119,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,120,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,121,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,122,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,123,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,124,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,125,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,126,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,127,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,128,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,129,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,130,2,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,131,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,132,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,133,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,134,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,135,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,136,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,137,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,138,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[70,139,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,140,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,141,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,142,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,143,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,144,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,145,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,146,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,147,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,148,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,149,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,150,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,151,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,152,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,153,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,154,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,155,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,156,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,157,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,158,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[70,159,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,160,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,161,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,162,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,163,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,164,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,165,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,166,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,167,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,168,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,169,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,170,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,171,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,172,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,173,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,174,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,175,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,176,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,177,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,178,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[70,179,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,180,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,181,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,182,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,183,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,184,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,185,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,186,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,187,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,188,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,189,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,190,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,191,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,192,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,193,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,194,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,195,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,196,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,197,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,198,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[70,199,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,0,0.71,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,1,0.72,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,2,0.73,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,3,0.74,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,4,0.75,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,5,0.76,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,6,0.77,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,7,0.78,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,8,0.79,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,9,0.8,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,10,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,11,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,12,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,13,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,14,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,15,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,16,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,17,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,18,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[71,19,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,20,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,21,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,22,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,23,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,24,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,25,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,26,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,27,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,28,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,29,1,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,30,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,31,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,32,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,33,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,34,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,35,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,36,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,37,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,38,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[71,39,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,40,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,41,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,42,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,43,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,44,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,45,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,46,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,47,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,48,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,49,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,50,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,51,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,52,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,53,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,54,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,55,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,56,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,57,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,58,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[71,59,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,60,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,61,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,62,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,63,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,64,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,65,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,66,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,67,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,68,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,69,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,70,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,71,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,72,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,73,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,74,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,75,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,76,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,77,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,78,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[71,79,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,80,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,81,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,82,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,83,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,84,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,85,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,86,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,87,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,88,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,89,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,90,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,91,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,92,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,93,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,94,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,95,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,96,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,97,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,98,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[71,99,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,100,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,101,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,102,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,103,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,104,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,105,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,106,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,107,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,108,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,109,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,110,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,111,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,112,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,113,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,114,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,115,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,116,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,117,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,118,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[71,119,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,120,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,121,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,122,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,123,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,124,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,125,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,126,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,127,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,128,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,129,2,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,130,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,131,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,132,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,133,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,134,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,135,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,136,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,137,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,138,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[71,139,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,140,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,141,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,142,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,143,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,144,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,145,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,146,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,147,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,148,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,149,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,150,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,151,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,152,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,153,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,154,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,155,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,156,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,157,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,158,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[71,159,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,160,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,161,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,162,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,163,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,164,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,165,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,166,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,167,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,168,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,169,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,170,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,171,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,172,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,173,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,174,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,175,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,176,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,177,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,178,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[71,179,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,180,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,181,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,182,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,183,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,184,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,185,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,186,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,187,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,188,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,189,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,190,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,191,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,192,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,193,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,194,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,195,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,196,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,197,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,198,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[71,199,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,0,0.72,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,1,0.73,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,2,0.74,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,3,0.75,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,4,0.76,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,5,0.77,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,6,0.78,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,7,0.79,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,8,0.8,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,9,0.81,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,10,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,11,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,12,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,13,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,14,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,15,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,16,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,17,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,18,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[72,19,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,20,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,21,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,22,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,23,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,24,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,25,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,26,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,27,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,28,1,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,29,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,30,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,31,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,32,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,33,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,34,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,35,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,36,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,37,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,38,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[72,39,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,40,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,41,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,42,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,43,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,44,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,45,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,46,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,47,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,48,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,49,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,50,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,51,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,52,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,53,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,54,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,55,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,56,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,57,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,58,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[72,59,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,60,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,61,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,62,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,63,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,64,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,65,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,66,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,67,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,68,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,69,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,70,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,71,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,72,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,73,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,74,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,75,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,76,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,77,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,78,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[72,79,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,80,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,81,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,82,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,83,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,84,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,85,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,86,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,87,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,88,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,89,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,90,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,91,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,92,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,93,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,94,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,95,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,96,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,97,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,98,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[72,99,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,100,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,101,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,102,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,103,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,104,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,105,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,106,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,107,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,108,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,109,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,110,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,111,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,112,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,113,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,114,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,115,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,116,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,117,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,118,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[72,119,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,120,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,121,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,122,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,123,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,124,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,125,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,126,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,127,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,128,2,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,129,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,130,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,131,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,132,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,133,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,134,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,135,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,136,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,137,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,138,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[72,139,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,140,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,141,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,142,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,143,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,144,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,145,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,146,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,147,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,148,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,149,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,150,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,151,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,152,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,153,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,154,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,155,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,156,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,157,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,158,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[72,159,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,160,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,161,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,162,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,163,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,164,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,165,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,166,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,167,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,168,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,169,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,170,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,171,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,172,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,173,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,174,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,175,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,176,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,177,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,178,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[72,179,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,180,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,181,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,182,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,183,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,184,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,185,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,186,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,187,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,188,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,189,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,190,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,191,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,192,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,193,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,194,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,195,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,196,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,197,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,198,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[72,199,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,0,0.73,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,1,0.74,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,2,0.75,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,3,0.76,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,4,0.77,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,5,0.78,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,6,0.79,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,7,0.8,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,8,0.81,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,9,0.82,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,10,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,11,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,12,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,13,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,14,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,15,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,16,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,17,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,18,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[73,19,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,20,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,21,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,22,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,23,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,24,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,25,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,26,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,27,1,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,28,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,29,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,30,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,31,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,32,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,33,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,34,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,35,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,36,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,37,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,38,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[73,39,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,40,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,41,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,42,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,43,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,44,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,45,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,46,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,47,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,48,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,49,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,50,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,51,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,52,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,53,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,54,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,55,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,56,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,57,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,58,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[73,59,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,60,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,61,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,62,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,63,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,64,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,65,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,66,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,67,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,68,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,69,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,70,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,71,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,72,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,73,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,74,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,75,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,76,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,77,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,78,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[73,79,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,80,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,81,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,82,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,83,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,84,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,85,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,86,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,87,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,88,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,89,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,90,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,91,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,92,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,93,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,94,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,95,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,96,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,97,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,98,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[73,99,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,100,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,101,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,102,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,103,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,104,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,105,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,106,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,107,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,108,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,109,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,110,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,111,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,112,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,113,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,114,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,115,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,116,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,117,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,118,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[73,119,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,120,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,121,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,122,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,123,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,124,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,125,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,126,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,127,2,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,128,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,129,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,130,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,131,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,132,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,133,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,134,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,135,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,136,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,137,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,138,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[73,139,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,140,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,141,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,142,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,143,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,144,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,145,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,146,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,147,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,148,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,149,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,150,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,151,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,152,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,153,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,154,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,155,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,156,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,157,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,158,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[73,159,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,160,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,161,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,162,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,163,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,164,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,165,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,166,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,167,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,168,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,169,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,170,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,171,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,172,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,173,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,174,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,175,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,176,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,177,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,178,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[73,179,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,180,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,181,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,182,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,183,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,184,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,185,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,186,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,187,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,188,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,189,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,190,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,191,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,192,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,193,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,194,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,195,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,196,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,197,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,198,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[73,199,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,0,0.74,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,1,0.75,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,2,0.76,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,3,0.77,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,4,0.78,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,5,0.79,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,6,0.8,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,7,0.81,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,8,0.82,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,9,0.83,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,10,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,11,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,12,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,13,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,14,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,15,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,16,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,17,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,18,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[74,19,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,20,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,21,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,22,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,23,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,24,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,25,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,26,1,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,27,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,28,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,29,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,30,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,31,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,32,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,33,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,34,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,35,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,36,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,37,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,38,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[74,39,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,40,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,41,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,42,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,43,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,44,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,45,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,46,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,47,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,48,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,49,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,50,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,51,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,52,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,53,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,54,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,55,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,56,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,57,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,58,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[74,59,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,60,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,61,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,62,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,63,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,64,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,65,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,66,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,67,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,68,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,69,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,70,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,71,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,72,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,73,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,74,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,75,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,76,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,77,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,78,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[74,79,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,80,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,81,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,82,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,83,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,84,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,85,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,86,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,87,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,88,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,89,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,90,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,91,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,92,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,93,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,94,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,95,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,96,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,97,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,98,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[74,99,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,100,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,101,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,102,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,103,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,104,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,105,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,106,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,107,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,108,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,109,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,110,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,111,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,112,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,113,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,114,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,115,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,116,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,117,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,118,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[74,119,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,120,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,121,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,122,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,123,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,124,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,125,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,126,2,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,127,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,128,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,129,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,130,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,131,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,132,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,133,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,134,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,135,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,136,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,137,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,138,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[74,139,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,140,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,141,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,142,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,143,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,144,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,145,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,146,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,147,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,148,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,149,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,150,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,151,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,152,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,153,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,154,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,155,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,156,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,157,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,158,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[74,159,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,160,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,161,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,162,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,163,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,164,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,165,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,166,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,167,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,168,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,169,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,170,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,171,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,172,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,173,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,174,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,175,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,176,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,177,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,178,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[74,179,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,180,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,181,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,182,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,183,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,184,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,185,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,186,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,187,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,188,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,189,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,190,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,191,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,192,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,193,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,194,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,195,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,196,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,197,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,198,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[74,199,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,0,0.75,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,1,0.76,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,2,0.77,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,3,0.78,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,4,0.79,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,5,0.8,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,6,0.81,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,7,0.82,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,8,0.83,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,9,0.84,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,10,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,11,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,12,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,13,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,14,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,15,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,16,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,17,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,18,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[75,19,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,20,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,21,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,22,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,23,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,24,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,25,1,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,26,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,27,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,28,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,29,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,30,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,31,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,32,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,33,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,34,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,35,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,36,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,37,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,38,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[75,39,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,40,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,41,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,42,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,43,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,44,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,45,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,46,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,47,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,48,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,49,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,50,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,51,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,52,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,53,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,54,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,55,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,56,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,57,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,58,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[75,59,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,60,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,61,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,62,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,63,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,64,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,65,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,66,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,67,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,68,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,69,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,70,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,71,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,72,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,73,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,74,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,75,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,76,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,77,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,78,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[75,79,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,80,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,81,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,82,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,83,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,84,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,85,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,86,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,87,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,88,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,89,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,90,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,91,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,92,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,93,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,94,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,95,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,96,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,97,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,98,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[75,99,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,100,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,101,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,102,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,103,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,104,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,105,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,106,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,107,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,108,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,109,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,110,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,111,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,112,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,113,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,114,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,115,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,116,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,117,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,118,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[75,119,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,120,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,121,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,122,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,123,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,124,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,125,2,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,126,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,127,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,128,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,129,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,130,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,131,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,132,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,133,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,134,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,135,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,136,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,137,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,138,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[75,139,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,140,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,141,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,142,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,143,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,144,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,145,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,146,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,147,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,148,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,149,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,150,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,151,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,152,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,153,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,154,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,155,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,156,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,157,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,158,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[75,159,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,160,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,161,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,162,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,163,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,164,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,165,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,166,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,167,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,168,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,169,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,170,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,171,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,172,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,173,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,174,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,175,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,176,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,177,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,178,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[75,179,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,180,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,181,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,182,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,183,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,184,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,185,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,186,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,187,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,188,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,189,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,190,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,191,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,192,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,193,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,194,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,195,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,196,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,197,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,198,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[75,199,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,0,0.76,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,1,0.77,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,2,0.78,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,3,0.79,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,4,0.8,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,5,0.81,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,6,0.82,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,7,0.83,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,8,0.84,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,9,0.85,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,10,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,11,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,12,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,13,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,14,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,15,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,16,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,17,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,18,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[76,19,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,20,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,21,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,22,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,23,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,24,1,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,25,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,26,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,27,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,28,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,29,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,30,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,31,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,32,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,33,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,34,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,35,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,36,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,37,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,38,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[76,39,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,40,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,41,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,42,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,43,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,44,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,45,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,46,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,47,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,48,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,49,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,50,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,51,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,52,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,53,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,54,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,55,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,56,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,57,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,58,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[76,59,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,60,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,61,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,62,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,63,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,64,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,65,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,66,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,67,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,68,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,69,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,70,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,71,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,72,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,73,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,74,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,75,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,76,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,77,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,78,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[76,79,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,80,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,81,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,82,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,83,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,84,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,85,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,86,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,87,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,88,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,89,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,90,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,91,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,92,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,93,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,94,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,95,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,96,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,97,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,98,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[76,99,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,100,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,101,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,102,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,103,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,104,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,105,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,106,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,107,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,108,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,109,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,110,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,111,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,112,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,113,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,114,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,115,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,116,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,117,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,118,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[76,119,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,120,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,121,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,122,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,123,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,124,2,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,125,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,126,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,127,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,128,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,129,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,130,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,131,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,132,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,133,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,134,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,135,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,136,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,137,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,138,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[76,139,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,140,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,141,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,142,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,143,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,144,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,145,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,146,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,147,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,148,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,149,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,150,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,151,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,152,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,153,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,154,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,155,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,156,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,157,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,158,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[76,159,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,160,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,161,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,162,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,163,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,164,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,165,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,166,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,167,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,168,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,169,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,170,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,171,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,172,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,173,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,174,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,175,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,176,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,177,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,178,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[76,179,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,180,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,181,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,182,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,183,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,184,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,185,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,186,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,187,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,188,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,189,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,190,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,191,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,192,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,193,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,194,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,195,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,196,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,197,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,198,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[76,199,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,0,0.77,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,1,0.78,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,2,0.79,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,3,0.8,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,4,0.81,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,5,0.82,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,6,0.83,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,7,0.84,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,8,0.85,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,9,0.86,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,10,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,11,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,12,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,13,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,14,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,15,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,16,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,17,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,18,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[77,19,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,20,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,21,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,22,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,23,1,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,24,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,25,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,26,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,27,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,28,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,29,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,30,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,31,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,32,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,33,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,34,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,35,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,36,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,37,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,38,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[77,39,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,40,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,41,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,42,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,43,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,44,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,45,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,46,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,47,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,48,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,49,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,50,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,51,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,52,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,53,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,54,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,55,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,56,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,57,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,58,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[77,59,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,60,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,61,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,62,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,63,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,64,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,65,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,66,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,67,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,68,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,69,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,70,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,71,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,72,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,73,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,74,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,75,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,76,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,77,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,78,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[77,79,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,80,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,81,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,82,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,83,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,84,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,85,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,86,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,87,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,88,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,89,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,90,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,91,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,92,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,93,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,94,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,95,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,96,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,97,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,98,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[77,99,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,100,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,101,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,102,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,103,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,104,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,105,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,106,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,107,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,108,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,109,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,110,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,111,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,112,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,113,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,114,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,115,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,116,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,117,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,118,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[77,119,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,120,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,121,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,122,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,123,2,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,124,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,125,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,126,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,127,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,128,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,129,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,130,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,131,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,132,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,133,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,134,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,135,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,136,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,137,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,138,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[77,139,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,140,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,141,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,142,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,143,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,144,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,145,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,146,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,147,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,148,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,149,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,150,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,151,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,152,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,153,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,154,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,155,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,156,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,157,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,158,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[77,159,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,160,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,161,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,162,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,163,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,164,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,165,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,166,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,167,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,168,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,169,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,170,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,171,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,172,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,173,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,174,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,175,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,176,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,177,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,178,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[77,179,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,180,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,181,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,182,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,183,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,184,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,185,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,186,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,187,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,188,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,189,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,190,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,191,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,192,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,193,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,194,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,195,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,196,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,197,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,198,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[77,199,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,0,0.78,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,1,0.79,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,2,0.8,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,3,0.81,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,4,0.82,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,5,0.83,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,6,0.84,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,7,0.85,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,8,0.86,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,9,0.87,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,10,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,11,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,12,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,13,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,14,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,15,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,16,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,17,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,18,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[78,19,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,20,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,21,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,22,1,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,23,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,24,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,25,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,26,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,27,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,28,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,29,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,30,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,31,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,32,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,33,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,34,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,35,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,36,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,37,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,38,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[78,39,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,40,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,41,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,42,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,43,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,44,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,45,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,46,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,47,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,48,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,49,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,50,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,51,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,52,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,53,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,54,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,55,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,56,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,57,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,58,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[78,59,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,60,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,61,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,62,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,63,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,64,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,65,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,66,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,67,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,68,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,69,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,70,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,71,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,72,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,73,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,74,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,75,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,76,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,77,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,78,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[78,79,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,80,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,81,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,82,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,83,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,84,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,85,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,86,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,87,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,88,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,89,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,90,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,91,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,92,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,93,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,94,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,95,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,96,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,97,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,98,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[78,99,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,100,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,101,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,102,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,103,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,104,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,105,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,106,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,107,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,108,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,109,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,110,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,111,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,112,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,113,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,114,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,115,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,116,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,117,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,118,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[78,119,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,120,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,121,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,122,2,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,123,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,124,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,125,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,126,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,127,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,128,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,129,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,130,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,131,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,132,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,133,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,134,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,135,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,136,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,137,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,138,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[78,139,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,140,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,141,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,142,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,143,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,144,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,145,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,146,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,147,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,148,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,149,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,150,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,151,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,152,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,153,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,154,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,155,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,156,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,157,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,158,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[78,159,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,160,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,161,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,162,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,163,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,164,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,165,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,166,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,167,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,168,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,169,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,170,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,171,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,172,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,173,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,174,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,175,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,176,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,177,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,178,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[78,179,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,180,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,181,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,182,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,183,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,184,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,185,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,186,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,187,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,188,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,189,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,190,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,191,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,192,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,193,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,194,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,195,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,196,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,197,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,198,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[78,199,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,0,0.79,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,1,0.8,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,2,0.81,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,3,0.82,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,4,0.83,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,5,0.84,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,6,0.85,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,7,0.86,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,8,0.87,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,9,0.88,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,10,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,11,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,12,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,13,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,14,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,15,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,16,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,17,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,18,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 60, pc_makepoint(1, ARRAY[79,19,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,20,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,21,1,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,22,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,23,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,24,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,25,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,26,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,27,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,28,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,29,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,30,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,31,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,32,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,33,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,34,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,35,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,36,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,37,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,38,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 61, pc_makepoint(1, ARRAY[79,39,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,40,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,41,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,42,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,43,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,44,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,45,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,46,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,47,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,48,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,49,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,50,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,51,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,52,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,53,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,54,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,55,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,56,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,57,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,58,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 62, pc_makepoint(1, ARRAY[79,59,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,60,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,61,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,62,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,63,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,64,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,65,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,66,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,67,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,68,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,69,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,70,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,71,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,72,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,73,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,74,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,75,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,76,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,77,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,78,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 63, pc_makepoint(1, ARRAY[79,79,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,80,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,81,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,82,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,83,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,84,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,85,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,86,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,87,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,88,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,89,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,90,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,91,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,92,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,93,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,94,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,95,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,96,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,97,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,98,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 64, pc_makepoint(1, ARRAY[79,99,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,100,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,101,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,102,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,103,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,104,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,105,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,106,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,107,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,108,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,109,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,110,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,111,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,112,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,113,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,114,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,115,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,116,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,117,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,118,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 65, pc_makepoint(1, ARRAY[79,119,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,120,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,121,2,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,122,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,123,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,124,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,125,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,126,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,127,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,128,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,129,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,130,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,131,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,132,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,133,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,134,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,135,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,136,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,137,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,138,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 66, pc_makepoint(1, ARRAY[79,139,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,140,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,141,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,142,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,143,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,144,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,145,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,146,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,147,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,148,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,149,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,150,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,151,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,152,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,153,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,154,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,155,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,156,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,157,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,158,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 67, pc_makepoint(1, ARRAY[79,159,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,160,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,161,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,162,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,163,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,164,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,165,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,166,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,167,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,168,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,169,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,170,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,171,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,172,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,173,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,174,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,175,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,176,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,177,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,178,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 68, pc_makepoint(1, ARRAY[79,179,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,180,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,181,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,182,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,183,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,184,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,185,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,186,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,187,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,188,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,189,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,190,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,191,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,192,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,193,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,194,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,195,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,196,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,197,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,198,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 69, pc_makepoint(1, ARRAY[79,199,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,0,0.8,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,1,0.81,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,2,0.82,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,3,0.83,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,4,0.84,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,5,0.85,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,6,0.86,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,7,0.87,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,8,0.88,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,9,0.89,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,10,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,11,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,12,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,13,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,14,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,15,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,16,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,17,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,18,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[80,19,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,20,1,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,21,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,22,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,23,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,24,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,25,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,26,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,27,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,28,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,29,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,30,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,31,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,32,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,33,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,34,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,35,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,36,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,37,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,38,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[80,39,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,40,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,41,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,42,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,43,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,44,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,45,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,46,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,47,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,48,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,49,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,50,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,51,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,52,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,53,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,54,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,55,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,56,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,57,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,58,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[80,59,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,60,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,61,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,62,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,63,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,64,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,65,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,66,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,67,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,68,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,69,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,70,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,71,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,72,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,73,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,74,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,75,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,76,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,77,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,78,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[80,79,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,80,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,81,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,82,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,83,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,84,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,85,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,86,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,87,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,88,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,89,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,90,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,91,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,92,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,93,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,94,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,95,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,96,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,97,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,98,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[80,99,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,100,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,101,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,102,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,103,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,104,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,105,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,106,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,107,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,108,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,109,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,110,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,111,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,112,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,113,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,114,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,115,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,116,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,117,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,118,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[80,119,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,120,2,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,121,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,122,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,123,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,124,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,125,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,126,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,127,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,128,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,129,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,130,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,131,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,132,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,133,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,134,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,135,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,136,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,137,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,138,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[80,139,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,140,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,141,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,142,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,143,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,144,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,145,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,146,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,147,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,148,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,149,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,150,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,151,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,152,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,153,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,154,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,155,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,156,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,157,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,158,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[80,159,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,160,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,161,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,162,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,163,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,164,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,165,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,166,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,167,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,168,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,169,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,170,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,171,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,172,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,173,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,174,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,175,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,176,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,177,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,178,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[80,179,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,180,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,181,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,182,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,183,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,184,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,185,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,186,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,187,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,188,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,189,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,190,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,191,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,192,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,193,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,194,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,195,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,196,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,197,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,198,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[80,199,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,0,0.81,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,1,0.82,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,2,0.83,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,3,0.84,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,4,0.85,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,5,0.86,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,6,0.87,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,7,0.88,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,8,0.89,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,9,0.9,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,10,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,11,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,12,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,13,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,14,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,15,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,16,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,17,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,18,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[81,19,1,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,20,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,21,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,22,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,23,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,24,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,25,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,26,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,27,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,28,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,29,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,30,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,31,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,32,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,33,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,34,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,35,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,36,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,37,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,38,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[81,39,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,40,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,41,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,42,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,43,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,44,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,45,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,46,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,47,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,48,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,49,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,50,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,51,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,52,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,53,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,54,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,55,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,56,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,57,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,58,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[81,59,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,60,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,61,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,62,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,63,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,64,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,65,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,66,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,67,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,68,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,69,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,70,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,71,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,72,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,73,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,74,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,75,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,76,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,77,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,78,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[81,79,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,80,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,81,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,82,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,83,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,84,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,85,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,86,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,87,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,88,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,89,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,90,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,91,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,92,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,93,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,94,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,95,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,96,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,97,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,98,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[81,99,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,100,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,101,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,102,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,103,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,104,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,105,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,106,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,107,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,108,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,109,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,110,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,111,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,112,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,113,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,114,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,115,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,116,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,117,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,118,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[81,119,2,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,120,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,121,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,122,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,123,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,124,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,125,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,126,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,127,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,128,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,129,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,130,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,131,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,132,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,133,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,134,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,135,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,136,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,137,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,138,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[81,139,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,140,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,141,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,142,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,143,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,144,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,145,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,146,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,147,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,148,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,149,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,150,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,151,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,152,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,153,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,154,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,155,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,156,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,157,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,158,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[81,159,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,160,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,161,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,162,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,163,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,164,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,165,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,166,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,167,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,168,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,169,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,170,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,171,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,172,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,173,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,174,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,175,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,176,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,177,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,178,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[81,179,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,180,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,181,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,182,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,183,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,184,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,185,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,186,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,187,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,188,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,189,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,190,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,191,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,192,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,193,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,194,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,195,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,196,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,197,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,198,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[81,199,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,0,0.82,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,1,0.83,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,2,0.84,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,3,0.85,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,4,0.86,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,5,0.87,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,6,0.88,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,7,0.89,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,8,0.9,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,9,0.91,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,10,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,11,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,12,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,13,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,14,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,15,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,16,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,17,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,18,1,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[82,19,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,20,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,21,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,22,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,23,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,24,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,25,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,26,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,27,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,28,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,29,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,30,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,31,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,32,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,33,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,34,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,35,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,36,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,37,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,38,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[82,39,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,40,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,41,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,42,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,43,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,44,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,45,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,46,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,47,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,48,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,49,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,50,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,51,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,52,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,53,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,54,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,55,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,56,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,57,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,58,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[82,59,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,60,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,61,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,62,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,63,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,64,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,65,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,66,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,67,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,68,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,69,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,70,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,71,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,72,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,73,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,74,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,75,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,76,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,77,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,78,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[82,79,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,80,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,81,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,82,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,83,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,84,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,85,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,86,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,87,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,88,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,89,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,90,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,91,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,92,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,93,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,94,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,95,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,96,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,97,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,98,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[82,99,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,100,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,101,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,102,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,103,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,104,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,105,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,106,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,107,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,108,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,109,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,110,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,111,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,112,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,113,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,114,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,115,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,116,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,117,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,118,2,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[82,119,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,120,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,121,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,122,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,123,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,124,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,125,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,126,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,127,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,128,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,129,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,130,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,131,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,132,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,133,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,134,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,135,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,136,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,137,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,138,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[82,139,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,140,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,141,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,142,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,143,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,144,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,145,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,146,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,147,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,148,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,149,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,150,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,151,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,152,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,153,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,154,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,155,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,156,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,157,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,158,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[82,159,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,160,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,161,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,162,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,163,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,164,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,165,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,166,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,167,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,168,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,169,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,170,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,171,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,172,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,173,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,174,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,175,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,176,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,177,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,178,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[82,179,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,180,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,181,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,182,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,183,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,184,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,185,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,186,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,187,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,188,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,189,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,190,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,191,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,192,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,193,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,194,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,195,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,196,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,197,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,198,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[82,199,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,0,0.83,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,1,0.84,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,2,0.85,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,3,0.86,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,4,0.87,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,5,0.88,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,6,0.89,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,7,0.9,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,8,0.91,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,9,0.92,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,10,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,11,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,12,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,13,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,14,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,15,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,16,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,17,1,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,18,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[83,19,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,20,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,21,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,22,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,23,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,24,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,25,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,26,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,27,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,28,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,29,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,30,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,31,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,32,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,33,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,34,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,35,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,36,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,37,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,38,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[83,39,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,40,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,41,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,42,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,43,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,44,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,45,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,46,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,47,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,48,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,49,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,50,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,51,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,52,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,53,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,54,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,55,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,56,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,57,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,58,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[83,59,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,60,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,61,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,62,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,63,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,64,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,65,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,66,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,67,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,68,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,69,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,70,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,71,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,72,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,73,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,74,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,75,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,76,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,77,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,78,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[83,79,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,80,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,81,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,82,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,83,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,84,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,85,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,86,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,87,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,88,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,89,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,90,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,91,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,92,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,93,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,94,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,95,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,96,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,97,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,98,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[83,99,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,100,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,101,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,102,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,103,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,104,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,105,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,106,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,107,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,108,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,109,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,110,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,111,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,112,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,113,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,114,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,115,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,116,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,117,2,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,118,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[83,119,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,120,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,121,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,122,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,123,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,124,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,125,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,126,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,127,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,128,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,129,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,130,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,131,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,132,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,133,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,134,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,135,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,136,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,137,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,138,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[83,139,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,140,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,141,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,142,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,143,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,144,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,145,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,146,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,147,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,148,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,149,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,150,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,151,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,152,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,153,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,154,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,155,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,156,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,157,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,158,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[83,159,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,160,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,161,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,162,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,163,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,164,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,165,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,166,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,167,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,168,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,169,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,170,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,171,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,172,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,173,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,174,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,175,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,176,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,177,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,178,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[83,179,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,180,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,181,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,182,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,183,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,184,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,185,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,186,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,187,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,188,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,189,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,190,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,191,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,192,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,193,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,194,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,195,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,196,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,197,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,198,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[83,199,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,0,0.84,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,1,0.85,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,2,0.86,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,3,0.87,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,4,0.88,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,5,0.89,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,6,0.9,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,7,0.91,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,8,0.92,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,9,0.93,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,10,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,11,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,12,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,13,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,14,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,15,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,16,1,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,17,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,18,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[84,19,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,20,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,21,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,22,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,23,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,24,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,25,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,26,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,27,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,28,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,29,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,30,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,31,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,32,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,33,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,34,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,35,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,36,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,37,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,38,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[84,39,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,40,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,41,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,42,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,43,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,44,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,45,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,46,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,47,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,48,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,49,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,50,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,51,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,52,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,53,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,54,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,55,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,56,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,57,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,58,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[84,59,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,60,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,61,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,62,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,63,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,64,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,65,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,66,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,67,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,68,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,69,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,70,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,71,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,72,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,73,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,74,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,75,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,76,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,77,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,78,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[84,79,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,80,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,81,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,82,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,83,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,84,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,85,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,86,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,87,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,88,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,89,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,90,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,91,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,92,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,93,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,94,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,95,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,96,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,97,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,98,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[84,99,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,100,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,101,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,102,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,103,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,104,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,105,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,106,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,107,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,108,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,109,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,110,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,111,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,112,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,113,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,114,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,115,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,116,2,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,117,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,118,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[84,119,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,120,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,121,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,122,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,123,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,124,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,125,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,126,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,127,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,128,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,129,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,130,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,131,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,132,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,133,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,134,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,135,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,136,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,137,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,138,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[84,139,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,140,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,141,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,142,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,143,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,144,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,145,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,146,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,147,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,148,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,149,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,150,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,151,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,152,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,153,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,154,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,155,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,156,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,157,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,158,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[84,159,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,160,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,161,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,162,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,163,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,164,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,165,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,166,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,167,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,168,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,169,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,170,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,171,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,172,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,173,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,174,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,175,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,176,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,177,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,178,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[84,179,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,180,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,181,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,182,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,183,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,184,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,185,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,186,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,187,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,188,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,189,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,190,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,191,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,192,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,193,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,194,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,195,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,196,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,197,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,198,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[84,199,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,0,0.85,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,1,0.86,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,2,0.87,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,3,0.88,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,4,0.89,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,5,0.9,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,6,0.91,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,7,0.92,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,8,0.93,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,9,0.94,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,10,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,11,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,12,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,13,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,14,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,15,1,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,16,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,17,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,18,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[85,19,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,20,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,21,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,22,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,23,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,24,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,25,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,26,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,27,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,28,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,29,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,30,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,31,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,32,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,33,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,34,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,35,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,36,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,37,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,38,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[85,39,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,40,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,41,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,42,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,43,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,44,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,45,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,46,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,47,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,48,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,49,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,50,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,51,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,52,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,53,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,54,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,55,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,56,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,57,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,58,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[85,59,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,60,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,61,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,62,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,63,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,64,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,65,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,66,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,67,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,68,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,69,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,70,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,71,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,72,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,73,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,74,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,75,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,76,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,77,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,78,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[85,79,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,80,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,81,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,82,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,83,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,84,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,85,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,86,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,87,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,88,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,89,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,90,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,91,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,92,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,93,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,94,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,95,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,96,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,97,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,98,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[85,99,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,100,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,101,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,102,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,103,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,104,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,105,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,106,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,107,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,108,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,109,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,110,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,111,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,112,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,113,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,114,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,115,2,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,116,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,117,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,118,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[85,119,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,120,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,121,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,122,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,123,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,124,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,125,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,126,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,127,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,128,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,129,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,130,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,131,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,132,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,133,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,134,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,135,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,136,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,137,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,138,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[85,139,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,140,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,141,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,142,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,143,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,144,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,145,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,146,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,147,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,148,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,149,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,150,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,151,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,152,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,153,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,154,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,155,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,156,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,157,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,158,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[85,159,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,160,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,161,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,162,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,163,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,164,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,165,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,166,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,167,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,168,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,169,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,170,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,171,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,172,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,173,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,174,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,175,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,176,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,177,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,178,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[85,179,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,180,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,181,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,182,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,183,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,184,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,185,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,186,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,187,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,188,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,189,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,190,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,191,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,192,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,193,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,194,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,195,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,196,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,197,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,198,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[85,199,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,0,0.86,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,1,0.87,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,2,0.88,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,3,0.89,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,4,0.9,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,5,0.91,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,6,0.92,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,7,0.93,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,8,0.94,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,9,0.95,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,10,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,11,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,12,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,13,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,14,1,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,15,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,16,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,17,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,18,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[86,19,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,20,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,21,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,22,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,23,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,24,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,25,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,26,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,27,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,28,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,29,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,30,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,31,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,32,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,33,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,34,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,35,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,36,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,37,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,38,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[86,39,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,40,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,41,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,42,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,43,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,44,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,45,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,46,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,47,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,48,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,49,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,50,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,51,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,52,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,53,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,54,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,55,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,56,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,57,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,58,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[86,59,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,60,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,61,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,62,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,63,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,64,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,65,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,66,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,67,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,68,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,69,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,70,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,71,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,72,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,73,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,74,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,75,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,76,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,77,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,78,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[86,79,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,80,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,81,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,82,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,83,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,84,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,85,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,86,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,87,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,88,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,89,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,90,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,91,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,92,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,93,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,94,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,95,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,96,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,97,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,98,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[86,99,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,100,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,101,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,102,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,103,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,104,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,105,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,106,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,107,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,108,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,109,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,110,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,111,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,112,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,113,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,114,2,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,115,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,116,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,117,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,118,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[86,119,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,120,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,121,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,122,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,123,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,124,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,125,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,126,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,127,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,128,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,129,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,130,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,131,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,132,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,133,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,134,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,135,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,136,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,137,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,138,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[86,139,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,140,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,141,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,142,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,143,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,144,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,145,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,146,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,147,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,148,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,149,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,150,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,151,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,152,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,153,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,154,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,155,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,156,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,157,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,158,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[86,159,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,160,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,161,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,162,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,163,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,164,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,165,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,166,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,167,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,168,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,169,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,170,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,171,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,172,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,173,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,174,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,175,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,176,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,177,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,178,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[86,179,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,180,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,181,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,182,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,183,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,184,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,185,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,186,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,187,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,188,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,189,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,190,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,191,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,192,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,193,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,194,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,195,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,196,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,197,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,198,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[86,199,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,0,0.87,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,1,0.88,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,2,0.89,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,3,0.9,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,4,0.91,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,5,0.92,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,6,0.93,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,7,0.94,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,8,0.95,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,9,0.96,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,10,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,11,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,12,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,13,1,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,14,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,15,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,16,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,17,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,18,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[87,19,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,20,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,21,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,22,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,23,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,24,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,25,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,26,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,27,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,28,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,29,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,30,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,31,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,32,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,33,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,34,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,35,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,36,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,37,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,38,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[87,39,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,40,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,41,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,42,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,43,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,44,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,45,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,46,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,47,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,48,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,49,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,50,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,51,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,52,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,53,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,54,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,55,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,56,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,57,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,58,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[87,59,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,60,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,61,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,62,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,63,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,64,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,65,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,66,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,67,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,68,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,69,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,70,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,71,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,72,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,73,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,74,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,75,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,76,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,77,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,78,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[87,79,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,80,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,81,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,82,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,83,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,84,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,85,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,86,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,87,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,88,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,89,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,90,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,91,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,92,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,93,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,94,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,95,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,96,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,97,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,98,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[87,99,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,100,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,101,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,102,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,103,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,104,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,105,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,106,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,107,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,108,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,109,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,110,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,111,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,112,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,113,2,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,114,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,115,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,116,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,117,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,118,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[87,119,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,120,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,121,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,122,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,123,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,124,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,125,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,126,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,127,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,128,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,129,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,130,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,131,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,132,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,133,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,134,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,135,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,136,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,137,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,138,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[87,139,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,140,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,141,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,142,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,143,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,144,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,145,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,146,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,147,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,148,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,149,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,150,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,151,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,152,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,153,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,154,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,155,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,156,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,157,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,158,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[87,159,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,160,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,161,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,162,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,163,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,164,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,165,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,166,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,167,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,168,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,169,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,170,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,171,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,172,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,173,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,174,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,175,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,176,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,177,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,178,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[87,179,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,180,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,181,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,182,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,183,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,184,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,185,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,186,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,187,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,188,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,189,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,190,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,191,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,192,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,193,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,194,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,195,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,196,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,197,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,198,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[87,199,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,0,0.88,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,1,0.89,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,2,0.9,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,3,0.91,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,4,0.92,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,5,0.93,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,6,0.94,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,7,0.95,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,8,0.96,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,9,0.97,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,10,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,11,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,12,1,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,13,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,14,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,15,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,16,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,17,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,18,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[88,19,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,20,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,21,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,22,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,23,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,24,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,25,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,26,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,27,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,28,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,29,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,30,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,31,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,32,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,33,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,34,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,35,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,36,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,37,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,38,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[88,39,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,40,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,41,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,42,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,43,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,44,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,45,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,46,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,47,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,48,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,49,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,50,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,51,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,52,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,53,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,54,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,55,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,56,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,57,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,58,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[88,59,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,60,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,61,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,62,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,63,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,64,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,65,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,66,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,67,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,68,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,69,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,70,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,71,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,72,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,73,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,74,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,75,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,76,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,77,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,78,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[88,79,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,80,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,81,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,82,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,83,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,84,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,85,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,86,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,87,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,88,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,89,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,90,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,91,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,92,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,93,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,94,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,95,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,96,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,97,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,98,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[88,99,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,100,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,101,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,102,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,103,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,104,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,105,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,106,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,107,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,108,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,109,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,110,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,111,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,112,2,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,113,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,114,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,115,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,116,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,117,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,118,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[88,119,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,120,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,121,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,122,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,123,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,124,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,125,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,126,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,127,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,128,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,129,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,130,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,131,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,132,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,133,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,134,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,135,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,136,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,137,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,138,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[88,139,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,140,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,141,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,142,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,143,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,144,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,145,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,146,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,147,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,148,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,149,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,150,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,151,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,152,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,153,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,154,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,155,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,156,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,157,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,158,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[88,159,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,160,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,161,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,162,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,163,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,164,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,165,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,166,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,167,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,168,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,169,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,170,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,171,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,172,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,173,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,174,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,175,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,176,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,177,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,178,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[88,179,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,180,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,181,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,182,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,183,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,184,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,185,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,186,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,187,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,188,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,189,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,190,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,191,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,192,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,193,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,194,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,195,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,196,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,197,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,198,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[88,199,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,0,0.89,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,1,0.9,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,2,0.91,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,3,0.92,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,4,0.93,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,5,0.94,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,6,0.95,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,7,0.96,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,8,0.97,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,9,0.98,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,10,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,11,1,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,12,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,13,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,14,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,15,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,16,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,17,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,18,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[89,19,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,20,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,21,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,22,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,23,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,24,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,25,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,26,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,27,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,28,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,29,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,30,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,31,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,32,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,33,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,34,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,35,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,36,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,37,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,38,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[89,39,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,40,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,41,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,42,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,43,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,44,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,45,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,46,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,47,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,48,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,49,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,50,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,51,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,52,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,53,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,54,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,55,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,56,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,57,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,58,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[89,59,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,60,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,61,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,62,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,63,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,64,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,65,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,66,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,67,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,68,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,69,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,70,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,71,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,72,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,73,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,74,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,75,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,76,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,77,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,78,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[89,79,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,80,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,81,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,82,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,83,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,84,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,85,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,86,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,87,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,88,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,89,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,90,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,91,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,92,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,93,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,94,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,95,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,96,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,97,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,98,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[89,99,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,100,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,101,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,102,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,103,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,104,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,105,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,106,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,107,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,108,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,109,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,110,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,111,2,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,112,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,113,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,114,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,115,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,116,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,117,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,118,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[89,119,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,120,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,121,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,122,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,123,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,124,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,125,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,126,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,127,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,128,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,129,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,130,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,131,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,132,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,133,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,134,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,135,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,136,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,137,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,138,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[89,139,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,140,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,141,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,142,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,143,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,144,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,145,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,146,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,147,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,148,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,149,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,150,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,151,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,152,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,153,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,154,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,155,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,156,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,157,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,158,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[89,159,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,160,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,161,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,162,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,163,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,164,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,165,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,166,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,167,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,168,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,169,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,170,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,171,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,172,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,173,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,174,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,175,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,176,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,177,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,178,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[89,179,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,180,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,181,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,182,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,183,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,184,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,185,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,186,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,187,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,188,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,189,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,190,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,191,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,192,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,193,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,194,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,195,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,196,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,197,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,198,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[89,199,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,0,0.9,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,1,0.91,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,2,0.92,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,3,0.93,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,4,0.94,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,5,0.95,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,6,0.96,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,7,0.97,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,8,0.98,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,9,0.99,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,10,1,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,11,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,12,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,13,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,14,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,15,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,16,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,17,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,18,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[90,19,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,20,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,21,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,22,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,23,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,24,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,25,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,26,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,27,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,28,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,29,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,30,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,31,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,32,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,33,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,34,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,35,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,36,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,37,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,38,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[90,39,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,40,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,41,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,42,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,43,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,44,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,45,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,46,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,47,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,48,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,49,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,50,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,51,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,52,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,53,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,54,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,55,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,56,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,57,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,58,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[90,59,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,60,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,61,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,62,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,63,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,64,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,65,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,66,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,67,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,68,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,69,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,70,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,71,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,72,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,73,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,74,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,75,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,76,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,77,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,78,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[90,79,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,80,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,81,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,82,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,83,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,84,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,85,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,86,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,87,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,88,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,89,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,90,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,91,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,92,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,93,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,94,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,95,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,96,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,97,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,98,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[90,99,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,100,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,101,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,102,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,103,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,104,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,105,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,106,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,107,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,108,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,109,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,110,2,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,111,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,112,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,113,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,114,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,115,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,116,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,117,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,118,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[90,119,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,120,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,121,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,122,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,123,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,124,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,125,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,126,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,127,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,128,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,129,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,130,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,131,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,132,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,133,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,134,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,135,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,136,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,137,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,138,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[90,139,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,140,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,141,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,142,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,143,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,144,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,145,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,146,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,147,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,148,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,149,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,150,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,151,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,152,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,153,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,154,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,155,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,156,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,157,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,158,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[90,159,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,160,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,161,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,162,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,163,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,164,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,165,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,166,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,167,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,168,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,169,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,170,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,171,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,172,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,173,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,174,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,175,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,176,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,177,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,178,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[90,179,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,180,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,181,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,182,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,183,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,184,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,185,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,186,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,187,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,188,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,189,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,190,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,191,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,192,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,193,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,194,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,195,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,196,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,197,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,198,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[90,199,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,0,0.91,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,1,0.92,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,2,0.93,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,3,0.94,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,4,0.95,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,5,0.96,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,6,0.97,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,7,0.98,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,8,0.99,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,9,1,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,10,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,11,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,12,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,13,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,14,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,15,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,16,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,17,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,18,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[91,19,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,20,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,21,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,22,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,23,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,24,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,25,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,26,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,27,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,28,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,29,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,30,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,31,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,32,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,33,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,34,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,35,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,36,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,37,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,38,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[91,39,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,40,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,41,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,42,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,43,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,44,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,45,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,46,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,47,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,48,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,49,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,50,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,51,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,52,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,53,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,54,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,55,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,56,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,57,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,58,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[91,59,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,60,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,61,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,62,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,63,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,64,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,65,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,66,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,67,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,68,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,69,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,70,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,71,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,72,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,73,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,74,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,75,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,76,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,77,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,78,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[91,79,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,80,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,81,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,82,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,83,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,84,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,85,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,86,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,87,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,88,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,89,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,90,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,91,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,92,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,93,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,94,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,95,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,96,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,97,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,98,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[91,99,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,100,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,101,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,102,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,103,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,104,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,105,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,106,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,107,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,108,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,109,2,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,110,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,111,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,112,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,113,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,114,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,115,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,116,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,117,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,118,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[91,119,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,120,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,121,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,122,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,123,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,124,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,125,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,126,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,127,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,128,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,129,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,130,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,131,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,132,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,133,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,134,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,135,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,136,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,137,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,138,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[91,139,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,140,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,141,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,142,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,143,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,144,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,145,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,146,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,147,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,148,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,149,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,150,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,151,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,152,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,153,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,154,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,155,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,156,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,157,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,158,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[91,159,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,160,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,161,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,162,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,163,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,164,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,165,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,166,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,167,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,168,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,169,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,170,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,171,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,172,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,173,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,174,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,175,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,176,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,177,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,178,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[91,179,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,180,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,181,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,182,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,183,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,184,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,185,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,186,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,187,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,188,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,189,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,190,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,191,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,192,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,193,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,194,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,195,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,196,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,197,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,198,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[91,199,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,0,0.92,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,1,0.93,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,2,0.94,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,3,0.95,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,4,0.96,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,5,0.97,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,6,0.98,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,7,0.99,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,8,1,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,9,1.01,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,10,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,11,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,12,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,13,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,14,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,15,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,16,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,17,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,18,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[92,19,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,20,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,21,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,22,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,23,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,24,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,25,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,26,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,27,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,28,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,29,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,30,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,31,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,32,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,33,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,34,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,35,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,36,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,37,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,38,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[92,39,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,40,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,41,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,42,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,43,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,44,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,45,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,46,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,47,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,48,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,49,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,50,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,51,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,52,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,53,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,54,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,55,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,56,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,57,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,58,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[92,59,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,60,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,61,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,62,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,63,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,64,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,65,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,66,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,67,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,68,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,69,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,70,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,71,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,72,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,73,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,74,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,75,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,76,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,77,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,78,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[92,79,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,80,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,81,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,82,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,83,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,84,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,85,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,86,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,87,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,88,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,89,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,90,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,91,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,92,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,93,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,94,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,95,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,96,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,97,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,98,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[92,99,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,100,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,101,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,102,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,103,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,104,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,105,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,106,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,107,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,108,2,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,109,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,110,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,111,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,112,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,113,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,114,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,115,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,116,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,117,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,118,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[92,119,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,120,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,121,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,122,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,123,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,124,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,125,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,126,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,127,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,128,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,129,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,130,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,131,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,132,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,133,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,134,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,135,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,136,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,137,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,138,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[92,139,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,140,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,141,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,142,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,143,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,144,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,145,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,146,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,147,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,148,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,149,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,150,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,151,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,152,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,153,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,154,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,155,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,156,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,157,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,158,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[92,159,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,160,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,161,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,162,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,163,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,164,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,165,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,166,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,167,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,168,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,169,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,170,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,171,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,172,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,173,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,174,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,175,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,176,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,177,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,178,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[92,179,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,180,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,181,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,182,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,183,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,184,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,185,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,186,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,187,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,188,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,189,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,190,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,191,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,192,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,193,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,194,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,195,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,196,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,197,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,198,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[92,199,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,0,0.93,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,1,0.94,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,2,0.95,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,3,0.96,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,4,0.97,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,5,0.98,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,6,0.99,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,7,1,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,8,1.01,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,9,1.02,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,10,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,11,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,12,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,13,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,14,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,15,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,16,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,17,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,18,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[93,19,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,20,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,21,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,22,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,23,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,24,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,25,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,26,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,27,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,28,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,29,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,30,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,31,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,32,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,33,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,34,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,35,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,36,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,37,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,38,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[93,39,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,40,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,41,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,42,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,43,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,44,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,45,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,46,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,47,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,48,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,49,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,50,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,51,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,52,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,53,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,54,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,55,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,56,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,57,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,58,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[93,59,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,60,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,61,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,62,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,63,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,64,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,65,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,66,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,67,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,68,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,69,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,70,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,71,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,72,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,73,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,74,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,75,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,76,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,77,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,78,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[93,79,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,80,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,81,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,82,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,83,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,84,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,85,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,86,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,87,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,88,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,89,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,90,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,91,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,92,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,93,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,94,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,95,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,96,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,97,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,98,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[93,99,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,100,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,101,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,102,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,103,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,104,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,105,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,106,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,107,2,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,108,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,109,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,110,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,111,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,112,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,113,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,114,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,115,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,116,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,117,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,118,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[93,119,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,120,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,121,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,122,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,123,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,124,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,125,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,126,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,127,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,128,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,129,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,130,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,131,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,132,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,133,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,134,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,135,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,136,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,137,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,138,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[93,139,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,140,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,141,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,142,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,143,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,144,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,145,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,146,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,147,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,148,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,149,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,150,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,151,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,152,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,153,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,154,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,155,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,156,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,157,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,158,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[93,159,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,160,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,161,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,162,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,163,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,164,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,165,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,166,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,167,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,168,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,169,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,170,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,171,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,172,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,173,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,174,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,175,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,176,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,177,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,178,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[93,179,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,180,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,181,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,182,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,183,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,184,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,185,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,186,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,187,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,188,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,189,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,190,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,191,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,192,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,193,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,194,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,195,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,196,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,197,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,198,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[93,199,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,0,0.94,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,1,0.95,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,2,0.96,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,3,0.97,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,4,0.98,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,5,0.99,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,6,1,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,7,1.01,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,8,1.02,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,9,1.03,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,10,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,11,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,12,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,13,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,14,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,15,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,16,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,17,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,18,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[94,19,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,20,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,21,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,22,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,23,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,24,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,25,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,26,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,27,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,28,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,29,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,30,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,31,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,32,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,33,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,34,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,35,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,36,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,37,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,38,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[94,39,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,40,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,41,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,42,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,43,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,44,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,45,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,46,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,47,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,48,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,49,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,50,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,51,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,52,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,53,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,54,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,55,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,56,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,57,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,58,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[94,59,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,60,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,61,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,62,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,63,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,64,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,65,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,66,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,67,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,68,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,69,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,70,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,71,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,72,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,73,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,74,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,75,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,76,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,77,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,78,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[94,79,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,80,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,81,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,82,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,83,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,84,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,85,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,86,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,87,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,88,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,89,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,90,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,91,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,92,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,93,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,94,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,95,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,96,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,97,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,98,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[94,99,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,100,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,101,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,102,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,103,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,104,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,105,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,106,2,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,107,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,108,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,109,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,110,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,111,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,112,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,113,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,114,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,115,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,116,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,117,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,118,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[94,119,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,120,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,121,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,122,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,123,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,124,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,125,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,126,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,127,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,128,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,129,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,130,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,131,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,132,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,133,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,134,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,135,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,136,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,137,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,138,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[94,139,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,140,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,141,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,142,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,143,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,144,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,145,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,146,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,147,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,148,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,149,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,150,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,151,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,152,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,153,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,154,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,155,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,156,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,157,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,158,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[94,159,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,160,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,161,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,162,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,163,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,164,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,165,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,166,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,167,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,168,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,169,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,170,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,171,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,172,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,173,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,174,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,175,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,176,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,177,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,178,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[94,179,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,180,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,181,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,182,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,183,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,184,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,185,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,186,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,187,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,188,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,189,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,190,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,191,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,192,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,193,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,194,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,195,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,196,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,197,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,198,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[94,199,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,0,0.95,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,1,0.96,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,2,0.97,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,3,0.98,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,4,0.99,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,5,1,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,6,1.01,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,7,1.02,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,8,1.03,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,9,1.04,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,10,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,11,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,12,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,13,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,14,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,15,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,16,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,17,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,18,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[95,19,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,20,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,21,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,22,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,23,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,24,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,25,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,26,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,27,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,28,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,29,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,30,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,31,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,32,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,33,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,34,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,35,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,36,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,37,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,38,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[95,39,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,40,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,41,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,42,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,43,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,44,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,45,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,46,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,47,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,48,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,49,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,50,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,51,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,52,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,53,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,54,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,55,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,56,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,57,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,58,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[95,59,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,60,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,61,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,62,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,63,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,64,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,65,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,66,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,67,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,68,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,69,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,70,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,71,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,72,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,73,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,74,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,75,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,76,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,77,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,78,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[95,79,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,80,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,81,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,82,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,83,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,84,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,85,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,86,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,87,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,88,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,89,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,90,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,91,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,92,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,93,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,94,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,95,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,96,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,97,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,98,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[95,99,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,100,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,101,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,102,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,103,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,104,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,105,2,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,106,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,107,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,108,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,109,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,110,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,111,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,112,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,113,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,114,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,115,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,116,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,117,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,118,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[95,119,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,120,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,121,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,122,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,123,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,124,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,125,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,126,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,127,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,128,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,129,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,130,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,131,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,132,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,133,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,134,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,135,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,136,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,137,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,138,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[95,139,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,140,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,141,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,142,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,143,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,144,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,145,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,146,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,147,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,148,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,149,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,150,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,151,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,152,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,153,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,154,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,155,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,156,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,157,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,158,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[95,159,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,160,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,161,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,162,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,163,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,164,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,165,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,166,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,167,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,168,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,169,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,170,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,171,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,172,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,173,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,174,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,175,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,176,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,177,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,178,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[95,179,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,180,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,181,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,182,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,183,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,184,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,185,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,186,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,187,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,188,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,189,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,190,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,191,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,192,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,193,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,194,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,195,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,196,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,197,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,198,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[95,199,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,0,0.96,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,1,0.97,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,2,0.98,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,3,0.99,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,4,1,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,5,1.01,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,6,1.02,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,7,1.03,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,8,1.04,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,9,1.05,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,10,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,11,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,12,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,13,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,14,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,15,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,16,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,17,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,18,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[96,19,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,20,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,21,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,22,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,23,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,24,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,25,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,26,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,27,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,28,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,29,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,30,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,31,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,32,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,33,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,34,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,35,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,36,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,37,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,38,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[96,39,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,40,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,41,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,42,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,43,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,44,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,45,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,46,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,47,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,48,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,49,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,50,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,51,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,52,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,53,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,54,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,55,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,56,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,57,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,58,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[96,59,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,60,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,61,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,62,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,63,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,64,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,65,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,66,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,67,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,68,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,69,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,70,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,71,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,72,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,73,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,74,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,75,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,76,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,77,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,78,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[96,79,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,80,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,81,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,82,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,83,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,84,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,85,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,86,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,87,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,88,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,89,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,90,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,91,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,92,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,93,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,94,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,95,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,96,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,97,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,98,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[96,99,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,100,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,101,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,102,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,103,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,104,2,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,105,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,106,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,107,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,108,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,109,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,110,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,111,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,112,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,113,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,114,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,115,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,116,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,117,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,118,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[96,119,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,120,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,121,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,122,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,123,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,124,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,125,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,126,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,127,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,128,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,129,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,130,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,131,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,132,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,133,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,134,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,135,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,136,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,137,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,138,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[96,139,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,140,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,141,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,142,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,143,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,144,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,145,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,146,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,147,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,148,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,149,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,150,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,151,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,152,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,153,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,154,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,155,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,156,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,157,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,158,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[96,159,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,160,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,161,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,162,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,163,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,164,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,165,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,166,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,167,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,168,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,169,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,170,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,171,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,172,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,173,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,174,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,175,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,176,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,177,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,178,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[96,179,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,180,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,181,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,182,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,183,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,184,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,185,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,186,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,187,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,188,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,189,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,190,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,191,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,192,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,193,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,194,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,195,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,196,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,197,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,198,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[96,199,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,0,0.97,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,1,0.98,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,2,0.99,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,3,1,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,4,1.01,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,5,1.02,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,6,1.03,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,7,1.04,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,8,1.05,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,9,1.06,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,10,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,11,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,12,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,13,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,14,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,15,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,16,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,17,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,18,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[97,19,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,20,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,21,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,22,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,23,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,24,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,25,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,26,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,27,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,28,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,29,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,30,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,31,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,32,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,33,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,34,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,35,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,36,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,37,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,38,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[97,39,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,40,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,41,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,42,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,43,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,44,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,45,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,46,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,47,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,48,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,49,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,50,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,51,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,52,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,53,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,54,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,55,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,56,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,57,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,58,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[97,59,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,60,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,61,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,62,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,63,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,64,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,65,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,66,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,67,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,68,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,69,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,70,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,71,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,72,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,73,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,74,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,75,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,76,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,77,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,78,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[97,79,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,80,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,81,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,82,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,83,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,84,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,85,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,86,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,87,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,88,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,89,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,90,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,91,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,92,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,93,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,94,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,95,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,96,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,97,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,98,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[97,99,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,100,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,101,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,102,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,103,2,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,104,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,105,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,106,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,107,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,108,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,109,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,110,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,111,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,112,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,113,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,114,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,115,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,116,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,117,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,118,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[97,119,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,120,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,121,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,122,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,123,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,124,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,125,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,126,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,127,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,128,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,129,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,130,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,131,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,132,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,133,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,134,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,135,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,136,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,137,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,138,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[97,139,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,140,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,141,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,142,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,143,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,144,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,145,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,146,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,147,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,148,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,149,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,150,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,151,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,152,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,153,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,154,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,155,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,156,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,157,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,158,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[97,159,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,160,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,161,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,162,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,163,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,164,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,165,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,166,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,167,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,168,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,169,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,170,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,171,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,172,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,173,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,174,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,175,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,176,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,177,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,178,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[97,179,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,180,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,181,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,182,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,183,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,184,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,185,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,186,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,187,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,188,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,189,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,190,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,191,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,192,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,193,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,194,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,195,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,196,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,197,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,198,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[97,199,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,0,0.98,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,1,0.99,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,2,1,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,3,1.01,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,4,1.02,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,5,1.03,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,6,1.04,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,7,1.05,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,8,1.06,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,9,1.07,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,10,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,11,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,12,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,13,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,14,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,15,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,16,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,17,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,18,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[98,19,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,20,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,21,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,22,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,23,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,24,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,25,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,26,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,27,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,28,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,29,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,30,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,31,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,32,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,33,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,34,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,35,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,36,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,37,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,38,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[98,39,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,40,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,41,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,42,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,43,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,44,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,45,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,46,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,47,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,48,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,49,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,50,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,51,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,52,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,53,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,54,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,55,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,56,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,57,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,58,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[98,59,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,60,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,61,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,62,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,63,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,64,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,65,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,66,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,67,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,68,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,69,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,70,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,71,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,72,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,73,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,74,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,75,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,76,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,77,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,78,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[98,79,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,80,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,81,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,82,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,83,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,84,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,85,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,86,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,87,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,88,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,89,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,90,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,91,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,92,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,93,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,94,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,95,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,96,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,97,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,98,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[98,99,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,100,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,101,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,102,2,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,103,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,104,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,105,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,106,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,107,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,108,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,109,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,110,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,111,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,112,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,113,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,114,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,115,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,116,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,117,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,118,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[98,119,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,120,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,121,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,122,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,123,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,124,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,125,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,126,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,127,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,128,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,129,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,130,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,131,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,132,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,133,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,134,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,135,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,136,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,137,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,138,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[98,139,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,140,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,141,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,142,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,143,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,144,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,145,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,146,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,147,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,148,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,149,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,150,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,151,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,152,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,153,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,154,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,155,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,156,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,157,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,158,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[98,159,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,160,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,161,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,162,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,163,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,164,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,165,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,166,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,167,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,168,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,169,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,170,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,171,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,172,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,173,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,174,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,175,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,176,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,177,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,178,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[98,179,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,180,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,181,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,182,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,183,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,184,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,185,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,186,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,187,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,188,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,189,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,190,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,191,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,192,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,193,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,194,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,195,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,196,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,197,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,198,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[98,199,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,0,0.99,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,1,1,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,2,1.01,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,3,1.02,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,4,1.03,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,5,1.04,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,6,1.05,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,7,1.06,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,8,1.07,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,9,1.08,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,10,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,11,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,12,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,13,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,14,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,15,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,16,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,17,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,18,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 80, pc_makepoint(1, ARRAY[99,19,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,20,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,21,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,22,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,23,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,24,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,25,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,26,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,27,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,28,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,29,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,30,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,31,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,32,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,33,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,34,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,35,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,36,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,37,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,38,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 81, pc_makepoint(1, ARRAY[99,39,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,40,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,41,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,42,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,43,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,44,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,45,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,46,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,47,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,48,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,49,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,50,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,51,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,52,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,53,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,54,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,55,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,56,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,57,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,58,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 82, pc_makepoint(1, ARRAY[99,59,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,60,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,61,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,62,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,63,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,64,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,65,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,66,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,67,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,68,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,69,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,70,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,71,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,72,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,73,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,74,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,75,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,76,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,77,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,78,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 83, pc_makepoint(1, ARRAY[99,79,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,80,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,81,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,82,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,83,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,84,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,85,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,86,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,87,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,88,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,89,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,90,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,91,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,92,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,93,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,94,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,95,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,96,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,97,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,98,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 84, pc_makepoint(1, ARRAY[99,99,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,100,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,101,2,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,102,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,103,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,104,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,105,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,106,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,107,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,108,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,109,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,110,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,111,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,112,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,113,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,114,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,115,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,116,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,117,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,118,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 85, pc_makepoint(1, ARRAY[99,119,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,120,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,121,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,122,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,123,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,124,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,125,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,126,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,127,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,128,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,129,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,130,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,131,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,132,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,133,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,134,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,135,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,136,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,137,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,138,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 86, pc_makepoint(1, ARRAY[99,139,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,140,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,141,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,142,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,143,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,144,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,145,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,146,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,147,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,148,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,149,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,150,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,151,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,152,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,153,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,154,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,155,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,156,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,157,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,158,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 87, pc_makepoint(1, ARRAY[99,159,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,160,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,161,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,162,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,163,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,164,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,165,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,166,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,167,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,168,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,169,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,170,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,171,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,172,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,173,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,174,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,175,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,176,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,177,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,178,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 88, pc_makepoint(1, ARRAY[99,179,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,180,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,181,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,182,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,183,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,184,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,185,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,186,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,187,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,188,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,189,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,190,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,191,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,192,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,193,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,194,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,195,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,196,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,197,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,198,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 89, pc_makepoint(1, ARRAY[99,199,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,0,1,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,1,1.01,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,2,1.02,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,3,1.03,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,4,1.04,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,5,1.05,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,6,1.06,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,7,1.07,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,8,1.08,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,9,1.09,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,10,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,11,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,12,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,13,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,14,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,15,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,16,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,17,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,18,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[100,19,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,20,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,21,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,22,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,23,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,24,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,25,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,26,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,27,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,28,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,29,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,30,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,31,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,32,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,33,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,34,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,35,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,36,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,37,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,38,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[100,39,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,40,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,41,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,42,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,43,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,44,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,45,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,46,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,47,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,48,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,49,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,50,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,51,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,52,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,53,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,54,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,55,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,56,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,57,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,58,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[100,59,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,60,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,61,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,62,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,63,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,64,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,65,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,66,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,67,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,68,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,69,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,70,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,71,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,72,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,73,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,74,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,75,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,76,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,77,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,78,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[100,79,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,80,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,81,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,82,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,83,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,84,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,85,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,86,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,87,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,88,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,89,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,90,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,91,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,92,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,93,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,94,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,95,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,96,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,97,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,98,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[100,99,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,100,2,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,101,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,102,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,103,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,104,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,105,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,106,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,107,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,108,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,109,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,110,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,111,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,112,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,113,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,114,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,115,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,116,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,117,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,118,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[100,119,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,120,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,121,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,122,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,123,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,124,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,125,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,126,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,127,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,128,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,129,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,130,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,131,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,132,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,133,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,134,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,135,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,136,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,137,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,138,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[100,139,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,140,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,141,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,142,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,143,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,144,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,145,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,146,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,147,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,148,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,149,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,150,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,151,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,152,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,153,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,154,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,155,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,156,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,157,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,158,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[100,159,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,160,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,161,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,162,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,163,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,164,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,165,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,166,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,167,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,168,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,169,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,170,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,171,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,172,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,173,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,174,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,175,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,176,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,177,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,178,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[100,179,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,180,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,181,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,182,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,183,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,184,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,185,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,186,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,187,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,188,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,189,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,190,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,191,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,192,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,193,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,194,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,195,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,196,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,197,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,198,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[100,199,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,0,1.01,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,1,1.02,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,2,1.03,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,3,1.04,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,4,1.05,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,5,1.06,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,6,1.07,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,7,1.08,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,8,1.09,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,9,1.1,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,10,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,11,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,12,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,13,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,14,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,15,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,16,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,17,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,18,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[101,19,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,20,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,21,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,22,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,23,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,24,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,25,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,26,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,27,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,28,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,29,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,30,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,31,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,32,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,33,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,34,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,35,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,36,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,37,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,38,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[101,39,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,40,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,41,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,42,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,43,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,44,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,45,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,46,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,47,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,48,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,49,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,50,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,51,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,52,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,53,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,54,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,55,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,56,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,57,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,58,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[101,59,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,60,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,61,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,62,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,63,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,64,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,65,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,66,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,67,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,68,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,69,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,70,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,71,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,72,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,73,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,74,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,75,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,76,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,77,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,78,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[101,79,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,80,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,81,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,82,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,83,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,84,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,85,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,86,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,87,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,88,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,89,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,90,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,91,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,92,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,93,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,94,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,95,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,96,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,97,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,98,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[101,99,2,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,100,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,101,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,102,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,103,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,104,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,105,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,106,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,107,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,108,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,109,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,110,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,111,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,112,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,113,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,114,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,115,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,116,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,117,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,118,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[101,119,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,120,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,121,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,122,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,123,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,124,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,125,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,126,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,127,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,128,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,129,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,130,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,131,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,132,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,133,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,134,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,135,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,136,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,137,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,138,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[101,139,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,140,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,141,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,142,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,143,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,144,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,145,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,146,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,147,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,148,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,149,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,150,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,151,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,152,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,153,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,154,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,155,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,156,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,157,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,158,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[101,159,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,160,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,161,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,162,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,163,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,164,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,165,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,166,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,167,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,168,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,169,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,170,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,171,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,172,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,173,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,174,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,175,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,176,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,177,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,178,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[101,179,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,180,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,181,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,182,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,183,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,184,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,185,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,186,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,187,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,188,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,189,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,190,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,191,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,192,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,193,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,194,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,195,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,196,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,197,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,198,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[101,199,3,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,0,1.02,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,1,1.03,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,2,1.04,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,3,1.05,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,4,1.06,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,5,1.07,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,6,1.08,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,7,1.09,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,8,1.1,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,9,1.11,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,10,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,11,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,12,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,13,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,14,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,15,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,16,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,17,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,18,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[102,19,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,20,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,21,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,22,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,23,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,24,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,25,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,26,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,27,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,28,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,29,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,30,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,31,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,32,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,33,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,34,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,35,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,36,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,37,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,38,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[102,39,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,40,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,41,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,42,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,43,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,44,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,45,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,46,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,47,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,48,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,49,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,50,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,51,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,52,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,53,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,54,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,55,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,56,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,57,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,58,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[102,59,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,60,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,61,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,62,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,63,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,64,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,65,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,66,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,67,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,68,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,69,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,70,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,71,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,72,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,73,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,74,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,75,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,76,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,77,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,78,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[102,79,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,80,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,81,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,82,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,83,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,84,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,85,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,86,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,87,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,88,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,89,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,90,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,91,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,92,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,93,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,94,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,95,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,96,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,97,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,98,2,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[102,99,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,100,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,101,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,102,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,103,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,104,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,105,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,106,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,107,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,108,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,109,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,110,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,111,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,112,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,113,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,114,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,115,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,116,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,117,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,118,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[102,119,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,120,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,121,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,122,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,123,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,124,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,125,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,126,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,127,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,128,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,129,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,130,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,131,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,132,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,133,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,134,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,135,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,136,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,137,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,138,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[102,139,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,140,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,141,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,142,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,143,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,144,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,145,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,146,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,147,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,148,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,149,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,150,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,151,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,152,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,153,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,154,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,155,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,156,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,157,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,158,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[102,159,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,160,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,161,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,162,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,163,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,164,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,165,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,166,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,167,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,168,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,169,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,170,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,171,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,172,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,173,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,174,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,175,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,176,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,177,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,178,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[102,179,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,180,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,181,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,182,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,183,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,184,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,185,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,186,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,187,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,188,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,189,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,190,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,191,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,192,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,193,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,194,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,195,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,196,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,197,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,198,3,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[102,199,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,0,1.03,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,1,1.04,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,2,1.05,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,3,1.06,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,4,1.07,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,5,1.08,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,6,1.09,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,7,1.1,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,8,1.11,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,9,1.12,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,10,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,11,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,12,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,13,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,14,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,15,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,16,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,17,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,18,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[103,19,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,20,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,21,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,22,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,23,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,24,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,25,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,26,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,27,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,28,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,29,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,30,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,31,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,32,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,33,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,34,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,35,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,36,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,37,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,38,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[103,39,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,40,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,41,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,42,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,43,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,44,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,45,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,46,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,47,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,48,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,49,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,50,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,51,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,52,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,53,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,54,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,55,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,56,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,57,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,58,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[103,59,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,60,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,61,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,62,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,63,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,64,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,65,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,66,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,67,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,68,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,69,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,70,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,71,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,72,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,73,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,74,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,75,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,76,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,77,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,78,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[103,79,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,80,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,81,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,82,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,83,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,84,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,85,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,86,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,87,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,88,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,89,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,90,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,91,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,92,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,93,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,94,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,95,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,96,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,97,2,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,98,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[103,99,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,100,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,101,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,102,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,103,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,104,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,105,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,106,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,107,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,108,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,109,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,110,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,111,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,112,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,113,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,114,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,115,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,116,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,117,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,118,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[103,119,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,120,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,121,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,122,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,123,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,124,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,125,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,126,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,127,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,128,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,129,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,130,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,131,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,132,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,133,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,134,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,135,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,136,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,137,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,138,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[103,139,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,140,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,141,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,142,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,143,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,144,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,145,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,146,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,147,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,148,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,149,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,150,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,151,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,152,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,153,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,154,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,155,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,156,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,157,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,158,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[103,159,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,160,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,161,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,162,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,163,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,164,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,165,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,166,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,167,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,168,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,169,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,170,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,171,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,172,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,173,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,174,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,175,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,176,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,177,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,178,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[103,179,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,180,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,181,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,182,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,183,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,184,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,185,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,186,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,187,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,188,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,189,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,190,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,191,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,192,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,193,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,194,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,195,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,196,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,197,3,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,198,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[103,199,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,0,1.04,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,1,1.05,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,2,1.06,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,3,1.07,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,4,1.08,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,5,1.09,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,6,1.1,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,7,1.11,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,8,1.12,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,9,1.13,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,10,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,11,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,12,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,13,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,14,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,15,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,16,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,17,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,18,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[104,19,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,20,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,21,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,22,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,23,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,24,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,25,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,26,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,27,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,28,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,29,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,30,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,31,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,32,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,33,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,34,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,35,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,36,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,37,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,38,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[104,39,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,40,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,41,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,42,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,43,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,44,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,45,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,46,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,47,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,48,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,49,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,50,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,51,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,52,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,53,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,54,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,55,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,56,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,57,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,58,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[104,59,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,60,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,61,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,62,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,63,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,64,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,65,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,66,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,67,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,68,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,69,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,70,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,71,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,72,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,73,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,74,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,75,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,76,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,77,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,78,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[104,79,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,80,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,81,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,82,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,83,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,84,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,85,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,86,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,87,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,88,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,89,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,90,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,91,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,92,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,93,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,94,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,95,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,96,2,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,97,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,98,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[104,99,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,100,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,101,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,102,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,103,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,104,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,105,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,106,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,107,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,108,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,109,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,110,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,111,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,112,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,113,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,114,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,115,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,116,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,117,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,118,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[104,119,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,120,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,121,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,122,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,123,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,124,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,125,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,126,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,127,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,128,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,129,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,130,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,131,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,132,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,133,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,134,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,135,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,136,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,137,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,138,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[104,139,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,140,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,141,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,142,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,143,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,144,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,145,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,146,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,147,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,148,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,149,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,150,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,151,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,152,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,153,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,154,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,155,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,156,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,157,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,158,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[104,159,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,160,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,161,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,162,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,163,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,164,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,165,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,166,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,167,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,168,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,169,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,170,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,171,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,172,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,173,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,174,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,175,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,176,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,177,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,178,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[104,179,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,180,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,181,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,182,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,183,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,184,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,185,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,186,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,187,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,188,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,189,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,190,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,191,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,192,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,193,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,194,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,195,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,196,3,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,197,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,198,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[104,199,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,0,1.05,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,1,1.06,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,2,1.07,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,3,1.08,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,4,1.09,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,5,1.1,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,6,1.11,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,7,1.12,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,8,1.13,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,9,1.14,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,10,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,11,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,12,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,13,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,14,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,15,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,16,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,17,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,18,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[105,19,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,20,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,21,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,22,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,23,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,24,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,25,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,26,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,27,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,28,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,29,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,30,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,31,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,32,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,33,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,34,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,35,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,36,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,37,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,38,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[105,39,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,40,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,41,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,42,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,43,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,44,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,45,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,46,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,47,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,48,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,49,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,50,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,51,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,52,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,53,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,54,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,55,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,56,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,57,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,58,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[105,59,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,60,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,61,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,62,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,63,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,64,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,65,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,66,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,67,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,68,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,69,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,70,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,71,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,72,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,73,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,74,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,75,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,76,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,77,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,78,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[105,79,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,80,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,81,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,82,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,83,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,84,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,85,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,86,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,87,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,88,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,89,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,90,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,91,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,92,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,93,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,94,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,95,2,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,96,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,97,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,98,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[105,99,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,100,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,101,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,102,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,103,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,104,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,105,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,106,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,107,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,108,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,109,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,110,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,111,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,112,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,113,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,114,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,115,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,116,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,117,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,118,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[105,119,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,120,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,121,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,122,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,123,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,124,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,125,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,126,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,127,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,128,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,129,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,130,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,131,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,132,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,133,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,134,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,135,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,136,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,137,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,138,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[105,139,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,140,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,141,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,142,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,143,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,144,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,145,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,146,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,147,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,148,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,149,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,150,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,151,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,152,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,153,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,154,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,155,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,156,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,157,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,158,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[105,159,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,160,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,161,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,162,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,163,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,164,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,165,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,166,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,167,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,168,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,169,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,170,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,171,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,172,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,173,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,174,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,175,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,176,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,177,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,178,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[105,179,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,180,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,181,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,182,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,183,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,184,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,185,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,186,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,187,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,188,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,189,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,190,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,191,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,192,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,193,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,194,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,195,3,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,196,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,197,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,198,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[105,199,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,0,1.06,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,1,1.07,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,2,1.08,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,3,1.09,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,4,1.1,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,5,1.11,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,6,1.12,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,7,1.13,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,8,1.14,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,9,1.15,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,10,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,11,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,12,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,13,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,14,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,15,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,16,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,17,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,18,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[106,19,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,20,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,21,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,22,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,23,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,24,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,25,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,26,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,27,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,28,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,29,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,30,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,31,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,32,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,33,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,34,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,35,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,36,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,37,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,38,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[106,39,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,40,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,41,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,42,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,43,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,44,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,45,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,46,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,47,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,48,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,49,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,50,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,51,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,52,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,53,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,54,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,55,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,56,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,57,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,58,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[106,59,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,60,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,61,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,62,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,63,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,64,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,65,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,66,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,67,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,68,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,69,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,70,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,71,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,72,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,73,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,74,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,75,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,76,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,77,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,78,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[106,79,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,80,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,81,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,82,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,83,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,84,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,85,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,86,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,87,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,88,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,89,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,90,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,91,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,92,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,93,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,94,2,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,95,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,96,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,97,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,98,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[106,99,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,100,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,101,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,102,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,103,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,104,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,105,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,106,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,107,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,108,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,109,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,110,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,111,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,112,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,113,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,114,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,115,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,116,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,117,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,118,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[106,119,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,120,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,121,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,122,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,123,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,124,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,125,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,126,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,127,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,128,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,129,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,130,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,131,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,132,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,133,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,134,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,135,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,136,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,137,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,138,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[106,139,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,140,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,141,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,142,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,143,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,144,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,145,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,146,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,147,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,148,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,149,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,150,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,151,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,152,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,153,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,154,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,155,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,156,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,157,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,158,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[106,159,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,160,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,161,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,162,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,163,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,164,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,165,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,166,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,167,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,168,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,169,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,170,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,171,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,172,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,173,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,174,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,175,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,176,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,177,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,178,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[106,179,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,180,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,181,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,182,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,183,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,184,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,185,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,186,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,187,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,188,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,189,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,190,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,191,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,192,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,193,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,194,3,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,195,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,196,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,197,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,198,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[106,199,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,0,1.07,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,1,1.08,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,2,1.09,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,3,1.1,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,4,1.11,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,5,1.12,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,6,1.13,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,7,1.14,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,8,1.15,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,9,1.16,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,10,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,11,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,12,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,13,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,14,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,15,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,16,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,17,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,18,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[107,19,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,20,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,21,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,22,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,23,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,24,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,25,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,26,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,27,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,28,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,29,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,30,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,31,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,32,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,33,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,34,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,35,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,36,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,37,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,38,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[107,39,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,40,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,41,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,42,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,43,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,44,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,45,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,46,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,47,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,48,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,49,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,50,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,51,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,52,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,53,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,54,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,55,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,56,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,57,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,58,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[107,59,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,60,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,61,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,62,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,63,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,64,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,65,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,66,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,67,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,68,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,69,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,70,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,71,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,72,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,73,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,74,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,75,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,76,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,77,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,78,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[107,79,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,80,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,81,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,82,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,83,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,84,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,85,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,86,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,87,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,88,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,89,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,90,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,91,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,92,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,93,2,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,94,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,95,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,96,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,97,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,98,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[107,99,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,100,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,101,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,102,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,103,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,104,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,105,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,106,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,107,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,108,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,109,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,110,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,111,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,112,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,113,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,114,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,115,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,116,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,117,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,118,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[107,119,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,120,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,121,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,122,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,123,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,124,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,125,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,126,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,127,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,128,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,129,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,130,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,131,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,132,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,133,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,134,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,135,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,136,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,137,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,138,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[107,139,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,140,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,141,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,142,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,143,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,144,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,145,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,146,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,147,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,148,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,149,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,150,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,151,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,152,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,153,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,154,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,155,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,156,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,157,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,158,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[107,159,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,160,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,161,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,162,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,163,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,164,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,165,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,166,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,167,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,168,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,169,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,170,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,171,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,172,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,173,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,174,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,175,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,176,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,177,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,178,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[107,179,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,180,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,181,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,182,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,183,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,184,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,185,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,186,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,187,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,188,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,189,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,190,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,191,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,192,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,193,3,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,194,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,195,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,196,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,197,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,198,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[107,199,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,0,1.08,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,1,1.09,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,2,1.1,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,3,1.11,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,4,1.12,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,5,1.13,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,6,1.14,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,7,1.15,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,8,1.16,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,9,1.17,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,10,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,11,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,12,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,13,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,14,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,15,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,16,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,17,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,18,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[108,19,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,20,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,21,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,22,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,23,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,24,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,25,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,26,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,27,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,28,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,29,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,30,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,31,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,32,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,33,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,34,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,35,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,36,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,37,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,38,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[108,39,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,40,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,41,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,42,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,43,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,44,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,45,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,46,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,47,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,48,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,49,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,50,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,51,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,52,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,53,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,54,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,55,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,56,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,57,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,58,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[108,59,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,60,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,61,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,62,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,63,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,64,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,65,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,66,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,67,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,68,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,69,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,70,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,71,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,72,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,73,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,74,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,75,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,76,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,77,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,78,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[108,79,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,80,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,81,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,82,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,83,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,84,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,85,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,86,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,87,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,88,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,89,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,90,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,91,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,92,2,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,93,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,94,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,95,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,96,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,97,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,98,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[108,99,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,100,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,101,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,102,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,103,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,104,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,105,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,106,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,107,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,108,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,109,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,110,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,111,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,112,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,113,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,114,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,115,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,116,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,117,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,118,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[108,119,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,120,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,121,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,122,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,123,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,124,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,125,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,126,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,127,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,128,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,129,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,130,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,131,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,132,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,133,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,134,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,135,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,136,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,137,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,138,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[108,139,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,140,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,141,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,142,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,143,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,144,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,145,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,146,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,147,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,148,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,149,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,150,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,151,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,152,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,153,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,154,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,155,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,156,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,157,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,158,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[108,159,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,160,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,161,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,162,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,163,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,164,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,165,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,166,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,167,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,168,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,169,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,170,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,171,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,172,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,173,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,174,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,175,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,176,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,177,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,178,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[108,179,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,180,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,181,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,182,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,183,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,184,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,185,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,186,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,187,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,188,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,189,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,190,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,191,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,192,3,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,193,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,194,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,195,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,196,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,197,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,198,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[108,199,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,0,1.09,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,1,1.1,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,2,1.11,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,3,1.12,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,4,1.13,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,5,1.14,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,6,1.15,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,7,1.16,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,8,1.17,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,9,1.18,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,10,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,11,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,12,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,13,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,14,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,15,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,16,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,17,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,18,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[109,19,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,20,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,21,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,22,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,23,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,24,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,25,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,26,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,27,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,28,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,29,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,30,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,31,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,32,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,33,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,34,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,35,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,36,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,37,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,38,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[109,39,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,40,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,41,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,42,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,43,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,44,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,45,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,46,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,47,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,48,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,49,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,50,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,51,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,52,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,53,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,54,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,55,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,56,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,57,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,58,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[109,59,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,60,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,61,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,62,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,63,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,64,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,65,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,66,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,67,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,68,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,69,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,70,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,71,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,72,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,73,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,74,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,75,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,76,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,77,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,78,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[109,79,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,80,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,81,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,82,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,83,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,84,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,85,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,86,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,87,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,88,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,89,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,90,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,91,2,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,92,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,93,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,94,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,95,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,96,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,97,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,98,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[109,99,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,100,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,101,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,102,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,103,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,104,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,105,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,106,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,107,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,108,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,109,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,110,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,111,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,112,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,113,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,114,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,115,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,116,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,117,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,118,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[109,119,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,120,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,121,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,122,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,123,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,124,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,125,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,126,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,127,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,128,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,129,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,130,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,131,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,132,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,133,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,134,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,135,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,136,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,137,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,138,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[109,139,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,140,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,141,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,142,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,143,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,144,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,145,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,146,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,147,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,148,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,149,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,150,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,151,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,152,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,153,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,154,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,155,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,156,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,157,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,158,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[109,159,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,160,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,161,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,162,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,163,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,164,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,165,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,166,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,167,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,168,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,169,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,170,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,171,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,172,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,173,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,174,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,175,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,176,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,177,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,178,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[109,179,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,180,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,181,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,182,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,183,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,184,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,185,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,186,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,187,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,188,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,189,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,190,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,191,3,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,192,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,193,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,194,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,195,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,196,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,197,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,198,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[109,199,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,0,1.1,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,1,1.11,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,2,1.12,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,3,1.13,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,4,1.14,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,5,1.15,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,6,1.16,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,7,1.17,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,8,1.18,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,9,1.19,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,10,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,11,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,12,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,13,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,14,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,15,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,16,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,17,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,18,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[110,19,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,20,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,21,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,22,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,23,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,24,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,25,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,26,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,27,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,28,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,29,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,30,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,31,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,32,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,33,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,34,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,35,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,36,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,37,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,38,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[110,39,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,40,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,41,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,42,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,43,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,44,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,45,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,46,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,47,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,48,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,49,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,50,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,51,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,52,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,53,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,54,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,55,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,56,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,57,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,58,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[110,59,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,60,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,61,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,62,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,63,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,64,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,65,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,66,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,67,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,68,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,69,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,70,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,71,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,72,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,73,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,74,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,75,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,76,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,77,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,78,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[110,79,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,80,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,81,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,82,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,83,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,84,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,85,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,86,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,87,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,88,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,89,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,90,2,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,91,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,92,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,93,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,94,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,95,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,96,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,97,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,98,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[110,99,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,100,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,101,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,102,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,103,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,104,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,105,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,106,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,107,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,108,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,109,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,110,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,111,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,112,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,113,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,114,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,115,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,116,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,117,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,118,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[110,119,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,120,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,121,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,122,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,123,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,124,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,125,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,126,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,127,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,128,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,129,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,130,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,131,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,132,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,133,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,134,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,135,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,136,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,137,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,138,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[110,139,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,140,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,141,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,142,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,143,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,144,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,145,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,146,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,147,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,148,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,149,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,150,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,151,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,152,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,153,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,154,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,155,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,156,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,157,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,158,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[110,159,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,160,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,161,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,162,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,163,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,164,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,165,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,166,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,167,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,168,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,169,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,170,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,171,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,172,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,173,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,174,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,175,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,176,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,177,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,178,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[110,179,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,180,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,181,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,182,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,183,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,184,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,185,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,186,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,187,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,188,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,189,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,190,3,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,191,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,192,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,193,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,194,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,195,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,196,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,197,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,198,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[110,199,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,0,1.11,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,1,1.12,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,2,1.13,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,3,1.14,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,4,1.15,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,5,1.16,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,6,1.17,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,7,1.18,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,8,1.19,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,9,1.2,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,10,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,11,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,12,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,13,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,14,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,15,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,16,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,17,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,18,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[111,19,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,20,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,21,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,22,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,23,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,24,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,25,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,26,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,27,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,28,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,29,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,30,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,31,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,32,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,33,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,34,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,35,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,36,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,37,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,38,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[111,39,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,40,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,41,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,42,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,43,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,44,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,45,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,46,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,47,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,48,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,49,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,50,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,51,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,52,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,53,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,54,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,55,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,56,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,57,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,58,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[111,59,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,60,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,61,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,62,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,63,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,64,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,65,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,66,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,67,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,68,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,69,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,70,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,71,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,72,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,73,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,74,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,75,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,76,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,77,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,78,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[111,79,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,80,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,81,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,82,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,83,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,84,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,85,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,86,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,87,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,88,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,89,2,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,90,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,91,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,92,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,93,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,94,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,95,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,96,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,97,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,98,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[111,99,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,100,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,101,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,102,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,103,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,104,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,105,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,106,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,107,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,108,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,109,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,110,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,111,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,112,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,113,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,114,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,115,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,116,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,117,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,118,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[111,119,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,120,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,121,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,122,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,123,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,124,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,125,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,126,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,127,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,128,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,129,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,130,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,131,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,132,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,133,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,134,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,135,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,136,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,137,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,138,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[111,139,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,140,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,141,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,142,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,143,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,144,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,145,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,146,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,147,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,148,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,149,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,150,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,151,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,152,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,153,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,154,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,155,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,156,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,157,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,158,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[111,159,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,160,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,161,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,162,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,163,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,164,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,165,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,166,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,167,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,168,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,169,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,170,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,171,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,172,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,173,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,174,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,175,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,176,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,177,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,178,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[111,179,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,180,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,181,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,182,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,183,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,184,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,185,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,186,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,187,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,188,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,189,3,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,190,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,191,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,192,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,193,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,194,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,195,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,196,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,197,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,198,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[111,199,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,0,1.12,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,1,1.13,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,2,1.14,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,3,1.15,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,4,1.16,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,5,1.17,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,6,1.18,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,7,1.19,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,8,1.2,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,9,1.21,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,10,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,11,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,12,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,13,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,14,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,15,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,16,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,17,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,18,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[112,19,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,20,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,21,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,22,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,23,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,24,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,25,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,26,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,27,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,28,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,29,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,30,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,31,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,32,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,33,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,34,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,35,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,36,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,37,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,38,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[112,39,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,40,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,41,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,42,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,43,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,44,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,45,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,46,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,47,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,48,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,49,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,50,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,51,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,52,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,53,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,54,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,55,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,56,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,57,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,58,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[112,59,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,60,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,61,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,62,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,63,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,64,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,65,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,66,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,67,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,68,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,69,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,70,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,71,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,72,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,73,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,74,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,75,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,76,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,77,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,78,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[112,79,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,80,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,81,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,82,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,83,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,84,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,85,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,86,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,87,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,88,2,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,89,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,90,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,91,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,92,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,93,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,94,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,95,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,96,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,97,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,98,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[112,99,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,100,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,101,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,102,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,103,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,104,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,105,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,106,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,107,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,108,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,109,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,110,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,111,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,112,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,113,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,114,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,115,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,116,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,117,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,118,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[112,119,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,120,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,121,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,122,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,123,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,124,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,125,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,126,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,127,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,128,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,129,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,130,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,131,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,132,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,133,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,134,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,135,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,136,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,137,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,138,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[112,139,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,140,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,141,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,142,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,143,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,144,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,145,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,146,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,147,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,148,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,149,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,150,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,151,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,152,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,153,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,154,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,155,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,156,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,157,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,158,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[112,159,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,160,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,161,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,162,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,163,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,164,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,165,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,166,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,167,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,168,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,169,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,170,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,171,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,172,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,173,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,174,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,175,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,176,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,177,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,178,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[112,179,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,180,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,181,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,182,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,183,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,184,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,185,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,186,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,187,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,188,3,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,189,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,190,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,191,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,192,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,193,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,194,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,195,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,196,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,197,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,198,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[112,199,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,0,1.13,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,1,1.14,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,2,1.15,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,3,1.16,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,4,1.17,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,5,1.18,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,6,1.19,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,7,1.2,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,8,1.21,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,9,1.22,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,10,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,11,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,12,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,13,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,14,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,15,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,16,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,17,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,18,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[113,19,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,20,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,21,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,22,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,23,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,24,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,25,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,26,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,27,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,28,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,29,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,30,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,31,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,32,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,33,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,34,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,35,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,36,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,37,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,38,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[113,39,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,40,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,41,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,42,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,43,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,44,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,45,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,46,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,47,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,48,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,49,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,50,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,51,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,52,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,53,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,54,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,55,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,56,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,57,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,58,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[113,59,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,60,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,61,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,62,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,63,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,64,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,65,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,66,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,67,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,68,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,69,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,70,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,71,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,72,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,73,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,74,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,75,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,76,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,77,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,78,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[113,79,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,80,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,81,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,82,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,83,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,84,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,85,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,86,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,87,2,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,88,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,89,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,90,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,91,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,92,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,93,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,94,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,95,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,96,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,97,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,98,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[113,99,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,100,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,101,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,102,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,103,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,104,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,105,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,106,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,107,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,108,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,109,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,110,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,111,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,112,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,113,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,114,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,115,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,116,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,117,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,118,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[113,119,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,120,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,121,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,122,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,123,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,124,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,125,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,126,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,127,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,128,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,129,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,130,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,131,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,132,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,133,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,134,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,135,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,136,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,137,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,138,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[113,139,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,140,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,141,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,142,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,143,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,144,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,145,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,146,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,147,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,148,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,149,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,150,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,151,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,152,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,153,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,154,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,155,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,156,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,157,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,158,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[113,159,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,160,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,161,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,162,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,163,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,164,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,165,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,166,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,167,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,168,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,169,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,170,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,171,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,172,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,173,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,174,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,175,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,176,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,177,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,178,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[113,179,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,180,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,181,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,182,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,183,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,184,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,185,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,186,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,187,3,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,188,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,189,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,190,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,191,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,192,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,193,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,194,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,195,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,196,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,197,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,198,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[113,199,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,0,1.14,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,1,1.15,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,2,1.16,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,3,1.17,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,4,1.18,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,5,1.19,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,6,1.2,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,7,1.21,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,8,1.22,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,9,1.23,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,10,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,11,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,12,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,13,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,14,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,15,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,16,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,17,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,18,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[114,19,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,20,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,21,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,22,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,23,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,24,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,25,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,26,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,27,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,28,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,29,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,30,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,31,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,32,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,33,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,34,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,35,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,36,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,37,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,38,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[114,39,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,40,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,41,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,42,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,43,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,44,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,45,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,46,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,47,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,48,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,49,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,50,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,51,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,52,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,53,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,54,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,55,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,56,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,57,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,58,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[114,59,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,60,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,61,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,62,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,63,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,64,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,65,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,66,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,67,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,68,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,69,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,70,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,71,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,72,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,73,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,74,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,75,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,76,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,77,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,78,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[114,79,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,80,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,81,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,82,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,83,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,84,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,85,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,86,2,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,87,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,88,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,89,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,90,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,91,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,92,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,93,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,94,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,95,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,96,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,97,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,98,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[114,99,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,100,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,101,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,102,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,103,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,104,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,105,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,106,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,107,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,108,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,109,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,110,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,111,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,112,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,113,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,114,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,115,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,116,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,117,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,118,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[114,119,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,120,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,121,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,122,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,123,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,124,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,125,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,126,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,127,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,128,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,129,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,130,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,131,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,132,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,133,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,134,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,135,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,136,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,137,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,138,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[114,139,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,140,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,141,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,142,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,143,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,144,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,145,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,146,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,147,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,148,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,149,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,150,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,151,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,152,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,153,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,154,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,155,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,156,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,157,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,158,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[114,159,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,160,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,161,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,162,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,163,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,164,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,165,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,166,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,167,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,168,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,169,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,170,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,171,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,172,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,173,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,174,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,175,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,176,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,177,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,178,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[114,179,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,180,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,181,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,182,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,183,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,184,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,185,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,186,3,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,187,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,188,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,189,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,190,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,191,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,192,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,193,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,194,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,195,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,196,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,197,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,198,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[114,199,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,0,1.15,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,1,1.16,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,2,1.17,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,3,1.18,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,4,1.19,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,5,1.2,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,6,1.21,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,7,1.22,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,8,1.23,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,9,1.24,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,10,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,11,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,12,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,13,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,14,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,15,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,16,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,17,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,18,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[115,19,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,20,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,21,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,22,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,23,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,24,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,25,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,26,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,27,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,28,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,29,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,30,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,31,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,32,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,33,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,34,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,35,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,36,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,37,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,38,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[115,39,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,40,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,41,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,42,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,43,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,44,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,45,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,46,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,47,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,48,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,49,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,50,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,51,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,52,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,53,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,54,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,55,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,56,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,57,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,58,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[115,59,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,60,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,61,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,62,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,63,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,64,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,65,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,66,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,67,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,68,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,69,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,70,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,71,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,72,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,73,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,74,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,75,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,76,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,77,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,78,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[115,79,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,80,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,81,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,82,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,83,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,84,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,85,2,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,86,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,87,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,88,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,89,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,90,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,91,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,92,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,93,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,94,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,95,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,96,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,97,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,98,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[115,99,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,100,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,101,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,102,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,103,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,104,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,105,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,106,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,107,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,108,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,109,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,110,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,111,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,112,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,113,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,114,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,115,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,116,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,117,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,118,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[115,119,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,120,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,121,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,122,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,123,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,124,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,125,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,126,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,127,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,128,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,129,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,130,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,131,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,132,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,133,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,134,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,135,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,136,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,137,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,138,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[115,139,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,140,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,141,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,142,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,143,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,144,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,145,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,146,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,147,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,148,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,149,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,150,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,151,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,152,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,153,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,154,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,155,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,156,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,157,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,158,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[115,159,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,160,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,161,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,162,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,163,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,164,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,165,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,166,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,167,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,168,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,169,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,170,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,171,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,172,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,173,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,174,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,175,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,176,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,177,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,178,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[115,179,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,180,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,181,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,182,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,183,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,184,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,185,3,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,186,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,187,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,188,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,189,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,190,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,191,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,192,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,193,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,194,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,195,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,196,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,197,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,198,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[115,199,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,0,1.16,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,1,1.17,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,2,1.18,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,3,1.19,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,4,1.2,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,5,1.21,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,6,1.22,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,7,1.23,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,8,1.24,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,9,1.25,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,10,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,11,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,12,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,13,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,14,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,15,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,16,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,17,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,18,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[116,19,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,20,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,21,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,22,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,23,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,24,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,25,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,26,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,27,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,28,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,29,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,30,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,31,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,32,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,33,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,34,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,35,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,36,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,37,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,38,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[116,39,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,40,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,41,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,42,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,43,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,44,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,45,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,46,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,47,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,48,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,49,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,50,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,51,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,52,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,53,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,54,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,55,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,56,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,57,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,58,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[116,59,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,60,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,61,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,62,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,63,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,64,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,65,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,66,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,67,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,68,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,69,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,70,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,71,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,72,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,73,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,74,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,75,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,76,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,77,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,78,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[116,79,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,80,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,81,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,82,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,83,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,84,2,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,85,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,86,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,87,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,88,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,89,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,90,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,91,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,92,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,93,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,94,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,95,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,96,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,97,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,98,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[116,99,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,100,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,101,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,102,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,103,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,104,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,105,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,106,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,107,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,108,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,109,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,110,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,111,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,112,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,113,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,114,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,115,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,116,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,117,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,118,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[116,119,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,120,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,121,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,122,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,123,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,124,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,125,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,126,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,127,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,128,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,129,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,130,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,131,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,132,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,133,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,134,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,135,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,136,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,137,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,138,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[116,139,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,140,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,141,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,142,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,143,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,144,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,145,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,146,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,147,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,148,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,149,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,150,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,151,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,152,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,153,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,154,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,155,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,156,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,157,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,158,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[116,159,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,160,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,161,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,162,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,163,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,164,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,165,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,166,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,167,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,168,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,169,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,170,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,171,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,172,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,173,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,174,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,175,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,176,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,177,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,178,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[116,179,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,180,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,181,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,182,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,183,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,184,3,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,185,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,186,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,187,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,188,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,189,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,190,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,191,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,192,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,193,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,194,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,195,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,196,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,197,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,198,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[116,199,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,0,1.17,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,1,1.18,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,2,1.19,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,3,1.2,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,4,1.21,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,5,1.22,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,6,1.23,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,7,1.24,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,8,1.25,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,9,1.26,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,10,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,11,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,12,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,13,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,14,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,15,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,16,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,17,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,18,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[117,19,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,20,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,21,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,22,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,23,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,24,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,25,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,26,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,27,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,28,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,29,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,30,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,31,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,32,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,33,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,34,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,35,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,36,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,37,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,38,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[117,39,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,40,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,41,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,42,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,43,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,44,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,45,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,46,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,47,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,48,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,49,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,50,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,51,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,52,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,53,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,54,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,55,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,56,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,57,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,58,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[117,59,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,60,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,61,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,62,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,63,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,64,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,65,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,66,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,67,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,68,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,69,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,70,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,71,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,72,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,73,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,74,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,75,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,76,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,77,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,78,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[117,79,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,80,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,81,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,82,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,83,2,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,84,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,85,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,86,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,87,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,88,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,89,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,90,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,91,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,92,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,93,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,94,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,95,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,96,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,97,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,98,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[117,99,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,100,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,101,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,102,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,103,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,104,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,105,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,106,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,107,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,108,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,109,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,110,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,111,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,112,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,113,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,114,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,115,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,116,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,117,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,118,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[117,119,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,120,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,121,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,122,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,123,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,124,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,125,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,126,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,127,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,128,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,129,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,130,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,131,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,132,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,133,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,134,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,135,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,136,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,137,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,138,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[117,139,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,140,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,141,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,142,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,143,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,144,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,145,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,146,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,147,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,148,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,149,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,150,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,151,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,152,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,153,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,154,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,155,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,156,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,157,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,158,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[117,159,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,160,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,161,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,162,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,163,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,164,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,165,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,166,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,167,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,168,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,169,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,170,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,171,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,172,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,173,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,174,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,175,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,176,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,177,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,178,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[117,179,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,180,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,181,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,182,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,183,3,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,184,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,185,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,186,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,187,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,188,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,189,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,190,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,191,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,192,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,193,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,194,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,195,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,196,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,197,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,198,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[117,199,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,0,1.18,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,1,1.19,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,2,1.2,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,3,1.21,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,4,1.22,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,5,1.23,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,6,1.24,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,7,1.25,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,8,1.26,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,9,1.27,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,10,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,11,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,12,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,13,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,14,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,15,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,16,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,17,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,18,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[118,19,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,20,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,21,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,22,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,23,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,24,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,25,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,26,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,27,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,28,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,29,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,30,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,31,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,32,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,33,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,34,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,35,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,36,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,37,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,38,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[118,39,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,40,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,41,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,42,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,43,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,44,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,45,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,46,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,47,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,48,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,49,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,50,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,51,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,52,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,53,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,54,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,55,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,56,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,57,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,58,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[118,59,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,60,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,61,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,62,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,63,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,64,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,65,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,66,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,67,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,68,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,69,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,70,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,71,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,72,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,73,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,74,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,75,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,76,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,77,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,78,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[118,79,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,80,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,81,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,82,2,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,83,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,84,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,85,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,86,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,87,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,88,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,89,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,90,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,91,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,92,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,93,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,94,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,95,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,96,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,97,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,98,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[118,99,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,100,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,101,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,102,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,103,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,104,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,105,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,106,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,107,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,108,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,109,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,110,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,111,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,112,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,113,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,114,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,115,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,116,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,117,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,118,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[118,119,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,120,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,121,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,122,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,123,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,124,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,125,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,126,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,127,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,128,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,129,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,130,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,131,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,132,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,133,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,134,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,135,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,136,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,137,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,138,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[118,139,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,140,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,141,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,142,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,143,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,144,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,145,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,146,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,147,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,148,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,149,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,150,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,151,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,152,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,153,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,154,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,155,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,156,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,157,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,158,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[118,159,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,160,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,161,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,162,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,163,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,164,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,165,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,166,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,167,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,168,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,169,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,170,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,171,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,172,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,173,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,174,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,175,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,176,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,177,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,178,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[118,179,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,180,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,181,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,182,3,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,183,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,184,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,185,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,186,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,187,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,188,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,189,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,190,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,191,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,192,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,193,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,194,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,195,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,196,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,197,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,198,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[118,199,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,0,1.19,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,1,1.2,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,2,1.21,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,3,1.22,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,4,1.23,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,5,1.24,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,6,1.25,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,7,1.26,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,8,1.27,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,9,1.28,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,10,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,11,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,12,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,13,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,14,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,15,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,16,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,17,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,18,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 100, pc_makepoint(1, ARRAY[119,19,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,20,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,21,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,22,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,23,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,24,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,25,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,26,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,27,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,28,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,29,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,30,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,31,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,32,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,33,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,34,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,35,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,36,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,37,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,38,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 101, pc_makepoint(1, ARRAY[119,39,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,40,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,41,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,42,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,43,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,44,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,45,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,46,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,47,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,48,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,49,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,50,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,51,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,52,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,53,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,54,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,55,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,56,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,57,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,58,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 102, pc_makepoint(1, ARRAY[119,59,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,60,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,61,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,62,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,63,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,64,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,65,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,66,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,67,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,68,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,69,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,70,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,71,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,72,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,73,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,74,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,75,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,76,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,77,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,78,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 103, pc_makepoint(1, ARRAY[119,79,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,80,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,81,2,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,82,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,83,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,84,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,85,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,86,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,87,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,88,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,89,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,90,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,91,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,92,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,93,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,94,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,95,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,96,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,97,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,98,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 104, pc_makepoint(1, ARRAY[119,99,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,100,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,101,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,102,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,103,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,104,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,105,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,106,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,107,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,108,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,109,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,110,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,111,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,112,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,113,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,114,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,115,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,116,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,117,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,118,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 105, pc_makepoint(1, ARRAY[119,119,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,120,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,121,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,122,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,123,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,124,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,125,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,126,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,127,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,128,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,129,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,130,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,131,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,132,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,133,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,134,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,135,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,136,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,137,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,138,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 106, pc_makepoint(1, ARRAY[119,139,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,140,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,141,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,142,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,143,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,144,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,145,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,146,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,147,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,148,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,149,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,150,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,151,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,152,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,153,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,154,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,155,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,156,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,157,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,158,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 107, pc_makepoint(1, ARRAY[119,159,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,160,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,161,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,162,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,163,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,164,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,165,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,166,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,167,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,168,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,169,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,170,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,171,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,172,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,173,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,174,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,175,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,176,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,177,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,178,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 108, pc_makepoint(1, ARRAY[119,179,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,180,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,181,3,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,182,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,183,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,184,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,185,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,186,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,187,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,188,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,189,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,190,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,191,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,192,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,193,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,194,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,195,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,196,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,197,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,198,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 109, pc_makepoint(1, ARRAY[119,199,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,0,1.2,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,1,1.21,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,2,1.22,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,3,1.23,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,4,1.24,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,5,1.25,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,6,1.26,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,7,1.27,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,8,1.28,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,9,1.29,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,10,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,11,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,12,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,13,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,14,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,15,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,16,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,17,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,18,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[120,19,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,20,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,21,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,22,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,23,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,24,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,25,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,26,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,27,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,28,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,29,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,30,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,31,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,32,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,33,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,34,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,35,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,36,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,37,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,38,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[120,39,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,40,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,41,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,42,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,43,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,44,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,45,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,46,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,47,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,48,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,49,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,50,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,51,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,52,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,53,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,54,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,55,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,56,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,57,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,58,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[120,59,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,60,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,61,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,62,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,63,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,64,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,65,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,66,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,67,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,68,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,69,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,70,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,71,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,72,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,73,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,74,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,75,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,76,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,77,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,78,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[120,79,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,80,2,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,81,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,82,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,83,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,84,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,85,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,86,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,87,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,88,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,89,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,90,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,91,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,92,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,93,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,94,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,95,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,96,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,97,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,98,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[120,99,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,100,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,101,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,102,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,103,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,104,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,105,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,106,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,107,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,108,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,109,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,110,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,111,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,112,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,113,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,114,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,115,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,116,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,117,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,118,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[120,119,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,120,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,121,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,122,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,123,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,124,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,125,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,126,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,127,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,128,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,129,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,130,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,131,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,132,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,133,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,134,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,135,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,136,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,137,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,138,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[120,139,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,140,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,141,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,142,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,143,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,144,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,145,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,146,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,147,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,148,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,149,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,150,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,151,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,152,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,153,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,154,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,155,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,156,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,157,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,158,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[120,159,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,160,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,161,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,162,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,163,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,164,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,165,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,166,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,167,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,168,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,169,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,170,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,171,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,172,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,173,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,174,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,175,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,176,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,177,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,178,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[120,179,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,180,3,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,181,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,182,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,183,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,184,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,185,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,186,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,187,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,188,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,189,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,190,3.1,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,191,3.11,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,192,3.12,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,193,3.13,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,194,3.14,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,195,3.15,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,196,3.16,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,197,3.17,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,198,3.18,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[120,199,3.19,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,0,1.21,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,1,1.22,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,2,1.23,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,3,1.24,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,4,1.25,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,5,1.26,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,6,1.27,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,7,1.28,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,8,1.29,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,9,1.3,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,10,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,11,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,12,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,13,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,14,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,15,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,16,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,17,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,18,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[121,19,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,20,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,21,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,22,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,23,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,24,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,25,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,26,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,27,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,28,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,29,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,30,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,31,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,32,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,33,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,34,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,35,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,36,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,37,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,38,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[121,39,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,40,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,41,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,42,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,43,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,44,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,45,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,46,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,47,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,48,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,49,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,50,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,51,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,52,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,53,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,54,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,55,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,56,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,57,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,58,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[121,59,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,60,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,61,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,62,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,63,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,64,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,65,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,66,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,67,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,68,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,69,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,70,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,71,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,72,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,73,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,74,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,75,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,76,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,77,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,78,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[121,79,2,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,80,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,81,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,82,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,83,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,84,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,85,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,86,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,87,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,88,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,89,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,90,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,91,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,92,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,93,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,94,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,95,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,96,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,97,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,98,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[121,99,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,100,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,101,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,102,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,103,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,104,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,105,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,106,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,107,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,108,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,109,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,110,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,111,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,112,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,113,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,114,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,115,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,116,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,117,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,118,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[121,119,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,120,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,121,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,122,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,123,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,124,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,125,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,126,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,127,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,128,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,129,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,130,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,131,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,132,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,133,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,134,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,135,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,136,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,137,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,138,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[121,139,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,140,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,141,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,142,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,143,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,144,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,145,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,146,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,147,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,148,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,149,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,150,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,151,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,152,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,153,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,154,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,155,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,156,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,157,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,158,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[121,159,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,160,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,161,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,162,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,163,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,164,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,165,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,166,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,167,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,168,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,169,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,170,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,171,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,172,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,173,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,174,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,175,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,176,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,177,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,178,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[121,179,3,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,180,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,181,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,182,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,183,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,184,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,185,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,186,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,187,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,188,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,189,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,190,3.11,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,191,3.12,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,192,3.13,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,193,3.14,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,194,3.15,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,195,3.16,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,196,3.17,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,197,3.18,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,198,3.19,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[121,199,3.2,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,0,1.22,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,1,1.23,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,2,1.24,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,3,1.25,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,4,1.26,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,5,1.27,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,6,1.28,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,7,1.29,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,8,1.3,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,9,1.31,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,10,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,11,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,12,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,13,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,14,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,15,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,16,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,17,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,18,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[122,19,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,20,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,21,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,22,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,23,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,24,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,25,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,26,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,27,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,28,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,29,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,30,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,31,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,32,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,33,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,34,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,35,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,36,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,37,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,38,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[122,39,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,40,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,41,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,42,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,43,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,44,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,45,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,46,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,47,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,48,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,49,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,50,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,51,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,52,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,53,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,54,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,55,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,56,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,57,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,58,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[122,59,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,60,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,61,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,62,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,63,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,64,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,65,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,66,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,67,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,68,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,69,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,70,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,71,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,72,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,73,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,74,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,75,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,76,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,77,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,78,2,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[122,79,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,80,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,81,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,82,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,83,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,84,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,85,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,86,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,87,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,88,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,89,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,90,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,91,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,92,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,93,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,94,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,95,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,96,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,97,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,98,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[122,99,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,100,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,101,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,102,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,103,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,104,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,105,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,106,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,107,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,108,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,109,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,110,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,111,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,112,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,113,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,114,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,115,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,116,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,117,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,118,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[122,119,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,120,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,121,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,122,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,123,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,124,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,125,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,126,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,127,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,128,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,129,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,130,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,131,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,132,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,133,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,134,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,135,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,136,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,137,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,138,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[122,139,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,140,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,141,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,142,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,143,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,144,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,145,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,146,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,147,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,148,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,149,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,150,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,151,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,152,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,153,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,154,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,155,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,156,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,157,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,158,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[122,159,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,160,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,161,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,162,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,163,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,164,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,165,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,166,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,167,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,168,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,169,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,170,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,171,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,172,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,173,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,174,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,175,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,176,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,177,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,178,3,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[122,179,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,180,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,181,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,182,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,183,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,184,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,185,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,186,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,187,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,188,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,189,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,190,3.12,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,191,3.13,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,192,3.14,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,193,3.15,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,194,3.16,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,195,3.17,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,196,3.18,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,197,3.19,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,198,3.2,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[122,199,3.21,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,0,1.23,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,1,1.24,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,2,1.25,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,3,1.26,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,4,1.27,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,5,1.28,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,6,1.29,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,7,1.3,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,8,1.31,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,9,1.32,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,10,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,11,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,12,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,13,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,14,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,15,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,16,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,17,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,18,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[123,19,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,20,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,21,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,22,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,23,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,24,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,25,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,26,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,27,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,28,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,29,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,30,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,31,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,32,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,33,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,34,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,35,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,36,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,37,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,38,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[123,39,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,40,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,41,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,42,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,43,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,44,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,45,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,46,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,47,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,48,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,49,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,50,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,51,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,52,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,53,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,54,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,55,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,56,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,57,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,58,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[123,59,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,60,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,61,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,62,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,63,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,64,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,65,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,66,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,67,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,68,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,69,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,70,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,71,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,72,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,73,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,74,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,75,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,76,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,77,2,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,78,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[123,79,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,80,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,81,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,82,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,83,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,84,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,85,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,86,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,87,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,88,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,89,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,90,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,91,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,92,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,93,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,94,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,95,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,96,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,97,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,98,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[123,99,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,100,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,101,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,102,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,103,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,104,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,105,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,106,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,107,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,108,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,109,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,110,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,111,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,112,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,113,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,114,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,115,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,116,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,117,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,118,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[123,119,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,120,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,121,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,122,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,123,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,124,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,125,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,126,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,127,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,128,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,129,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,130,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,131,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,132,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,133,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,134,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,135,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,136,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,137,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,138,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[123,139,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,140,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,141,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,142,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,143,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,144,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,145,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,146,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,147,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,148,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,149,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,150,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,151,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,152,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,153,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,154,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,155,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,156,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,157,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,158,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[123,159,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,160,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,161,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,162,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,163,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,164,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,165,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,166,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,167,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,168,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,169,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,170,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,171,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,172,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,173,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,174,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,175,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,176,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,177,3,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,178,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[123,179,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,180,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,181,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,182,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,183,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,184,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,185,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,186,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,187,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,188,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,189,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,190,3.13,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,191,3.14,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,192,3.15,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,193,3.16,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,194,3.17,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,195,3.18,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,196,3.19,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,197,3.2,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,198,3.21,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[123,199,3.22,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,0,1.24,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,1,1.25,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,2,1.26,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,3,1.27,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,4,1.28,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,5,1.29,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,6,1.3,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,7,1.31,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,8,1.32,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,9,1.33,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,10,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,11,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,12,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,13,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,14,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,15,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,16,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,17,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,18,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[124,19,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,20,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,21,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,22,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,23,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,24,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,25,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,26,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,27,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,28,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,29,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,30,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,31,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,32,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,33,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,34,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,35,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,36,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,37,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,38,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[124,39,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,40,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,41,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,42,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,43,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,44,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,45,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,46,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,47,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,48,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,49,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,50,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,51,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,52,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,53,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,54,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,55,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,56,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,57,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,58,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[124,59,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,60,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,61,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,62,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,63,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,64,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,65,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,66,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,67,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,68,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,69,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,70,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,71,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,72,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,73,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,74,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,75,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,76,2,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,77,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,78,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[124,79,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,80,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,81,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,82,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,83,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,84,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,85,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,86,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,87,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,88,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,89,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,90,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,91,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,92,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,93,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,94,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,95,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,96,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,97,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,98,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[124,99,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,100,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,101,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,102,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,103,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,104,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,105,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,106,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,107,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,108,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,109,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,110,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,111,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,112,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,113,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,114,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,115,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,116,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,117,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,118,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[124,119,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,120,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,121,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,122,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,123,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,124,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,125,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,126,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,127,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,128,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,129,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,130,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,131,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,132,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,133,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,134,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,135,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,136,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,137,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,138,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[124,139,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,140,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,141,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,142,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,143,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,144,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,145,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,146,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,147,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,148,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,149,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,150,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,151,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,152,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,153,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,154,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,155,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,156,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,157,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,158,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[124,159,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,160,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,161,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,162,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,163,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,164,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,165,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,166,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,167,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,168,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,169,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,170,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,171,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,172,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,173,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,174,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,175,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,176,3,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,177,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,178,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[124,179,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,180,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,181,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,182,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,183,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,184,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,185,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,186,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,187,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,188,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,189,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,190,3.14,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,191,3.15,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,192,3.16,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,193,3.17,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,194,3.18,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,195,3.19,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,196,3.2,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,197,3.21,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,198,3.22,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[124,199,3.23,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,0,1.25,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,1,1.26,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,2,1.27,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,3,1.28,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,4,1.29,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,5,1.3,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,6,1.31,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,7,1.32,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,8,1.33,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,9,1.34,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,10,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,11,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,12,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,13,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,14,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,15,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,16,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,17,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,18,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[125,19,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,20,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,21,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,22,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,23,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,24,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,25,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,26,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,27,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,28,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,29,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,30,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,31,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,32,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,33,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,34,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,35,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,36,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,37,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,38,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[125,39,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,40,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,41,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,42,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,43,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,44,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,45,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,46,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,47,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,48,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,49,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,50,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,51,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,52,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,53,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,54,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,55,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,56,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,57,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,58,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[125,59,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,60,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,61,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,62,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,63,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,64,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,65,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,66,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,67,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,68,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,69,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,70,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,71,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,72,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,73,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,74,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,75,2,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,76,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,77,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,78,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[125,79,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,80,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,81,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,82,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,83,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,84,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,85,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,86,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,87,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,88,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,89,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,90,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,91,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,92,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,93,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,94,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,95,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,96,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,97,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,98,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[125,99,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,100,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,101,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,102,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,103,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,104,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,105,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,106,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,107,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,108,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,109,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,110,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,111,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,112,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,113,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,114,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,115,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,116,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,117,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,118,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[125,119,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,120,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,121,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,122,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,123,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,124,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,125,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,126,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,127,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,128,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,129,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,130,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,131,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,132,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,133,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,134,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,135,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,136,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,137,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,138,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[125,139,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,140,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,141,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,142,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,143,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,144,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,145,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,146,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,147,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,148,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,149,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,150,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,151,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,152,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,153,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,154,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,155,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,156,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,157,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,158,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[125,159,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,160,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,161,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,162,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,163,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,164,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,165,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,166,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,167,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,168,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,169,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,170,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,171,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,172,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,173,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,174,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,175,3,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,176,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,177,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,178,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[125,179,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,180,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,181,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,182,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,183,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,184,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,185,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,186,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,187,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,188,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,189,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,190,3.15,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,191,3.16,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,192,3.17,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,193,3.18,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,194,3.19,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,195,3.2,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,196,3.21,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,197,3.22,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,198,3.23,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[125,199,3.24,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,0,1.26,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,1,1.27,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,2,1.28,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,3,1.29,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,4,1.3,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,5,1.31,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,6,1.32,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,7,1.33,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,8,1.34,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,9,1.35,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,10,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,11,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,12,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,13,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,14,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,15,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,16,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,17,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,18,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[126,19,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,20,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,21,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,22,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,23,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,24,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,25,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,26,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,27,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,28,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,29,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,30,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,31,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,32,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,33,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,34,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,35,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,36,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,37,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,38,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[126,39,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,40,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,41,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,42,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,43,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,44,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,45,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,46,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,47,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,48,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,49,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,50,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,51,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,52,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,53,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,54,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,55,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,56,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,57,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,58,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[126,59,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,60,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,61,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,62,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,63,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,64,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,65,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,66,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,67,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,68,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,69,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,70,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,71,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,72,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,73,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,74,2,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,75,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,76,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,77,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,78,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[126,79,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,80,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,81,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,82,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,83,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,84,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,85,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,86,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,87,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,88,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,89,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,90,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,91,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,92,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,93,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,94,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,95,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,96,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,97,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,98,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[126,99,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,100,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,101,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,102,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,103,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,104,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,105,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,106,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,107,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,108,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,109,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,110,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,111,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,112,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,113,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,114,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,115,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,116,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,117,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,118,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[126,119,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,120,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,121,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,122,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,123,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,124,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,125,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,126,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,127,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,128,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,129,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,130,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,131,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,132,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,133,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,134,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,135,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,136,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,137,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,138,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[126,139,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,140,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,141,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,142,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,143,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,144,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,145,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,146,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,147,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,148,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,149,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,150,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,151,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,152,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,153,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,154,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,155,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,156,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,157,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,158,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[126,159,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,160,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,161,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,162,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,163,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,164,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,165,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,166,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,167,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,168,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,169,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,170,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,171,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,172,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,173,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,174,3,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,175,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,176,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,177,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,178,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[126,179,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,180,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,181,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,182,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,183,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,184,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,185,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,186,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,187,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,188,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,189,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,190,3.16,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,191,3.17,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,192,3.18,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,193,3.19,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,194,3.2,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,195,3.21,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,196,3.22,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,197,3.23,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,198,3.24,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[126,199,3.25,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,0,1.27,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,1,1.28,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,2,1.29,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,3,1.3,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,4,1.31,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,5,1.32,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,6,1.33,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,7,1.34,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,8,1.35,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,9,1.36,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,10,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,11,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,12,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,13,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,14,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,15,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,16,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,17,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,18,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[127,19,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,20,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,21,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,22,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,23,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,24,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,25,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,26,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,27,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,28,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,29,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,30,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,31,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,32,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,33,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,34,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,35,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,36,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,37,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,38,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[127,39,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,40,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,41,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,42,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,43,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,44,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,45,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,46,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,47,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,48,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,49,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,50,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,51,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,52,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,53,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,54,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,55,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,56,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,57,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,58,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[127,59,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,60,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,61,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,62,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,63,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,64,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,65,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,66,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,67,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,68,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,69,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,70,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,71,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,72,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,73,2,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,74,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,75,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,76,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,77,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,78,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[127,79,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,80,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,81,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,82,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,83,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,84,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,85,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,86,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,87,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,88,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,89,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,90,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,91,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,92,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,93,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,94,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,95,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,96,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,97,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,98,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[127,99,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,100,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,101,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,102,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,103,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,104,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,105,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,106,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,107,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,108,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,109,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,110,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,111,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,112,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,113,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,114,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,115,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,116,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,117,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,118,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[127,119,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,120,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,121,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,122,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,123,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,124,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,125,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,126,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,127,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,128,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,129,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,130,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,131,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,132,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,133,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,134,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,135,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,136,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,137,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,138,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[127,139,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,140,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,141,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,142,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,143,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,144,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,145,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,146,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,147,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,148,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,149,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,150,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,151,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,152,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,153,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,154,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,155,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,156,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,157,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,158,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[127,159,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,160,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,161,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,162,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,163,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,164,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,165,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,166,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,167,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,168,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,169,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,170,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,171,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,172,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,173,3,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,174,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,175,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,176,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,177,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,178,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[127,179,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,180,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,181,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,182,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,183,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,184,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,185,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,186,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,187,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,188,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,189,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,190,3.17,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,191,3.18,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,192,3.19,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,193,3.2,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,194,3.21,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,195,3.22,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,196,3.23,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,197,3.24,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,198,3.25,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[127,199,3.26,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,0,1.28,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,1,1.29,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,2,1.3,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,3,1.31,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,4,1.32,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,5,1.33,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,6,1.34,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,7,1.35,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,8,1.36,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,9,1.37,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,10,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,11,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,12,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,13,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,14,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,15,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,16,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,17,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,18,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[128,19,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,20,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,21,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,22,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,23,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,24,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,25,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,26,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,27,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,28,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,29,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,30,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,31,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,32,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,33,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,34,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,35,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,36,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,37,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,38,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[128,39,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,40,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,41,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,42,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,43,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,44,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,45,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,46,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,47,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,48,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,49,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,50,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,51,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,52,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,53,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,54,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,55,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,56,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,57,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,58,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[128,59,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,60,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,61,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,62,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,63,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,64,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,65,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,66,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,67,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,68,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,69,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,70,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,71,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,72,2,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,73,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,74,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,75,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,76,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,77,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,78,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[128,79,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,80,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,81,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,82,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,83,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,84,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,85,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,86,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,87,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,88,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,89,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,90,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,91,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,92,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,93,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,94,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,95,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,96,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,97,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,98,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[128,99,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,100,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,101,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,102,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,103,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,104,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,105,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,106,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,107,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,108,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,109,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,110,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,111,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,112,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,113,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,114,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,115,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,116,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,117,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,118,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[128,119,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,120,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,121,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,122,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,123,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,124,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,125,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,126,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,127,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,128,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,129,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,130,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,131,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,132,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,133,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,134,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,135,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,136,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,137,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,138,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[128,139,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,140,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,141,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,142,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,143,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,144,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,145,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,146,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,147,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,148,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,149,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,150,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,151,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,152,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,153,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,154,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,155,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,156,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,157,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,158,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[128,159,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,160,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,161,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,162,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,163,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,164,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,165,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,166,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,167,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,168,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,169,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,170,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,171,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,172,3,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,173,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,174,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,175,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,176,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,177,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,178,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[128,179,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,180,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,181,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,182,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,183,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,184,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,185,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,186,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,187,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,188,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,189,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,190,3.18,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,191,3.19,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,192,3.2,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,193,3.21,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,194,3.22,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,195,3.23,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,196,3.24,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,197,3.25,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,198,3.26,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[128,199,3.27,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,0,1.29,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,1,1.3,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,2,1.31,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,3,1.32,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,4,1.33,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,5,1.34,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,6,1.35,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,7,1.36,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,8,1.37,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,9,1.38,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,10,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,11,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,12,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,13,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,14,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,15,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,16,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,17,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,18,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[129,19,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,20,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,21,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,22,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,23,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,24,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,25,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,26,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,27,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,28,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,29,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,30,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,31,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,32,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,33,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,34,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,35,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,36,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,37,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,38,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[129,39,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,40,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,41,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,42,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,43,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,44,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,45,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,46,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,47,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,48,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,49,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,50,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,51,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,52,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,53,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,54,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,55,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,56,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,57,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,58,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[129,59,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,60,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,61,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,62,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,63,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,64,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,65,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,66,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,67,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,68,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,69,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,70,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,71,2,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,72,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,73,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,74,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,75,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,76,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,77,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,78,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[129,79,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,80,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,81,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,82,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,83,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,84,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,85,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,86,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,87,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,88,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,89,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,90,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,91,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,92,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,93,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,94,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,95,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,96,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,97,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,98,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[129,99,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,100,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,101,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,102,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,103,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,104,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,105,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,106,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,107,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,108,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,109,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,110,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,111,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,112,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,113,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,114,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,115,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,116,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,117,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,118,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[129,119,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,120,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,121,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,122,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,123,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,124,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,125,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,126,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,127,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,128,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,129,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,130,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,131,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,132,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,133,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,134,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,135,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,136,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,137,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,138,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[129,139,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,140,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,141,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,142,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,143,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,144,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,145,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,146,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,147,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,148,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,149,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,150,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,151,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,152,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,153,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,154,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,155,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,156,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,157,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,158,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[129,159,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,160,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,161,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,162,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,163,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,164,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,165,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,166,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,167,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,168,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,169,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,170,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,171,3,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,172,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,173,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,174,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,175,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,176,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,177,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,178,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[129,179,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,180,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,181,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,182,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,183,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,184,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,185,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,186,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,187,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,188,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,189,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,190,3.19,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,191,3.2,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,192,3.21,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,193,3.22,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,194,3.23,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,195,3.24,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,196,3.25,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,197,3.26,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,198,3.27,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[129,199,3.28,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,0,1.3,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,1,1.31,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,2,1.32,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,3,1.33,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,4,1.34,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,5,1.35,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,6,1.36,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,7,1.37,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,8,1.38,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,9,1.39,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,10,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,11,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,12,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,13,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,14,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,15,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,16,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,17,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,18,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[130,19,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,20,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,21,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,22,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,23,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,24,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,25,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,26,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,27,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,28,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,29,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,30,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,31,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,32,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,33,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,34,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,35,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,36,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,37,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,38,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[130,39,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,40,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,41,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,42,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,43,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,44,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,45,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,46,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,47,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,48,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,49,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,50,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,51,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,52,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,53,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,54,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,55,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,56,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,57,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,58,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[130,59,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,60,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,61,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,62,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,63,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,64,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,65,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,66,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,67,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,68,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,69,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,70,2,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,71,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,72,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,73,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,74,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,75,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,76,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,77,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,78,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[130,79,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,80,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,81,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,82,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,83,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,84,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,85,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,86,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,87,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,88,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,89,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,90,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,91,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,92,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,93,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,94,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,95,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,96,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,97,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,98,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[130,99,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,100,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,101,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,102,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,103,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,104,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,105,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,106,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,107,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,108,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,109,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,110,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,111,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,112,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,113,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,114,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,115,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,116,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,117,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,118,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[130,119,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,120,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,121,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,122,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,123,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,124,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,125,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,126,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,127,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,128,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,129,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,130,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,131,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,132,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,133,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,134,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,135,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,136,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,137,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,138,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[130,139,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,140,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,141,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,142,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,143,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,144,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,145,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,146,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,147,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,148,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,149,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,150,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,151,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,152,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,153,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,154,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,155,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,156,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,157,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,158,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[130,159,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,160,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,161,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,162,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,163,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,164,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,165,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,166,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,167,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,168,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,169,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,170,3,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,171,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,172,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,173,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,174,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,175,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,176,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,177,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,178,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[130,179,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,180,3.1,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,181,3.11,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,182,3.12,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,183,3.13,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,184,3.14,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,185,3.15,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,186,3.16,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,187,3.17,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,188,3.18,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,189,3.19,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,190,3.2,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,191,3.21,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,192,3.22,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,193,3.23,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,194,3.24,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,195,3.25,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,196,3.26,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,197,3.27,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,198,3.28,0]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[130,199,3.29,0]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,0,1.31,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,1,1.32,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,2,1.33,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,3,1.34,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,4,1.35,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,5,1.36,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,6,1.37,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,7,1.38,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,8,1.39,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,9,1.4,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,10,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,11,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,12,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,13,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,14,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,15,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,16,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,17,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,18,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[131,19,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,20,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,21,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,22,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,23,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,24,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,25,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,26,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,27,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,28,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,29,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,30,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,31,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,32,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,33,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,34,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,35,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,36,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,37,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,38,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[131,39,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,40,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,41,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,42,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,43,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,44,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,45,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,46,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,47,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,48,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,49,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,50,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,51,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,52,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,53,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,54,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,55,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,56,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,57,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,58,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[131,59,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,60,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,61,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,62,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,63,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,64,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,65,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,66,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,67,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,68,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,69,2,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,70,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,71,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,72,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,73,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,74,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,75,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,76,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,77,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,78,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[131,79,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,80,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,81,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,82,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,83,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,84,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,85,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,86,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,87,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,88,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,89,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,90,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,91,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,92,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,93,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,94,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,95,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,96,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,97,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,98,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[131,99,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,100,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,101,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,102,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,103,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,104,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,105,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,106,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,107,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,108,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,109,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,110,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,111,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,112,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,113,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,114,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,115,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,116,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,117,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,118,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[131,119,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,120,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,121,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,122,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,123,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,124,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,125,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,126,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,127,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,128,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,129,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,130,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,131,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,132,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,133,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,134,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,135,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,136,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,137,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,138,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[131,139,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,140,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,141,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,142,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,143,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,144,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,145,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,146,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,147,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,148,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,149,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,150,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,151,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,152,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,153,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,154,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,155,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,156,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,157,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,158,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[131,159,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,160,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,161,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,162,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,163,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,164,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,165,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,166,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,167,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,168,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,169,3,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,170,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,171,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,172,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,173,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,174,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,175,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,176,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,177,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,178,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[131,179,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,180,3.11,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,181,3.12,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,182,3.13,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,183,3.14,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,184,3.15,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,185,3.16,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,186,3.17,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,187,3.18,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,188,3.19,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,189,3.2,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,190,3.21,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,191,3.22,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,192,3.23,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,193,3.24,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,194,3.25,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,195,3.26,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,196,3.27,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,197,3.28,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,198,3.29,1]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[131,199,3.3,1]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,0,1.32,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,1,1.33,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,2,1.34,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,3,1.35,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,4,1.36,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,5,1.37,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,6,1.38,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,7,1.39,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,8,1.4,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,9,1.41,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,10,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,11,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,12,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,13,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,14,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,15,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,16,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,17,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,18,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[132,19,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,20,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,21,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,22,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,23,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,24,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,25,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,26,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,27,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,28,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,29,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,30,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,31,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,32,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,33,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,34,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,35,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,36,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,37,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,38,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[132,39,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,40,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,41,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,42,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,43,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,44,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,45,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,46,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,47,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,48,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,49,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,50,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,51,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,52,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,53,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,54,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,55,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,56,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,57,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,58,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[132,59,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,60,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,61,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,62,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,63,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,64,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,65,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,66,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,67,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,68,2,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,69,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,70,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,71,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,72,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,73,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,74,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,75,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,76,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,77,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,78,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[132,79,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,80,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,81,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,82,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,83,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,84,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,85,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,86,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,87,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,88,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,89,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,90,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,91,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,92,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,93,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,94,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,95,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,96,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,97,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,98,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[132,99,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,100,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,101,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,102,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,103,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,104,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,105,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,106,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,107,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,108,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,109,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,110,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,111,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,112,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,113,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,114,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,115,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,116,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,117,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,118,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[132,119,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,120,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,121,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,122,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,123,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,124,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,125,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,126,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,127,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,128,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,129,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,130,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,131,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,132,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,133,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,134,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,135,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,136,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,137,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,138,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[132,139,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,140,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,141,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,142,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,143,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,144,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,145,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,146,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,147,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,148,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,149,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,150,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,151,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,152,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,153,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,154,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,155,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,156,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,157,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,158,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[132,159,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,160,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,161,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,162,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,163,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,164,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,165,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,166,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,167,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,168,3,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,169,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,170,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,171,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,172,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,173,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,174,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,175,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,176,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,177,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,178,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[132,179,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,180,3.12,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,181,3.13,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,182,3.14,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,183,3.15,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,184,3.16,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,185,3.17,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,186,3.18,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,187,3.19,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,188,3.2,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,189,3.21,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,190,3.22,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,191,3.23,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,192,3.24,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,193,3.25,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,194,3.26,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,195,3.27,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,196,3.28,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,197,3.29,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,198,3.3,2]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[132,199,3.31,2]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,0,1.33,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,1,1.34,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,2,1.35,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,3,1.36,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,4,1.37,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,5,1.38,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,6,1.39,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,7,1.4,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,8,1.41,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,9,1.42,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,10,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,11,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,12,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,13,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,14,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,15,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,16,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,17,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,18,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[133,19,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,20,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,21,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,22,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,23,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,24,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,25,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,26,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,27,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,28,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,29,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,30,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,31,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,32,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,33,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,34,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,35,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,36,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,37,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,38,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[133,39,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,40,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,41,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,42,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,43,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,44,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,45,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,46,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,47,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,48,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,49,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,50,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,51,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,52,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,53,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,54,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,55,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,56,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,57,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,58,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[133,59,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,60,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,61,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,62,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,63,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,64,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,65,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,66,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,67,2,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,68,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,69,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,70,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,71,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,72,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,73,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,74,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,75,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,76,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,77,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,78,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[133,79,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,80,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,81,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,82,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,83,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,84,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,85,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,86,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,87,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,88,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,89,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,90,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,91,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,92,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,93,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,94,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,95,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,96,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,97,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,98,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[133,99,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,100,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,101,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,102,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,103,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,104,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,105,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,106,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,107,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,108,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,109,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,110,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,111,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,112,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,113,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,114,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,115,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,116,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,117,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,118,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[133,119,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,120,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,121,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,122,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,123,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,124,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,125,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,126,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,127,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,128,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,129,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,130,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,131,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,132,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,133,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,134,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,135,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,136,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,137,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,138,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[133,139,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,140,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,141,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,142,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,143,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,144,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,145,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,146,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,147,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,148,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,149,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,150,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,151,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,152,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,153,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,154,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,155,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,156,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,157,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,158,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[133,159,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,160,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,161,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,162,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,163,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,164,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,165,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,166,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,167,3,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,168,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,169,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,170,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,171,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,172,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,173,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,174,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,175,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,176,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,177,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,178,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[133,179,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,180,3.13,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,181,3.14,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,182,3.15,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,183,3.16,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,184,3.17,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,185,3.18,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,186,3.19,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,187,3.2,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,188,3.21,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,189,3.22,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,190,3.23,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,191,3.24,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,192,3.25,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,193,3.26,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,194,3.27,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,195,3.28,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,196,3.29,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,197,3.3,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,198,3.31,3]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[133,199,3.32,3]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,0,1.34,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,1,1.35,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,2,1.36,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,3,1.37,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,4,1.38,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,5,1.39,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,6,1.4,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,7,1.41,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,8,1.42,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,9,1.43,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,10,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,11,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,12,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,13,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,14,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,15,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,16,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,17,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,18,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[134,19,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,20,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,21,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,22,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,23,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,24,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,25,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,26,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,27,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,28,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,29,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,30,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,31,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,32,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,33,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,34,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,35,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,36,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,37,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,38,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[134,39,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,40,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,41,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,42,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,43,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,44,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,45,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,46,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,47,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,48,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,49,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,50,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,51,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,52,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,53,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,54,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,55,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,56,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,57,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,58,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[134,59,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,60,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,61,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,62,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,63,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,64,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,65,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,66,2,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,67,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,68,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,69,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,70,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,71,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,72,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,73,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,74,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,75,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,76,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,77,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,78,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[134,79,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,80,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,81,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,82,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,83,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,84,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,85,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,86,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,87,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,88,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,89,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,90,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,91,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,92,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,93,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,94,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,95,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,96,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,97,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,98,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[134,99,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,100,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,101,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,102,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,103,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,104,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,105,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,106,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,107,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,108,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,109,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,110,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,111,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,112,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,113,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,114,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,115,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,116,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,117,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,118,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[134,119,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,120,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,121,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,122,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,123,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,124,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,125,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,126,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,127,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,128,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,129,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,130,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,131,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,132,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,133,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,134,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,135,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,136,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,137,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,138,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[134,139,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,140,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,141,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,142,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,143,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,144,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,145,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,146,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,147,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,148,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,149,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,150,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,151,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,152,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,153,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,154,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,155,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,156,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,157,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,158,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[134,159,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,160,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,161,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,162,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,163,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,164,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,165,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,166,3,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,167,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,168,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,169,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,170,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,171,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,172,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,173,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,174,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,175,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,176,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,177,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,178,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[134,179,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,180,3.14,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,181,3.15,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,182,3.16,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,183,3.17,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,184,3.18,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,185,3.19,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,186,3.2,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,187,3.21,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,188,3.22,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,189,3.23,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,190,3.24,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,191,3.25,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,192,3.26,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,193,3.27,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,194,3.28,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,195,3.29,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,196,3.3,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,197,3.31,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,198,3.32,4]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[134,199,3.33,4]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,0,1.35,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,1,1.36,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,2,1.37,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,3,1.38,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,4,1.39,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,5,1.4,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,6,1.41,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,7,1.42,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,8,1.43,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,9,1.44,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,10,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,11,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,12,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,13,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,14,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,15,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,16,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,17,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,18,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[135,19,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,20,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,21,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,22,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,23,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,24,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,25,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,26,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,27,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,28,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,29,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,30,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,31,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,32,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,33,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,34,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,35,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,36,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,37,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,38,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[135,39,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,40,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,41,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,42,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,43,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,44,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,45,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,46,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,47,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,48,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,49,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,50,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,51,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,52,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,53,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,54,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,55,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,56,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,57,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,58,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[135,59,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,60,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,61,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,62,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,63,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,64,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,65,2,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,66,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,67,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,68,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,69,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,70,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,71,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,72,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,73,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,74,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,75,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,76,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,77,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,78,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[135,79,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,80,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,81,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,82,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,83,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,84,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,85,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,86,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,87,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,88,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,89,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,90,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,91,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,92,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,93,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,94,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,95,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,96,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,97,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,98,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[135,99,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,100,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,101,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,102,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,103,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,104,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,105,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,106,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,107,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,108,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,109,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,110,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,111,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,112,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,113,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,114,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,115,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,116,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,117,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,118,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[135,119,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,120,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,121,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,122,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,123,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,124,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,125,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,126,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,127,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,128,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,129,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,130,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,131,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,132,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,133,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,134,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,135,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,136,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,137,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,138,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[135,139,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,140,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,141,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,142,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,143,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,144,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,145,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,146,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,147,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,148,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,149,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,150,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,151,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,152,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,153,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,154,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,155,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,156,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,157,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,158,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[135,159,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,160,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,161,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,162,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,163,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,164,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,165,3,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,166,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,167,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,168,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,169,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,170,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,171,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,172,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,173,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,174,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,175,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,176,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,177,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,178,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[135,179,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,180,3.15,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,181,3.16,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,182,3.17,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,183,3.18,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,184,3.19,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,185,3.2,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,186,3.21,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,187,3.22,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,188,3.23,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,189,3.24,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,190,3.25,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,191,3.26,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,192,3.27,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,193,3.28,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,194,3.29,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,195,3.3,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,196,3.31,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,197,3.32,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,198,3.33,5]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[135,199,3.34,5]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,0,1.36,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,1,1.37,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,2,1.38,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,3,1.39,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,4,1.4,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,5,1.41,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,6,1.42,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,7,1.43,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,8,1.44,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,9,1.45,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,10,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,11,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,12,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,13,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,14,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,15,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,16,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,17,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,18,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[136,19,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,20,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,21,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,22,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,23,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,24,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,25,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,26,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,27,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,28,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,29,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,30,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,31,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,32,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,33,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,34,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,35,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,36,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,37,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,38,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[136,39,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,40,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,41,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,42,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,43,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,44,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,45,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,46,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,47,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,48,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,49,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,50,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,51,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,52,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,53,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,54,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,55,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,56,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,57,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,58,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[136,59,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,60,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,61,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,62,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,63,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,64,2,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,65,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,66,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,67,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,68,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,69,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,70,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,71,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,72,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,73,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,74,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,75,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,76,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,77,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,78,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[136,79,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,80,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,81,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,82,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,83,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,84,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,85,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,86,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,87,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,88,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,89,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,90,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,91,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,92,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,93,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,94,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,95,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,96,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,97,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,98,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[136,99,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,100,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,101,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,102,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,103,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,104,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,105,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,106,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,107,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,108,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,109,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,110,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,111,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,112,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,113,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,114,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,115,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,116,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,117,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,118,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[136,119,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,120,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,121,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,122,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,123,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,124,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,125,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,126,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,127,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,128,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,129,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,130,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,131,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,132,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,133,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,134,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,135,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,136,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,137,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,138,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[136,139,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,140,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,141,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,142,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,143,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,144,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,145,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,146,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,147,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,148,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,149,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,150,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,151,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,152,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,153,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,154,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,155,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,156,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,157,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,158,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[136,159,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,160,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,161,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,162,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,163,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,164,3,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,165,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,166,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,167,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,168,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,169,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,170,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,171,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,172,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,173,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,174,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,175,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,176,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,177,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,178,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[136,179,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,180,3.16,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,181,3.17,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,182,3.18,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,183,3.19,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,184,3.2,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,185,3.21,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,186,3.22,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,187,3.23,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,188,3.24,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,189,3.25,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,190,3.26,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,191,3.27,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,192,3.28,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,193,3.29,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,194,3.3,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,195,3.31,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,196,3.32,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,197,3.33,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,198,3.34,6]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[136,199,3.35,6]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,0,1.37,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,1,1.38,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,2,1.39,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,3,1.4,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,4,1.41,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,5,1.42,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,6,1.43,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,7,1.44,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,8,1.45,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,9,1.46,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,10,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,11,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,12,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,13,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,14,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,15,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,16,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,17,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,18,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[137,19,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,20,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,21,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,22,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,23,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,24,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,25,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,26,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,27,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,28,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,29,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,30,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,31,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,32,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,33,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,34,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,35,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,36,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,37,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,38,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[137,39,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,40,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,41,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,42,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,43,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,44,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,45,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,46,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,47,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,48,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,49,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,50,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,51,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,52,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,53,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,54,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,55,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,56,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,57,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,58,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[137,59,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,60,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,61,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,62,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,63,2,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,64,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,65,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,66,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,67,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,68,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,69,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,70,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,71,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,72,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,73,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,74,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,75,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,76,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,77,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,78,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[137,79,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,80,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,81,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,82,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,83,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,84,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,85,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,86,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,87,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,88,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,89,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,90,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,91,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,92,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,93,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,94,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,95,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,96,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,97,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,98,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[137,99,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,100,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,101,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,102,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,103,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,104,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,105,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,106,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,107,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,108,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,109,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,110,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,111,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,112,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,113,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,114,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,115,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,116,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,117,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,118,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[137,119,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,120,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,121,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,122,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,123,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,124,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,125,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,126,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,127,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,128,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,129,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,130,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,131,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,132,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,133,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,134,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,135,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,136,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,137,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,138,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[137,139,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,140,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,141,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,142,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,143,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,144,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,145,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,146,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,147,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,148,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,149,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,150,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,151,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,152,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,153,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,154,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,155,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,156,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,157,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,158,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[137,159,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,160,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,161,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,162,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,163,3,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,164,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,165,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,166,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,167,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,168,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,169,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,170,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,171,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,172,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,173,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,174,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,175,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,176,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,177,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,178,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[137,179,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,180,3.17,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,181,3.18,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,182,3.19,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,183,3.2,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,184,3.21,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,185,3.22,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,186,3.23,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,187,3.24,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,188,3.25,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,189,3.26,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,190,3.27,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,191,3.28,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,192,3.29,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,193,3.3,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,194,3.31,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,195,3.32,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,196,3.33,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,197,3.34,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,198,3.35,7]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[137,199,3.36,7]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,0,1.38,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,1,1.39,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,2,1.4,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,3,1.41,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,4,1.42,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,5,1.43,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,6,1.44,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,7,1.45,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,8,1.46,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,9,1.47,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,10,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,11,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,12,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,13,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,14,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,15,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,16,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,17,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,18,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[138,19,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,20,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,21,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,22,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,23,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,24,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,25,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,26,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,27,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,28,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,29,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,30,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,31,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,32,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,33,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,34,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,35,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,36,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,37,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,38,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[138,39,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,40,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,41,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,42,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,43,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,44,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,45,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,46,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,47,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,48,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,49,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,50,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,51,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,52,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,53,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,54,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,55,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,56,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,57,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,58,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[138,59,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,60,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,61,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,62,2,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,63,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,64,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,65,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,66,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,67,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,68,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,69,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,70,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,71,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,72,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,73,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,74,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,75,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,76,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,77,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,78,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[138,79,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,80,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,81,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,82,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,83,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,84,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,85,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,86,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,87,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,88,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,89,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,90,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,91,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,92,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,93,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,94,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,95,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,96,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,97,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,98,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[138,99,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,100,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,101,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,102,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,103,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,104,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,105,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,106,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,107,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,108,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,109,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,110,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,111,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,112,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,113,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,114,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,115,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,116,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,117,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,118,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[138,119,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,120,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,121,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,122,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,123,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,124,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,125,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,126,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,127,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,128,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,129,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,130,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,131,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,132,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,133,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,134,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,135,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,136,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,137,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,138,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[138,139,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,140,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,141,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,142,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,143,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,144,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,145,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,146,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,147,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,148,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,149,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,150,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,151,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,152,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,153,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,154,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,155,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,156,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,157,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,158,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[138,159,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,160,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,161,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,162,3,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,163,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,164,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,165,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,166,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,167,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,168,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,169,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,170,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,171,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,172,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,173,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,174,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,175,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,176,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,177,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,178,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[138,179,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,180,3.18,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,181,3.19,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,182,3.2,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,183,3.21,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,184,3.22,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,185,3.23,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,186,3.24,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,187,3.25,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,188,3.26,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,189,3.27,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,190,3.28,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,191,3.29,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,192,3.3,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,193,3.31,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,194,3.32,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,195,3.33,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,196,3.34,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,197,3.35,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,198,3.36,8]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[138,199,3.37,8]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,0,1.39,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,1,1.4,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,2,1.41,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,3,1.42,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,4,1.43,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,5,1.44,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,6,1.45,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,7,1.46,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,8,1.47,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,9,1.48,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,10,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,11,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,12,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,13,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,14,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,15,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,16,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,17,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,18,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 120, pc_makepoint(1, ARRAY[139,19,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,20,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,21,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,22,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,23,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,24,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,25,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,26,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,27,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,28,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,29,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,30,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,31,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,32,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,33,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,34,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,35,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,36,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,37,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,38,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 121, pc_makepoint(1, ARRAY[139,39,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,40,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,41,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,42,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,43,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,44,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,45,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,46,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,47,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,48,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,49,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,50,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,51,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,52,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,53,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,54,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,55,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,56,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,57,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,58,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 122, pc_makepoint(1, ARRAY[139,59,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,60,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,61,2,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,62,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,63,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,64,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,65,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,66,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,67,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,68,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,69,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,70,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,71,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,72,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,73,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,74,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,75,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,76,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,77,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,78,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 123, pc_makepoint(1, ARRAY[139,79,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,80,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,81,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,82,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,83,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,84,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,85,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,86,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,87,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,88,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,89,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,90,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,91,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,92,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,93,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,94,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,95,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,96,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,97,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,98,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 124, pc_makepoint(1, ARRAY[139,99,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,100,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,101,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,102,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,103,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,104,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,105,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,106,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,107,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,108,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,109,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,110,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,111,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,112,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,113,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,114,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,115,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,116,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,117,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,118,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 125, pc_makepoint(1, ARRAY[139,119,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,120,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,121,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,122,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,123,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,124,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,125,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,126,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,127,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,128,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,129,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,130,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,131,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,132,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,133,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,134,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,135,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,136,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,137,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,138,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 126, pc_makepoint(1, ARRAY[139,139,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,140,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,141,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,142,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,143,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,144,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,145,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,146,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,147,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,148,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,149,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,150,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,151,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,152,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,153,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,154,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,155,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,156,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,157,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,158,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 127, pc_makepoint(1, ARRAY[139,159,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,160,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,161,3,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,162,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,163,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,164,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,165,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,166,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,167,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,168,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,169,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,170,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,171,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,172,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,173,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,174,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,175,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,176,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,177,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,178,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 128, pc_makepoint(1, ARRAY[139,179,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,180,3.19,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,181,3.2,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,182,3.21,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,183,3.22,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,184,3.23,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,185,3.24,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,186,3.25,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,187,3.26,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,188,3.27,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,189,3.28,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,190,3.29,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,191,3.3,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,192,3.31,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,193,3.32,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,194,3.33,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,195,3.34,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,196,3.35,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,197,3.36,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,198,3.37,9]) );
insert into pts_collection ( gid, pt ) values ( 129, pc_makepoint(1, ARRAY[139,199,3.38,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,0,1.4,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,1,1.41,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,2,1.42,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,3,1.43,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,4,1.44,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,5,1.45,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,6,1.46,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,7,1.47,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,8,1.48,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,9,1.49,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,10,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,11,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,12,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,13,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,14,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,15,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,16,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,17,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,18,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[140,19,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,20,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,21,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,22,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,23,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,24,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,25,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,26,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,27,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,28,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,29,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,30,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,31,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,32,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,33,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,34,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,35,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,36,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,37,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,38,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[140,39,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,40,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,41,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,42,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,43,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,44,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,45,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,46,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,47,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,48,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,49,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,50,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,51,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,52,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,53,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,54,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,55,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,56,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,57,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,58,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[140,59,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,60,2,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,61,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,62,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,63,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,64,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,65,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,66,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,67,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,68,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,69,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,70,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,71,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,72,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,73,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,74,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,75,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,76,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,77,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,78,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[140,79,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,80,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,81,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,82,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,83,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,84,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,85,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,86,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,87,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,88,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,89,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,90,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,91,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,92,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,93,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,94,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,95,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,96,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,97,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,98,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[140,99,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,100,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,101,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,102,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,103,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,104,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,105,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,106,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,107,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,108,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,109,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,110,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,111,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,112,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,113,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,114,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,115,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,116,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,117,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,118,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[140,119,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,120,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,121,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,122,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,123,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,124,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,125,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,126,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,127,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,128,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,129,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,130,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,131,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,132,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,133,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,134,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,135,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,136,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,137,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,138,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[140,139,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,140,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,141,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,142,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,143,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,144,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,145,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,146,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,147,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,148,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,149,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,150,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,151,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,152,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,153,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,154,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,155,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,156,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,157,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,158,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[140,159,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,160,3,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,161,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,162,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,163,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,164,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,165,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,166,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,167,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,168,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,169,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,170,3.1,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,171,3.11,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,172,3.12,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,173,3.13,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,174,3.14,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,175,3.15,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,176,3.16,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,177,3.17,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,178,3.18,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[140,179,3.19,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,180,3.2,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,181,3.21,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,182,3.22,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,183,3.23,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,184,3.24,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,185,3.25,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,186,3.26,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,187,3.27,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,188,3.28,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,189,3.29,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,190,3.3,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,191,3.31,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,192,3.32,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,193,3.33,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,194,3.34,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,195,3.35,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,196,3.36,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,197,3.37,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,198,3.38,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[140,199,3.39,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,0,1.41,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,1,1.42,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,2,1.43,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,3,1.44,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,4,1.45,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,5,1.46,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,6,1.47,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,7,1.48,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,8,1.49,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,9,1.5,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,10,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,11,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,12,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,13,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,14,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,15,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,16,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,17,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,18,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[141,19,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,20,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,21,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,22,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,23,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,24,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,25,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,26,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,27,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,28,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,29,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,30,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,31,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,32,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,33,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,34,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,35,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,36,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,37,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,38,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[141,39,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,40,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,41,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,42,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,43,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,44,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,45,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,46,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,47,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,48,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,49,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,50,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,51,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,52,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,53,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,54,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,55,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,56,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,57,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,58,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[141,59,2,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,60,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,61,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,62,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,63,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,64,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,65,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,66,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,67,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,68,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,69,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,70,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,71,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,72,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,73,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,74,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,75,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,76,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,77,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,78,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[141,79,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,80,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,81,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,82,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,83,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,84,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,85,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,86,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,87,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,88,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,89,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,90,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,91,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,92,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,93,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,94,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,95,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,96,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,97,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,98,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[141,99,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,100,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,101,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,102,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,103,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,104,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,105,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,106,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,107,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,108,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,109,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,110,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,111,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,112,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,113,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,114,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,115,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,116,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,117,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,118,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[141,119,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,120,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,121,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,122,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,123,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,124,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,125,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,126,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,127,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,128,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,129,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,130,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,131,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,132,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,133,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,134,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,135,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,136,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,137,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,138,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[141,139,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,140,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,141,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,142,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,143,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,144,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,145,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,146,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,147,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,148,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,149,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,150,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,151,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,152,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,153,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,154,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,155,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,156,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,157,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,158,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[141,159,3,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,160,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,161,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,162,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,163,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,164,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,165,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,166,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,167,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,168,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,169,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,170,3.11,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,171,3.12,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,172,3.13,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,173,3.14,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,174,3.15,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,175,3.16,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,176,3.17,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,177,3.18,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,178,3.19,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[141,179,3.2,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,180,3.21,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,181,3.22,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,182,3.23,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,183,3.24,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,184,3.25,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,185,3.26,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,186,3.27,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,187,3.28,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,188,3.29,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,189,3.3,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,190,3.31,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,191,3.32,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,192,3.33,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,193,3.34,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,194,3.35,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,195,3.36,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,196,3.37,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,197,3.38,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,198,3.39,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[141,199,3.4,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,0,1.42,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,1,1.43,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,2,1.44,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,3,1.45,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,4,1.46,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,5,1.47,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,6,1.48,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,7,1.49,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,8,1.5,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,9,1.51,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,10,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,11,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,12,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,13,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,14,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,15,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,16,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,17,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,18,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[142,19,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,20,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,21,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,22,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,23,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,24,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,25,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,26,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,27,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,28,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,29,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,30,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,31,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,32,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,33,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,34,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,35,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,36,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,37,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,38,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[142,39,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,40,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,41,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,42,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,43,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,44,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,45,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,46,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,47,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,48,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,49,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,50,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,51,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,52,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,53,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,54,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,55,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,56,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,57,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,58,2,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[142,59,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,60,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,61,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,62,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,63,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,64,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,65,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,66,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,67,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,68,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,69,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,70,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,71,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,72,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,73,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,74,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,75,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,76,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,77,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,78,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[142,79,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,80,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,81,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,82,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,83,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,84,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,85,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,86,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,87,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,88,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,89,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,90,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,91,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,92,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,93,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,94,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,95,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,96,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,97,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,98,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[142,99,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,100,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,101,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,102,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,103,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,104,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,105,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,106,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,107,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,108,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,109,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,110,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,111,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,112,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,113,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,114,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,115,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,116,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,117,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,118,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[142,119,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,120,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,121,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,122,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,123,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,124,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,125,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,126,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,127,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,128,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,129,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,130,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,131,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,132,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,133,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,134,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,135,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,136,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,137,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,138,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[142,139,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,140,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,141,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,142,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,143,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,144,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,145,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,146,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,147,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,148,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,149,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,150,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,151,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,152,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,153,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,154,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,155,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,156,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,157,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,158,3,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[142,159,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,160,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,161,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,162,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,163,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,164,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,165,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,166,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,167,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,168,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,169,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,170,3.12,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,171,3.13,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,172,3.14,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,173,3.15,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,174,3.16,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,175,3.17,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,176,3.18,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,177,3.19,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,178,3.2,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[142,179,3.21,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,180,3.22,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,181,3.23,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,182,3.24,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,183,3.25,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,184,3.26,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,185,3.27,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,186,3.28,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,187,3.29,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,188,3.3,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,189,3.31,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,190,3.32,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,191,3.33,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,192,3.34,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,193,3.35,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,194,3.36,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,195,3.37,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,196,3.38,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,197,3.39,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,198,3.4,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[142,199,3.41,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,0,1.43,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,1,1.44,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,2,1.45,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,3,1.46,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,4,1.47,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,5,1.48,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,6,1.49,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,7,1.5,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,8,1.51,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,9,1.52,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,10,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,11,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,12,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,13,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,14,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,15,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,16,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,17,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,18,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[143,19,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,20,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,21,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,22,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,23,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,24,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,25,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,26,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,27,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,28,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,29,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,30,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,31,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,32,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,33,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,34,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,35,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,36,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,37,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,38,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[143,39,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,40,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,41,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,42,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,43,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,44,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,45,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,46,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,47,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,48,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,49,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,50,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,51,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,52,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,53,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,54,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,55,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,56,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,57,2,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,58,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[143,59,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,60,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,61,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,62,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,63,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,64,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,65,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,66,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,67,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,68,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,69,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,70,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,71,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,72,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,73,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,74,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,75,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,76,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,77,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,78,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[143,79,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,80,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,81,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,82,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,83,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,84,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,85,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,86,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,87,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,88,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,89,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,90,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,91,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,92,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,93,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,94,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,95,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,96,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,97,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,98,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[143,99,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,100,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,101,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,102,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,103,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,104,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,105,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,106,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,107,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,108,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,109,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,110,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,111,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,112,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,113,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,114,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,115,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,116,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,117,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,118,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[143,119,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,120,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,121,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,122,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,123,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,124,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,125,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,126,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,127,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,128,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,129,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,130,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,131,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,132,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,133,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,134,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,135,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,136,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,137,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,138,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[143,139,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,140,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,141,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,142,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,143,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,144,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,145,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,146,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,147,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,148,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,149,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,150,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,151,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,152,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,153,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,154,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,155,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,156,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,157,3,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,158,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[143,159,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,160,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,161,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,162,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,163,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,164,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,165,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,166,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,167,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,168,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,169,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,170,3.13,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,171,3.14,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,172,3.15,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,173,3.16,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,174,3.17,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,175,3.18,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,176,3.19,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,177,3.2,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,178,3.21,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[143,179,3.22,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,180,3.23,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,181,3.24,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,182,3.25,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,183,3.26,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,184,3.27,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,185,3.28,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,186,3.29,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,187,3.3,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,188,3.31,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,189,3.32,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,190,3.33,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,191,3.34,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,192,3.35,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,193,3.36,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,194,3.37,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,195,3.38,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,196,3.39,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,197,3.4,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,198,3.41,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[143,199,3.42,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,0,1.44,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,1,1.45,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,2,1.46,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,3,1.47,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,4,1.48,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,5,1.49,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,6,1.5,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,7,1.51,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,8,1.52,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,9,1.53,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,10,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,11,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,12,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,13,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,14,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,15,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,16,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,17,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,18,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[144,19,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,20,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,21,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,22,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,23,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,24,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,25,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,26,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,27,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,28,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,29,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,30,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,31,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,32,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,33,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,34,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,35,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,36,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,37,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,38,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[144,39,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,40,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,41,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,42,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,43,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,44,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,45,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,46,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,47,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,48,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,49,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,50,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,51,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,52,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,53,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,54,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,55,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,56,2,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,57,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,58,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[144,59,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,60,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,61,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,62,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,63,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,64,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,65,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,66,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,67,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,68,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,69,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,70,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,71,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,72,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,73,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,74,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,75,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,76,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,77,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,78,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[144,79,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,80,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,81,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,82,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,83,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,84,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,85,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,86,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,87,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,88,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,89,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,90,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,91,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,92,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,93,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,94,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,95,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,96,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,97,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,98,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[144,99,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,100,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,101,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,102,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,103,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,104,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,105,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,106,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,107,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,108,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,109,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,110,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,111,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,112,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,113,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,114,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,115,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,116,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,117,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,118,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[144,119,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,120,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,121,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,122,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,123,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,124,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,125,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,126,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,127,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,128,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,129,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,130,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,131,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,132,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,133,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,134,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,135,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,136,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,137,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,138,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[144,139,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,140,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,141,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,142,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,143,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,144,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,145,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,146,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,147,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,148,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,149,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,150,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,151,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,152,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,153,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,154,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,155,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,156,3,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,157,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,158,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[144,159,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,160,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,161,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,162,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,163,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,164,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,165,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,166,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,167,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,168,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,169,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,170,3.14,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,171,3.15,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,172,3.16,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,173,3.17,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,174,3.18,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,175,3.19,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,176,3.2,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,177,3.21,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,178,3.22,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[144,179,3.23,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,180,3.24,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,181,3.25,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,182,3.26,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,183,3.27,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,184,3.28,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,185,3.29,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,186,3.3,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,187,3.31,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,188,3.32,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,189,3.33,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,190,3.34,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,191,3.35,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,192,3.36,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,193,3.37,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,194,3.38,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,195,3.39,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,196,3.4,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,197,3.41,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,198,3.42,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[144,199,3.43,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,0,1.45,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,1,1.46,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,2,1.47,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,3,1.48,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,4,1.49,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,5,1.5,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,6,1.51,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,7,1.52,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,8,1.53,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,9,1.54,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,10,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,11,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,12,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,13,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,14,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,15,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,16,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,17,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,18,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[145,19,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,20,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,21,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,22,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,23,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,24,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,25,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,26,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,27,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,28,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,29,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,30,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,31,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,32,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,33,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,34,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,35,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,36,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,37,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,38,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[145,39,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,40,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,41,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,42,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,43,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,44,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,45,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,46,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,47,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,48,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,49,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,50,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,51,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,52,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,53,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,54,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,55,2,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,56,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,57,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,58,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[145,59,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,60,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,61,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,62,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,63,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,64,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,65,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,66,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,67,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,68,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,69,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,70,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,71,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,72,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,73,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,74,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,75,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,76,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,77,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,78,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[145,79,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,80,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,81,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,82,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,83,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,84,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,85,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,86,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,87,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,88,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,89,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,90,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,91,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,92,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,93,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,94,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,95,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,96,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,97,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,98,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[145,99,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,100,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,101,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,102,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,103,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,104,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,105,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,106,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,107,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,108,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,109,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,110,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,111,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,112,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,113,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,114,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,115,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,116,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,117,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,118,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[145,119,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,120,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,121,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,122,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,123,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,124,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,125,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,126,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,127,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,128,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,129,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,130,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,131,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,132,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,133,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,134,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,135,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,136,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,137,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,138,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[145,139,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,140,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,141,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,142,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,143,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,144,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,145,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,146,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,147,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,148,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,149,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,150,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,151,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,152,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,153,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,154,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,155,3,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,156,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,157,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,158,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[145,159,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,160,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,161,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,162,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,163,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,164,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,165,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,166,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,167,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,168,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,169,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,170,3.15,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,171,3.16,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,172,3.17,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,173,3.18,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,174,3.19,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,175,3.2,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,176,3.21,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,177,3.22,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,178,3.23,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[145,179,3.24,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,180,3.25,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,181,3.26,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,182,3.27,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,183,3.28,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,184,3.29,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,185,3.3,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,186,3.31,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,187,3.32,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,188,3.33,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,189,3.34,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,190,3.35,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,191,3.36,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,192,3.37,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,193,3.38,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,194,3.39,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,195,3.4,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,196,3.41,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,197,3.42,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,198,3.43,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[145,199,3.44,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,0,1.46,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,1,1.47,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,2,1.48,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,3,1.49,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,4,1.5,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,5,1.51,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,6,1.52,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,7,1.53,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,8,1.54,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,9,1.55,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,10,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,11,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,12,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,13,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,14,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,15,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,16,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,17,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,18,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[146,19,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,20,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,21,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,22,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,23,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,24,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,25,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,26,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,27,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,28,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,29,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,30,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,31,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,32,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,33,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,34,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,35,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,36,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,37,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,38,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[146,39,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,40,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,41,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,42,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,43,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,44,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,45,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,46,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,47,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,48,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,49,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,50,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,51,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,52,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,53,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,54,2,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,55,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,56,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,57,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,58,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[146,59,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,60,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,61,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,62,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,63,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,64,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,65,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,66,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,67,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,68,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,69,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,70,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,71,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,72,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,73,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,74,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,75,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,76,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,77,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,78,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[146,79,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,80,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,81,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,82,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,83,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,84,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,85,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,86,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,87,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,88,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,89,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,90,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,91,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,92,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,93,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,94,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,95,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,96,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,97,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,98,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[146,99,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,100,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,101,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,102,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,103,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,104,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,105,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,106,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,107,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,108,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,109,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,110,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,111,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,112,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,113,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,114,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,115,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,116,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,117,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,118,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[146,119,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,120,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,121,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,122,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,123,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,124,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,125,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,126,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,127,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,128,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,129,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,130,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,131,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,132,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,133,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,134,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,135,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,136,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,137,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,138,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[146,139,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,140,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,141,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,142,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,143,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,144,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,145,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,146,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,147,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,148,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,149,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,150,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,151,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,152,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,153,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,154,3,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,155,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,156,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,157,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,158,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[146,159,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,160,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,161,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,162,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,163,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,164,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,165,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,166,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,167,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,168,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,169,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,170,3.16,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,171,3.17,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,172,3.18,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,173,3.19,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,174,3.2,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,175,3.21,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,176,3.22,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,177,3.23,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,178,3.24,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[146,179,3.25,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,180,3.26,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,181,3.27,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,182,3.28,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,183,3.29,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,184,3.3,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,185,3.31,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,186,3.32,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,187,3.33,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,188,3.34,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,189,3.35,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,190,3.36,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,191,3.37,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,192,3.38,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,193,3.39,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,194,3.4,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,195,3.41,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,196,3.42,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,197,3.43,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,198,3.44,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[146,199,3.45,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,0,1.47,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,1,1.48,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,2,1.49,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,3,1.5,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,4,1.51,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,5,1.52,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,6,1.53,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,7,1.54,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,8,1.55,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,9,1.56,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,10,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,11,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,12,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,13,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,14,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,15,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,16,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,17,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,18,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[147,19,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,20,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,21,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,22,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,23,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,24,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,25,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,26,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,27,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,28,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,29,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,30,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,31,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,32,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,33,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,34,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,35,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,36,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,37,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,38,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[147,39,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,40,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,41,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,42,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,43,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,44,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,45,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,46,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,47,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,48,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,49,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,50,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,51,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,52,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,53,2,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,54,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,55,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,56,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,57,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,58,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[147,59,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,60,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,61,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,62,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,63,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,64,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,65,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,66,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,67,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,68,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,69,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,70,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,71,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,72,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,73,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,74,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,75,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,76,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,77,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,78,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[147,79,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,80,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,81,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,82,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,83,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,84,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,85,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,86,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,87,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,88,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,89,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,90,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,91,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,92,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,93,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,94,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,95,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,96,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,97,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,98,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[147,99,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,100,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,101,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,102,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,103,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,104,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,105,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,106,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,107,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,108,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,109,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,110,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,111,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,112,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,113,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,114,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,115,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,116,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,117,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,118,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[147,119,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,120,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,121,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,122,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,123,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,124,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,125,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,126,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,127,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,128,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,129,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,130,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,131,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,132,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,133,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,134,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,135,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,136,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,137,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,138,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[147,139,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,140,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,141,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,142,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,143,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,144,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,145,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,146,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,147,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,148,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,149,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,150,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,151,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,152,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,153,3,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,154,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,155,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,156,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,157,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,158,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[147,159,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,160,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,161,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,162,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,163,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,164,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,165,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,166,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,167,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,168,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,169,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,170,3.17,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,171,3.18,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,172,3.19,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,173,3.2,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,174,3.21,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,175,3.22,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,176,3.23,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,177,3.24,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,178,3.25,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[147,179,3.26,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,180,3.27,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,181,3.28,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,182,3.29,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,183,3.3,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,184,3.31,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,185,3.32,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,186,3.33,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,187,3.34,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,188,3.35,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,189,3.36,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,190,3.37,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,191,3.38,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,192,3.39,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,193,3.4,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,194,3.41,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,195,3.42,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,196,3.43,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,197,3.44,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,198,3.45,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[147,199,3.46,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,0,1.48,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,1,1.49,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,2,1.5,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,3,1.51,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,4,1.52,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,5,1.53,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,6,1.54,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,7,1.55,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,8,1.56,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,9,1.57,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,10,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,11,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,12,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,13,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,14,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,15,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,16,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,17,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,18,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[148,19,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,20,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,21,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,22,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,23,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,24,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,25,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,26,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,27,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,28,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,29,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,30,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,31,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,32,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,33,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,34,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,35,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,36,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,37,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,38,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[148,39,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,40,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,41,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,42,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,43,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,44,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,45,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,46,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,47,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,48,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,49,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,50,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,51,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,52,2,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,53,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,54,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,55,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,56,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,57,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,58,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[148,59,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,60,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,61,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,62,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,63,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,64,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,65,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,66,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,67,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,68,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,69,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,70,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,71,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,72,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,73,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,74,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,75,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,76,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,77,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,78,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[148,79,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,80,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,81,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,82,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,83,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,84,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,85,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,86,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,87,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,88,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,89,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,90,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,91,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,92,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,93,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,94,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,95,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,96,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,97,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,98,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[148,99,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,100,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,101,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,102,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,103,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,104,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,105,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,106,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,107,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,108,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,109,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,110,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,111,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,112,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,113,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,114,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,115,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,116,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,117,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,118,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[148,119,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,120,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,121,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,122,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,123,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,124,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,125,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,126,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,127,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,128,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,129,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,130,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,131,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,132,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,133,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,134,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,135,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,136,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,137,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,138,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[148,139,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,140,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,141,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,142,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,143,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,144,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,145,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,146,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,147,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,148,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,149,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,150,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,151,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,152,3,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,153,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,154,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,155,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,156,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,157,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,158,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[148,159,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,160,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,161,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,162,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,163,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,164,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,165,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,166,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,167,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,168,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,169,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,170,3.18,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,171,3.19,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,172,3.2,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,173,3.21,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,174,3.22,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,175,3.23,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,176,3.24,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,177,3.25,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,178,3.26,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[148,179,3.27,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,180,3.28,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,181,3.29,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,182,3.3,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,183,3.31,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,184,3.32,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,185,3.33,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,186,3.34,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,187,3.35,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,188,3.36,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,189,3.37,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,190,3.38,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,191,3.39,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,192,3.4,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,193,3.41,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,194,3.42,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,195,3.43,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,196,3.44,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,197,3.45,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,198,3.46,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[148,199,3.47,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,0,1.49,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,1,1.5,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,2,1.51,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,3,1.52,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,4,1.53,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,5,1.54,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,6,1.55,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,7,1.56,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,8,1.57,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,9,1.58,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,10,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,11,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,12,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,13,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,14,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,15,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,16,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,17,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,18,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[149,19,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,20,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,21,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,22,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,23,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,24,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,25,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,26,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,27,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,28,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,29,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,30,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,31,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,32,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,33,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,34,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,35,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,36,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,37,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,38,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[149,39,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,40,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,41,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,42,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,43,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,44,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,45,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,46,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,47,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,48,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,49,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,50,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,51,2,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,52,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,53,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,54,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,55,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,56,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,57,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,58,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[149,59,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,60,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,61,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,62,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,63,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,64,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,65,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,66,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,67,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,68,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,69,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,70,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,71,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,72,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,73,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,74,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,75,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,76,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,77,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,78,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[149,79,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,80,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,81,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,82,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,83,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,84,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,85,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,86,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,87,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,88,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,89,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,90,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,91,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,92,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,93,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,94,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,95,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,96,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,97,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,98,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[149,99,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,100,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,101,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,102,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,103,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,104,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,105,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,106,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,107,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,108,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,109,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,110,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,111,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,112,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,113,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,114,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,115,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,116,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,117,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,118,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[149,119,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,120,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,121,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,122,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,123,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,124,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,125,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,126,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,127,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,128,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,129,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,130,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,131,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,132,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,133,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,134,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,135,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,136,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,137,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,138,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[149,139,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,140,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,141,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,142,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,143,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,144,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,145,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,146,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,147,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,148,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,149,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,150,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,151,3,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,152,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,153,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,154,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,155,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,156,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,157,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,158,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[149,159,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,160,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,161,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,162,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,163,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,164,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,165,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,166,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,167,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,168,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,169,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,170,3.19,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,171,3.2,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,172,3.21,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,173,3.22,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,174,3.23,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,175,3.24,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,176,3.25,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,177,3.26,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,178,3.27,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[149,179,3.28,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,180,3.29,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,181,3.3,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,182,3.31,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,183,3.32,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,184,3.33,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,185,3.34,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,186,3.35,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,187,3.36,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,188,3.37,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,189,3.38,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,190,3.39,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,191,3.4,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,192,3.41,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,193,3.42,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,194,3.43,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,195,3.44,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,196,3.45,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,197,3.46,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,198,3.47,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[149,199,3.48,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,0,1.5,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,1,1.51,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,2,1.52,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,3,1.53,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,4,1.54,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,5,1.55,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,6,1.56,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,7,1.57,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,8,1.58,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,9,1.59,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,10,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,11,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,12,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,13,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,14,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,15,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,16,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,17,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,18,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[150,19,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,20,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,21,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,22,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,23,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,24,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,25,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,26,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,27,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,28,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,29,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,30,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,31,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,32,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,33,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,34,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,35,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,36,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,37,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,38,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[150,39,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,40,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,41,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,42,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,43,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,44,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,45,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,46,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,47,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,48,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,49,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,50,2,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,51,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,52,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,53,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,54,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,55,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,56,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,57,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,58,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[150,59,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,60,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,61,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,62,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,63,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,64,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,65,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,66,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,67,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,68,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,69,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,70,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,71,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,72,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,73,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,74,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,75,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,76,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,77,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,78,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[150,79,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,80,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,81,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,82,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,83,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,84,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,85,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,86,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,87,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,88,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,89,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,90,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,91,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,92,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,93,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,94,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,95,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,96,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,97,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,98,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[150,99,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,100,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,101,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,102,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,103,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,104,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,105,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,106,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,107,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,108,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,109,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,110,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,111,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,112,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,113,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,114,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,115,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,116,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,117,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,118,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[150,119,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,120,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,121,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,122,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,123,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,124,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,125,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,126,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,127,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,128,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,129,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,130,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,131,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,132,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,133,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,134,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,135,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,136,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,137,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,138,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[150,139,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,140,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,141,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,142,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,143,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,144,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,145,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,146,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,147,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,148,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,149,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,150,3,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,151,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,152,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,153,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,154,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,155,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,156,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,157,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,158,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[150,159,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,160,3.1,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,161,3.11,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,162,3.12,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,163,3.13,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,164,3.14,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,165,3.15,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,166,3.16,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,167,3.17,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,168,3.18,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,169,3.19,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,170,3.2,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,171,3.21,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,172,3.22,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,173,3.23,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,174,3.24,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,175,3.25,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,176,3.26,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,177,3.27,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,178,3.28,0]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[150,179,3.29,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,180,3.3,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,181,3.31,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,182,3.32,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,183,3.33,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,184,3.34,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,185,3.35,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,186,3.36,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,187,3.37,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,188,3.38,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,189,3.39,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,190,3.4,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,191,3.41,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,192,3.42,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,193,3.43,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,194,3.44,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,195,3.45,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,196,3.46,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,197,3.47,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,198,3.48,0]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[150,199,3.49,0]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,0,1.51,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,1,1.52,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,2,1.53,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,3,1.54,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,4,1.55,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,5,1.56,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,6,1.57,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,7,1.58,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,8,1.59,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,9,1.6,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,10,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,11,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,12,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,13,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,14,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,15,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,16,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,17,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,18,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[151,19,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,20,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,21,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,22,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,23,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,24,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,25,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,26,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,27,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,28,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,29,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,30,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,31,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,32,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,33,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,34,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,35,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,36,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,37,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,38,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[151,39,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,40,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,41,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,42,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,43,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,44,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,45,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,46,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,47,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,48,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,49,2,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,50,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,51,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,52,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,53,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,54,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,55,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,56,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,57,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,58,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[151,59,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,60,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,61,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,62,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,63,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,64,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,65,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,66,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,67,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,68,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,69,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,70,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,71,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,72,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,73,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,74,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,75,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,76,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,77,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,78,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[151,79,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,80,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,81,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,82,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,83,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,84,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,85,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,86,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,87,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,88,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,89,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,90,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,91,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,92,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,93,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,94,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,95,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,96,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,97,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,98,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[151,99,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,100,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,101,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,102,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,103,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,104,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,105,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,106,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,107,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,108,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,109,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,110,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,111,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,112,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,113,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,114,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,115,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,116,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,117,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,118,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[151,119,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,120,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,121,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,122,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,123,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,124,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,125,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,126,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,127,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,128,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,129,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,130,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,131,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,132,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,133,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,134,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,135,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,136,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,137,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,138,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[151,139,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,140,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,141,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,142,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,143,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,144,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,145,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,146,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,147,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,148,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,149,3,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,150,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,151,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,152,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,153,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,154,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,155,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,156,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,157,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,158,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[151,159,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,160,3.11,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,161,3.12,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,162,3.13,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,163,3.14,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,164,3.15,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,165,3.16,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,166,3.17,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,167,3.18,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,168,3.19,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,169,3.2,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,170,3.21,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,171,3.22,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,172,3.23,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,173,3.24,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,174,3.25,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,175,3.26,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,176,3.27,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,177,3.28,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,178,3.29,1]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[151,179,3.3,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,180,3.31,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,181,3.32,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,182,3.33,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,183,3.34,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,184,3.35,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,185,3.36,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,186,3.37,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,187,3.38,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,188,3.39,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,189,3.4,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,190,3.41,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,191,3.42,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,192,3.43,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,193,3.44,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,194,3.45,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,195,3.46,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,196,3.47,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,197,3.48,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,198,3.49,1]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[151,199,3.5,1]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,0,1.52,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,1,1.53,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,2,1.54,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,3,1.55,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,4,1.56,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,5,1.57,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,6,1.58,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,7,1.59,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,8,1.6,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,9,1.61,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,10,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,11,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,12,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,13,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,14,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,15,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,16,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,17,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,18,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[152,19,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,20,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,21,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,22,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,23,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,24,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,25,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,26,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,27,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,28,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,29,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,30,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,31,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,32,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,33,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,34,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,35,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,36,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,37,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,38,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[152,39,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,40,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,41,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,42,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,43,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,44,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,45,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,46,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,47,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,48,2,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,49,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,50,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,51,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,52,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,53,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,54,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,55,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,56,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,57,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,58,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[152,59,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,60,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,61,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,62,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,63,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,64,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,65,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,66,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,67,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,68,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,69,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,70,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,71,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,72,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,73,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,74,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,75,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,76,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,77,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,78,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[152,79,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,80,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,81,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,82,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,83,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,84,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,85,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,86,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,87,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,88,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,89,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,90,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,91,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,92,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,93,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,94,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,95,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,96,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,97,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,98,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[152,99,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,100,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,101,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,102,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,103,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,104,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,105,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,106,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,107,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,108,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,109,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,110,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,111,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,112,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,113,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,114,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,115,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,116,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,117,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,118,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[152,119,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,120,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,121,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,122,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,123,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,124,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,125,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,126,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,127,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,128,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,129,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,130,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,131,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,132,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,133,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,134,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,135,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,136,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,137,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,138,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[152,139,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,140,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,141,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,142,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,143,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,144,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,145,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,146,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,147,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,148,3,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,149,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,150,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,151,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,152,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,153,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,154,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,155,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,156,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,157,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,158,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[152,159,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,160,3.12,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,161,3.13,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,162,3.14,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,163,3.15,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,164,3.16,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,165,3.17,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,166,3.18,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,167,3.19,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,168,3.2,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,169,3.21,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,170,3.22,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,171,3.23,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,172,3.24,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,173,3.25,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,174,3.26,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,175,3.27,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,176,3.28,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,177,3.29,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,178,3.3,2]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[152,179,3.31,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,180,3.32,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,181,3.33,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,182,3.34,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,183,3.35,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,184,3.36,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,185,3.37,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,186,3.38,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,187,3.39,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,188,3.4,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,189,3.41,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,190,3.42,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,191,3.43,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,192,3.44,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,193,3.45,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,194,3.46,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,195,3.47,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,196,3.48,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,197,3.49,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,198,3.5,2]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[152,199,3.51,2]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,0,1.53,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,1,1.54,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,2,1.55,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,3,1.56,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,4,1.57,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,5,1.58,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,6,1.59,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,7,1.6,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,8,1.61,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,9,1.62,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,10,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,11,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,12,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,13,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,14,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,15,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,16,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,17,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,18,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[153,19,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,20,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,21,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,22,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,23,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,24,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,25,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,26,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,27,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,28,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,29,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,30,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,31,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,32,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,33,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,34,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,35,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,36,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,37,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,38,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[153,39,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,40,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,41,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,42,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,43,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,44,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,45,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,46,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,47,2,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,48,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,49,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,50,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,51,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,52,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,53,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,54,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,55,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,56,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,57,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,58,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[153,59,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,60,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,61,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,62,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,63,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,64,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,65,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,66,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,67,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,68,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,69,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,70,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,71,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,72,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,73,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,74,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,75,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,76,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,77,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,78,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[153,79,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,80,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,81,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,82,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,83,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,84,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,85,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,86,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,87,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,88,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,89,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,90,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,91,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,92,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,93,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,94,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,95,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,96,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,97,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,98,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[153,99,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,100,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,101,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,102,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,103,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,104,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,105,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,106,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,107,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,108,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,109,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,110,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,111,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,112,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,113,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,114,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,115,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,116,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,117,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,118,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[153,119,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,120,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,121,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,122,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,123,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,124,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,125,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,126,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,127,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,128,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,129,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,130,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,131,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,132,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,133,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,134,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,135,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,136,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,137,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,138,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[153,139,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,140,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,141,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,142,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,143,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,144,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,145,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,146,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,147,3,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,148,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,149,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,150,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,151,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,152,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,153,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,154,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,155,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,156,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,157,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,158,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[153,159,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,160,3.13,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,161,3.14,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,162,3.15,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,163,3.16,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,164,3.17,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,165,3.18,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,166,3.19,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,167,3.2,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,168,3.21,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,169,3.22,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,170,3.23,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,171,3.24,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,172,3.25,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,173,3.26,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,174,3.27,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,175,3.28,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,176,3.29,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,177,3.3,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,178,3.31,3]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[153,179,3.32,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,180,3.33,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,181,3.34,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,182,3.35,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,183,3.36,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,184,3.37,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,185,3.38,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,186,3.39,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,187,3.4,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,188,3.41,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,189,3.42,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,190,3.43,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,191,3.44,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,192,3.45,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,193,3.46,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,194,3.47,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,195,3.48,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,196,3.49,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,197,3.5,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,198,3.51,3]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[153,199,3.52,3]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,0,1.54,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,1,1.55,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,2,1.56,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,3,1.57,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,4,1.58,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,5,1.59,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,6,1.6,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,7,1.61,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,8,1.62,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,9,1.63,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,10,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,11,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,12,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,13,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,14,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,15,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,16,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,17,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,18,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[154,19,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,20,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,21,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,22,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,23,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,24,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,25,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,26,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,27,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,28,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,29,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,30,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,31,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,32,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,33,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,34,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,35,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,36,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,37,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,38,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[154,39,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,40,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,41,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,42,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,43,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,44,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,45,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,46,2,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,47,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,48,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,49,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,50,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,51,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,52,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,53,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,54,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,55,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,56,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,57,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,58,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[154,59,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,60,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,61,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,62,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,63,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,64,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,65,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,66,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,67,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,68,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,69,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,70,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,71,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,72,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,73,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,74,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,75,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,76,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,77,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,78,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[154,79,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,80,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,81,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,82,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,83,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,84,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,85,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,86,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,87,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,88,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,89,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,90,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,91,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,92,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,93,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,94,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,95,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,96,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,97,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,98,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[154,99,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,100,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,101,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,102,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,103,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,104,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,105,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,106,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,107,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,108,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,109,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,110,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,111,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,112,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,113,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,114,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,115,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,116,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,117,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,118,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[154,119,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,120,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,121,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,122,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,123,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,124,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,125,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,126,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,127,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,128,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,129,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,130,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,131,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,132,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,133,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,134,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,135,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,136,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,137,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,138,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[154,139,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,140,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,141,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,142,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,143,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,144,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,145,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,146,3,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,147,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,148,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,149,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,150,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,151,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,152,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,153,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,154,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,155,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,156,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,157,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,158,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[154,159,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,160,3.14,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,161,3.15,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,162,3.16,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,163,3.17,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,164,3.18,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,165,3.19,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,166,3.2,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,167,3.21,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,168,3.22,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,169,3.23,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,170,3.24,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,171,3.25,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,172,3.26,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,173,3.27,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,174,3.28,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,175,3.29,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,176,3.3,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,177,3.31,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,178,3.32,4]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[154,179,3.33,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,180,3.34,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,181,3.35,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,182,3.36,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,183,3.37,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,184,3.38,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,185,3.39,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,186,3.4,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,187,3.41,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,188,3.42,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,189,3.43,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,190,3.44,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,191,3.45,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,192,3.46,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,193,3.47,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,194,3.48,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,195,3.49,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,196,3.5,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,197,3.51,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,198,3.52,4]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[154,199,3.53,4]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,0,1.55,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,1,1.56,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,2,1.57,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,3,1.58,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,4,1.59,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,5,1.6,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,6,1.61,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,7,1.62,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,8,1.63,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,9,1.64,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,10,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,11,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,12,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,13,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,14,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,15,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,16,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,17,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,18,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[155,19,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,20,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,21,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,22,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,23,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,24,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,25,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,26,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,27,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,28,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,29,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,30,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,31,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,32,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,33,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,34,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,35,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,36,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,37,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,38,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[155,39,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,40,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,41,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,42,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,43,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,44,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,45,2,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,46,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,47,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,48,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,49,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,50,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,51,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,52,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,53,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,54,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,55,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,56,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,57,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,58,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[155,59,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,60,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,61,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,62,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,63,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,64,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,65,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,66,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,67,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,68,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,69,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,70,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,71,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,72,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,73,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,74,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,75,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,76,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,77,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,78,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[155,79,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,80,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,81,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,82,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,83,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,84,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,85,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,86,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,87,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,88,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,89,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,90,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,91,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,92,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,93,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,94,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,95,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,96,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,97,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,98,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[155,99,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,100,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,101,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,102,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,103,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,104,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,105,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,106,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,107,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,108,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,109,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,110,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,111,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,112,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,113,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,114,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,115,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,116,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,117,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,118,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[155,119,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,120,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,121,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,122,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,123,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,124,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,125,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,126,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,127,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,128,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,129,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,130,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,131,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,132,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,133,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,134,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,135,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,136,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,137,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,138,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[155,139,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,140,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,141,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,142,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,143,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,144,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,145,3,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,146,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,147,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,148,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,149,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,150,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,151,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,152,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,153,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,154,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,155,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,156,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,157,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,158,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[155,159,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,160,3.15,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,161,3.16,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,162,3.17,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,163,3.18,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,164,3.19,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,165,3.2,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,166,3.21,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,167,3.22,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,168,3.23,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,169,3.24,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,170,3.25,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,171,3.26,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,172,3.27,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,173,3.28,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,174,3.29,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,175,3.3,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,176,3.31,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,177,3.32,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,178,3.33,5]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[155,179,3.34,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,180,3.35,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,181,3.36,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,182,3.37,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,183,3.38,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,184,3.39,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,185,3.4,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,186,3.41,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,187,3.42,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,188,3.43,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,189,3.44,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,190,3.45,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,191,3.46,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,192,3.47,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,193,3.48,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,194,3.49,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,195,3.5,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,196,3.51,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,197,3.52,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,198,3.53,5]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[155,199,3.54,5]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,0,1.56,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,1,1.57,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,2,1.58,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,3,1.59,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,4,1.6,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,5,1.61,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,6,1.62,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,7,1.63,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,8,1.64,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,9,1.65,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,10,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,11,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,12,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,13,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,14,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,15,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,16,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,17,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,18,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[156,19,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,20,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,21,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,22,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,23,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,24,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,25,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,26,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,27,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,28,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,29,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,30,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,31,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,32,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,33,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,34,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,35,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,36,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,37,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,38,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[156,39,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,40,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,41,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,42,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,43,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,44,2,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,45,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,46,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,47,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,48,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,49,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,50,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,51,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,52,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,53,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,54,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,55,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,56,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,57,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,58,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[156,59,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,60,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,61,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,62,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,63,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,64,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,65,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,66,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,67,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,68,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,69,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,70,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,71,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,72,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,73,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,74,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,75,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,76,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,77,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,78,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[156,79,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,80,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,81,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,82,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,83,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,84,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,85,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,86,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,87,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,88,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,89,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,90,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,91,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,92,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,93,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,94,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,95,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,96,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,97,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,98,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[156,99,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,100,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,101,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,102,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,103,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,104,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,105,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,106,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,107,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,108,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,109,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,110,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,111,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,112,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,113,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,114,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,115,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,116,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,117,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,118,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[156,119,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,120,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,121,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,122,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,123,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,124,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,125,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,126,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,127,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,128,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,129,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,130,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,131,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,132,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,133,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,134,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,135,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,136,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,137,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,138,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[156,139,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,140,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,141,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,142,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,143,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,144,3,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,145,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,146,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,147,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,148,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,149,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,150,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,151,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,152,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,153,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,154,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,155,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,156,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,157,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,158,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[156,159,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,160,3.16,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,161,3.17,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,162,3.18,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,163,3.19,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,164,3.2,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,165,3.21,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,166,3.22,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,167,3.23,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,168,3.24,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,169,3.25,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,170,3.26,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,171,3.27,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,172,3.28,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,173,3.29,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,174,3.3,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,175,3.31,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,176,3.32,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,177,3.33,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,178,3.34,6]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[156,179,3.35,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,180,3.36,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,181,3.37,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,182,3.38,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,183,3.39,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,184,3.4,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,185,3.41,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,186,3.42,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,187,3.43,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,188,3.44,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,189,3.45,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,190,3.46,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,191,3.47,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,192,3.48,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,193,3.49,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,194,3.5,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,195,3.51,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,196,3.52,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,197,3.53,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,198,3.54,6]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[156,199,3.55,6]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,0,1.57,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,1,1.58,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,2,1.59,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,3,1.6,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,4,1.61,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,5,1.62,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,6,1.63,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,7,1.64,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,8,1.65,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,9,1.66,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,10,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,11,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,12,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,13,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,14,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,15,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,16,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,17,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,18,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[157,19,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,20,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,21,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,22,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,23,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,24,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,25,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,26,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,27,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,28,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,29,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,30,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,31,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,32,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,33,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,34,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,35,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,36,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,37,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,38,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[157,39,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,40,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,41,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,42,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,43,2,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,44,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,45,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,46,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,47,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,48,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,49,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,50,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,51,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,52,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,53,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,54,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,55,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,56,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,57,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,58,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[157,59,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,60,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,61,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,62,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,63,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,64,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,65,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,66,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,67,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,68,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,69,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,70,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,71,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,72,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,73,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,74,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,75,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,76,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,77,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,78,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[157,79,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,80,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,81,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,82,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,83,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,84,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,85,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,86,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,87,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,88,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,89,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,90,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,91,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,92,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,93,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,94,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,95,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,96,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,97,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,98,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[157,99,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,100,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,101,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,102,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,103,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,104,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,105,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,106,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,107,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,108,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,109,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,110,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,111,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,112,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,113,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,114,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,115,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,116,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,117,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,118,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[157,119,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,120,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,121,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,122,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,123,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,124,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,125,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,126,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,127,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,128,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,129,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,130,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,131,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,132,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,133,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,134,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,135,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,136,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,137,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,138,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[157,139,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,140,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,141,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,142,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,143,3,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,144,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,145,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,146,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,147,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,148,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,149,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,150,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,151,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,152,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,153,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,154,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,155,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,156,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,157,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,158,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[157,159,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,160,3.17,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,161,3.18,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,162,3.19,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,163,3.2,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,164,3.21,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,165,3.22,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,166,3.23,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,167,3.24,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,168,3.25,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,169,3.26,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,170,3.27,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,171,3.28,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,172,3.29,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,173,3.3,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,174,3.31,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,175,3.32,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,176,3.33,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,177,3.34,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,178,3.35,7]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[157,179,3.36,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,180,3.37,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,181,3.38,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,182,3.39,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,183,3.4,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,184,3.41,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,185,3.42,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,186,3.43,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,187,3.44,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,188,3.45,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,189,3.46,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,190,3.47,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,191,3.48,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,192,3.49,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,193,3.5,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,194,3.51,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,195,3.52,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,196,3.53,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,197,3.54,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,198,3.55,7]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[157,199,3.56,7]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,0,1.58,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,1,1.59,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,2,1.6,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,3,1.61,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,4,1.62,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,5,1.63,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,6,1.64,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,7,1.65,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,8,1.66,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,9,1.67,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,10,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,11,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,12,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,13,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,14,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,15,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,16,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,17,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,18,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[158,19,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,20,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,21,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,22,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,23,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,24,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,25,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,26,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,27,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,28,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,29,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,30,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,31,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,32,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,33,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,34,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,35,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,36,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,37,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,38,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[158,39,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,40,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,41,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,42,2,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,43,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,44,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,45,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,46,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,47,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,48,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,49,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,50,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,51,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,52,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,53,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,54,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,55,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,56,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,57,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,58,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[158,59,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,60,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,61,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,62,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,63,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,64,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,65,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,66,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,67,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,68,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,69,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,70,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,71,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,72,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,73,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,74,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,75,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,76,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,77,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,78,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[158,79,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,80,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,81,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,82,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,83,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,84,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,85,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,86,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,87,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,88,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,89,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,90,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,91,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,92,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,93,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,94,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,95,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,96,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,97,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,98,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[158,99,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,100,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,101,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,102,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,103,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,104,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,105,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,106,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,107,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,108,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,109,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,110,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,111,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,112,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,113,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,114,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,115,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,116,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,117,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,118,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[158,119,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,120,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,121,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,122,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,123,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,124,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,125,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,126,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,127,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,128,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,129,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,130,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,131,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,132,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,133,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,134,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,135,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,136,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,137,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,138,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[158,139,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,140,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,141,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,142,3,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,143,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,144,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,145,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,146,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,147,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,148,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,149,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,150,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,151,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,152,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,153,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,154,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,155,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,156,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,157,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,158,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[158,159,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,160,3.18,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,161,3.19,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,162,3.2,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,163,3.21,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,164,3.22,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,165,3.23,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,166,3.24,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,167,3.25,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,168,3.26,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,169,3.27,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,170,3.28,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,171,3.29,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,172,3.3,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,173,3.31,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,174,3.32,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,175,3.33,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,176,3.34,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,177,3.35,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,178,3.36,8]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[158,179,3.37,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,180,3.38,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,181,3.39,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,182,3.4,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,183,3.41,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,184,3.42,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,185,3.43,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,186,3.44,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,187,3.45,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,188,3.46,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,189,3.47,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,190,3.48,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,191,3.49,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,192,3.5,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,193,3.51,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,194,3.52,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,195,3.53,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,196,3.54,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,197,3.55,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,198,3.56,8]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[158,199,3.57,8]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,0,1.59,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,1,1.6,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,2,1.61,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,3,1.62,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,4,1.63,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,5,1.64,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,6,1.65,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,7,1.66,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,8,1.67,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,9,1.68,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,10,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,11,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,12,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,13,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,14,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,15,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,16,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,17,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,18,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 140, pc_makepoint(1, ARRAY[159,19,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,20,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,21,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,22,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,23,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,24,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,25,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,26,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,27,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,28,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,29,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,30,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,31,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,32,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,33,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,34,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,35,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,36,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,37,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,38,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 141, pc_makepoint(1, ARRAY[159,39,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,40,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,41,2,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,42,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,43,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,44,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,45,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,46,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,47,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,48,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,49,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,50,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,51,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,52,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,53,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,54,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,55,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,56,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,57,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,58,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 142, pc_makepoint(1, ARRAY[159,59,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,60,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,61,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,62,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,63,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,64,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,65,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,66,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,67,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,68,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,69,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,70,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,71,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,72,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,73,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,74,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,75,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,76,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,77,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,78,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 143, pc_makepoint(1, ARRAY[159,79,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,80,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,81,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,82,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,83,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,84,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,85,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,86,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,87,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,88,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,89,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,90,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,91,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,92,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,93,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,94,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,95,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,96,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,97,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,98,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 144, pc_makepoint(1, ARRAY[159,99,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,100,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,101,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,102,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,103,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,104,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,105,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,106,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,107,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,108,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,109,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,110,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,111,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,112,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,113,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,114,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,115,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,116,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,117,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,118,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 145, pc_makepoint(1, ARRAY[159,119,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,120,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,121,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,122,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,123,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,124,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,125,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,126,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,127,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,128,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,129,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,130,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,131,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,132,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,133,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,134,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,135,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,136,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,137,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,138,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 146, pc_makepoint(1, ARRAY[159,139,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,140,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,141,3,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,142,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,143,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,144,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,145,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,146,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,147,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,148,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,149,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,150,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,151,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,152,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,153,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,154,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,155,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,156,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,157,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,158,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 147, pc_makepoint(1, ARRAY[159,159,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,160,3.19,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,161,3.2,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,162,3.21,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,163,3.22,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,164,3.23,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,165,3.24,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,166,3.25,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,167,3.26,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,168,3.27,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,169,3.28,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,170,3.29,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,171,3.3,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,172,3.31,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,173,3.32,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,174,3.33,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,175,3.34,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,176,3.35,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,177,3.36,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,178,3.37,9]) );
insert into pts_collection ( gid, pt ) values ( 148, pc_makepoint(1, ARRAY[159,179,3.38,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,180,3.39,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,181,3.4,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,182,3.41,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,183,3.42,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,184,3.43,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,185,3.44,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,186,3.45,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,187,3.46,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,188,3.47,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,189,3.48,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,190,3.49,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,191,3.5,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,192,3.51,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,193,3.52,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,194,3.53,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,195,3.54,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,196,3.55,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,197,3.56,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,198,3.57,9]) );
insert into pts_collection ( gid, pt ) values ( 149, pc_makepoint(1, ARRAY[159,199,3.58,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,0,1.6,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,1,1.61,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,2,1.62,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,3,1.63,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,4,1.64,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,5,1.65,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,6,1.66,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,7,1.67,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,8,1.68,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,9,1.69,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,10,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,11,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,12,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,13,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,14,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,15,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,16,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,17,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,18,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[160,19,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,20,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,21,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,22,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,23,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,24,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,25,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,26,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,27,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,28,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,29,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,30,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,31,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,32,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,33,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,34,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,35,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,36,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,37,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,38,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[160,39,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,40,2,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,41,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,42,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,43,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,44,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,45,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,46,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,47,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,48,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,49,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,50,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,51,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,52,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,53,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,54,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,55,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,56,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,57,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,58,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[160,59,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,60,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,61,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,62,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,63,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,64,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,65,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,66,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,67,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,68,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,69,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,70,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,71,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,72,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,73,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,74,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,75,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,76,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,77,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,78,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[160,79,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,80,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,81,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,82,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,83,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,84,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,85,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,86,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,87,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,88,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,89,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,90,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,91,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,92,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,93,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,94,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,95,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,96,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,97,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,98,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[160,99,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,100,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,101,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,102,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,103,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,104,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,105,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,106,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,107,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,108,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,109,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,110,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,111,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,112,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,113,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,114,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,115,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,116,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,117,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,118,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[160,119,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,120,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,121,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,122,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,123,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,124,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,125,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,126,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,127,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,128,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,129,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,130,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,131,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,132,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,133,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,134,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,135,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,136,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,137,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,138,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[160,139,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,140,3,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,141,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,142,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,143,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,144,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,145,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,146,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,147,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,148,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,149,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,150,3.1,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,151,3.11,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,152,3.12,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,153,3.13,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,154,3.14,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,155,3.15,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,156,3.16,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,157,3.17,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,158,3.18,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[160,159,3.19,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,160,3.2,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,161,3.21,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,162,3.22,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,163,3.23,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,164,3.24,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,165,3.25,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,166,3.26,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,167,3.27,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,168,3.28,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,169,3.29,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,170,3.3,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,171,3.31,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,172,3.32,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,173,3.33,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,174,3.34,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,175,3.35,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,176,3.36,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,177,3.37,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,178,3.38,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[160,179,3.39,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,180,3.4,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,181,3.41,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,182,3.42,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,183,3.43,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,184,3.44,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,185,3.45,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,186,3.46,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,187,3.47,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,188,3.48,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,189,3.49,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,190,3.5,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,191,3.51,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,192,3.52,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,193,3.53,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,194,3.54,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,195,3.55,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,196,3.56,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,197,3.57,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,198,3.58,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[160,199,3.59,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,0,1.61,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,1,1.62,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,2,1.63,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,3,1.64,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,4,1.65,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,5,1.66,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,6,1.67,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,7,1.68,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,8,1.69,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,9,1.7,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,10,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,11,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,12,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,13,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,14,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,15,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,16,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,17,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,18,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[161,19,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,20,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,21,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,22,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,23,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,24,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,25,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,26,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,27,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,28,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,29,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,30,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,31,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,32,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,33,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,34,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,35,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,36,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,37,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,38,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[161,39,2,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,40,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,41,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,42,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,43,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,44,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,45,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,46,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,47,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,48,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,49,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,50,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,51,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,52,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,53,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,54,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,55,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,56,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,57,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,58,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[161,59,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,60,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,61,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,62,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,63,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,64,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,65,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,66,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,67,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,68,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,69,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,70,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,71,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,72,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,73,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,74,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,75,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,76,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,77,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,78,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[161,79,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,80,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,81,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,82,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,83,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,84,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,85,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,86,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,87,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,88,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,89,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,90,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,91,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,92,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,93,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,94,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,95,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,96,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,97,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,98,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[161,99,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,100,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,101,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,102,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,103,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,104,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,105,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,106,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,107,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,108,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,109,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,110,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,111,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,112,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,113,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,114,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,115,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,116,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,117,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,118,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[161,119,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,120,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,121,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,122,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,123,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,124,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,125,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,126,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,127,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,128,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,129,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,130,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,131,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,132,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,133,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,134,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,135,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,136,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,137,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,138,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[161,139,3,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,140,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,141,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,142,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,143,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,144,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,145,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,146,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,147,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,148,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,149,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,150,3.11,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,151,3.12,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,152,3.13,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,153,3.14,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,154,3.15,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,155,3.16,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,156,3.17,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,157,3.18,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,158,3.19,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[161,159,3.2,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,160,3.21,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,161,3.22,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,162,3.23,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,163,3.24,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,164,3.25,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,165,3.26,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,166,3.27,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,167,3.28,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,168,3.29,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,169,3.3,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,170,3.31,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,171,3.32,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,172,3.33,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,173,3.34,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,174,3.35,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,175,3.36,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,176,3.37,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,177,3.38,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,178,3.39,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[161,179,3.4,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,180,3.41,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,181,3.42,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,182,3.43,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,183,3.44,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,184,3.45,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,185,3.46,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,186,3.47,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,187,3.48,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,188,3.49,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,189,3.5,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,190,3.51,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,191,3.52,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,192,3.53,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,193,3.54,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,194,3.55,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,195,3.56,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,196,3.57,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,197,3.58,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,198,3.59,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[161,199,3.6,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,0,1.62,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,1,1.63,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,2,1.64,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,3,1.65,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,4,1.66,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,5,1.67,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,6,1.68,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,7,1.69,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,8,1.7,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,9,1.71,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,10,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,11,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,12,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,13,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,14,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,15,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,16,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,17,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,18,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[162,19,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,20,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,21,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,22,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,23,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,24,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,25,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,26,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,27,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,28,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,29,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,30,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,31,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,32,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,33,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,34,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,35,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,36,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,37,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,38,2,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[162,39,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,40,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,41,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,42,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,43,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,44,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,45,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,46,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,47,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,48,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,49,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,50,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,51,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,52,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,53,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,54,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,55,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,56,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,57,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,58,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[162,59,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,60,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,61,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,62,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,63,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,64,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,65,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,66,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,67,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,68,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,69,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,70,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,71,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,72,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,73,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,74,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,75,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,76,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,77,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,78,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[162,79,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,80,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,81,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,82,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,83,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,84,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,85,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,86,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,87,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,88,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,89,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,90,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,91,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,92,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,93,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,94,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,95,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,96,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,97,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,98,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[162,99,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,100,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,101,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,102,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,103,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,104,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,105,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,106,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,107,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,108,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,109,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,110,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,111,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,112,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,113,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,114,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,115,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,116,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,117,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,118,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[162,119,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,120,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,121,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,122,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,123,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,124,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,125,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,126,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,127,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,128,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,129,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,130,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,131,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,132,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,133,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,134,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,135,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,136,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,137,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,138,3,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[162,139,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,140,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,141,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,142,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,143,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,144,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,145,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,146,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,147,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,148,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,149,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,150,3.12,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,151,3.13,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,152,3.14,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,153,3.15,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,154,3.16,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,155,3.17,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,156,3.18,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,157,3.19,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,158,3.2,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[162,159,3.21,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,160,3.22,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,161,3.23,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,162,3.24,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,163,3.25,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,164,3.26,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,165,3.27,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,166,3.28,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,167,3.29,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,168,3.3,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,169,3.31,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,170,3.32,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,171,3.33,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,172,3.34,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,173,3.35,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,174,3.36,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,175,3.37,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,176,3.38,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,177,3.39,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,178,3.4,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[162,179,3.41,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,180,3.42,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,181,3.43,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,182,3.44,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,183,3.45,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,184,3.46,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,185,3.47,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,186,3.48,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,187,3.49,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,188,3.5,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,189,3.51,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,190,3.52,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,191,3.53,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,192,3.54,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,193,3.55,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,194,3.56,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,195,3.57,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,196,3.58,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,197,3.59,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,198,3.6,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[162,199,3.61,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,0,1.63,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,1,1.64,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,2,1.65,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,3,1.66,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,4,1.67,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,5,1.68,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,6,1.69,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,7,1.7,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,8,1.71,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,9,1.72,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,10,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,11,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,12,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,13,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,14,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,15,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,16,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,17,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,18,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[163,19,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,20,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,21,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,22,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,23,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,24,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,25,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,26,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,27,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,28,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,29,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,30,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,31,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,32,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,33,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,34,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,35,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,36,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,37,2,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,38,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[163,39,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,40,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,41,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,42,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,43,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,44,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,45,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,46,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,47,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,48,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,49,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,50,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,51,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,52,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,53,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,54,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,55,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,56,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,57,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,58,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[163,59,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,60,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,61,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,62,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,63,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,64,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,65,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,66,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,67,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,68,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,69,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,70,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,71,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,72,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,73,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,74,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,75,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,76,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,77,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,78,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[163,79,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,80,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,81,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,82,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,83,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,84,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,85,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,86,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,87,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,88,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,89,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,90,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,91,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,92,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,93,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,94,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,95,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,96,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,97,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,98,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[163,99,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,100,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,101,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,102,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,103,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,104,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,105,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,106,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,107,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,108,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,109,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,110,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,111,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,112,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,113,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,114,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,115,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,116,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,117,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,118,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[163,119,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,120,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,121,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,122,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,123,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,124,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,125,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,126,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,127,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,128,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,129,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,130,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,131,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,132,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,133,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,134,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,135,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,136,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,137,3,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,138,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[163,139,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,140,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,141,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,142,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,143,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,144,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,145,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,146,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,147,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,148,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,149,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,150,3.13,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,151,3.14,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,152,3.15,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,153,3.16,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,154,3.17,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,155,3.18,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,156,3.19,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,157,3.2,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,158,3.21,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[163,159,3.22,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,160,3.23,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,161,3.24,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,162,3.25,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,163,3.26,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,164,3.27,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,165,3.28,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,166,3.29,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,167,3.3,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,168,3.31,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,169,3.32,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,170,3.33,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,171,3.34,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,172,3.35,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,173,3.36,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,174,3.37,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,175,3.38,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,176,3.39,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,177,3.4,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,178,3.41,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[163,179,3.42,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,180,3.43,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,181,3.44,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,182,3.45,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,183,3.46,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,184,3.47,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,185,3.48,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,186,3.49,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,187,3.5,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,188,3.51,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,189,3.52,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,190,3.53,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,191,3.54,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,192,3.55,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,193,3.56,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,194,3.57,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,195,3.58,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,196,3.59,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,197,3.6,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,198,3.61,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[163,199,3.62,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,0,1.64,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,1,1.65,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,2,1.66,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,3,1.67,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,4,1.68,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,5,1.69,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,6,1.7,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,7,1.71,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,8,1.72,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,9,1.73,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,10,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,11,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,12,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,13,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,14,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,15,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,16,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,17,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,18,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[164,19,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,20,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,21,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,22,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,23,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,24,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,25,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,26,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,27,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,28,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,29,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,30,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,31,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,32,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,33,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,34,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,35,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,36,2,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,37,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,38,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[164,39,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,40,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,41,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,42,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,43,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,44,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,45,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,46,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,47,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,48,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,49,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,50,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,51,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,52,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,53,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,54,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,55,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,56,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,57,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,58,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[164,59,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,60,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,61,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,62,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,63,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,64,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,65,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,66,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,67,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,68,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,69,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,70,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,71,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,72,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,73,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,74,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,75,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,76,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,77,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,78,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[164,79,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,80,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,81,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,82,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,83,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,84,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,85,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,86,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,87,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,88,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,89,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,90,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,91,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,92,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,93,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,94,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,95,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,96,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,97,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,98,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[164,99,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,100,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,101,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,102,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,103,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,104,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,105,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,106,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,107,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,108,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,109,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,110,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,111,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,112,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,113,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,114,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,115,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,116,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,117,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,118,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[164,119,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,120,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,121,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,122,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,123,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,124,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,125,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,126,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,127,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,128,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,129,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,130,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,131,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,132,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,133,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,134,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,135,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,136,3,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,137,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,138,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[164,139,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,140,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,141,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,142,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,143,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,144,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,145,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,146,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,147,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,148,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,149,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,150,3.14,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,151,3.15,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,152,3.16,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,153,3.17,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,154,3.18,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,155,3.19,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,156,3.2,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,157,3.21,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,158,3.22,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[164,159,3.23,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,160,3.24,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,161,3.25,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,162,3.26,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,163,3.27,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,164,3.28,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,165,3.29,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,166,3.3,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,167,3.31,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,168,3.32,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,169,3.33,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,170,3.34,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,171,3.35,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,172,3.36,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,173,3.37,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,174,3.38,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,175,3.39,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,176,3.4,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,177,3.41,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,178,3.42,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[164,179,3.43,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,180,3.44,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,181,3.45,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,182,3.46,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,183,3.47,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,184,3.48,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,185,3.49,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,186,3.5,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,187,3.51,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,188,3.52,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,189,3.53,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,190,3.54,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,191,3.55,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,192,3.56,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,193,3.57,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,194,3.58,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,195,3.59,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,196,3.6,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,197,3.61,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,198,3.62,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[164,199,3.63,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,0,1.65,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,1,1.66,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,2,1.67,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,3,1.68,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,4,1.69,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,5,1.7,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,6,1.71,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,7,1.72,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,8,1.73,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,9,1.74,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,10,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,11,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,12,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,13,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,14,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,15,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,16,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,17,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,18,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[165,19,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,20,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,21,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,22,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,23,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,24,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,25,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,26,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,27,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,28,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,29,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,30,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,31,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,32,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,33,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,34,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,35,2,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,36,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,37,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,38,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[165,39,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,40,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,41,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,42,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,43,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,44,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,45,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,46,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,47,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,48,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,49,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,50,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,51,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,52,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,53,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,54,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,55,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,56,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,57,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,58,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[165,59,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,60,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,61,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,62,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,63,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,64,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,65,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,66,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,67,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,68,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,69,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,70,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,71,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,72,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,73,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,74,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,75,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,76,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,77,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,78,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[165,79,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,80,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,81,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,82,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,83,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,84,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,85,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,86,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,87,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,88,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,89,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,90,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,91,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,92,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,93,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,94,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,95,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,96,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,97,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,98,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[165,99,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,100,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,101,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,102,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,103,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,104,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,105,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,106,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,107,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,108,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,109,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,110,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,111,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,112,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,113,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,114,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,115,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,116,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,117,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,118,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[165,119,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,120,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,121,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,122,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,123,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,124,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,125,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,126,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,127,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,128,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,129,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,130,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,131,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,132,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,133,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,134,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,135,3,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,136,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,137,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,138,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[165,139,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,140,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,141,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,142,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,143,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,144,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,145,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,146,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,147,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,148,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,149,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,150,3.15,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,151,3.16,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,152,3.17,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,153,3.18,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,154,3.19,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,155,3.2,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,156,3.21,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,157,3.22,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,158,3.23,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[165,159,3.24,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,160,3.25,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,161,3.26,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,162,3.27,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,163,3.28,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,164,3.29,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,165,3.3,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,166,3.31,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,167,3.32,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,168,3.33,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,169,3.34,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,170,3.35,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,171,3.36,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,172,3.37,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,173,3.38,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,174,3.39,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,175,3.4,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,176,3.41,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,177,3.42,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,178,3.43,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[165,179,3.44,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,180,3.45,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,181,3.46,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,182,3.47,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,183,3.48,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,184,3.49,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,185,3.5,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,186,3.51,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,187,3.52,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,188,3.53,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,189,3.54,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,190,3.55,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,191,3.56,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,192,3.57,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,193,3.58,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,194,3.59,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,195,3.6,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,196,3.61,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,197,3.62,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,198,3.63,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[165,199,3.64,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,0,1.66,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,1,1.67,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,2,1.68,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,3,1.69,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,4,1.7,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,5,1.71,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,6,1.72,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,7,1.73,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,8,1.74,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,9,1.75,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,10,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,11,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,12,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,13,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,14,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,15,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,16,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,17,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,18,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[166,19,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,20,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,21,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,22,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,23,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,24,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,25,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,26,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,27,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,28,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,29,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,30,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,31,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,32,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,33,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,34,2,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,35,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,36,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,37,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,38,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[166,39,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,40,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,41,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,42,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,43,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,44,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,45,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,46,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,47,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,48,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,49,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,50,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,51,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,52,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,53,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,54,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,55,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,56,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,57,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,58,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[166,59,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,60,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,61,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,62,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,63,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,64,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,65,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,66,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,67,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,68,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,69,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,70,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,71,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,72,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,73,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,74,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,75,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,76,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,77,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,78,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[166,79,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,80,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,81,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,82,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,83,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,84,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,85,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,86,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,87,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,88,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,89,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,90,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,91,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,92,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,93,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,94,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,95,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,96,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,97,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,98,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[166,99,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,100,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,101,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,102,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,103,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,104,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,105,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,106,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,107,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,108,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,109,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,110,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,111,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,112,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,113,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,114,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,115,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,116,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,117,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,118,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[166,119,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,120,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,121,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,122,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,123,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,124,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,125,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,126,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,127,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,128,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,129,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,130,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,131,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,132,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,133,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,134,3,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,135,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,136,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,137,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,138,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[166,139,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,140,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,141,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,142,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,143,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,144,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,145,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,146,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,147,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,148,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,149,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,150,3.16,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,151,3.17,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,152,3.18,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,153,3.19,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,154,3.2,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,155,3.21,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,156,3.22,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,157,3.23,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,158,3.24,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[166,159,3.25,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,160,3.26,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,161,3.27,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,162,3.28,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,163,3.29,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,164,3.3,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,165,3.31,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,166,3.32,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,167,3.33,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,168,3.34,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,169,3.35,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,170,3.36,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,171,3.37,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,172,3.38,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,173,3.39,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,174,3.4,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,175,3.41,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,176,3.42,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,177,3.43,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,178,3.44,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[166,179,3.45,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,180,3.46,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,181,3.47,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,182,3.48,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,183,3.49,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,184,3.5,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,185,3.51,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,186,3.52,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,187,3.53,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,188,3.54,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,189,3.55,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,190,3.56,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,191,3.57,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,192,3.58,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,193,3.59,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,194,3.6,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,195,3.61,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,196,3.62,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,197,3.63,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,198,3.64,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[166,199,3.65,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,0,1.67,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,1,1.68,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,2,1.69,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,3,1.7,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,4,1.71,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,5,1.72,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,6,1.73,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,7,1.74,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,8,1.75,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,9,1.76,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,10,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,11,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,12,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,13,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,14,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,15,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,16,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,17,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,18,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[167,19,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,20,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,21,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,22,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,23,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,24,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,25,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,26,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,27,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,28,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,29,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,30,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,31,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,32,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,33,2,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,34,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,35,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,36,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,37,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,38,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[167,39,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,40,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,41,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,42,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,43,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,44,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,45,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,46,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,47,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,48,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,49,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,50,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,51,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,52,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,53,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,54,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,55,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,56,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,57,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,58,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[167,59,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,60,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,61,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,62,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,63,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,64,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,65,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,66,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,67,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,68,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,69,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,70,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,71,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,72,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,73,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,74,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,75,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,76,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,77,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,78,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[167,79,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,80,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,81,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,82,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,83,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,84,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,85,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,86,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,87,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,88,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,89,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,90,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,91,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,92,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,93,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,94,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,95,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,96,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,97,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,98,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[167,99,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,100,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,101,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,102,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,103,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,104,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,105,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,106,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,107,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,108,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,109,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,110,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,111,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,112,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,113,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,114,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,115,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,116,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,117,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,118,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[167,119,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,120,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,121,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,122,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,123,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,124,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,125,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,126,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,127,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,128,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,129,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,130,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,131,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,132,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,133,3,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,134,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,135,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,136,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,137,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,138,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[167,139,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,140,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,141,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,142,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,143,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,144,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,145,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,146,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,147,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,148,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,149,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,150,3.17,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,151,3.18,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,152,3.19,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,153,3.2,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,154,3.21,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,155,3.22,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,156,3.23,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,157,3.24,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,158,3.25,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[167,159,3.26,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,160,3.27,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,161,3.28,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,162,3.29,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,163,3.3,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,164,3.31,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,165,3.32,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,166,3.33,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,167,3.34,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,168,3.35,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,169,3.36,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,170,3.37,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,171,3.38,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,172,3.39,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,173,3.4,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,174,3.41,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,175,3.42,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,176,3.43,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,177,3.44,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,178,3.45,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[167,179,3.46,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,180,3.47,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,181,3.48,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,182,3.49,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,183,3.5,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,184,3.51,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,185,3.52,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,186,3.53,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,187,3.54,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,188,3.55,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,189,3.56,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,190,3.57,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,191,3.58,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,192,3.59,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,193,3.6,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,194,3.61,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,195,3.62,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,196,3.63,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,197,3.64,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,198,3.65,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[167,199,3.66,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,0,1.68,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,1,1.69,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,2,1.7,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,3,1.71,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,4,1.72,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,5,1.73,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,6,1.74,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,7,1.75,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,8,1.76,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,9,1.77,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,10,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,11,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,12,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,13,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,14,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,15,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,16,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,17,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,18,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[168,19,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,20,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,21,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,22,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,23,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,24,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,25,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,26,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,27,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,28,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,29,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,30,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,31,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,32,2,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,33,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,34,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,35,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,36,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,37,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,38,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[168,39,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,40,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,41,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,42,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,43,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,44,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,45,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,46,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,47,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,48,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,49,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,50,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,51,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,52,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,53,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,54,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,55,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,56,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,57,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,58,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[168,59,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,60,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,61,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,62,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,63,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,64,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,65,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,66,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,67,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,68,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,69,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,70,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,71,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,72,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,73,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,74,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,75,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,76,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,77,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,78,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[168,79,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,80,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,81,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,82,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,83,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,84,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,85,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,86,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,87,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,88,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,89,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,90,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,91,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,92,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,93,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,94,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,95,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,96,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,97,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,98,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[168,99,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,100,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,101,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,102,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,103,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,104,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,105,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,106,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,107,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,108,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,109,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,110,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,111,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,112,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,113,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,114,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,115,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,116,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,117,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,118,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[168,119,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,120,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,121,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,122,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,123,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,124,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,125,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,126,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,127,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,128,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,129,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,130,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,131,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,132,3,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,133,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,134,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,135,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,136,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,137,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,138,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[168,139,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,140,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,141,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,142,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,143,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,144,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,145,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,146,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,147,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,148,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,149,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,150,3.18,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,151,3.19,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,152,3.2,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,153,3.21,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,154,3.22,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,155,3.23,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,156,3.24,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,157,3.25,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,158,3.26,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[168,159,3.27,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,160,3.28,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,161,3.29,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,162,3.3,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,163,3.31,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,164,3.32,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,165,3.33,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,166,3.34,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,167,3.35,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,168,3.36,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,169,3.37,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,170,3.38,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,171,3.39,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,172,3.4,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,173,3.41,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,174,3.42,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,175,3.43,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,176,3.44,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,177,3.45,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,178,3.46,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[168,179,3.47,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,180,3.48,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,181,3.49,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,182,3.5,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,183,3.51,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,184,3.52,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,185,3.53,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,186,3.54,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,187,3.55,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,188,3.56,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,189,3.57,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,190,3.58,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,191,3.59,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,192,3.6,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,193,3.61,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,194,3.62,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,195,3.63,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,196,3.64,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,197,3.65,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,198,3.66,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[168,199,3.67,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,0,1.69,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,1,1.7,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,2,1.71,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,3,1.72,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,4,1.73,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,5,1.74,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,6,1.75,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,7,1.76,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,8,1.77,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,9,1.78,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,10,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,11,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,12,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,13,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,14,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,15,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,16,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,17,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,18,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[169,19,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,20,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,21,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,22,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,23,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,24,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,25,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,26,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,27,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,28,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,29,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,30,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,31,2,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,32,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,33,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,34,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,35,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,36,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,37,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,38,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[169,39,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,40,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,41,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,42,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,43,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,44,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,45,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,46,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,47,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,48,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,49,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,50,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,51,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,52,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,53,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,54,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,55,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,56,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,57,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,58,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[169,59,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,60,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,61,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,62,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,63,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,64,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,65,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,66,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,67,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,68,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,69,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,70,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,71,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,72,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,73,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,74,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,75,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,76,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,77,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,78,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[169,79,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,80,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,81,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,82,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,83,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,84,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,85,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,86,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,87,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,88,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,89,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,90,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,91,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,92,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,93,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,94,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,95,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,96,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,97,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,98,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[169,99,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,100,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,101,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,102,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,103,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,104,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,105,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,106,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,107,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,108,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,109,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,110,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,111,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,112,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,113,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,114,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,115,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,116,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,117,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,118,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[169,119,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,120,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,121,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,122,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,123,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,124,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,125,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,126,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,127,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,128,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,129,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,130,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,131,3,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,132,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,133,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,134,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,135,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,136,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,137,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,138,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[169,139,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,140,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,141,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,142,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,143,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,144,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,145,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,146,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,147,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,148,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,149,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,150,3.19,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,151,3.2,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,152,3.21,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,153,3.22,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,154,3.23,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,155,3.24,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,156,3.25,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,157,3.26,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,158,3.27,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[169,159,3.28,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,160,3.29,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,161,3.3,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,162,3.31,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,163,3.32,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,164,3.33,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,165,3.34,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,166,3.35,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,167,3.36,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,168,3.37,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,169,3.38,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,170,3.39,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,171,3.4,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,172,3.41,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,173,3.42,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,174,3.43,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,175,3.44,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,176,3.45,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,177,3.46,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,178,3.47,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[169,179,3.48,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,180,3.49,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,181,3.5,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,182,3.51,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,183,3.52,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,184,3.53,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,185,3.54,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,186,3.55,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,187,3.56,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,188,3.57,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,189,3.58,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,190,3.59,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,191,3.6,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,192,3.61,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,193,3.62,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,194,3.63,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,195,3.64,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,196,3.65,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,197,3.66,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,198,3.67,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[169,199,3.68,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,0,1.7,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,1,1.71,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,2,1.72,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,3,1.73,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,4,1.74,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,5,1.75,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,6,1.76,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,7,1.77,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,8,1.78,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,9,1.79,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,10,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,11,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,12,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,13,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,14,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,15,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,16,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,17,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,18,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[170,19,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,20,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,21,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,22,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,23,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,24,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,25,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,26,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,27,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,28,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,29,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,30,2,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,31,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,32,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,33,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,34,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,35,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,36,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,37,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,38,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[170,39,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,40,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,41,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,42,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,43,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,44,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,45,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,46,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,47,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,48,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,49,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,50,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,51,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,52,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,53,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,54,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,55,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,56,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,57,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,58,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[170,59,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,60,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,61,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,62,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,63,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,64,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,65,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,66,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,67,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,68,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,69,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,70,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,71,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,72,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,73,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,74,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,75,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,76,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,77,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,78,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[170,79,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,80,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,81,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,82,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,83,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,84,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,85,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,86,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,87,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,88,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,89,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,90,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,91,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,92,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,93,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,94,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,95,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,96,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,97,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,98,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[170,99,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,100,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,101,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,102,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,103,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,104,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,105,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,106,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,107,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,108,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,109,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,110,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,111,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,112,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,113,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,114,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,115,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,116,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,117,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,118,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[170,119,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,120,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,121,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,122,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,123,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,124,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,125,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,126,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,127,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,128,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,129,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,130,3,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,131,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,132,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,133,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,134,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,135,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,136,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,137,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,138,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[170,139,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,140,3.1,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,141,3.11,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,142,3.12,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,143,3.13,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,144,3.14,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,145,3.15,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,146,3.16,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,147,3.17,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,148,3.18,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,149,3.19,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,150,3.2,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,151,3.21,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,152,3.22,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,153,3.23,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,154,3.24,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,155,3.25,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,156,3.26,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,157,3.27,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,158,3.28,0]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[170,159,3.29,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,160,3.3,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,161,3.31,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,162,3.32,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,163,3.33,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,164,3.34,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,165,3.35,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,166,3.36,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,167,3.37,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,168,3.38,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,169,3.39,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,170,3.4,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,171,3.41,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,172,3.42,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,173,3.43,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,174,3.44,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,175,3.45,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,176,3.46,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,177,3.47,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,178,3.48,0]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[170,179,3.49,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,180,3.5,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,181,3.51,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,182,3.52,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,183,3.53,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,184,3.54,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,185,3.55,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,186,3.56,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,187,3.57,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,188,3.58,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,189,3.59,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,190,3.6,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,191,3.61,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,192,3.62,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,193,3.63,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,194,3.64,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,195,3.65,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,196,3.66,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,197,3.67,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,198,3.68,0]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[170,199,3.69,0]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,0,1.71,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,1,1.72,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,2,1.73,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,3,1.74,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,4,1.75,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,5,1.76,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,6,1.77,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,7,1.78,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,8,1.79,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,9,1.8,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,10,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,11,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,12,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,13,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,14,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,15,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,16,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,17,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,18,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[171,19,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,20,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,21,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,22,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,23,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,24,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,25,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,26,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,27,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,28,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,29,2,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,30,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,31,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,32,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,33,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,34,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,35,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,36,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,37,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,38,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[171,39,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,40,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,41,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,42,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,43,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,44,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,45,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,46,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,47,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,48,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,49,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,50,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,51,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,52,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,53,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,54,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,55,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,56,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,57,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,58,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[171,59,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,60,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,61,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,62,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,63,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,64,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,65,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,66,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,67,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,68,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,69,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,70,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,71,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,72,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,73,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,74,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,75,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,76,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,77,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,78,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[171,79,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,80,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,81,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,82,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,83,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,84,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,85,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,86,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,87,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,88,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,89,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,90,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,91,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,92,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,93,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,94,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,95,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,96,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,97,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,98,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[171,99,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,100,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,101,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,102,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,103,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,104,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,105,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,106,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,107,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,108,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,109,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,110,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,111,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,112,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,113,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,114,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,115,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,116,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,117,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,118,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[171,119,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,120,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,121,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,122,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,123,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,124,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,125,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,126,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,127,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,128,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,129,3,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,130,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,131,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,132,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,133,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,134,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,135,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,136,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,137,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,138,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[171,139,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,140,3.11,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,141,3.12,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,142,3.13,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,143,3.14,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,144,3.15,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,145,3.16,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,146,3.17,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,147,3.18,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,148,3.19,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,149,3.2,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,150,3.21,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,151,3.22,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,152,3.23,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,153,3.24,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,154,3.25,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,155,3.26,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,156,3.27,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,157,3.28,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,158,3.29,1]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[171,159,3.3,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,160,3.31,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,161,3.32,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,162,3.33,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,163,3.34,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,164,3.35,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,165,3.36,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,166,3.37,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,167,3.38,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,168,3.39,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,169,3.4,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,170,3.41,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,171,3.42,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,172,3.43,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,173,3.44,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,174,3.45,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,175,3.46,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,176,3.47,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,177,3.48,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,178,3.49,1]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[171,179,3.5,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,180,3.51,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,181,3.52,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,182,3.53,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,183,3.54,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,184,3.55,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,185,3.56,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,186,3.57,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,187,3.58,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,188,3.59,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,189,3.6,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,190,3.61,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,191,3.62,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,192,3.63,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,193,3.64,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,194,3.65,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,195,3.66,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,196,3.67,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,197,3.68,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,198,3.69,1]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[171,199,3.7,1]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,0,1.72,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,1,1.73,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,2,1.74,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,3,1.75,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,4,1.76,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,5,1.77,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,6,1.78,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,7,1.79,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,8,1.8,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,9,1.81,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,10,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,11,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,12,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,13,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,14,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,15,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,16,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,17,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,18,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[172,19,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,20,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,21,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,22,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,23,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,24,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,25,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,26,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,27,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,28,2,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,29,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,30,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,31,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,32,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,33,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,34,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,35,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,36,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,37,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,38,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[172,39,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,40,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,41,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,42,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,43,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,44,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,45,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,46,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,47,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,48,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,49,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,50,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,51,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,52,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,53,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,54,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,55,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,56,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,57,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,58,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[172,59,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,60,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,61,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,62,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,63,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,64,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,65,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,66,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,67,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,68,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,69,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,70,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,71,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,72,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,73,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,74,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,75,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,76,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,77,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,78,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[172,79,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,80,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,81,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,82,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,83,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,84,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,85,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,86,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,87,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,88,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,89,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,90,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,91,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,92,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,93,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,94,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,95,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,96,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,97,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,98,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[172,99,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,100,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,101,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,102,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,103,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,104,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,105,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,106,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,107,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,108,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,109,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,110,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,111,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,112,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,113,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,114,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,115,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,116,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,117,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,118,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[172,119,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,120,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,121,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,122,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,123,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,124,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,125,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,126,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,127,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,128,3,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,129,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,130,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,131,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,132,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,133,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,134,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,135,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,136,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,137,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,138,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[172,139,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,140,3.12,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,141,3.13,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,142,3.14,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,143,3.15,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,144,3.16,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,145,3.17,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,146,3.18,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,147,3.19,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,148,3.2,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,149,3.21,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,150,3.22,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,151,3.23,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,152,3.24,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,153,3.25,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,154,3.26,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,155,3.27,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,156,3.28,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,157,3.29,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,158,3.3,2]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[172,159,3.31,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,160,3.32,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,161,3.33,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,162,3.34,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,163,3.35,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,164,3.36,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,165,3.37,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,166,3.38,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,167,3.39,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,168,3.4,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,169,3.41,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,170,3.42,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,171,3.43,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,172,3.44,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,173,3.45,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,174,3.46,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,175,3.47,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,176,3.48,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,177,3.49,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,178,3.5,2]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[172,179,3.51,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,180,3.52,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,181,3.53,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,182,3.54,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,183,3.55,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,184,3.56,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,185,3.57,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,186,3.58,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,187,3.59,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,188,3.6,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,189,3.61,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,190,3.62,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,191,3.63,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,192,3.64,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,193,3.65,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,194,3.66,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,195,3.67,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,196,3.68,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,197,3.69,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,198,3.7,2]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[172,199,3.71,2]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,0,1.73,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,1,1.74,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,2,1.75,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,3,1.76,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,4,1.77,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,5,1.78,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,6,1.79,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,7,1.8,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,8,1.81,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,9,1.82,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,10,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,11,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,12,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,13,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,14,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,15,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,16,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,17,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,18,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[173,19,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,20,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,21,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,22,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,23,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,24,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,25,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,26,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,27,2,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,28,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,29,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,30,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,31,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,32,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,33,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,34,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,35,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,36,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,37,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,38,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[173,39,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,40,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,41,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,42,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,43,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,44,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,45,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,46,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,47,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,48,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,49,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,50,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,51,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,52,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,53,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,54,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,55,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,56,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,57,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,58,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[173,59,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,60,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,61,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,62,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,63,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,64,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,65,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,66,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,67,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,68,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,69,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,70,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,71,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,72,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,73,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,74,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,75,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,76,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,77,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,78,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[173,79,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,80,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,81,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,82,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,83,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,84,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,85,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,86,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,87,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,88,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,89,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,90,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,91,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,92,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,93,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,94,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,95,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,96,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,97,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,98,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[173,99,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,100,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,101,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,102,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,103,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,104,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,105,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,106,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,107,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,108,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,109,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,110,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,111,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,112,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,113,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,114,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,115,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,116,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,117,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,118,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[173,119,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,120,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,121,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,122,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,123,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,124,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,125,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,126,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,127,3,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,128,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,129,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,130,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,131,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,132,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,133,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,134,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,135,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,136,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,137,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,138,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[173,139,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,140,3.13,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,141,3.14,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,142,3.15,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,143,3.16,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,144,3.17,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,145,3.18,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,146,3.19,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,147,3.2,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,148,3.21,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,149,3.22,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,150,3.23,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,151,3.24,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,152,3.25,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,153,3.26,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,154,3.27,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,155,3.28,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,156,3.29,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,157,3.3,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,158,3.31,3]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[173,159,3.32,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,160,3.33,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,161,3.34,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,162,3.35,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,163,3.36,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,164,3.37,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,165,3.38,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,166,3.39,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,167,3.4,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,168,3.41,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,169,3.42,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,170,3.43,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,171,3.44,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,172,3.45,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,173,3.46,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,174,3.47,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,175,3.48,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,176,3.49,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,177,3.5,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,178,3.51,3]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[173,179,3.52,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,180,3.53,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,181,3.54,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,182,3.55,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,183,3.56,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,184,3.57,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,185,3.58,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,186,3.59,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,187,3.6,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,188,3.61,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,189,3.62,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,190,3.63,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,191,3.64,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,192,3.65,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,193,3.66,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,194,3.67,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,195,3.68,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,196,3.69,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,197,3.7,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,198,3.71,3]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[173,199,3.72,3]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,0,1.74,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,1,1.75,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,2,1.76,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,3,1.77,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,4,1.78,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,5,1.79,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,6,1.8,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,7,1.81,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,8,1.82,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,9,1.83,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,10,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,11,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,12,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,13,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,14,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,15,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,16,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,17,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,18,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[174,19,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,20,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,21,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,22,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,23,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,24,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,25,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,26,2,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,27,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,28,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,29,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,30,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,31,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,32,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,33,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,34,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,35,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,36,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,37,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,38,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[174,39,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,40,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,41,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,42,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,43,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,44,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,45,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,46,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,47,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,48,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,49,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,50,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,51,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,52,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,53,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,54,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,55,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,56,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,57,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,58,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[174,59,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,60,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,61,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,62,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,63,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,64,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,65,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,66,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,67,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,68,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,69,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,70,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,71,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,72,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,73,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,74,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,75,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,76,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,77,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,78,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[174,79,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,80,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,81,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,82,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,83,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,84,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,85,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,86,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,87,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,88,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,89,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,90,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,91,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,92,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,93,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,94,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,95,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,96,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,97,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,98,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[174,99,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,100,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,101,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,102,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,103,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,104,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,105,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,106,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,107,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,108,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,109,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,110,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,111,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,112,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,113,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,114,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,115,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,116,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,117,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,118,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[174,119,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,120,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,121,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,122,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,123,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,124,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,125,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,126,3,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,127,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,128,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,129,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,130,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,131,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,132,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,133,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,134,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,135,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,136,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,137,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,138,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[174,139,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,140,3.14,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,141,3.15,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,142,3.16,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,143,3.17,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,144,3.18,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,145,3.19,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,146,3.2,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,147,3.21,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,148,3.22,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,149,3.23,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,150,3.24,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,151,3.25,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,152,3.26,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,153,3.27,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,154,3.28,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,155,3.29,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,156,3.3,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,157,3.31,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,158,3.32,4]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[174,159,3.33,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,160,3.34,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,161,3.35,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,162,3.36,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,163,3.37,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,164,3.38,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,165,3.39,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,166,3.4,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,167,3.41,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,168,3.42,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,169,3.43,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,170,3.44,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,171,3.45,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,172,3.46,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,173,3.47,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,174,3.48,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,175,3.49,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,176,3.5,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,177,3.51,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,178,3.52,4]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[174,179,3.53,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,180,3.54,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,181,3.55,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,182,3.56,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,183,3.57,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,184,3.58,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,185,3.59,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,186,3.6,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,187,3.61,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,188,3.62,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,189,3.63,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,190,3.64,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,191,3.65,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,192,3.66,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,193,3.67,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,194,3.68,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,195,3.69,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,196,3.7,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,197,3.71,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,198,3.72,4]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[174,199,3.73,4]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,0,1.75,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,1,1.76,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,2,1.77,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,3,1.78,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,4,1.79,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,5,1.8,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,6,1.81,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,7,1.82,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,8,1.83,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,9,1.84,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,10,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,11,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,12,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,13,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,14,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,15,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,16,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,17,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,18,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[175,19,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,20,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,21,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,22,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,23,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,24,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,25,2,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,26,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,27,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,28,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,29,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,30,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,31,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,32,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,33,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,34,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,35,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,36,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,37,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,38,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[175,39,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,40,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,41,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,42,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,43,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,44,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,45,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,46,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,47,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,48,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,49,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,50,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,51,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,52,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,53,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,54,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,55,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,56,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,57,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,58,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[175,59,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,60,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,61,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,62,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,63,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,64,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,65,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,66,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,67,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,68,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,69,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,70,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,71,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,72,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,73,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,74,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,75,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,76,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,77,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,78,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[175,79,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,80,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,81,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,82,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,83,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,84,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,85,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,86,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,87,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,88,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,89,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,90,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,91,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,92,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,93,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,94,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,95,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,96,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,97,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,98,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[175,99,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,100,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,101,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,102,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,103,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,104,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,105,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,106,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,107,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,108,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,109,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,110,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,111,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,112,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,113,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,114,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,115,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,116,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,117,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,118,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[175,119,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,120,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,121,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,122,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,123,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,124,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,125,3,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,126,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,127,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,128,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,129,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,130,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,131,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,132,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,133,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,134,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,135,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,136,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,137,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,138,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[175,139,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,140,3.15,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,141,3.16,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,142,3.17,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,143,3.18,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,144,3.19,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,145,3.2,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,146,3.21,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,147,3.22,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,148,3.23,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,149,3.24,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,150,3.25,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,151,3.26,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,152,3.27,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,153,3.28,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,154,3.29,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,155,3.3,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,156,3.31,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,157,3.32,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,158,3.33,5]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[175,159,3.34,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,160,3.35,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,161,3.36,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,162,3.37,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,163,3.38,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,164,3.39,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,165,3.4,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,166,3.41,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,167,3.42,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,168,3.43,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,169,3.44,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,170,3.45,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,171,3.46,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,172,3.47,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,173,3.48,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,174,3.49,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,175,3.5,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,176,3.51,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,177,3.52,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,178,3.53,5]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[175,179,3.54,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,180,3.55,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,181,3.56,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,182,3.57,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,183,3.58,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,184,3.59,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,185,3.6,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,186,3.61,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,187,3.62,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,188,3.63,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,189,3.64,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,190,3.65,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,191,3.66,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,192,3.67,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,193,3.68,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,194,3.69,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,195,3.7,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,196,3.71,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,197,3.72,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,198,3.73,5]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[175,199,3.74,5]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,0,1.76,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,1,1.77,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,2,1.78,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,3,1.79,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,4,1.8,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,5,1.81,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,6,1.82,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,7,1.83,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,8,1.84,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,9,1.85,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,10,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,11,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,12,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,13,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,14,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,15,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,16,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,17,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,18,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[176,19,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,20,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,21,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,22,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,23,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,24,2,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,25,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,26,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,27,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,28,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,29,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,30,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,31,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,32,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,33,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,34,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,35,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,36,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,37,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,38,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[176,39,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,40,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,41,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,42,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,43,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,44,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,45,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,46,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,47,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,48,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,49,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,50,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,51,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,52,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,53,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,54,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,55,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,56,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,57,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,58,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[176,59,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,60,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,61,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,62,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,63,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,64,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,65,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,66,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,67,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,68,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,69,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,70,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,71,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,72,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,73,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,74,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,75,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,76,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,77,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,78,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[176,79,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,80,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,81,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,82,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,83,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,84,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,85,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,86,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,87,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,88,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,89,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,90,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,91,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,92,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,93,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,94,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,95,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,96,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,97,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,98,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[176,99,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,100,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,101,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,102,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,103,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,104,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,105,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,106,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,107,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,108,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,109,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,110,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,111,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,112,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,113,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,114,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,115,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,116,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,117,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,118,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[176,119,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,120,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,121,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,122,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,123,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,124,3,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,125,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,126,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,127,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,128,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,129,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,130,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,131,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,132,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,133,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,134,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,135,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,136,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,137,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,138,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[176,139,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,140,3.16,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,141,3.17,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,142,3.18,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,143,3.19,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,144,3.2,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,145,3.21,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,146,3.22,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,147,3.23,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,148,3.24,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,149,3.25,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,150,3.26,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,151,3.27,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,152,3.28,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,153,3.29,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,154,3.3,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,155,3.31,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,156,3.32,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,157,3.33,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,158,3.34,6]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[176,159,3.35,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,160,3.36,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,161,3.37,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,162,3.38,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,163,3.39,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,164,3.4,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,165,3.41,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,166,3.42,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,167,3.43,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,168,3.44,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,169,3.45,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,170,3.46,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,171,3.47,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,172,3.48,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,173,3.49,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,174,3.5,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,175,3.51,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,176,3.52,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,177,3.53,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,178,3.54,6]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[176,179,3.55,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,180,3.56,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,181,3.57,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,182,3.58,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,183,3.59,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,184,3.6,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,185,3.61,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,186,3.62,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,187,3.63,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,188,3.64,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,189,3.65,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,190,3.66,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,191,3.67,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,192,3.68,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,193,3.69,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,194,3.7,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,195,3.71,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,196,3.72,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,197,3.73,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,198,3.74,6]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[176,199,3.75,6]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,0,1.77,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,1,1.78,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,2,1.79,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,3,1.8,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,4,1.81,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,5,1.82,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,6,1.83,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,7,1.84,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,8,1.85,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,9,1.86,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,10,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,11,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,12,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,13,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,14,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,15,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,16,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,17,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,18,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[177,19,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,20,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,21,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,22,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,23,2,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,24,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,25,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,26,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,27,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,28,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,29,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,30,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,31,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,32,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,33,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,34,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,35,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,36,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,37,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,38,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[177,39,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,40,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,41,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,42,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,43,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,44,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,45,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,46,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,47,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,48,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,49,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,50,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,51,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,52,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,53,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,54,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,55,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,56,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,57,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,58,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[177,59,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,60,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,61,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,62,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,63,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,64,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,65,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,66,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,67,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,68,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,69,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,70,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,71,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,72,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,73,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,74,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,75,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,76,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,77,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,78,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[177,79,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,80,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,81,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,82,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,83,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,84,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,85,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,86,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,87,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,88,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,89,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,90,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,91,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,92,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,93,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,94,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,95,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,96,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,97,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,98,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[177,99,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,100,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,101,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,102,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,103,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,104,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,105,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,106,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,107,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,108,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,109,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,110,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,111,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,112,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,113,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,114,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,115,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,116,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,117,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,118,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[177,119,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,120,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,121,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,122,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,123,3,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,124,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,125,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,126,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,127,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,128,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,129,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,130,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,131,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,132,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,133,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,134,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,135,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,136,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,137,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,138,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[177,139,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,140,3.17,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,141,3.18,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,142,3.19,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,143,3.2,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,144,3.21,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,145,3.22,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,146,3.23,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,147,3.24,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,148,3.25,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,149,3.26,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,150,3.27,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,151,3.28,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,152,3.29,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,153,3.3,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,154,3.31,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,155,3.32,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,156,3.33,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,157,3.34,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,158,3.35,7]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[177,159,3.36,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,160,3.37,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,161,3.38,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,162,3.39,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,163,3.4,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,164,3.41,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,165,3.42,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,166,3.43,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,167,3.44,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,168,3.45,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,169,3.46,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,170,3.47,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,171,3.48,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,172,3.49,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,173,3.5,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,174,3.51,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,175,3.52,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,176,3.53,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,177,3.54,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,178,3.55,7]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[177,179,3.56,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,180,3.57,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,181,3.58,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,182,3.59,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,183,3.6,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,184,3.61,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,185,3.62,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,186,3.63,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,187,3.64,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,188,3.65,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,189,3.66,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,190,3.67,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,191,3.68,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,192,3.69,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,193,3.7,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,194,3.71,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,195,3.72,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,196,3.73,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,197,3.74,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,198,3.75,7]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[177,199,3.76,7]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,0,1.78,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,1,1.79,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,2,1.8,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,3,1.81,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,4,1.82,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,5,1.83,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,6,1.84,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,7,1.85,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,8,1.86,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,9,1.87,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,10,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,11,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,12,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,13,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,14,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,15,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,16,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,17,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,18,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[178,19,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,20,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,21,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,22,2,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,23,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,24,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,25,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,26,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,27,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,28,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,29,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,30,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,31,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,32,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,33,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,34,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,35,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,36,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,37,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,38,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[178,39,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,40,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,41,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,42,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,43,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,44,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,45,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,46,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,47,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,48,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,49,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,50,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,51,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,52,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,53,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,54,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,55,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,56,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,57,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,58,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[178,59,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,60,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,61,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,62,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,63,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,64,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,65,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,66,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,67,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,68,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,69,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,70,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,71,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,72,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,73,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,74,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,75,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,76,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,77,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,78,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[178,79,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,80,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,81,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,82,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,83,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,84,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,85,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,86,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,87,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,88,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,89,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,90,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,91,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,92,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,93,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,94,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,95,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,96,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,97,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,98,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[178,99,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,100,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,101,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,102,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,103,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,104,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,105,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,106,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,107,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,108,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,109,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,110,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,111,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,112,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,113,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,114,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,115,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,116,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,117,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,118,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[178,119,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,120,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,121,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,122,3,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,123,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,124,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,125,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,126,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,127,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,128,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,129,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,130,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,131,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,132,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,133,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,134,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,135,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,136,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,137,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,138,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[178,139,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,140,3.18,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,141,3.19,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,142,3.2,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,143,3.21,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,144,3.22,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,145,3.23,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,146,3.24,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,147,3.25,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,148,3.26,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,149,3.27,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,150,3.28,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,151,3.29,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,152,3.3,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,153,3.31,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,154,3.32,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,155,3.33,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,156,3.34,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,157,3.35,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,158,3.36,8]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[178,159,3.37,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,160,3.38,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,161,3.39,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,162,3.4,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,163,3.41,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,164,3.42,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,165,3.43,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,166,3.44,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,167,3.45,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,168,3.46,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,169,3.47,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,170,3.48,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,171,3.49,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,172,3.5,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,173,3.51,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,174,3.52,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,175,3.53,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,176,3.54,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,177,3.55,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,178,3.56,8]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[178,179,3.57,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,180,3.58,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,181,3.59,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,182,3.6,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,183,3.61,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,184,3.62,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,185,3.63,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,186,3.64,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,187,3.65,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,188,3.66,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,189,3.67,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,190,3.68,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,191,3.69,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,192,3.7,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,193,3.71,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,194,3.72,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,195,3.73,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,196,3.74,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,197,3.75,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,198,3.76,8]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[178,199,3.77,8]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,0,1.79,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,1,1.8,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,2,1.81,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,3,1.82,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,4,1.83,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,5,1.84,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,6,1.85,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,7,1.86,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,8,1.87,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,9,1.88,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,10,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,11,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,12,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,13,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,14,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,15,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,16,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,17,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,18,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 160, pc_makepoint(1, ARRAY[179,19,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,20,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,21,2,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,22,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,23,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,24,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,25,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,26,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,27,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,28,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,29,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,30,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,31,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,32,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,33,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,34,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,35,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,36,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,37,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,38,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 161, pc_makepoint(1, ARRAY[179,39,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,40,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,41,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,42,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,43,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,44,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,45,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,46,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,47,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,48,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,49,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,50,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,51,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,52,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,53,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,54,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,55,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,56,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,57,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,58,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 162, pc_makepoint(1, ARRAY[179,59,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,60,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,61,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,62,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,63,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,64,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,65,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,66,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,67,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,68,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,69,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,70,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,71,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,72,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,73,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,74,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,75,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,76,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,77,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,78,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 163, pc_makepoint(1, ARRAY[179,79,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,80,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,81,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,82,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,83,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,84,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,85,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,86,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,87,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,88,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,89,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,90,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,91,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,92,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,93,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,94,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,95,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,96,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,97,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,98,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 164, pc_makepoint(1, ARRAY[179,99,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,100,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,101,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,102,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,103,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,104,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,105,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,106,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,107,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,108,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,109,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,110,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,111,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,112,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,113,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,114,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,115,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,116,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,117,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,118,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 165, pc_makepoint(1, ARRAY[179,119,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,120,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,121,3,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,122,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,123,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,124,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,125,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,126,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,127,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,128,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,129,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,130,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,131,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,132,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,133,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,134,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,135,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,136,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,137,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,138,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 166, pc_makepoint(1, ARRAY[179,139,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,140,3.19,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,141,3.2,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,142,3.21,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,143,3.22,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,144,3.23,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,145,3.24,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,146,3.25,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,147,3.26,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,148,3.27,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,149,3.28,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,150,3.29,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,151,3.3,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,152,3.31,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,153,3.32,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,154,3.33,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,155,3.34,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,156,3.35,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,157,3.36,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,158,3.37,9]) );
insert into pts_collection ( gid, pt ) values ( 167, pc_makepoint(1, ARRAY[179,159,3.38,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,160,3.39,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,161,3.4,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,162,3.41,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,163,3.42,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,164,3.43,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,165,3.44,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,166,3.45,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,167,3.46,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,168,3.47,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,169,3.48,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,170,3.49,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,171,3.5,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,172,3.51,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,173,3.52,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,174,3.53,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,175,3.54,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,176,3.55,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,177,3.56,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,178,3.57,9]) );
insert into pts_collection ( gid, pt ) values ( 168, pc_makepoint(1, ARRAY[179,179,3.58,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,180,3.59,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,181,3.6,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,182,3.61,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,183,3.62,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,184,3.63,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,185,3.64,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,186,3.65,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,187,3.66,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,188,3.67,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,189,3.68,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,190,3.69,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,191,3.7,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,192,3.71,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,193,3.72,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,194,3.73,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,195,3.74,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,196,3.75,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,197,3.76,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,198,3.77,9]) );
insert into pts_collection ( gid, pt ) values ( 169, pc_makepoint(1, ARRAY[179,199,3.78,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,0,1.8,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,1,1.81,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,2,1.82,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,3,1.83,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,4,1.84,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,5,1.85,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,6,1.86,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,7,1.87,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,8,1.88,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,9,1.89,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,10,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,11,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,12,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,13,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,14,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,15,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,16,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,17,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,18,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[180,19,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,20,2,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,21,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,22,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,23,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,24,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,25,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,26,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,27,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,28,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,29,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,30,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,31,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,32,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,33,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,34,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,35,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,36,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,37,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,38,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[180,39,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,40,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,41,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,42,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,43,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,44,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,45,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,46,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,47,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,48,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,49,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,50,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,51,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,52,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,53,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,54,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,55,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,56,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,57,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,58,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[180,59,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,60,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,61,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,62,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,63,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,64,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,65,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,66,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,67,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,68,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,69,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,70,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,71,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,72,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,73,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,74,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,75,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,76,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,77,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,78,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[180,79,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,80,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,81,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,82,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,83,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,84,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,85,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,86,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,87,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,88,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,89,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,90,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,91,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,92,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,93,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,94,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,95,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,96,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,97,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,98,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[180,99,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,100,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,101,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,102,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,103,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,104,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,105,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,106,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,107,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,108,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,109,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,110,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,111,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,112,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,113,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,114,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,115,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,116,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,117,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,118,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[180,119,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,120,3,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,121,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,122,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,123,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,124,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,125,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,126,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,127,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,128,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,129,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,130,3.1,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,131,3.11,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,132,3.12,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,133,3.13,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,134,3.14,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,135,3.15,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,136,3.16,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,137,3.17,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,138,3.18,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[180,139,3.19,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,140,3.2,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,141,3.21,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,142,3.22,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,143,3.23,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,144,3.24,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,145,3.25,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,146,3.26,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,147,3.27,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,148,3.28,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,149,3.29,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,150,3.3,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,151,3.31,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,152,3.32,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,153,3.33,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,154,3.34,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,155,3.35,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,156,3.36,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,157,3.37,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,158,3.38,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[180,159,3.39,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,160,3.4,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,161,3.41,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,162,3.42,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,163,3.43,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,164,3.44,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,165,3.45,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,166,3.46,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,167,3.47,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,168,3.48,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,169,3.49,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,170,3.5,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,171,3.51,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,172,3.52,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,173,3.53,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,174,3.54,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,175,3.55,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,176,3.56,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,177,3.57,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,178,3.58,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[180,179,3.59,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,180,3.6,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,181,3.61,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,182,3.62,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,183,3.63,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,184,3.64,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,185,3.65,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,186,3.66,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,187,3.67,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,188,3.68,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,189,3.69,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,190,3.7,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,191,3.71,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,192,3.72,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,193,3.73,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,194,3.74,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,195,3.75,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,196,3.76,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,197,3.77,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,198,3.78,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[180,199,3.79,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,0,1.81,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,1,1.82,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,2,1.83,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,3,1.84,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,4,1.85,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,5,1.86,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,6,1.87,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,7,1.88,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,8,1.89,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,9,1.9,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,10,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,11,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,12,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,13,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,14,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,15,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,16,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,17,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,18,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[181,19,2,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,20,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,21,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,22,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,23,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,24,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,25,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,26,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,27,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,28,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,29,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,30,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,31,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,32,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,33,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,34,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,35,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,36,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,37,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,38,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[181,39,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,40,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,41,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,42,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,43,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,44,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,45,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,46,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,47,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,48,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,49,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,50,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,51,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,52,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,53,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,54,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,55,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,56,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,57,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,58,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[181,59,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,60,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,61,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,62,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,63,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,64,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,65,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,66,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,67,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,68,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,69,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,70,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,71,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,72,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,73,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,74,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,75,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,76,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,77,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,78,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[181,79,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,80,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,81,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,82,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,83,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,84,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,85,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,86,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,87,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,88,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,89,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,90,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,91,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,92,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,93,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,94,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,95,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,96,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,97,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,98,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[181,99,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,100,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,101,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,102,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,103,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,104,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,105,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,106,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,107,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,108,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,109,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,110,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,111,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,112,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,113,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,114,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,115,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,116,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,117,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,118,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[181,119,3,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,120,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,121,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,122,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,123,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,124,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,125,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,126,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,127,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,128,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,129,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,130,3.11,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,131,3.12,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,132,3.13,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,133,3.14,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,134,3.15,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,135,3.16,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,136,3.17,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,137,3.18,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,138,3.19,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[181,139,3.2,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,140,3.21,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,141,3.22,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,142,3.23,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,143,3.24,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,144,3.25,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,145,3.26,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,146,3.27,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,147,3.28,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,148,3.29,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,149,3.3,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,150,3.31,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,151,3.32,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,152,3.33,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,153,3.34,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,154,3.35,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,155,3.36,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,156,3.37,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,157,3.38,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,158,3.39,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[181,159,3.4,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,160,3.41,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,161,3.42,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,162,3.43,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,163,3.44,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,164,3.45,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,165,3.46,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,166,3.47,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,167,3.48,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,168,3.49,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,169,3.5,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,170,3.51,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,171,3.52,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,172,3.53,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,173,3.54,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,174,3.55,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,175,3.56,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,176,3.57,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,177,3.58,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,178,3.59,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[181,179,3.6,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,180,3.61,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,181,3.62,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,182,3.63,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,183,3.64,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,184,3.65,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,185,3.66,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,186,3.67,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,187,3.68,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,188,3.69,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,189,3.7,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,190,3.71,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,191,3.72,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,192,3.73,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,193,3.74,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,194,3.75,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,195,3.76,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,196,3.77,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,197,3.78,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,198,3.79,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[181,199,3.8,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,0,1.82,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,1,1.83,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,2,1.84,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,3,1.85,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,4,1.86,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,5,1.87,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,6,1.88,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,7,1.89,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,8,1.9,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,9,1.91,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,10,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,11,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,12,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,13,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,14,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,15,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,16,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,17,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,18,2,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[182,19,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,20,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,21,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,22,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,23,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,24,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,25,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,26,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,27,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,28,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,29,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,30,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,31,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,32,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,33,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,34,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,35,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,36,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,37,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,38,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[182,39,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,40,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,41,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,42,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,43,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,44,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,45,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,46,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,47,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,48,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,49,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,50,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,51,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,52,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,53,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,54,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,55,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,56,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,57,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,58,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[182,59,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,60,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,61,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,62,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,63,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,64,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,65,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,66,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,67,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,68,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,69,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,70,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,71,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,72,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,73,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,74,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,75,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,76,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,77,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,78,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[182,79,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,80,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,81,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,82,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,83,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,84,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,85,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,86,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,87,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,88,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,89,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,90,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,91,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,92,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,93,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,94,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,95,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,96,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,97,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,98,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[182,99,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,100,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,101,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,102,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,103,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,104,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,105,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,106,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,107,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,108,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,109,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,110,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,111,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,112,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,113,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,114,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,115,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,116,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,117,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,118,3,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[182,119,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,120,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,121,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,122,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,123,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,124,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,125,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,126,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,127,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,128,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,129,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,130,3.12,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,131,3.13,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,132,3.14,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,133,3.15,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,134,3.16,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,135,3.17,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,136,3.18,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,137,3.19,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,138,3.2,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[182,139,3.21,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,140,3.22,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,141,3.23,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,142,3.24,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,143,3.25,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,144,3.26,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,145,3.27,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,146,3.28,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,147,3.29,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,148,3.3,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,149,3.31,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,150,3.32,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,151,3.33,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,152,3.34,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,153,3.35,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,154,3.36,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,155,3.37,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,156,3.38,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,157,3.39,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,158,3.4,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[182,159,3.41,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,160,3.42,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,161,3.43,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,162,3.44,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,163,3.45,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,164,3.46,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,165,3.47,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,166,3.48,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,167,3.49,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,168,3.5,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,169,3.51,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,170,3.52,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,171,3.53,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,172,3.54,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,173,3.55,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,174,3.56,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,175,3.57,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,176,3.58,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,177,3.59,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,178,3.6,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[182,179,3.61,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,180,3.62,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,181,3.63,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,182,3.64,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,183,3.65,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,184,3.66,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,185,3.67,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,186,3.68,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,187,3.69,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,188,3.7,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,189,3.71,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,190,3.72,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,191,3.73,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,192,3.74,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,193,3.75,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,194,3.76,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,195,3.77,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,196,3.78,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,197,3.79,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,198,3.8,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[182,199,3.81,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,0,1.83,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,1,1.84,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,2,1.85,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,3,1.86,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,4,1.87,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,5,1.88,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,6,1.89,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,7,1.9,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,8,1.91,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,9,1.92,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,10,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,11,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,12,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,13,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,14,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,15,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,16,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,17,2,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,18,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[183,19,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,20,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,21,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,22,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,23,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,24,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,25,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,26,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,27,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,28,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,29,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,30,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,31,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,32,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,33,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,34,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,35,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,36,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,37,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,38,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[183,39,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,40,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,41,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,42,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,43,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,44,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,45,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,46,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,47,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,48,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,49,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,50,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,51,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,52,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,53,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,54,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,55,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,56,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,57,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,58,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[183,59,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,60,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,61,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,62,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,63,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,64,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,65,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,66,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,67,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,68,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,69,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,70,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,71,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,72,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,73,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,74,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,75,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,76,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,77,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,78,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[183,79,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,80,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,81,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,82,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,83,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,84,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,85,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,86,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,87,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,88,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,89,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,90,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,91,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,92,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,93,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,94,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,95,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,96,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,97,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,98,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[183,99,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,100,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,101,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,102,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,103,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,104,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,105,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,106,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,107,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,108,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,109,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,110,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,111,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,112,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,113,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,114,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,115,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,116,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,117,3,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,118,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[183,119,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,120,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,121,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,122,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,123,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,124,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,125,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,126,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,127,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,128,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,129,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,130,3.13,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,131,3.14,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,132,3.15,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,133,3.16,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,134,3.17,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,135,3.18,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,136,3.19,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,137,3.2,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,138,3.21,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[183,139,3.22,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,140,3.23,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,141,3.24,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,142,3.25,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,143,3.26,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,144,3.27,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,145,3.28,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,146,3.29,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,147,3.3,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,148,3.31,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,149,3.32,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,150,3.33,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,151,3.34,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,152,3.35,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,153,3.36,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,154,3.37,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,155,3.38,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,156,3.39,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,157,3.4,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,158,3.41,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[183,159,3.42,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,160,3.43,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,161,3.44,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,162,3.45,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,163,3.46,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,164,3.47,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,165,3.48,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,166,3.49,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,167,3.5,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,168,3.51,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,169,3.52,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,170,3.53,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,171,3.54,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,172,3.55,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,173,3.56,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,174,3.57,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,175,3.58,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,176,3.59,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,177,3.6,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,178,3.61,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[183,179,3.62,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,180,3.63,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,181,3.64,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,182,3.65,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,183,3.66,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,184,3.67,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,185,3.68,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,186,3.69,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,187,3.7,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,188,3.71,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,189,3.72,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,190,3.73,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,191,3.74,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,192,3.75,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,193,3.76,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,194,3.77,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,195,3.78,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,196,3.79,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,197,3.8,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,198,3.81,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[183,199,3.82,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,0,1.84,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,1,1.85,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,2,1.86,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,3,1.87,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,4,1.88,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,5,1.89,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,6,1.9,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,7,1.91,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,8,1.92,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,9,1.93,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,10,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,11,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,12,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,13,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,14,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,15,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,16,2,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,17,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,18,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[184,19,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,20,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,21,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,22,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,23,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,24,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,25,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,26,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,27,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,28,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,29,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,30,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,31,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,32,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,33,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,34,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,35,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,36,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,37,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,38,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[184,39,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,40,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,41,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,42,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,43,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,44,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,45,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,46,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,47,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,48,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,49,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,50,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,51,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,52,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,53,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,54,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,55,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,56,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,57,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,58,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[184,59,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,60,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,61,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,62,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,63,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,64,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,65,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,66,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,67,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,68,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,69,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,70,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,71,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,72,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,73,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,74,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,75,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,76,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,77,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,78,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[184,79,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,80,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,81,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,82,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,83,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,84,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,85,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,86,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,87,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,88,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,89,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,90,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,91,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,92,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,93,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,94,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,95,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,96,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,97,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,98,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[184,99,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,100,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,101,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,102,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,103,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,104,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,105,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,106,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,107,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,108,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,109,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,110,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,111,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,112,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,113,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,114,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,115,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,116,3,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,117,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,118,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[184,119,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,120,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,121,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,122,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,123,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,124,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,125,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,126,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,127,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,128,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,129,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,130,3.14,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,131,3.15,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,132,3.16,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,133,3.17,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,134,3.18,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,135,3.19,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,136,3.2,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,137,3.21,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,138,3.22,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[184,139,3.23,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,140,3.24,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,141,3.25,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,142,3.26,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,143,3.27,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,144,3.28,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,145,3.29,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,146,3.3,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,147,3.31,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,148,3.32,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,149,3.33,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,150,3.34,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,151,3.35,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,152,3.36,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,153,3.37,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,154,3.38,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,155,3.39,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,156,3.4,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,157,3.41,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,158,3.42,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[184,159,3.43,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,160,3.44,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,161,3.45,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,162,3.46,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,163,3.47,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,164,3.48,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,165,3.49,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,166,3.5,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,167,3.51,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,168,3.52,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,169,3.53,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,170,3.54,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,171,3.55,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,172,3.56,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,173,3.57,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,174,3.58,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,175,3.59,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,176,3.6,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,177,3.61,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,178,3.62,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[184,179,3.63,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,180,3.64,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,181,3.65,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,182,3.66,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,183,3.67,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,184,3.68,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,185,3.69,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,186,3.7,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,187,3.71,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,188,3.72,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,189,3.73,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,190,3.74,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,191,3.75,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,192,3.76,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,193,3.77,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,194,3.78,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,195,3.79,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,196,3.8,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,197,3.81,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,198,3.82,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[184,199,3.83,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,0,1.85,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,1,1.86,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,2,1.87,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,3,1.88,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,4,1.89,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,5,1.9,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,6,1.91,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,7,1.92,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,8,1.93,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,9,1.94,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,10,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,11,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,12,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,13,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,14,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,15,2,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,16,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,17,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,18,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[185,19,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,20,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,21,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,22,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,23,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,24,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,25,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,26,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,27,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,28,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,29,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,30,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,31,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,32,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,33,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,34,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,35,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,36,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,37,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,38,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[185,39,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,40,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,41,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,42,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,43,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,44,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,45,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,46,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,47,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,48,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,49,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,50,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,51,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,52,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,53,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,54,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,55,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,56,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,57,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,58,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[185,59,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,60,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,61,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,62,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,63,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,64,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,65,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,66,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,67,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,68,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,69,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,70,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,71,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,72,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,73,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,74,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,75,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,76,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,77,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,78,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[185,79,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,80,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,81,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,82,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,83,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,84,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,85,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,86,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,87,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,88,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,89,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,90,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,91,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,92,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,93,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,94,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,95,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,96,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,97,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,98,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[185,99,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,100,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,101,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,102,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,103,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,104,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,105,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,106,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,107,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,108,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,109,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,110,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,111,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,112,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,113,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,114,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,115,3,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,116,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,117,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,118,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[185,119,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,120,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,121,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,122,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,123,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,124,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,125,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,126,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,127,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,128,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,129,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,130,3.15,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,131,3.16,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,132,3.17,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,133,3.18,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,134,3.19,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,135,3.2,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,136,3.21,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,137,3.22,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,138,3.23,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[185,139,3.24,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,140,3.25,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,141,3.26,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,142,3.27,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,143,3.28,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,144,3.29,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,145,3.3,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,146,3.31,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,147,3.32,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,148,3.33,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,149,3.34,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,150,3.35,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,151,3.36,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,152,3.37,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,153,3.38,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,154,3.39,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,155,3.4,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,156,3.41,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,157,3.42,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,158,3.43,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[185,159,3.44,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,160,3.45,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,161,3.46,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,162,3.47,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,163,3.48,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,164,3.49,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,165,3.5,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,166,3.51,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,167,3.52,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,168,3.53,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,169,3.54,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,170,3.55,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,171,3.56,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,172,3.57,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,173,3.58,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,174,3.59,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,175,3.6,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,176,3.61,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,177,3.62,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,178,3.63,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[185,179,3.64,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,180,3.65,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,181,3.66,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,182,3.67,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,183,3.68,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,184,3.69,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,185,3.7,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,186,3.71,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,187,3.72,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,188,3.73,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,189,3.74,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,190,3.75,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,191,3.76,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,192,3.77,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,193,3.78,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,194,3.79,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,195,3.8,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,196,3.81,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,197,3.82,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,198,3.83,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[185,199,3.84,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,0,1.86,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,1,1.87,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,2,1.88,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,3,1.89,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,4,1.9,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,5,1.91,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,6,1.92,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,7,1.93,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,8,1.94,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,9,1.95,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,10,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,11,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,12,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,13,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,14,2,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,15,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,16,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,17,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,18,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[186,19,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,20,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,21,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,22,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,23,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,24,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,25,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,26,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,27,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,28,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,29,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,30,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,31,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,32,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,33,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,34,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,35,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,36,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,37,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,38,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[186,39,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,40,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,41,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,42,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,43,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,44,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,45,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,46,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,47,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,48,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,49,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,50,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,51,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,52,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,53,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,54,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,55,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,56,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,57,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,58,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[186,59,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,60,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,61,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,62,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,63,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,64,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,65,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,66,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,67,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,68,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,69,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,70,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,71,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,72,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,73,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,74,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,75,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,76,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,77,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,78,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[186,79,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,80,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,81,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,82,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,83,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,84,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,85,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,86,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,87,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,88,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,89,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,90,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,91,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,92,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,93,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,94,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,95,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,96,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,97,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,98,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[186,99,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,100,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,101,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,102,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,103,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,104,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,105,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,106,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,107,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,108,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,109,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,110,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,111,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,112,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,113,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,114,3,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,115,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,116,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,117,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,118,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[186,119,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,120,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,121,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,122,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,123,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,124,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,125,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,126,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,127,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,128,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,129,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,130,3.16,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,131,3.17,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,132,3.18,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,133,3.19,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,134,3.2,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,135,3.21,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,136,3.22,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,137,3.23,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,138,3.24,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[186,139,3.25,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,140,3.26,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,141,3.27,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,142,3.28,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,143,3.29,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,144,3.3,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,145,3.31,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,146,3.32,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,147,3.33,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,148,3.34,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,149,3.35,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,150,3.36,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,151,3.37,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,152,3.38,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,153,3.39,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,154,3.4,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,155,3.41,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,156,3.42,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,157,3.43,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,158,3.44,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[186,159,3.45,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,160,3.46,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,161,3.47,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,162,3.48,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,163,3.49,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,164,3.5,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,165,3.51,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,166,3.52,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,167,3.53,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,168,3.54,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,169,3.55,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,170,3.56,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,171,3.57,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,172,3.58,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,173,3.59,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,174,3.6,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,175,3.61,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,176,3.62,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,177,3.63,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,178,3.64,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[186,179,3.65,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,180,3.66,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,181,3.67,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,182,3.68,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,183,3.69,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,184,3.7,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,185,3.71,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,186,3.72,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,187,3.73,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,188,3.74,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,189,3.75,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,190,3.76,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,191,3.77,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,192,3.78,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,193,3.79,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,194,3.8,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,195,3.81,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,196,3.82,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,197,3.83,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,198,3.84,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[186,199,3.85,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,0,1.87,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,1,1.88,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,2,1.89,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,3,1.9,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,4,1.91,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,5,1.92,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,6,1.93,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,7,1.94,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,8,1.95,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,9,1.96,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,10,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,11,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,12,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,13,2,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,14,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,15,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,16,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,17,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,18,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[187,19,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,20,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,21,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,22,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,23,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,24,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,25,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,26,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,27,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,28,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,29,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,30,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,31,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,32,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,33,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,34,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,35,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,36,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,37,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,38,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[187,39,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,40,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,41,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,42,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,43,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,44,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,45,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,46,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,47,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,48,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,49,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,50,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,51,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,52,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,53,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,54,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,55,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,56,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,57,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,58,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[187,59,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,60,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,61,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,62,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,63,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,64,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,65,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,66,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,67,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,68,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,69,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,70,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,71,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,72,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,73,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,74,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,75,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,76,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,77,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,78,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[187,79,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,80,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,81,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,82,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,83,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,84,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,85,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,86,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,87,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,88,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,89,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,90,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,91,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,92,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,93,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,94,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,95,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,96,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,97,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,98,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[187,99,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,100,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,101,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,102,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,103,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,104,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,105,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,106,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,107,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,108,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,109,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,110,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,111,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,112,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,113,3,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,114,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,115,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,116,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,117,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,118,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[187,119,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,120,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,121,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,122,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,123,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,124,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,125,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,126,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,127,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,128,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,129,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,130,3.17,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,131,3.18,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,132,3.19,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,133,3.2,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,134,3.21,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,135,3.22,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,136,3.23,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,137,3.24,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,138,3.25,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[187,139,3.26,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,140,3.27,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,141,3.28,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,142,3.29,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,143,3.3,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,144,3.31,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,145,3.32,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,146,3.33,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,147,3.34,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,148,3.35,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,149,3.36,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,150,3.37,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,151,3.38,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,152,3.39,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,153,3.4,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,154,3.41,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,155,3.42,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,156,3.43,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,157,3.44,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,158,3.45,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[187,159,3.46,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,160,3.47,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,161,3.48,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,162,3.49,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,163,3.5,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,164,3.51,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,165,3.52,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,166,3.53,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,167,3.54,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,168,3.55,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,169,3.56,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,170,3.57,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,171,3.58,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,172,3.59,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,173,3.6,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,174,3.61,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,175,3.62,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,176,3.63,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,177,3.64,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,178,3.65,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[187,179,3.66,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,180,3.67,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,181,3.68,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,182,3.69,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,183,3.7,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,184,3.71,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,185,3.72,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,186,3.73,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,187,3.74,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,188,3.75,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,189,3.76,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,190,3.77,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,191,3.78,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,192,3.79,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,193,3.8,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,194,3.81,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,195,3.82,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,196,3.83,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,197,3.84,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,198,3.85,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[187,199,3.86,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,0,1.88,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,1,1.89,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,2,1.9,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,3,1.91,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,4,1.92,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,5,1.93,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,6,1.94,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,7,1.95,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,8,1.96,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,9,1.97,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,10,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,11,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,12,2,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,13,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,14,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,15,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,16,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,17,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,18,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[188,19,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,20,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,21,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,22,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,23,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,24,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,25,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,26,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,27,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,28,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,29,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,30,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,31,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,32,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,33,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,34,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,35,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,36,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,37,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,38,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[188,39,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,40,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,41,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,42,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,43,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,44,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,45,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,46,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,47,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,48,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,49,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,50,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,51,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,52,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,53,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,54,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,55,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,56,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,57,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,58,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[188,59,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,60,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,61,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,62,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,63,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,64,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,65,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,66,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,67,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,68,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,69,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,70,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,71,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,72,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,73,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,74,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,75,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,76,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,77,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,78,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[188,79,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,80,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,81,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,82,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,83,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,84,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,85,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,86,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,87,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,88,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,89,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,90,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,91,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,92,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,93,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,94,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,95,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,96,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,97,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,98,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[188,99,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,100,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,101,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,102,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,103,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,104,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,105,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,106,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,107,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,108,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,109,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,110,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,111,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,112,3,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,113,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,114,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,115,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,116,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,117,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,118,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[188,119,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,120,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,121,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,122,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,123,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,124,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,125,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,126,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,127,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,128,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,129,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,130,3.18,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,131,3.19,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,132,3.2,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,133,3.21,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,134,3.22,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,135,3.23,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,136,3.24,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,137,3.25,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,138,3.26,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[188,139,3.27,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,140,3.28,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,141,3.29,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,142,3.3,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,143,3.31,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,144,3.32,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,145,3.33,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,146,3.34,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,147,3.35,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,148,3.36,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,149,3.37,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,150,3.38,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,151,3.39,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,152,3.4,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,153,3.41,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,154,3.42,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,155,3.43,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,156,3.44,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,157,3.45,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,158,3.46,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[188,159,3.47,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,160,3.48,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,161,3.49,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,162,3.5,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,163,3.51,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,164,3.52,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,165,3.53,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,166,3.54,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,167,3.55,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,168,3.56,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,169,3.57,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,170,3.58,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,171,3.59,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,172,3.6,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,173,3.61,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,174,3.62,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,175,3.63,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,176,3.64,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,177,3.65,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,178,3.66,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[188,179,3.67,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,180,3.68,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,181,3.69,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,182,3.7,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,183,3.71,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,184,3.72,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,185,3.73,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,186,3.74,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,187,3.75,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,188,3.76,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,189,3.77,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,190,3.78,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,191,3.79,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,192,3.8,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,193,3.81,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,194,3.82,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,195,3.83,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,196,3.84,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,197,3.85,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,198,3.86,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[188,199,3.87,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,0,1.89,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,1,1.9,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,2,1.91,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,3,1.92,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,4,1.93,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,5,1.94,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,6,1.95,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,7,1.96,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,8,1.97,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,9,1.98,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,10,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,11,2,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,12,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,13,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,14,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,15,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,16,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,17,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,18,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[189,19,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,20,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,21,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,22,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,23,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,24,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,25,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,26,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,27,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,28,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,29,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,30,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,31,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,32,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,33,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,34,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,35,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,36,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,37,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,38,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[189,39,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,40,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,41,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,42,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,43,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,44,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,45,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,46,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,47,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,48,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,49,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,50,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,51,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,52,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,53,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,54,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,55,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,56,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,57,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,58,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[189,59,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,60,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,61,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,62,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,63,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,64,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,65,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,66,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,67,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,68,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,69,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,70,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,71,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,72,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,73,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,74,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,75,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,76,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,77,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,78,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[189,79,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,80,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,81,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,82,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,83,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,84,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,85,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,86,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,87,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,88,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,89,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,90,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,91,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,92,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,93,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,94,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,95,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,96,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,97,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,98,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[189,99,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,100,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,101,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,102,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,103,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,104,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,105,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,106,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,107,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,108,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,109,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,110,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,111,3,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,112,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,113,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,114,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,115,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,116,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,117,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,118,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[189,119,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,120,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,121,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,122,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,123,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,124,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,125,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,126,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,127,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,128,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,129,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,130,3.19,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,131,3.2,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,132,3.21,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,133,3.22,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,134,3.23,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,135,3.24,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,136,3.25,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,137,3.26,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,138,3.27,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[189,139,3.28,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,140,3.29,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,141,3.3,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,142,3.31,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,143,3.32,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,144,3.33,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,145,3.34,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,146,3.35,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,147,3.36,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,148,3.37,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,149,3.38,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,150,3.39,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,151,3.4,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,152,3.41,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,153,3.42,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,154,3.43,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,155,3.44,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,156,3.45,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,157,3.46,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,158,3.47,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[189,159,3.48,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,160,3.49,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,161,3.5,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,162,3.51,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,163,3.52,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,164,3.53,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,165,3.54,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,166,3.55,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,167,3.56,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,168,3.57,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,169,3.58,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,170,3.59,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,171,3.6,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,172,3.61,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,173,3.62,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,174,3.63,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,175,3.64,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,176,3.65,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,177,3.66,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,178,3.67,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[189,179,3.68,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,180,3.69,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,181,3.7,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,182,3.71,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,183,3.72,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,184,3.73,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,185,3.74,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,186,3.75,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,187,3.76,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,188,3.77,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,189,3.78,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,190,3.79,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,191,3.8,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,192,3.81,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,193,3.82,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,194,3.83,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,195,3.84,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,196,3.85,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,197,3.86,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,198,3.87,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[189,199,3.88,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,0,1.9,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,1,1.91,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,2,1.92,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,3,1.93,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,4,1.94,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,5,1.95,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,6,1.96,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,7,1.97,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,8,1.98,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,9,1.99,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,10,2,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,11,2.01,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,12,2.02,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,13,2.03,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,14,2.04,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,15,2.05,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,16,2.06,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,17,2.07,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,18,2.08,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[190,19,2.09,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,20,2.1,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,21,2.11,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,22,2.12,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,23,2.13,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,24,2.14,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,25,2.15,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,26,2.16,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,27,2.17,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,28,2.18,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,29,2.19,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,30,2.2,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,31,2.21,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,32,2.22,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,33,2.23,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,34,2.24,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,35,2.25,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,36,2.26,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,37,2.27,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,38,2.28,0]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[190,39,2.29,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,40,2.3,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,41,2.31,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,42,2.32,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,43,2.33,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,44,2.34,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,45,2.35,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,46,2.36,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,47,2.37,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,48,2.38,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,49,2.39,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,50,2.4,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,51,2.41,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,52,2.42,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,53,2.43,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,54,2.44,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,55,2.45,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,56,2.46,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,57,2.47,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,58,2.48,0]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[190,59,2.49,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,60,2.5,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,61,2.51,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,62,2.52,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,63,2.53,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,64,2.54,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,65,2.55,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,66,2.56,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,67,2.57,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,68,2.58,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,69,2.59,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,70,2.6,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,71,2.61,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,72,2.62,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,73,2.63,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,74,2.64,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,75,2.65,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,76,2.66,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,77,2.67,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,78,2.68,0]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[190,79,2.69,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,80,2.7,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,81,2.71,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,82,2.72,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,83,2.73,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,84,2.74,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,85,2.75,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,86,2.76,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,87,2.77,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,88,2.78,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,89,2.79,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,90,2.8,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,91,2.81,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,92,2.82,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,93,2.83,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,94,2.84,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,95,2.85,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,96,2.86,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,97,2.87,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,98,2.88,0]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[190,99,2.89,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,100,2.9,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,101,2.91,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,102,2.92,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,103,2.93,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,104,2.94,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,105,2.95,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,106,2.96,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,107,2.97,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,108,2.98,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,109,2.99,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,110,3,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,111,3.01,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,112,3.02,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,113,3.03,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,114,3.04,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,115,3.05,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,116,3.06,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,117,3.07,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,118,3.08,0]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[190,119,3.09,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,120,3.1,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,121,3.11,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,122,3.12,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,123,3.13,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,124,3.14,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,125,3.15,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,126,3.16,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,127,3.17,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,128,3.18,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,129,3.19,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,130,3.2,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,131,3.21,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,132,3.22,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,133,3.23,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,134,3.24,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,135,3.25,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,136,3.26,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,137,3.27,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,138,3.28,0]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[190,139,3.29,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,140,3.3,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,141,3.31,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,142,3.32,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,143,3.33,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,144,3.34,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,145,3.35,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,146,3.36,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,147,3.37,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,148,3.38,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,149,3.39,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,150,3.4,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,151,3.41,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,152,3.42,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,153,3.43,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,154,3.44,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,155,3.45,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,156,3.46,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,157,3.47,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,158,3.48,0]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[190,159,3.49,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,160,3.5,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,161,3.51,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,162,3.52,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,163,3.53,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,164,3.54,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,165,3.55,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,166,3.56,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,167,3.57,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,168,3.58,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,169,3.59,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,170,3.6,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,171,3.61,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,172,3.62,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,173,3.63,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,174,3.64,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,175,3.65,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,176,3.66,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,177,3.67,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,178,3.68,0]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[190,179,3.69,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,180,3.7,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,181,3.71,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,182,3.72,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,183,3.73,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,184,3.74,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,185,3.75,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,186,3.76,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,187,3.77,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,188,3.78,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,189,3.79,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,190,3.8,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,191,3.81,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,192,3.82,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,193,3.83,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,194,3.84,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,195,3.85,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,196,3.86,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,197,3.87,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,198,3.88,0]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[190,199,3.89,0]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,0,1.91,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,1,1.92,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,2,1.93,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,3,1.94,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,4,1.95,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,5,1.96,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,6,1.97,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,7,1.98,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,8,1.99,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,9,2,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,10,2.01,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,11,2.02,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,12,2.03,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,13,2.04,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,14,2.05,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,15,2.06,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,16,2.07,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,17,2.08,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,18,2.09,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[191,19,2.1,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,20,2.11,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,21,2.12,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,22,2.13,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,23,2.14,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,24,2.15,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,25,2.16,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,26,2.17,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,27,2.18,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,28,2.19,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,29,2.2,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,30,2.21,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,31,2.22,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,32,2.23,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,33,2.24,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,34,2.25,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,35,2.26,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,36,2.27,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,37,2.28,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,38,2.29,1]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[191,39,2.3,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,40,2.31,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,41,2.32,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,42,2.33,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,43,2.34,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,44,2.35,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,45,2.36,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,46,2.37,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,47,2.38,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,48,2.39,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,49,2.4,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,50,2.41,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,51,2.42,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,52,2.43,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,53,2.44,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,54,2.45,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,55,2.46,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,56,2.47,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,57,2.48,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,58,2.49,1]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[191,59,2.5,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,60,2.51,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,61,2.52,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,62,2.53,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,63,2.54,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,64,2.55,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,65,2.56,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,66,2.57,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,67,2.58,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,68,2.59,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,69,2.6,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,70,2.61,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,71,2.62,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,72,2.63,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,73,2.64,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,74,2.65,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,75,2.66,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,76,2.67,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,77,2.68,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,78,2.69,1]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[191,79,2.7,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,80,2.71,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,81,2.72,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,82,2.73,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,83,2.74,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,84,2.75,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,85,2.76,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,86,2.77,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,87,2.78,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,88,2.79,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,89,2.8,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,90,2.81,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,91,2.82,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,92,2.83,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,93,2.84,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,94,2.85,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,95,2.86,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,96,2.87,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,97,2.88,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,98,2.89,1]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[191,99,2.9,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,100,2.91,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,101,2.92,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,102,2.93,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,103,2.94,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,104,2.95,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,105,2.96,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,106,2.97,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,107,2.98,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,108,2.99,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,109,3,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,110,3.01,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,111,3.02,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,112,3.03,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,113,3.04,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,114,3.05,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,115,3.06,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,116,3.07,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,117,3.08,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,118,3.09,1]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[191,119,3.1,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,120,3.11,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,121,3.12,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,122,3.13,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,123,3.14,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,124,3.15,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,125,3.16,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,126,3.17,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,127,3.18,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,128,3.19,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,129,3.2,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,130,3.21,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,131,3.22,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,132,3.23,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,133,3.24,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,134,3.25,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,135,3.26,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,136,3.27,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,137,3.28,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,138,3.29,1]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[191,139,3.3,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,140,3.31,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,141,3.32,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,142,3.33,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,143,3.34,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,144,3.35,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,145,3.36,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,146,3.37,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,147,3.38,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,148,3.39,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,149,3.4,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,150,3.41,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,151,3.42,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,152,3.43,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,153,3.44,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,154,3.45,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,155,3.46,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,156,3.47,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,157,3.48,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,158,3.49,1]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[191,159,3.5,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,160,3.51,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,161,3.52,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,162,3.53,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,163,3.54,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,164,3.55,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,165,3.56,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,166,3.57,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,167,3.58,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,168,3.59,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,169,3.6,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,170,3.61,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,171,3.62,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,172,3.63,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,173,3.64,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,174,3.65,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,175,3.66,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,176,3.67,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,177,3.68,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,178,3.69,1]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[191,179,3.7,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,180,3.71,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,181,3.72,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,182,3.73,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,183,3.74,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,184,3.75,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,185,3.76,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,186,3.77,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,187,3.78,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,188,3.79,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,189,3.8,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,190,3.81,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,191,3.82,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,192,3.83,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,193,3.84,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,194,3.85,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,195,3.86,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,196,3.87,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,197,3.88,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,198,3.89,1]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[191,199,3.9,1]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,0,1.92,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,1,1.93,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,2,1.94,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,3,1.95,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,4,1.96,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,5,1.97,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,6,1.98,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,7,1.99,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,8,2,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,9,2.01,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,10,2.02,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,11,2.03,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,12,2.04,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,13,2.05,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,14,2.06,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,15,2.07,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,16,2.08,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,17,2.09,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,18,2.1,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[192,19,2.11,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,20,2.12,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,21,2.13,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,22,2.14,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,23,2.15,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,24,2.16,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,25,2.17,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,26,2.18,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,27,2.19,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,28,2.2,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,29,2.21,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,30,2.22,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,31,2.23,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,32,2.24,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,33,2.25,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,34,2.26,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,35,2.27,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,36,2.28,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,37,2.29,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,38,2.3,2]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[192,39,2.31,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,40,2.32,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,41,2.33,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,42,2.34,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,43,2.35,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,44,2.36,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,45,2.37,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,46,2.38,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,47,2.39,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,48,2.4,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,49,2.41,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,50,2.42,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,51,2.43,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,52,2.44,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,53,2.45,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,54,2.46,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,55,2.47,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,56,2.48,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,57,2.49,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,58,2.5,2]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[192,59,2.51,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,60,2.52,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,61,2.53,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,62,2.54,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,63,2.55,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,64,2.56,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,65,2.57,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,66,2.58,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,67,2.59,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,68,2.6,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,69,2.61,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,70,2.62,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,71,2.63,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,72,2.64,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,73,2.65,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,74,2.66,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,75,2.67,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,76,2.68,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,77,2.69,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,78,2.7,2]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[192,79,2.71,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,80,2.72,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,81,2.73,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,82,2.74,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,83,2.75,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,84,2.76,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,85,2.77,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,86,2.78,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,87,2.79,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,88,2.8,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,89,2.81,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,90,2.82,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,91,2.83,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,92,2.84,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,93,2.85,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,94,2.86,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,95,2.87,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,96,2.88,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,97,2.89,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,98,2.9,2]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[192,99,2.91,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,100,2.92,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,101,2.93,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,102,2.94,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,103,2.95,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,104,2.96,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,105,2.97,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,106,2.98,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,107,2.99,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,108,3,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,109,3.01,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,110,3.02,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,111,3.03,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,112,3.04,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,113,3.05,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,114,3.06,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,115,3.07,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,116,3.08,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,117,3.09,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,118,3.1,2]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[192,119,3.11,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,120,3.12,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,121,3.13,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,122,3.14,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,123,3.15,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,124,3.16,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,125,3.17,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,126,3.18,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,127,3.19,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,128,3.2,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,129,3.21,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,130,3.22,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,131,3.23,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,132,3.24,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,133,3.25,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,134,3.26,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,135,3.27,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,136,3.28,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,137,3.29,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,138,3.3,2]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[192,139,3.31,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,140,3.32,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,141,3.33,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,142,3.34,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,143,3.35,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,144,3.36,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,145,3.37,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,146,3.38,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,147,3.39,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,148,3.4,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,149,3.41,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,150,3.42,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,151,3.43,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,152,3.44,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,153,3.45,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,154,3.46,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,155,3.47,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,156,3.48,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,157,3.49,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,158,3.5,2]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[192,159,3.51,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,160,3.52,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,161,3.53,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,162,3.54,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,163,3.55,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,164,3.56,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,165,3.57,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,166,3.58,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,167,3.59,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,168,3.6,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,169,3.61,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,170,3.62,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,171,3.63,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,172,3.64,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,173,3.65,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,174,3.66,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,175,3.67,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,176,3.68,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,177,3.69,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,178,3.7,2]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[192,179,3.71,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,180,3.72,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,181,3.73,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,182,3.74,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,183,3.75,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,184,3.76,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,185,3.77,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,186,3.78,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,187,3.79,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,188,3.8,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,189,3.81,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,190,3.82,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,191,3.83,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,192,3.84,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,193,3.85,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,194,3.86,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,195,3.87,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,196,3.88,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,197,3.89,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,198,3.9,2]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[192,199,3.91,2]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,0,1.93,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,1,1.94,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,2,1.95,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,3,1.96,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,4,1.97,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,5,1.98,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,6,1.99,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,7,2,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,8,2.01,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,9,2.02,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,10,2.03,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,11,2.04,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,12,2.05,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,13,2.06,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,14,2.07,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,15,2.08,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,16,2.09,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,17,2.1,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,18,2.11,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[193,19,2.12,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,20,2.13,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,21,2.14,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,22,2.15,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,23,2.16,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,24,2.17,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,25,2.18,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,26,2.19,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,27,2.2,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,28,2.21,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,29,2.22,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,30,2.23,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,31,2.24,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,32,2.25,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,33,2.26,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,34,2.27,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,35,2.28,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,36,2.29,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,37,2.3,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,38,2.31,3]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[193,39,2.32,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,40,2.33,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,41,2.34,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,42,2.35,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,43,2.36,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,44,2.37,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,45,2.38,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,46,2.39,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,47,2.4,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,48,2.41,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,49,2.42,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,50,2.43,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,51,2.44,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,52,2.45,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,53,2.46,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,54,2.47,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,55,2.48,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,56,2.49,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,57,2.5,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,58,2.51,3]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[193,59,2.52,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,60,2.53,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,61,2.54,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,62,2.55,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,63,2.56,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,64,2.57,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,65,2.58,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,66,2.59,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,67,2.6,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,68,2.61,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,69,2.62,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,70,2.63,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,71,2.64,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,72,2.65,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,73,2.66,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,74,2.67,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,75,2.68,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,76,2.69,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,77,2.7,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,78,2.71,3]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[193,79,2.72,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,80,2.73,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,81,2.74,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,82,2.75,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,83,2.76,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,84,2.77,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,85,2.78,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,86,2.79,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,87,2.8,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,88,2.81,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,89,2.82,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,90,2.83,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,91,2.84,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,92,2.85,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,93,2.86,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,94,2.87,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,95,2.88,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,96,2.89,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,97,2.9,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,98,2.91,3]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[193,99,2.92,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,100,2.93,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,101,2.94,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,102,2.95,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,103,2.96,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,104,2.97,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,105,2.98,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,106,2.99,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,107,3,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,108,3.01,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,109,3.02,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,110,3.03,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,111,3.04,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,112,3.05,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,113,3.06,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,114,3.07,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,115,3.08,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,116,3.09,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,117,3.1,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,118,3.11,3]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[193,119,3.12,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,120,3.13,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,121,3.14,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,122,3.15,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,123,3.16,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,124,3.17,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,125,3.18,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,126,3.19,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,127,3.2,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,128,3.21,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,129,3.22,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,130,3.23,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,131,3.24,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,132,3.25,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,133,3.26,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,134,3.27,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,135,3.28,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,136,3.29,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,137,3.3,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,138,3.31,3]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[193,139,3.32,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,140,3.33,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,141,3.34,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,142,3.35,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,143,3.36,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,144,3.37,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,145,3.38,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,146,3.39,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,147,3.4,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,148,3.41,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,149,3.42,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,150,3.43,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,151,3.44,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,152,3.45,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,153,3.46,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,154,3.47,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,155,3.48,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,156,3.49,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,157,3.5,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,158,3.51,3]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[193,159,3.52,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,160,3.53,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,161,3.54,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,162,3.55,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,163,3.56,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,164,3.57,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,165,3.58,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,166,3.59,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,167,3.6,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,168,3.61,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,169,3.62,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,170,3.63,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,171,3.64,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,172,3.65,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,173,3.66,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,174,3.67,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,175,3.68,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,176,3.69,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,177,3.7,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,178,3.71,3]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[193,179,3.72,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,180,3.73,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,181,3.74,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,182,3.75,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,183,3.76,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,184,3.77,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,185,3.78,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,186,3.79,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,187,3.8,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,188,3.81,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,189,3.82,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,190,3.83,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,191,3.84,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,192,3.85,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,193,3.86,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,194,3.87,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,195,3.88,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,196,3.89,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,197,3.9,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,198,3.91,3]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[193,199,3.92,3]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,0,1.94,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,1,1.95,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,2,1.96,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,3,1.97,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,4,1.98,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,5,1.99,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,6,2,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,7,2.01,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,8,2.02,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,9,2.03,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,10,2.04,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,11,2.05,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,12,2.06,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,13,2.07,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,14,2.08,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,15,2.09,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,16,2.1,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,17,2.11,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,18,2.12,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[194,19,2.13,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,20,2.14,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,21,2.15,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,22,2.16,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,23,2.17,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,24,2.18,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,25,2.19,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,26,2.2,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,27,2.21,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,28,2.22,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,29,2.23,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,30,2.24,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,31,2.25,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,32,2.26,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,33,2.27,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,34,2.28,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,35,2.29,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,36,2.3,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,37,2.31,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,38,2.32,4]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[194,39,2.33,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,40,2.34,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,41,2.35,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,42,2.36,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,43,2.37,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,44,2.38,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,45,2.39,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,46,2.4,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,47,2.41,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,48,2.42,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,49,2.43,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,50,2.44,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,51,2.45,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,52,2.46,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,53,2.47,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,54,2.48,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,55,2.49,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,56,2.5,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,57,2.51,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,58,2.52,4]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[194,59,2.53,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,60,2.54,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,61,2.55,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,62,2.56,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,63,2.57,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,64,2.58,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,65,2.59,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,66,2.6,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,67,2.61,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,68,2.62,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,69,2.63,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,70,2.64,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,71,2.65,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,72,2.66,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,73,2.67,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,74,2.68,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,75,2.69,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,76,2.7,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,77,2.71,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,78,2.72,4]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[194,79,2.73,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,80,2.74,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,81,2.75,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,82,2.76,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,83,2.77,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,84,2.78,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,85,2.79,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,86,2.8,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,87,2.81,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,88,2.82,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,89,2.83,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,90,2.84,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,91,2.85,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,92,2.86,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,93,2.87,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,94,2.88,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,95,2.89,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,96,2.9,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,97,2.91,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,98,2.92,4]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[194,99,2.93,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,100,2.94,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,101,2.95,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,102,2.96,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,103,2.97,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,104,2.98,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,105,2.99,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,106,3,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,107,3.01,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,108,3.02,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,109,3.03,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,110,3.04,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,111,3.05,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,112,3.06,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,113,3.07,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,114,3.08,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,115,3.09,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,116,3.1,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,117,3.11,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,118,3.12,4]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[194,119,3.13,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,120,3.14,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,121,3.15,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,122,3.16,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,123,3.17,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,124,3.18,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,125,3.19,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,126,3.2,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,127,3.21,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,128,3.22,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,129,3.23,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,130,3.24,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,131,3.25,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,132,3.26,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,133,3.27,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,134,3.28,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,135,3.29,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,136,3.3,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,137,3.31,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,138,3.32,4]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[194,139,3.33,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,140,3.34,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,141,3.35,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,142,3.36,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,143,3.37,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,144,3.38,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,145,3.39,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,146,3.4,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,147,3.41,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,148,3.42,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,149,3.43,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,150,3.44,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,151,3.45,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,152,3.46,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,153,3.47,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,154,3.48,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,155,3.49,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,156,3.5,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,157,3.51,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,158,3.52,4]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[194,159,3.53,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,160,3.54,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,161,3.55,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,162,3.56,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,163,3.57,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,164,3.58,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,165,3.59,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,166,3.6,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,167,3.61,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,168,3.62,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,169,3.63,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,170,3.64,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,171,3.65,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,172,3.66,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,173,3.67,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,174,3.68,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,175,3.69,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,176,3.7,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,177,3.71,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,178,3.72,4]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[194,179,3.73,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,180,3.74,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,181,3.75,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,182,3.76,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,183,3.77,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,184,3.78,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,185,3.79,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,186,3.8,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,187,3.81,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,188,3.82,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,189,3.83,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,190,3.84,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,191,3.85,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,192,3.86,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,193,3.87,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,194,3.88,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,195,3.89,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,196,3.9,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,197,3.91,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,198,3.92,4]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[194,199,3.93,4]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,0,1.95,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,1,1.96,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,2,1.97,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,3,1.98,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,4,1.99,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,5,2,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,6,2.01,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,7,2.02,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,8,2.03,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,9,2.04,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,10,2.05,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,11,2.06,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,12,2.07,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,13,2.08,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,14,2.09,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,15,2.1,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,16,2.11,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,17,2.12,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,18,2.13,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[195,19,2.14,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,20,2.15,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,21,2.16,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,22,2.17,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,23,2.18,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,24,2.19,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,25,2.2,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,26,2.21,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,27,2.22,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,28,2.23,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,29,2.24,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,30,2.25,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,31,2.26,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,32,2.27,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,33,2.28,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,34,2.29,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,35,2.3,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,36,2.31,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,37,2.32,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,38,2.33,5]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[195,39,2.34,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,40,2.35,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,41,2.36,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,42,2.37,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,43,2.38,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,44,2.39,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,45,2.4,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,46,2.41,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,47,2.42,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,48,2.43,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,49,2.44,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,50,2.45,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,51,2.46,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,52,2.47,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,53,2.48,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,54,2.49,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,55,2.5,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,56,2.51,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,57,2.52,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,58,2.53,5]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[195,59,2.54,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,60,2.55,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,61,2.56,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,62,2.57,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,63,2.58,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,64,2.59,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,65,2.6,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,66,2.61,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,67,2.62,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,68,2.63,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,69,2.64,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,70,2.65,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,71,2.66,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,72,2.67,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,73,2.68,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,74,2.69,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,75,2.7,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,76,2.71,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,77,2.72,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,78,2.73,5]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[195,79,2.74,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,80,2.75,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,81,2.76,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,82,2.77,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,83,2.78,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,84,2.79,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,85,2.8,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,86,2.81,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,87,2.82,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,88,2.83,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,89,2.84,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,90,2.85,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,91,2.86,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,92,2.87,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,93,2.88,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,94,2.89,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,95,2.9,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,96,2.91,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,97,2.92,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,98,2.93,5]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[195,99,2.94,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,100,2.95,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,101,2.96,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,102,2.97,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,103,2.98,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,104,2.99,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,105,3,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,106,3.01,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,107,3.02,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,108,3.03,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,109,3.04,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,110,3.05,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,111,3.06,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,112,3.07,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,113,3.08,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,114,3.09,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,115,3.1,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,116,3.11,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,117,3.12,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,118,3.13,5]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[195,119,3.14,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,120,3.15,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,121,3.16,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,122,3.17,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,123,3.18,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,124,3.19,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,125,3.2,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,126,3.21,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,127,3.22,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,128,3.23,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,129,3.24,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,130,3.25,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,131,3.26,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,132,3.27,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,133,3.28,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,134,3.29,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,135,3.3,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,136,3.31,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,137,3.32,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,138,3.33,5]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[195,139,3.34,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,140,3.35,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,141,3.36,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,142,3.37,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,143,3.38,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,144,3.39,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,145,3.4,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,146,3.41,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,147,3.42,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,148,3.43,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,149,3.44,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,150,3.45,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,151,3.46,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,152,3.47,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,153,3.48,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,154,3.49,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,155,3.5,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,156,3.51,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,157,3.52,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,158,3.53,5]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[195,159,3.54,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,160,3.55,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,161,3.56,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,162,3.57,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,163,3.58,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,164,3.59,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,165,3.6,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,166,3.61,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,167,3.62,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,168,3.63,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,169,3.64,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,170,3.65,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,171,3.66,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,172,3.67,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,173,3.68,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,174,3.69,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,175,3.7,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,176,3.71,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,177,3.72,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,178,3.73,5]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[195,179,3.74,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,180,3.75,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,181,3.76,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,182,3.77,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,183,3.78,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,184,3.79,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,185,3.8,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,186,3.81,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,187,3.82,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,188,3.83,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,189,3.84,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,190,3.85,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,191,3.86,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,192,3.87,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,193,3.88,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,194,3.89,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,195,3.9,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,196,3.91,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,197,3.92,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,198,3.93,5]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[195,199,3.94,5]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,0,1.96,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,1,1.97,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,2,1.98,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,3,1.99,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,4,2,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,5,2.01,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,6,2.02,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,7,2.03,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,8,2.04,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,9,2.05,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,10,2.06,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,11,2.07,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,12,2.08,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,13,2.09,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,14,2.1,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,15,2.11,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,16,2.12,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,17,2.13,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,18,2.14,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[196,19,2.15,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,20,2.16,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,21,2.17,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,22,2.18,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,23,2.19,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,24,2.2,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,25,2.21,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,26,2.22,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,27,2.23,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,28,2.24,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,29,2.25,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,30,2.26,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,31,2.27,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,32,2.28,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,33,2.29,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,34,2.3,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,35,2.31,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,36,2.32,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,37,2.33,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,38,2.34,6]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[196,39,2.35,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,40,2.36,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,41,2.37,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,42,2.38,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,43,2.39,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,44,2.4,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,45,2.41,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,46,2.42,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,47,2.43,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,48,2.44,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,49,2.45,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,50,2.46,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,51,2.47,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,52,2.48,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,53,2.49,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,54,2.5,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,55,2.51,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,56,2.52,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,57,2.53,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,58,2.54,6]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[196,59,2.55,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,60,2.56,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,61,2.57,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,62,2.58,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,63,2.59,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,64,2.6,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,65,2.61,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,66,2.62,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,67,2.63,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,68,2.64,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,69,2.65,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,70,2.66,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,71,2.67,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,72,2.68,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,73,2.69,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,74,2.7,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,75,2.71,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,76,2.72,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,77,2.73,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,78,2.74,6]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[196,79,2.75,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,80,2.76,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,81,2.77,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,82,2.78,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,83,2.79,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,84,2.8,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,85,2.81,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,86,2.82,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,87,2.83,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,88,2.84,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,89,2.85,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,90,2.86,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,91,2.87,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,92,2.88,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,93,2.89,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,94,2.9,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,95,2.91,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,96,2.92,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,97,2.93,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,98,2.94,6]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[196,99,2.95,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,100,2.96,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,101,2.97,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,102,2.98,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,103,2.99,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,104,3,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,105,3.01,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,106,3.02,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,107,3.03,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,108,3.04,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,109,3.05,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,110,3.06,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,111,3.07,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,112,3.08,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,113,3.09,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,114,3.1,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,115,3.11,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,116,3.12,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,117,3.13,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,118,3.14,6]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[196,119,3.15,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,120,3.16,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,121,3.17,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,122,3.18,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,123,3.19,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,124,3.2,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,125,3.21,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,126,3.22,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,127,3.23,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,128,3.24,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,129,3.25,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,130,3.26,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,131,3.27,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,132,3.28,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,133,3.29,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,134,3.3,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,135,3.31,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,136,3.32,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,137,3.33,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,138,3.34,6]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[196,139,3.35,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,140,3.36,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,141,3.37,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,142,3.38,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,143,3.39,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,144,3.4,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,145,3.41,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,146,3.42,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,147,3.43,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,148,3.44,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,149,3.45,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,150,3.46,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,151,3.47,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,152,3.48,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,153,3.49,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,154,3.5,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,155,3.51,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,156,3.52,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,157,3.53,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,158,3.54,6]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[196,159,3.55,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,160,3.56,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,161,3.57,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,162,3.58,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,163,3.59,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,164,3.6,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,165,3.61,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,166,3.62,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,167,3.63,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,168,3.64,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,169,3.65,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,170,3.66,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,171,3.67,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,172,3.68,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,173,3.69,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,174,3.7,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,175,3.71,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,176,3.72,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,177,3.73,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,178,3.74,6]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[196,179,3.75,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,180,3.76,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,181,3.77,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,182,3.78,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,183,3.79,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,184,3.8,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,185,3.81,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,186,3.82,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,187,3.83,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,188,3.84,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,189,3.85,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,190,3.86,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,191,3.87,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,192,3.88,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,193,3.89,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,194,3.9,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,195,3.91,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,196,3.92,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,197,3.93,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,198,3.94,6]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[196,199,3.95,6]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,0,1.97,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,1,1.98,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,2,1.99,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,3,2,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,4,2.01,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,5,2.02,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,6,2.03,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,7,2.04,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,8,2.05,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,9,2.06,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,10,2.07,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,11,2.08,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,12,2.09,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,13,2.1,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,14,2.11,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,15,2.12,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,16,2.13,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,17,2.14,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,18,2.15,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[197,19,2.16,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,20,2.17,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,21,2.18,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,22,2.19,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,23,2.2,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,24,2.21,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,25,2.22,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,26,2.23,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,27,2.24,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,28,2.25,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,29,2.26,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,30,2.27,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,31,2.28,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,32,2.29,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,33,2.3,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,34,2.31,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,35,2.32,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,36,2.33,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,37,2.34,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,38,2.35,7]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[197,39,2.36,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,40,2.37,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,41,2.38,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,42,2.39,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,43,2.4,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,44,2.41,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,45,2.42,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,46,2.43,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,47,2.44,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,48,2.45,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,49,2.46,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,50,2.47,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,51,2.48,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,52,2.49,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,53,2.5,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,54,2.51,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,55,2.52,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,56,2.53,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,57,2.54,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,58,2.55,7]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[197,59,2.56,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,60,2.57,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,61,2.58,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,62,2.59,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,63,2.6,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,64,2.61,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,65,2.62,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,66,2.63,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,67,2.64,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,68,2.65,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,69,2.66,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,70,2.67,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,71,2.68,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,72,2.69,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,73,2.7,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,74,2.71,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,75,2.72,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,76,2.73,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,77,2.74,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,78,2.75,7]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[197,79,2.76,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,80,2.77,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,81,2.78,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,82,2.79,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,83,2.8,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,84,2.81,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,85,2.82,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,86,2.83,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,87,2.84,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,88,2.85,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,89,2.86,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,90,2.87,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,91,2.88,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,92,2.89,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,93,2.9,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,94,2.91,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,95,2.92,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,96,2.93,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,97,2.94,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,98,2.95,7]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[197,99,2.96,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,100,2.97,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,101,2.98,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,102,2.99,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,103,3,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,104,3.01,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,105,3.02,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,106,3.03,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,107,3.04,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,108,3.05,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,109,3.06,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,110,3.07,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,111,3.08,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,112,3.09,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,113,3.1,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,114,3.11,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,115,3.12,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,116,3.13,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,117,3.14,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,118,3.15,7]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[197,119,3.16,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,120,3.17,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,121,3.18,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,122,3.19,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,123,3.2,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,124,3.21,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,125,3.22,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,126,3.23,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,127,3.24,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,128,3.25,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,129,3.26,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,130,3.27,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,131,3.28,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,132,3.29,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,133,3.3,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,134,3.31,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,135,3.32,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,136,3.33,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,137,3.34,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,138,3.35,7]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[197,139,3.36,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,140,3.37,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,141,3.38,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,142,3.39,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,143,3.4,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,144,3.41,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,145,3.42,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,146,3.43,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,147,3.44,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,148,3.45,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,149,3.46,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,150,3.47,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,151,3.48,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,152,3.49,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,153,3.5,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,154,3.51,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,155,3.52,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,156,3.53,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,157,3.54,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,158,3.55,7]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[197,159,3.56,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,160,3.57,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,161,3.58,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,162,3.59,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,163,3.6,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,164,3.61,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,165,3.62,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,166,3.63,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,167,3.64,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,168,3.65,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,169,3.66,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,170,3.67,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,171,3.68,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,172,3.69,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,173,3.7,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,174,3.71,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,175,3.72,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,176,3.73,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,177,3.74,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,178,3.75,7]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[197,179,3.76,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,180,3.77,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,181,3.78,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,182,3.79,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,183,3.8,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,184,3.81,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,185,3.82,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,186,3.83,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,187,3.84,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,188,3.85,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,189,3.86,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,190,3.87,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,191,3.88,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,192,3.89,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,193,3.9,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,194,3.91,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,195,3.92,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,196,3.93,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,197,3.94,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,198,3.95,7]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[197,199,3.96,7]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,0,1.98,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,1,1.99,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,2,2,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,3,2.01,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,4,2.02,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,5,2.03,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,6,2.04,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,7,2.05,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,8,2.06,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,9,2.07,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,10,2.08,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,11,2.09,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,12,2.1,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,13,2.11,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,14,2.12,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,15,2.13,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,16,2.14,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,17,2.15,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,18,2.16,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[198,19,2.17,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,20,2.18,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,21,2.19,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,22,2.2,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,23,2.21,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,24,2.22,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,25,2.23,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,26,2.24,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,27,2.25,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,28,2.26,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,29,2.27,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,30,2.28,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,31,2.29,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,32,2.3,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,33,2.31,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,34,2.32,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,35,2.33,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,36,2.34,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,37,2.35,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,38,2.36,8]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[198,39,2.37,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,40,2.38,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,41,2.39,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,42,2.4,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,43,2.41,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,44,2.42,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,45,2.43,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,46,2.44,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,47,2.45,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,48,2.46,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,49,2.47,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,50,2.48,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,51,2.49,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,52,2.5,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,53,2.51,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,54,2.52,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,55,2.53,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,56,2.54,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,57,2.55,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,58,2.56,8]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[198,59,2.57,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,60,2.58,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,61,2.59,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,62,2.6,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,63,2.61,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,64,2.62,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,65,2.63,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,66,2.64,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,67,2.65,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,68,2.66,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,69,2.67,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,70,2.68,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,71,2.69,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,72,2.7,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,73,2.71,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,74,2.72,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,75,2.73,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,76,2.74,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,77,2.75,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,78,2.76,8]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[198,79,2.77,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,80,2.78,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,81,2.79,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,82,2.8,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,83,2.81,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,84,2.82,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,85,2.83,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,86,2.84,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,87,2.85,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,88,2.86,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,89,2.87,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,90,2.88,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,91,2.89,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,92,2.9,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,93,2.91,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,94,2.92,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,95,2.93,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,96,2.94,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,97,2.95,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,98,2.96,8]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[198,99,2.97,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,100,2.98,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,101,2.99,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,102,3,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,103,3.01,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,104,3.02,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,105,3.03,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,106,3.04,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,107,3.05,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,108,3.06,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,109,3.07,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,110,3.08,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,111,3.09,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,112,3.1,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,113,3.11,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,114,3.12,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,115,3.13,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,116,3.14,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,117,3.15,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,118,3.16,8]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[198,119,3.17,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,120,3.18,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,121,3.19,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,122,3.2,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,123,3.21,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,124,3.22,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,125,3.23,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,126,3.24,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,127,3.25,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,128,3.26,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,129,3.27,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,130,3.28,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,131,3.29,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,132,3.3,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,133,3.31,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,134,3.32,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,135,3.33,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,136,3.34,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,137,3.35,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,138,3.36,8]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[198,139,3.37,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,140,3.38,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,141,3.39,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,142,3.4,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,143,3.41,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,144,3.42,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,145,3.43,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,146,3.44,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,147,3.45,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,148,3.46,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,149,3.47,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,150,3.48,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,151,3.49,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,152,3.5,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,153,3.51,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,154,3.52,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,155,3.53,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,156,3.54,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,157,3.55,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,158,3.56,8]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[198,159,3.57,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,160,3.58,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,161,3.59,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,162,3.6,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,163,3.61,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,164,3.62,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,165,3.63,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,166,3.64,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,167,3.65,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,168,3.66,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,169,3.67,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,170,3.68,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,171,3.69,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,172,3.7,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,173,3.71,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,174,3.72,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,175,3.73,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,176,3.74,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,177,3.75,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,178,3.76,8]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[198,179,3.77,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,180,3.78,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,181,3.79,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,182,3.8,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,183,3.81,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,184,3.82,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,185,3.83,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,186,3.84,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,187,3.85,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,188,3.86,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,189,3.87,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,190,3.88,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,191,3.89,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,192,3.9,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,193,3.91,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,194,3.92,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,195,3.93,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,196,3.94,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,197,3.95,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,198,3.96,8]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[198,199,3.97,8]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,0,1.99,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,1,2,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,2,2.01,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,3,2.02,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,4,2.03,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,5,2.04,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,6,2.05,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,7,2.06,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,8,2.07,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,9,2.08,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,10,2.09,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,11,2.1,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,12,2.11,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,13,2.12,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,14,2.13,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,15,2.14,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,16,2.15,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,17,2.16,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,18,2.17,9]) );
insert into pts_collection ( gid, pt ) values ( 180, pc_makepoint(1, ARRAY[199,19,2.18,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,20,2.19,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,21,2.2,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,22,2.21,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,23,2.22,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,24,2.23,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,25,2.24,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,26,2.25,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,27,2.26,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,28,2.27,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,29,2.28,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,30,2.29,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,31,2.3,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,32,2.31,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,33,2.32,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,34,2.33,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,35,2.34,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,36,2.35,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,37,2.36,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,38,2.37,9]) );
insert into pts_collection ( gid, pt ) values ( 181, pc_makepoint(1, ARRAY[199,39,2.38,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,40,2.39,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,41,2.4,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,42,2.41,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,43,2.42,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,44,2.43,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,45,2.44,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,46,2.45,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,47,2.46,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,48,2.47,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,49,2.48,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,50,2.49,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,51,2.5,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,52,2.51,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,53,2.52,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,54,2.53,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,55,2.54,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,56,2.55,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,57,2.56,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,58,2.57,9]) );
insert into pts_collection ( gid, pt ) values ( 182, pc_makepoint(1, ARRAY[199,59,2.58,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,60,2.59,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,61,2.6,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,62,2.61,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,63,2.62,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,64,2.63,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,65,2.64,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,66,2.65,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,67,2.66,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,68,2.67,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,69,2.68,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,70,2.69,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,71,2.7,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,72,2.71,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,73,2.72,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,74,2.73,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,75,2.74,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,76,2.75,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,77,2.76,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,78,2.77,9]) );
insert into pts_collection ( gid, pt ) values ( 183, pc_makepoint(1, ARRAY[199,79,2.78,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,80,2.79,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,81,2.8,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,82,2.81,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,83,2.82,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,84,2.83,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,85,2.84,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,86,2.85,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,87,2.86,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,88,2.87,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,89,2.88,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,90,2.89,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,91,2.9,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,92,2.91,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,93,2.92,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,94,2.93,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,95,2.94,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,96,2.95,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,97,2.96,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,98,2.97,9]) );
insert into pts_collection ( gid, pt ) values ( 184, pc_makepoint(1, ARRAY[199,99,2.98,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,100,2.99,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,101,3,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,102,3.01,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,103,3.02,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,104,3.03,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,105,3.04,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,106,3.05,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,107,3.06,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,108,3.07,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,109,3.08,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,110,3.09,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,111,3.1,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,112,3.11,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,113,3.12,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,114,3.13,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,115,3.14,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,116,3.15,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,117,3.16,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,118,3.17,9]) );
insert into pts_collection ( gid, pt ) values ( 185, pc_makepoint(1, ARRAY[199,119,3.18,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,120,3.19,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,121,3.2,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,122,3.21,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,123,3.22,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,124,3.23,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,125,3.24,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,126,3.25,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,127,3.26,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,128,3.27,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,129,3.28,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,130,3.29,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,131,3.3,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,132,3.31,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,133,3.32,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,134,3.33,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,135,3.34,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,136,3.35,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,137,3.36,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,138,3.37,9]) );
insert into pts_collection ( gid, pt ) values ( 186, pc_makepoint(1, ARRAY[199,139,3.38,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,140,3.39,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,141,3.4,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,142,3.41,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,143,3.42,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,144,3.43,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,145,3.44,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,146,3.45,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,147,3.46,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,148,3.47,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,149,3.48,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,150,3.49,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,151,3.5,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,152,3.51,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,153,3.52,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,154,3.53,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,155,3.54,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,156,3.55,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,157,3.56,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,158,3.57,9]) );
insert into pts_collection ( gid, pt ) values ( 187, pc_makepoint(1, ARRAY[199,159,3.58,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,160,3.59,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,161,3.6,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,162,3.61,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,163,3.62,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,164,3.63,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,165,3.64,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,166,3.65,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,167,3.66,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,168,3.67,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,169,3.68,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,170,3.69,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,171,3.7,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,172,3.71,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,173,3.72,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,174,3.73,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,175,3.74,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,176,3.75,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,177,3.76,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,178,3.77,9]) );
insert into pts_collection ( gid, pt ) values ( 188, pc_makepoint(1, ARRAY[199,179,3.78,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,180,3.79,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,181,3.8,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,182,3.81,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,183,3.82,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,184,3.83,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,185,3.84,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,186,3.85,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,187,3.86,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,188,3.87,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,189,3.88,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,190,3.89,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,191,3.9,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,192,3.91,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,193,3.92,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,194,3.93,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,195,3.94,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,196,3.95,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,197,3.96,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,198,3.97,9]) );
insert into pts_collection ( gid, pt ) values ( 189, pc_makepoint(1, ARRAY[199,199,3.98,9]) );
commit;
